#!/sbin/ksh

#  failure for:
#  sandia29

args=
start_at=1
bin_dir=$(dirname $0)
if [ -d $bin_dir/../VDT ] ; then
        vdt_dir=$bin_dir/../VDT
elif [ -d $bin_dir/../../VDT ] ; then 
        vdt_dir=$bin_dir/../../VDT
fi

#
#  usage function
#
function usage_and_exit {
cat <<EOF
Usage: $(basename $0) + optional arguments
   -start_at n
        start testing with input # n
   -vdt_dir d
        VDT installation directory
EOF
   exit 0
}

#
#  get the additional arguments
#
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
      -start_at)
         if [ $# -ge 1 ] ; then
            start_at=$1; shift
         else
            usage_and_exit
         fi
         ;;
      -vdt_dir)
         if [ $# -ge 1 ] ; then
            vdt_dir=$1; shift
         else
            usage_and_exit
         fi
         ;;
      -help)
        usage_and_exit
        ;;
      *)
         aargs=$args
         args="$a $aargs"
         ;;
   esac
done

# MANDIBLE 

testsuite="blockstack blockhole part1 awheel cubes twocubes cogwheel0 wheel twocylip brekeke1 sandia32"

log=demo.log
oldlog=$log.$$
intrlog=$log.intr

echo "\nVDT (C) 1998, Petr Krysl\n"

echo "Please, wait: preparing demo ..."

if [ -f $log ] ; then
        echo Moving $log into $oldlog
        mv $log $oldlog
fi

if [ -f "$oldlog" ] ; then
        tags=$(egrep 'VDT test @' $oldlog | sed -e "s/# *VDT/\\\n# VDT/g")
else

        t="# VDT test @ bcube @ 0.02 @ 8 @ 12 @ 8 @ 12 @ 7 @ # VDT test @ slitblock @ 0.02 @ 16 @ 28 @ 16 @ 28 @ 23 @ # VDT test @ blockhole1 @ 0.03 @ 20 @ 40 @ 20 @ 40 @ 38 @ # VDT test @ bracket @ 0.03 @ 18 @ 32 @ 18 @ 32 @ 37 @ # VDT test @ sandia8 @ 0.03 @ 34 @ 64 @ 34 @ 64 @ 83 @ # VDT test @ sandia11 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 116 @ # VDT test @ sandia13 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 114 @ # VDT test @ sandia18 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 116 @ # VDT test @ sandia2 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 117 @ # VDT test @ sandia30 @ 0.04 @ 12 @ 20 @ 17 @ 30 @ 32 @ # VDT test @ sandia3 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 118 @ # VDT test @ sandia4 @ 0.05 @ 46 @ 88 @ 46 @ 88 @ 115 @ # VDT test @ sandia10 @ 0.06 @ 46 @ 88 @ 46 @ 88 @ 116 @ # VDT test @ sandia9 @ 0.06 @ 46 @ 88 @ 46 @ 88 @ 113 @ # VDT test @ sandia29 @ 0.07 @ 20 @ 36 @ 25 @ 46 @ 51 @ # VDT test @ sandia25 @ 0.08 @ 18 @ 32 @ 23 @ 42 @ 44 @ # VDT test @ sandia31 @ 0.08 @ 18 @ 32 @ 28 @ 52 @ 62 @ # VDT test @ sandia5 @ 0.08 @ 26 @ 48 @ 34 @ 64 @ 75 @ # VDT test @ blockstack @ 0.1 @ 40 @ 76 @ 40 @ 76 @ 90 @ # VDT test @ sandia7 @ 0.11 @ 46 @ 88 @ 50 @ 96 @ 109 @ # VDT test @ blockhole @ 0.14 @ 56 @ 112 @ 56 @ 112 @ 90 @ # VDT test @ sandia27 @ 0.14 @ 94 @ 184 @ 96 @ 188 @ 251 @ # VDT test @ bracket1 @ 0.16 @ 243 @ 482 @ 243 @ 482 @ 685 @ # VDT test @ prism1 @ 0.16 @ 316 @ 488 @ 316 @ 488 @ 1111 @ # VDT test @ sandia26 @ 0.17 @ 90 @ 176 @ 94 @ 184 @ 245 @ # VDT test @ part1 @ 0.18 @ 64 @ 124 @ 64 @ 124 @ 98 @ # VDT test @ clover @ 0.2 @ 387 @ 770 @ 387 @ 770 @ 966 @ # VDT test @ ring1 @ 0.23 @ 400 @ 800 @ 400 @ 800 @ 1083 @ # VDT test @ sphere5 @ 0.23 @ 436 @ 568 @ 436 @ 568 @ 1708 @ # VDT test @ sandia16 @ 0.29 @ 194 @ 384 @ 207 @ 410 @ 580 @ # VDT test @ prism3 @ 0.3 @ 494 @ 528 @ 494 @ 528 @ 2053 @ # VDT test @ vertebra @ 0.3 @ 72 @ 140 @ 82 @ 160 @ 201 @ # VDT test @ sandia15 @ 0.34 @ 206 @ 408 @ 224 @ 444 @ 634 @ # VDT test @ hollowcube @ 0.39 @ 876 @ 1768 @ 876 @ 1768 @ 2262 @ # VDT test @ joint @ 0.43 @ 808 @ 1090 @ 808 @ 1090 @ 3172 @ # VDT test @ slitshell @ 0.42 @ 348 @ 692 @ 354 @ 704 @ 992 @ # VDT test @ cog0 @ 0.44 @ 336 @ 672 @ 336 @ 672 @ 572 @ # VDT test @ funny2 @ 0.45 @ 373 @ 746 @ 375 @ 750 @ 793 @ # VDT test @ sandia20 @ 0.49 @ 326 @ 648 @ 326 @ 648 @ 1077 @ # VDT test @ splitshell1 @ 0.48 @ 352 @ 700 @ 360 @ 716 @ 1003 @ # VDT test @ awheel @ 0.56 @ 216 @ 480 @ 216 @ 480 @ 534 @ # VDT test @ sgi @ 0.55 @ 1314 @ 2208 @ 1314 @ 2208 @ 3795 @ # VDT test @ crack @ 0.59 @ 965 @ 508 @ 965 @ 508 @ 4880 @ # VDT test @ cage1 @ 0.64 @ 456 @ 968 @ 460 @ 976 @ 1130 @ # VDT test @ holecyl1 @ 0.67 @ 708 @ 1416 @ 708 @ 1416 @ 2148 @ # VDT test @ prism2 @ 0.68 @ 1394 @ 1896 @ 1394 @ 1896 @ 5265 @ # VDT test @ cubes @ 0.67 @ 54 @ 104 @ 90 @ 176 @ 194 @ # VDT test @ maze1 @ 0.7 @ 76 @ 140 @ 123 @ 234 @ 235 @ # VDT test @ aholecyl @ 0.71 @ 738 @ 1484 @ 738 @ 1484 @ 2213 @ # VDT test @ awheel1 @ 0.74 @ 1534 @ 3116 @ 1538 @ 3124 @ 4171 @ # VDT test @ castlefiner @ 0.83 @ 1176 @ 1106 @ 1176 @ 1106 @ 4973 @ # VDT test @ sandia22 @ 0.84 @ 480 @ 956 @ 480 @ 956 @ 1618 @ # VDT test @ nofancy1 @ 0.87 @ 456 @ 912 @ 456 @ 912 @ 1368 @ # VDT test @ aholecyl2 @ 0.89 @ 993 @ 2034 @ 1001 @ 2050 @ 2980 @ # VDT test @ aholecyl1 @ 0.89 @ 820 @ 1652 @ 828 @ 1668 @ 2548 @ # VDT test @ flangedpipe @ 14.61 @ 1538 @ 3140 @ 1613 @ 3290 @ 4133 @ # VDT test @ realmosk @ 35.43 @ 22683 @ 7290 @ 22834 @ 7592 @ 19478 @ # VDT test @ sandia1 @ 258.27 @ 79031 @ 13644 @ 79280 @ 14142 @ 81996 @ # VDT test @ twocyl @ 13.67 @ 92 @ 180 @ 172 @ 340 @ 451 @ # VDT test @ twocylip @ 4.68 @ 108 @ 180 @ 211 @ 386 @ 601 @ # VDT test @ cyls @ 1 @ 835 @ 1670 @ 844 @ 1688 @ 2530 @ # VDT test @ aholecyl3 @ 1.01 @ 1073 @ 2190 @ 1085 @ 2214 @ 3232 @ # VDT test @ twocubes @ 1.02 @ 16 @ 24 @ 78 @ 148 @ 165 @ # VDT test @ sandia14 @ 1.03 @ 58 @ 112 @ 109 @ 214 @ 272 @ # VDT test @ yawheel @ 1.09 @ 1732 @ 3508 @ 1733 @ 3510 @ 4946 @ # VDT test @ elementaler @ 1.1 @ 1214 @ 2444 @ 1217 @ 2450 @ 3382 @ # VDT test @ pierced1 @ 1.17 @ 506 @ 1004 @ 514 @ 1020 @ 1783 @ # VDT test @ cogwheel0 @ 1.19 @ 1216 @ 2448 @ 1216 @ 2448 @ 3320 @ # VDT test @ funny6 @ 1.18 @ 845 @ 1694 @ 864 @ 1732 @ 2179 @ # VDT test @ junction @ 1.24 @ 1843 @ 3256 @ 1843 @ 3256 @ 6133 @ # VDT test @ spherewcutouts @ 1.24 @ 508 @ 1008 @ 533 @ 1058 @ 1757 @ # VDT test @ apierce @ 1.38 @ 1359 @ 2726 @ 1360 @ 2728 @ 4228 @ # VDT test @ funny3 @ 1.36 @ 652 @ 1308 @ 662 @ 1328 @ 1539 @ # VDT test @ crankshaft @ 1.41 @ 1673 @ 2068 @ 1673 @ 2068 @ 6740 @ # VDT test @ bridge3 @ 1.43 @ 1286 @ 2568 @ 1308 @ 2612 @ 4005 @ # VDT test @ dumbbell @ 1.47 @ 583 @ 1162 @ 593 @ 1182 @ 1659 @ # VDT test @ castle @ 1.57 @ 57 @ 110 @ 84 @ 164 @ 237 @ # VDT test @ cage2 @ 1.61 @ 1240 @ 2536 @ 1257 @ 2570 @ 3082 @ # VDT test @ blocks2 @ 1.72 @ 190 @ 384 @ 259 @ 522 @ 718 @ # VDT test @ vejir @ 1.81 @ 577 @ 1150 @ 670 @ 1336 @ 1911 @ # VDT test @ opencage @ 1.87 @ 1263 @ 2566 @ 1301 @ 2642 @ 3242 @ # VDT test @ cog1 @ 2.12 @ 2561 @ 5122 @ 2568 @ 5136 @ 7049 @ # VDT test @ sandia21 @ 2.14 @ 1496 @ 2992 @ 1496 @ 2992 @ 4867 @ # VDT test @ wheel @ 2.17 @ 624 @ 1248 @ 624 @ 1248 @ 1591 @ # VDT test @ slab2 @ 2.3 @ 580 @ 1196 @ 661 @ 1358 @ 1691 @ # VDT test @ bumps1 @ 2.31 @ 297 @ 590 @ 308 @ 612 @ 771 @ # VDT test @ apart1 @ 2.32 @ 1186 @ 2368 @ 1189 @ 2374 @ 3710 @ # VDT test @ funny @ 2.32 @ 2447 @ 4894 @ 2447 @ 4894 @ 7126 @ # VDT test @ pipeline3 @ 2.36 @ 2899 @ 1878 @ 2899 @ 1878 @ 14474 @ # VDT test @ mosk1 @ 2.47 @ 1160 @ 2320 @ 1198 @ 2396 @ 3156 @ # VDT test @ cheese @ 2.46 @ 1475 @ 2978 @ 1503 @ 3034 @ 3972 @ # VDT test @ bridge1 @ 2.58 @ 166 @ 328 @ 305 @ 606 @ 907 @ # VDT test @ blocks1 @ 2.59 @ 54 @ 112 @ 148 @ 300 @ 427 @ # VDT test @ funny5 @ 2.72 @ 689 @ 1386 @ 727 @ 1462 @ 1709 @ # VDT test @ bumps2 @ 2.69 @ 631 @ 1254 @ 667 @ 1326 @ 2293 @ # VDT test @ cage @ 2.9 @ 144 @ 344 @ 175 @ 406 @ 502 @ # VDT test @ rosette @ 2.91 @ 104 @ 204 @ 174 @ 344 @ 469 @ # VDT test @ funny4 @ 3.42 @ 665 @ 1338 @ 689 @ 1386 @ 1637 @ # VDT test @ ringlet @ 3.51 @ 1778 @ 3564 @ 1792 @ 3592 @ 5306 @ # VDT test @ pierce @ 4.03 @ 1344 @ 2688 @ 1366 @ 2732 @ 4286 @ # VDT test @ sandia24 @ 4.12 @ 1230 @ 2456 @ 1242 @ 2480 @ 3813 @ # VDT test @ cogwheel @ 5.17 @ 5474 @ 10948 @ 5477 @ 10954 @ 17485 @ # VDT test @ pierced2 @ 5.13 @ 1184 @ 2360 @ 1210 @ 2412 @ 3944 @ # VDT test @ funny7 @ 6.01 @ 931 @ 1870 @ 986 @ 1980 @ 2491 @ # VDT test @ brekeke1 @ 6.51 @ 528 @ 1060 @ 564 @ 1132 @ 1328 @ # VDT test @ axle @ 6.78 @ 3744 @ 7580 @ 3745 @ 7582 @ 11792 @ # VDT test @ brekeke @ 8.17 @ 180 @ 364 @ 328 @ 660 @ 792 @ # VDT test @ sandia32 @ 8.92 @ 2192 @ 4388 @ 2197 @ 4398 @ 6956 @ # VDT test @ sandia28 @ 9.98 @ 979 @ 1954 @ 980 @ 1956 @ 2920 @ # VDT test @ dumbbell1 @ 10.6 @ 703 @ 1402 @ 736 @ 1468 @ 2290 @ # VDT test @ MANDIBLE @ 11.93 @ 1453 @ 2902 @ 1553 @ 3102 @ 4818 @ # VDT test @ ring @ 12.93 @ 212 @ 424 @ 263 @ 526 @ 692 @ # VDT test @ junction1 @ 28.01 @ 27862 @ 22500 @ 27862 @ 22500 @ 133318 @ # VDT test @ archfiner1 @ 81.87 @ 2410 @ 4836 @ 2591 @ 5198 @ 9937 @ # VDT test @  mol @ 15.69  @ 16939 @  3002 @  17037  @  3198  @  19197 @"
      tags=$(echo "$t" | sed -e "s/# *VDT/\\\n# VDT/g")
fi


function find_tag {
        input=$1; shift
        where="$*"
        t="$(echo $where | sed -e "s/# *VDT/\\\n# VDT/g" | egrep "VDT *test *\@ *$input *\@.*")"
        echo "$t"
}


function expected_time {
        l="$*"
        if [ x"$l" != x ] ; then
                t=$(echo $l | cut -d@ -f3)
        else
                t="?"
        fi
        echo "$t"
}


function extract_field {
        aline=$1; shift
        which=$1
        if [ x"$aline" != x ] ; then
                extract_field_result=$(echo $aline | cut -d@ -f$which)
        else
                extract_field_result="?"
        fi
        echo "$extract_field_result"
}

function consumed_time {
        aline=$1
        echo $(extract_field "$aline" 3)
}

function num_of_input_points {
        aline=$1
        echo $(extract_field "$aline" 4)
}

function num_of_input_bsfs {
        aline=$1
        echo $(extract_field "$aline" 5)
}

function num_of_output_points {
        aline=$1
        echo $(extract_field "$aline" 6)
}

function num_of_output_bsfs {
        aline=$1
        echo $(extract_field "$aline" 7)
}


function num_of_output_tets {
        aline=$1
        echo $(extract_field "$aline" 8)
}



function kill_vdt {
        echo "Killing vdt"
        echo "Moving current log \"$log\" to \"$intrlog\""
        mv $log $intrlog
        if [ -f  $oldlog ] ; then 
           echo "Copying old log \"$oldlog\" to \"$log\""
           cp $oldlog $log
        fi
        exit 0
}

trap "kill_vdt SIGINT $file" 2 
trap "kill_vdt SIGKILL $file" 9 
trap "kill_vdt SIGTERM $file" 15


function order_by_time_to_completion {
        list=$*
        time_to_completion_list=""
        for n in $list
        do
                t=$(expected_time $(find_tag $n "$tags"))
                if [ x"$t" = x"?" ] ; then
                        ttcl="$time_to_completion_list"
                        time_to_completion_list="$ttcl\n$n:1e6"
                else
                        ttcl="$time_to_completion_list"
                        time_to_completion_list="$ttcl\n$n:$t"
                fi
        done
        result=$(echo "$time_to_completion_list" | sort | uniq | sort -t: -k2 -n | cut -d: -f1)
        echo $result
}


testsuite=$(order_by_time_to_completion $testsuite)


serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
done

echo Running test for $(($serial-$start_at+1)) inputs    
echo ""

echo "VDT test started at $(date +%H:%M:%S)"
echo ""

serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
        if [ $serial -ge $start_at ] ; then
          echo "Mesh #${serial}: $n, expected time: $(expected_time $(find_tag $n "$tags")) seconds, started at $(date +%H:%M:%S)"
          echo "" >> $log 2>&1 
          echo "<<<<<<<<<<<<<<<< Problem #${serial}: $n >>>>>>>>>>>>>>>>" >> $log 2>&1 
          vdt -giface -shift_method 1 -i $vdt_dir/inputs/$n $(eval echo $args) >> $log 2>&1 
          if [ $? -ne 0 ] ; then
                  echo "***** ERROR ($(basename $0)); see the log $log"
                  exit 1
          fi
          echo "" >> $log 2>&1 
          l=$(find_tag $n "$(cat $log | sed -e "s/# *VDT/\\\n# VDT/g")")
          echo "   Input:  $(num_of_input_points "$l") points, $(num_of_input_bsfs "$l") boundary facets"
          echo "   Output: $(num_of_output_points "$l") points, $(num_of_output_bsfs "$l") boundary facets, $(num_of_output_tets "$l") tets."
          echo "   Time consumed: $(consumed_time "$l") seconds"
          echo ""
        fi
done
echo "VDT demo finished at $(date +%H:%M:%S)"