
#include <stdlib.h>
#include <stdio.h>
#include "randP.h"
#include "getoptP.h"



#ifndef MAX_GEN_PT 
# define MAX_GEN_PT 1000
#endif

static int max_gen_pt = MAX_GEN_PT;
static int report_opts = FALSE;

static void 
usage (int argc, char **argv)
{
  char *option, *usage;

  fprintf (stderr, "usage: %s -i file + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  exit (0);
}

static void 
process_options (int argc, char **argv)
{
  int help = FALSE;

#define DOSTRINGZ(s) #s
#define STRINGZ(s) DOSTRINGZ(s)
#define SET_BOOL(name)                                          \
  {                                                             \
    if (report_opts) { fprintf (stderr, "Set: %s\n", #name); }  \
    name = TRUE;                                                \
  }
#define SET_WITH_ARG(name, format)                                        \
  {                                                                       \
    if (report_opts) { fprintf (stderr, "Set: %s = %" #format "\n",       \
                                STRINGZ(name), name); }                   \
  }
  
  GET_OPT_BOOL (argc, argv, "-report_opts", SET_BOOL(report_opts),
                "prints options as they are being set");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL(help),
                "prints usage for each option recognized");
  
  GET_OPT_ARG (argc, argv, "-max_gen_pt", int, max_gen_pt,
               SET_WITH_ARG (max_gen_pt, d),
               "max of points to generate");
  
  if (help)
    usage (argc, argv);
}

int
main (int argc, char **argv)
{
  double r = 1;
  double x, y, z;
  int n, i;
  int id = 0;

  process_options (argc, argv);

  n = pow ((double) max_gen_pt, (double) 1/3) + 1;
  printf ("mesh_size %g\n", 2*r / n);
  for (i = 0; i < max_gen_pt; i++) {
    x = r * (-1 + 2 * rand_0_to_1 ());
    y = r * (-1 + 2 * rand_0_to_1 ());
    z = r * (-1 + 2 * rand_0_to_1 ());
    
    if (x*x + y*y + z*z <= r*r) {
      printf ("pt %d %g %g %g\n", id++, x, y, z);
    }

  }

  return 0;
}
