

#include <stdio.h>
#include <stdlib.h>
#include "ioP.h"              /* Ckit library */
#include "tokensP.h"          /* Ckit library */

static int npts = 0, nfacets = 0;
static tokens_parser PARSER = NULL;
static double toknum = 0;
static const char *tokstring = NULL;

#define TOKNUM(which)                                                   \
  {                                                                     \
    if (   tokens_total_of_tokens (PARSER) < which                      \
        || tokens_type_of_token (PARSER, which) != TOKENS_NUMBER) {     \
      char buf[512];                                                    \
      sprintf (buf, "Line %d: Expecting numger at pos %d, got \"%s\"",  \
               tokens_curr_line_num (PARSER), which,                    \
               tokens_curr_line_contents (PARSER));                     \
      errexit (buf);                                                    \
    }                                                                   \
    toknum = tokens_token_as_double (PARSER, which);                    \
  }  

#define TOKSTR(which)                                                   \
  {                                                                     \
    if (   tokens_total_of_tokens (PARSER) < which                      \
        || tokens_type_of_token (PARSER, which) != TOKENS_STRING) {     \
      char buf[512];                                                    \
      sprintf (buf, "Line %d: Expecting string at pos %d, got \"%s\"",  \
               tokens_curr_line_num (PARSER), which,                    \
               tokens_curr_line_contents (PARSER));                     \
      errexit (buf);                                                    \
    }                                                                   \
    tokstring = tokens_token_as_string (PARSER, which);                 \
  }  

#define NEXT_LINE                               \
  {                                             \
    if (!tokens_next_line (PARSER)) {           \
      errexit ("Failed to read stdin");         \
    }                                           \
  }

static int
errexit_ (const char *s, const char *file, int line)
{
  fprintf (stderr, "Error: %s (l.%d in %s)\n", s, line, file);
  exit (0);
  return 1;
}
#define errexit(s) errexit_ (s, __FILE__, __LINE__)

#define NL "\n"

static void
read_header (void)
{
  NEXT_LINE; TOKSTR (1);
  printf ("! session: %s" NL, tokstring);
  NEXT_LINE; TOKNUM (1);
  printf ("! partition: %d" NL, (int)toknum);
  NEXT_LINE; TOKNUM (1);
  npts = (int) toknum;
  printf ("! # points: %d" NL, npts);
  NEXT_LINE; TOKNUM (1);
  nfacets = (int) toknum;
  printf ("! # facets: %d" NL, nfacets);
  NEXT_LINE; TOKNUM (1);
  printf ("! space dimensions: %d" NL, (int) toknum);
  NEXT_LINE; TOKNUM (1);
  printf ("! nodes / element: %d" NL, (int) toknum);
  NEXT_LINE; TOKNUM (1);
  printf ("! element type: %d" NL, (int) toknum);
  NEXT_LINE; TOKSTR (1);
  printf ("! format: %s" NL, tokstring);
  printf ("mesh_size 1" NL);
}

static void
read_pts (void)
{
  int i;
  double x, y, z, h;

  for (i = 0; i < npts; i++) {
    NEXT_LINE; 
    TOKNUM (1); x = toknum;
    TOKNUM (2); y = toknum;
    TOKNUM (3); z = toknum;
    TOKNUM (4); h = toknum;
    printf ("pt %d %g %g %g %g" NL, i, x, y, z, h);
  }
}


static void
read_facets (void)
{
  int i;
  int n[3];
  int surf_id = 1;

  for (i = 0; i < nfacets; i++) {
    NEXT_LINE; 
    TOKNUM (1); n[0] = toknum;
    TOKNUM (2); n[1] = toknum; 
    TOKNUM (3); n[2] = toknum;
    /* Note: swapping two points to achieve the correct orientation */
    printf ("bsf %d %d %d %d" NL, n[0], n[2], n[1], surf_id);
  }
}


int 
main (void)
{
  PARSER = tokens_new_parser ();
  if (!tokens_open_file (PARSER, "-"))
    errexit ("Failed to open stdin");
  
  ckit_io_set_comment_char (tokens_get_iobuf (PARSER), '#');
  
  read_header ();
  read_pts ();
  read_facets ();

  printf ("random_pt_shift 1e-5" NL);

  tokens_close_curr_file (PARSER);

  return 0;
}
