#!/bin/bash

args=
start_at=1
bin_dir=$(dirname $0)
if [ -d $bin_dir/../VDT ] ; then
        vdt_dir=$bin_dir/../VDT
elif [ -d $bin_dir/../../VDT ] ; then 
        vdt_dir=$bin_dir/../../VDT
fi

vdt_exec=vdt

#
#  usage function
#
function usage_and_exit {
cat <<EOF
Usage: $(basename $0) + optional arguments
   -start_at n
        start testing with input # n
   -vdt_dir d
        VDT installation directory
   -vdt_exec b
        VDT executable to test
EOF
   exit 0
}

#
#  get the additional arguments
#
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
      -start_at)
         if [ $# -ge 1 ] ; then
            start_at=$1; shift
         else
            usage_and_exit
         fi
         ;;
      -vdt_dir)
         if [ $# -ge 1 ] ; then
            vdt_dir=$1; shift
         else
            usage_and_exit
         fi
         ;;
      -vdt_exec)
         if [ $# -ge 1 ] ; then
            vdt_exec=$1; shift
         else
            usage_and_exit
         fi
         ;;
      -help)
        usage_and_exit
        ;;
      *)
         aargs=$args
         args="$aargs $a"
         ;;
   esac
done

#  test failed for
failedfor=""
if [ x"$failedfor" != x ] ; then
   echo ""
   echo "***** WARNING ($(basename $0)): Test failed before for: $failedfor"
fi

#  test suite
testsuite="ahub alucagepipe alucage blocktorus tori comb combref combref1 blockwfh1 cranwbrain hand beethoven bhovenrep boxes donut triceratops ckanesphere ckaneplate block1 block0.5 block0.35 block0.25 nestedcubes brekeke cheese twocubes cage2 opencage longcone cage1 maze1 rosette aholecyl aholecyl1 aholecyl2 aholecyl3 archfiner1 awheel awheel1 axle blockhole blockhole1 blockstack bracket bracket1 bubbles bumps2 cage can2 castle castlefiner clover cog0 cogwheel cone coneref coneref1 cylinder  mech twocyl twocylip cogwheel0 crack crankshaft cubes cyls elementaler funny funny2 funny4 funny6 holecyl1 hollowcube joint junction MANDIBLE mandible  nofancy1 part1 pierced1 pierced2 pipeline3 prism1 prism2 prism3 realmosk ring ring1 bcube doubtet schonhardt revschonhardt faceknife knife3 knife2 indpint gsphere scale1 mazewheel scale2 scale3 scale5 scale2ref manyreg two ringlet sandia1 sandia2 sandia3 sandia4 sandia5 sandia7 sandia8 sandia9 sandia10 sandia11 sandia13 sandia14 sandia15 sandia16 sandia18 sandia20 sandia21 sandia22 sandia24 sandia25 sandia26 sandia27 sandia28 sandia30 sandia31 sandia32 sgi sgi1 slab2 slitblock slitshell sphere5 spherewcutouts splitshell1  vejir vertebra wheel yawheel apierce bumps1 cog1 funny3 apart1 flangedpipe blocks1 blocks2 brekeke1 bridge1 bridge3 dumbbell dumbbell1 funny5 funny7 junction1 mol mosk1 nofancy1 pierce pierced2 pipeline3 realmosk tire tfire rosette sandia5 sandia7 sandia8 sandia14 sandia15 sandia20 sandia22 sandia24 spiral sputter1 sandia25 sandia26 sandia27 sandia28 sandia29 sandia30 sandia31 sandia32 sgi slab2 slitblock slitshell sphere5 spherewcutouts splitshell1 vejir vertebra wheel yawheel hollowtorus hollowtorus1 hollowtorus2 hollowtorus3 strangepropeller nicesphere nicespheresmall 2cubes blade l"

log=test.log
oldlog=$log.$$
reflog=test.ref
intrlog=$log.intr

echo "\nVDT (C) 1998--2000, Petr Krysl\n"

echo "Please, wait: preparing test ..."

if [ -f $log ] ; then
        echo Moving $log into $oldlog
        mv $log $oldlog
fi

if [ -f "$oldlog" ] ; then
    echo Using timing info from $oldlog
    tags=$(egrep 'VDT test @' $oldlog | sed -e "s/# *VDT/\\\n# VDT/g")
elif [ -f "$reflog" ] ; then
    echo Using timing info from $reflog
    tags=$(egrep 'VDT test @' $reflog | sed -e "s/# *VDT/\\\n# VDT/g")
else
    echo No timing info available
    tags="?"
fi


function find_tag {
        input=$1; shift
        where="$*"
        t="$(echo $where | sed -e "s/# *VDT/\\\n# VDT/g" | egrep "VDT *test *\@ *$input *\@.*")"
        echo "$t"
}


function expected_time {
        l="$*"
        if [ x"$l" != x ] ; then
                t=$(echo $l | cut -d@ -f3)
        else
                t="?"
        fi
        echo "$t"
}


function extract_field {
        aline=$1; shift
        which=$1
        if [ x"$aline" != x ] ; then
                extract_field_result=$(echo $aline | cut -d@ -f$which)
        else
                extract_field_result="?"
        fi
        echo "$extract_field_result"
}

function consumed_time {
        aline=$1
        echo $(extract_field "$aline" 3)
}

function num_of_input_points {
        aline=$1
        echo $(extract_field "$aline" 4)
}

function num_of_input_bsfs {
        aline=$1
        echo $(extract_field "$aline" 5)
}

function num_of_output_points {
        aline=$1
        echo $(extract_field "$aline" 6)
}

function num_of_output_bsfs {
        aline=$1
        echo $(extract_field "$aline" 7)
}


function num_of_output_tets {
        aline=$1
        echo $(extract_field "$aline" 8)
}



function kill_vdt {
        echo "Killing vdt"
        echo "Moving current log \"$log\" to \"$intrlog\""
        mv $log $intrlog
        if [ -f  $oldlog ] ; then 
           echo "Copying old log \"$oldlog\" to \"$log\""
           cp $oldlog $log
        fi
        exit 0
}

trap "kill_vdt SIGINT $file" 2 
trap "kill_vdt SIGKILL $file" 9 
trap "kill_vdt SIGTERM $file" 15


function order_by_time_to_completion {
        list=$*
        time_to_completion_list=""
        for n in $list
        do
                t=$(expected_time $(find_tag $n "$tags"))
                if [ x"$t" = x"?" ] ; then
                        ttcl="$time_to_completion_list"
                        time_to_completion_list="$ttcl\n$n:1e6"
                else
                        ttcl="$time_to_completion_list"
                        time_to_completion_list="$ttcl\n$n:$t"
                fi
        done
        s=$(echo "$time_to_completion_list" | sort | uniq)
        result="?"
        if [ $(echo "$s" | wc -w) -gt 2 ] ; then
            result=$(echo "$s" | sort -n -t : -k 2 | cut -d: -f1)
        fi
        echo $result
}


testsuite=$(order_by_time_to_completion $testsuite)


serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
done

arch=$(vdtarch.sh)
echo ""
echo Architecture: $arch
echo ""

echo Running test for $(($serial-$start_at+1)) inputs    
echo ""

echo "VDT test started at $(date +%H:%M:%S)"
echo ""

echo "=======================================================================" >>  $log 2>&1 
echo "=                                                                     =" >>  $log 2>&1 
echo "  VDT run on architecture $arch, started $(date)" >>  $log 2>&1 
echo "      Started $(date)" >>  $log 2>&1 
echo "=                                                                     =" >>  $log 2>&1 
echo "=======================================================================" >>  $log 2>&1 

serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
        if [ $serial -ge $start_at ] ; then
          date=$(date +%H:%M:%S)
          expt="$(expected_time $(find_tag $n "$tags"))"
          echo "Mesh #${serial}: $n, expected time: $expt seconds, started at $date"
          echo "" >> $log 2>&1 
          echo "<<<<<<<<<<<<<<<< Problem #${serial}: $n >>>>>>>>>>>>>>>>" >> $log 2>&1 
          input=$vdt_dir/inputs/RCS/${n%%.vin}.vin
          vdt.sh -vdt_exec $vdt_exec -report_opts -test_mode -i $input $(eval echo $args) >> $log 2>&1 
          if [ $? -ne 0 ] ; then
                  echo "***** ERROR ($(basename $0)); see the log $log"
                  exit 1
          fi
          echo "" >> $log 2>&1 
          l=$(find_tag ${n%%.vin} "$(cat $log | sed -e "s/# *VDT/\\\n# VDT/g" | egrep '\# VDT')")
          nip=$(num_of_input_points "$l")
          nibf=$(num_of_input_bsfs "$l")
          echo "   Input:  $nip points, $nibf boundary facets"
          nop=$(num_of_output_points "$l")
          nobf=$(num_of_output_bsfs "$l")
          not=$(num_of_output_tets "$l")
          echo "   Output: $nop points, $nobf boundary facets, $not tets."
          ct=$(consumed_time "$l")
          if [ x"$expt" != x'?' ] ; then
             pt=$(echo "$ct / $expt * 100" | bc -l | cut -d. -f1)
             if [ $pt -le 75 -o $pt -ge 150 ] ; then
                echo "   Time consumed: $ct seconds ($pt %)" 
             else
                echo "   Time consumed: $ct seconds"
             fi
          else
             echo "   Time consumed: $ct seconds"
          fi
          echo ""
        fi
done

echo "VDT test finished at $(date +%H:%M:%S)"
