
/* 
      Variational Delaunay Triangulation
             (C) 1998-2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdt.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdt.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdt) 
 
#define TETPTSARE(tet, n0, n1, n2, n3)          \
  (                                             \
   (((int) tet->pt[0]->id) == n0 ||             \
    ((int) tet->pt[0]->id) == n1 ||             \
    ((int) tet->pt[0]->id) == n2 ||             \
    ((int) tet->pt[0]->id) == n3)               \
   &&                                           \
   (((int) tet->pt[1]->id) == n0 ||             \
    ((int) tet->pt[1]->id) == n1 ||             \
    ((int) tet->pt[1]->id) == n2 ||             \
    ((int) tet->pt[1]->id) == n3)               \
   &&                                           \
   (((int) tet->pt[2]->id) == n0 ||             \
    ((int) tet->pt[2]->id) == n1 ||             \
    ((int) tet->pt[2]->id) == n2 ||             \
    ((int) tet->pt[2]->id) == n3)               \
   &&                                           \
   (((int) tet->pt[3]->id) == n0 ||             \
    ((int) tet->pt[3]->id) == n1 ||             \
    ((int) tet->pt[3]->id) == n2 ||             \
    ((int) tet->pt[3]->id) == n3)               \
   )


static void
reset_quality (vdt_t *vdt)
{
  vdt->quality.quality_measure            = VDT_GAMMA_QUALITY;
  vdt->quality.sliver_dihed_angle         = VDT_SLIVER_DIHED_ANGLE;
  vdt->quality.sliver_quality             = 0.001;
  vdt->quality.nranges                    = 0;
  vdt->quality.quality_summary            = NULL;
  vdt->quality.angle_quality_summary      = NULL;
  vdt->quality.num_non_positive_vol_tets  = 0;
  vdt->quality.num_small_dihed_angle_tets = 0;
  vdt->quality.min_quality                = 1;
  vdt->quality.max_quality                = 0;
  vdt->quality.min_volume                 = FLT_MAX;
  vdt->quality.max_volume                 = 0;
  vdt->quality.min_dihed_angle            = 0;
  vdt->quality.max_dihed_angle            = M_PI;
}

/*
   Create a vdt context.
 */
vdt_t *
vdt_new (void)
{
  vdt_t *vdt;
  CkitPoint3
    infty = {FLT_MAX,FLT_MAX,FLT_MAX},
    mininfty = {-FLT_MAX,-FLT_MAX,-FLT_MAX};  
  int i;
  
  vdt_restart_rand_0_to_1 ();
  
  vdt = NEWP_OBJ (vdt_t); CHECK_MEM (vdt);

  vdt->input                    = NULL;
  vdt->input_version            = NULL;
  vdt->input_true_name          = NULL;
  vdt->pt_as_read               = make_list ();
  vdt->bsf_as_read              = make_list ();
  vdt->pt_to_insert_as_if_gen   = make_list ();
  vdt->regd                     = dict_create_dictionary ();
  vdt->pointd                   = dict_create_dictionary ();
  vdt->pointl                   = make_list ();
  vdt->max_pt_id                = 0;
  vdt->input_max_pt_id          = 0;
  G3D_INIT_BOX (vdt->bbox, infty);
  G3D_UPDT_BOX (vdt->bbox, mininfty);
  vdt->tetl                     = make_list ();
  vdt->bsfl                     = make_list ();
  vdt->def_mesh_size            = 0;
  vdt->min_mesh_size            = FLT_MAX;
  vdt->max_mesh_size            = 0;
#if GIFACE
  vdt->gmodel                   = NULL;
#endif
  vdt->num_active_points        = 0;
  vdt->tot_active_points        = 1000;
  vdt->active_pt                = NEWP_OBJ_ARRAY (vdt->tot_active_points, PTP);
  CHECK_MEM (vdt->active_pt);
  vdt->active_loc               = NEWP_OBJ_ARRAY (vdt->tot_active_points,
                                                  CkitPoint3);
  CHECK_MEM (vdt->active_loc);
#if VDT_PIRAT
  vdt->pt_pirat                 = NULL;
#endif
#if VDT_BUCKETS
  vdt->buckets                  = NULL;
#endif
  vdt->bucket_size              = 0;
  vdt->fflist                   = make_list ();
  FOR (i, 4) {
    vdt->stop_tet.pt[i]     = NULL;
    vdt->stop_tet.adjtet[i] = NULL;
  }
  vdt->new_pt                   = NULL;
  vdt->surf_mod_pass            = 0;
  vdt->ot                       = NULL;
  vdt->bdry_pt_shift_mult       = VDT_BDRY_PT_DEFAULT_SHIFT_MULT;
  vdt->gen_pt_shift_mult        = VDT_GEN_PT_DEFAULT_SHIFT_MULT;
#if VDT_USE_HIERLOBS
  vdt->hierlobs                 = NULL;
#else
  vdt->bsflobs                  = NULL;
#endif
  vdt->near_surf_test           = VDT_NEAR_SURF_TEST_DIST;
  vdt->del_stat.num_bsf_updates          = 0;
  vdt->del_stat.lp_stat.sum_unknowns     = 0;
  vdt->del_stat.lp_stat.sum_evals        = 0;
  vdt->del_stat.lp_stat.max_num_unknowns = 0;
  vdt->lp_size_limit            = 0;
  vdt->allow_flips              = TRUE;
  vdt->gen_interior_pts         = FALSE;
  vdt->gen_near_surf_pts        = FALSE;
  vdt->pdebug                   = FALSE;
  vdt->maxdebug                 = FALSE;
#if GIFACE
  vdt->gdebug                   = FALSE;
  vdt->giface                   = FALSE;
#endif
  vdt->test_mode                = FALSE;
  vdt->keep_slivers             = FALSE;
  vdt->dump_tet_wrl             = FALSE;
  vdt->dump_tet_geom            = FALSE;
  vdt->dump_tet_egf             = FALSE;
  vdt->dump_bsf_egf             = FALSE;
  vdt->dump_t3d                 = FALSE;
  vdt->dump_data                = FALSE;
  reset_quality (vdt);
  vdt->check_interrupt.max_ctet = 1000;
  vdt->check_interrupt.ctet     = 0;
  vdt->check_interrupt.func     = NULL;
  vdt->report_opts              = FALSE;
  vdt->reset_locs               = TRUE;
  vdt->last_error_msg           = NULL;
  vdt->retri_to_desliver        = FALSE;
  vdt->sliver_ccl               = make_list ();
  vdt->inject_surf_proj         = FALSE;
  vdt->eval_mesh_size_from_pt   = FALSE;
  vdt->eval_mesh_size_func      = NULL;
  vdt->eval_mesh_size_func_data = NULL;
  vdt->orient_initial_af        = FALSE;

  return vdt;
}

void
vdt_set_last_error_msg (vdt_t *vdt, char *msg)
{
  if (vdt->last_error_msg) {
    NEWP_FREE_OBJ (vdt->last_error_msg);
  }
  if (msg) {
    NEWP_DUPL_STRING (vdt->last_error_msg, msg,
                      { errexit ("Not enough memory"); });
    
  }
}
  
void
vdt_perturb_loc (double shift_mult, double mesh_size, CkitPoint3 *loc)
{
  CkitPoint3 rv;
  double sm = shift_mult * mesh_size;
  rv.x = vdt_rand_0_to_1 (0) * sm;
  rv.y = vdt_rand_0_to_1 (1) * sm;
  rv.z = vdt_rand_0_to_1 (2) * sm;
  VECMAC_2_VEC ((*loc), +=, rv);
}


void
vdt_attach_pt (vdt_t *vdt, vdt_pt_t *pt)
{
  vdt->max_pt_id = max (vdt->max_pt_id, pt->id);
  dict_add_obj_handle_name (vdt->pointd, pt, pt->id);
  add_to_tail (vdt->pointl, pt);
#if VDT_PIRAT
  if (!vdt->pt_pirat) {
    vdt->pt_pirat = pirat_new_pirat (250, 1000 * DBL_EPSILON);
    CHECK_MEM (vdt->pt_pirat);
  }
  pirat_add_pt (vdt->pt_pirat, pt, &pt->loc);
#endif
#if VDT_BUCKETS
  if (!vdt->buckets) {
    if (vdt->bucket_size < vdt->def_mesh_size)
      vdt->bucket_size = 1.5 * vdt->def_mesh_size;
    vdt->buckets = vdt_bucket_new (&vdt->bbox, vdt->bucket_size);
  }
  vdt_bucket_add_pt (vdt->buckets, pt);
#endif
  vdt->new_pt    = pt;
}

#if VDT_USE_HIERLOBS
static int
bsf_close_by (vdt_t *vdt, G3D_box_t *box, CkitPoint3 *loc, double mind)
{
  BSFP bsf;
  G3D_plane_t plane;
  double d;
  lobs_3d_front_end lobs;

  switch (vdt->near_surf_test) {
  case VDT_NEAR_SURF_TEST_CIRCUMSPHERE:
    {
      lobs_3d_search (vdt->hierlobs,
                      loc->x, loc->y, loc->z, loc->x, loc->y, loc->z);
      lobs = lobs_3d_first_overlapping_object (vdt->hierlobs);
      while (lobs) {
        lobs_3d_search (lobs, loc->x, loc->y, loc->z, loc->x, loc->y, loc->z);
        if (lobs_3d_num_of_objects_found (lobs) > 0) return TRUE;
        lobs = lobs_3d_next_overlapping_object (vdt->hierlobs);
      }
    }
    break;
  case VDT_NEAR_SURF_TEST_DIST:
  default:
    {
      lobs_3d_search (vdt->hierlobs,
                      box->lower.x, box->lower.y, box->lower.z,
                      box->upper.x, box->upper.y, box->upper.z);
      lobs = lobs_3d_first_overlapping_object (vdt->hierlobs);
      while (lobs) {
        lobs_3d_search (lobs,
                        box->lower.x, box->lower.y, box->lower.z,
                        box->upper.x, box->upper.y, box->upper.z);
        bsf = lobs_3d_first_overlapping_object (lobs);
        while (bsf) {
          G3D_plane_consts_3 (&plane,
                              &bsf->pt[0]->loc,
                              &bsf->pt[1]->loc,
                              &bsf->pt[2]->loc);
          d = fabs (G3D_signed_dist_from_plane (&plane, loc));
          if (d < mind)
            return TRUE;
          bsf = lobs_3d_next_overlapping_object (lobs);
        }
        lobs = lobs_3d_next_overlapping_object (vdt->hierlobs);
      }
    }
    break;
  }
  return FALSE;
}
#else
static int
bsf_close_by (vdt_t *vdt, G3D_box_t *box, CkitPoint3 *loc, double mind)
{
  BSFP bsf;
  G3D_plane_t plane;
  double d;

  switch (vdt->near_surf_test) {
  case VDT_NEAR_SURF_TEST_CIRCUMSPHERE:
    {
      lobs_3d_search (vdt->bsflobs,
                      loc->x, loc->y, loc->z, loc->x, loc->y, loc->z);
      if (lobs_3d_num_of_objects_found (vdt->bsflobs) > 0) return TRUE;
    }
    break;
  case VDT_NEAR_SURF_TEST_DIST:
  default:
    {
      lobs_3d_search (vdt->bsflobs,
                      box->lower.x, box->lower.y, box->lower.z,
                      box->upper.x, box->upper.y, box->upper.z);
      
      bsf = lobs_3d_first_overlapping_object (vdt->bsflobs);
      while (bsf) {
        G3D_plane_consts_3 (&plane,
                            &bsf->pt[0]->loc,
                            &bsf->pt[1]->loc,
                            &bsf->pt[2]->loc);
        d = fabs (G3D_signed_dist_from_plane (&plane, loc));
        if (d < mind)
          return TRUE;
        bsf = lobs_3d_next_overlapping_object (vdt->bsflobs);
      }
    }
    break;
  }
  return FALSE;
}
#endif


static int 
any_pt_in_box (vdt_t *vdt, G3D_box_t *box)
{
#if VDT_PIRAT
  if (!pirat_search_range (vdt->pt_pirat, box, 1))
    return TRUE;

  return (vdt->pt_pirat->num_of_points_found > 0);
#endif
#if VDT_BUCKETS
  return vdt_bucket_search_box (vdt->buckets, box, 1) > 0;
#endif
}




int
vdt_gen_pt (void *data,
            CkitPoint3 *loc,
            double mesh_size,
            double mindist,
            double bdry_mindist)
{
  vdt_t *vdt = (vdt_t *)data;
  CkitPoint3 ploc = (*loc);
  PTP pt;

  vdt_perturb_loc (vdt->gen_pt_shift_mult, mesh_size, &ploc); 
  if (G3D_PT_INSIDE_BOX (vdt->gen_space, ploc)) {
    G3D_box_t box;
    G3D_INIT_BOX (box, ploc);
    G3D_INFLATE_BOX (box, mindist);
    if (!any_pt_in_box (vdt, &box)) {
      if (vdt->near_surf_test == VDT_NEAR_SURF_TEST_IGNORE ||
          (!bsf_close_by (vdt, &box, &ploc, bdry_mindist))) {
        vdt->max_pt_id = (void *)((char *)vdt->max_pt_id + 1);
        if ((pt = vdt_pt_new ((*vdt).max_pt_id, &ploc, mesh_size))
             != NULL) {
          vdt_attach_pt (vdt, pt);
          return VDT_PT_GEN_ACCEPTED;
        } else
          return VDT_PT_GEN_FAILED;
      } else {
        return VDT_PT_GEN_TOO_CLOSE_TO_BDRY;
      }
    } else {
      return VDT_PT_GEN_TOO_CLOSE_TO_ANOTHER;
    }
  } else {
    return VDT_PT_GEN_OUTSIDE_OF_GENSPACE;
  }
}

void
vdt_attach_tet (vdt_t *vdt, vdt_tet_t *tet) 
{
  char buf[128];
  vdt_tet_t *otet;
#define HAVE_PT(apt, atet)                        \
  ((apt == atet->pt[0]) ||                        \
   (apt == atet->pt[1]) ||                        \
   (apt == atet->pt[2]) ||                        \
   (apt == atet->pt[3]))
#define IDENTICAL_TETS(tet, otet)               \
  (HAVE_PT(tet->pt[0], otet) &&                 \
   HAVE_PT(tet->pt[1], otet) &&                 \
   HAVE_PT(tet->pt[2], otet) &&                 \
   HAVE_PT(tet->pt[3], otet))

  if (vdt->pdebug) {
    char buf[64];
    sprintf (buf, "Added tet pt = { %d %d %d %d }",
             (int) tet->pt[0]->id, (int) tet->pt[1]->id,
             (int) tet->pt[2]->id, (int) tet->pt[3]->id);
    logit (buf);
    if (tet->pt[0] == tet->pt[1] ||
        tet->pt[0] == tet->pt[2] ||
        tet->pt[0] == tet->pt[3] ||
        tet->pt[1] == tet->pt[2] ||
        tet->pt[1] == tet->pt[3] ||
        tet->pt[2] == tet->pt[3]) {
      sprintf (buf, "Invalid tet pt = { %d %d %d %d }",
               (int) tet->pt[0]->id, (int) tet->pt[1]->id,
               (int) tet->pt[2]->id, (int) tet->pt[3]->id);
      warn (buf);
    }
  }
#if GIFACE
  if (vdt->gdebug) {
    vdt_tet_generate_graphics (tet, vdt);
    vdt_refresh_display (vdt);
    ESIEventLoop (vdt->pdebug, "attached tet");
  }
#endif

  if (vdt->maxdebug) {
    otet = NULL;
    while ((otet = (TETP) get_list_next (vdt->tetl, otet))) {
      if (IDENTICAL_TETS (tet, otet)) {
        sprintf (buf, "Identical tets, pt = { %d %d %d %d }",
                 tet->pt[0]->id, tet->pt[1]->id,
                 tet->pt[2]->id, tet->pt[3]->id);
        warn (buf);
      }
    }
  }
    
  /* this is the only necessary action */
  add_to_tail (vdt->tetl, tet); vdt->ntets++;
}

static void
update_regd (vdt_t *vdt, void *reg_id[2])
{
  if (reg_id[0] != VDT_INVALID_REG_ID) 
    if (dict_find_obj_with_handle_name (vdt->regd, reg_id[0]) == NULL) {
      dict_add_obj_handle_name (vdt->regd, reg_id[0], reg_id[0]);
    }
  if (reg_id[1] != VDT_INVALID_REG_ID) 
    if (dict_find_obj_with_handle_name (vdt->regd, reg_id[1]) == NULL) {
      dict_add_obj_handle_name (vdt->regd, reg_id[1], reg_id[1]);
    }
}

void
vdt_attach_bsf (vdt_t *vdt, vdt_bsf_t *bsf)
{
  add_to_head (vdt->bsfl, bsf);
  update_regd (vdt, bsf->reg_id);
}

#if GIFACE

/*
   Generate graphics.
 */
void
vdt_generate_graphics (vdt_t *vdt)
{
  vdt_pt_t *pt;
  vdt_bsf_t *bsf;
  vdt_tet_t *tet;

  /* draw the input points */
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    vdt_pt_generate_graphics (pt, vdt);
  }

  /* draw the input boundary surface faces */
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    vdt_bsf_generate_graphics (bsf, vdt); 
  }
  
  /* draw the output tets  */
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    vdt_tet_generate_graphics (tet, vdt); 
  }
  
#if 0
  /* draw the octree */
  if (vdt->ot && vdt->gdebug)
    vdt_ot_generate_graphics (vdt->ot, vdt->gmodel);
#endif
}
#endif

/*
   Find active points.
   The box specified is searched for points.
 */
int /* number of found points */
vdt_pts_in_box (vdt_t *vdt, G3D_box_t *box, int max_found)
{
  vdt_pt_t *pt;
  int i, n;
  
  vdt->search_stat.num_searches++;
  
#if VDT_PIRAT
  if (!pirat_search_range (vdt->pt_pirat, box,
                           (max_found > 0 ? max_found : INT_MAX)))
    return 0;

  vdt->num_active_points = n = vdt->pt_pirat->num_of_points_found;
  if (vdt->num_active_points <= 0)
    return 0;

  if (vdt->num_active_points >= vdt->tot_active_points) {
    vdt->tot_active_points = vdt->num_active_points;
    vdt->active_loc = NEWP_REALLOC_OBJ_ARRAY (vdt->active_loc,
                                              vdt->tot_active_points,
                                              CkitPoint3);
    CHECK_MEM (vdt->active_loc);
    vdt->active_pt = NEWP_REALLOC_OBJ_ARRAY (vdt->active_pt,
                                             vdt->tot_active_points,
                                             vdt_pt_t *);
    CHECK_MEM (vdt->active_pt);
  }
  
  FOR (i, n) {
    pt = (vdt_pt_t *) vdt->pt_pirat->entities_found[i];
    vdt->active_pt[i]     = pt;
    vdt->active_loc[i]    = pt->loc;
  }
#endif
#if VDT_BUCKETS
  vdt->num_active_points
    = n = vdt_bucket_search_box (vdt->buckets, box,
                                 (max_found > 0 ? max_found : INT_MAX));
  if (n <= 0)
    return 0;

  if (vdt->num_active_points >= vdt->tot_active_points) {
    vdt->tot_active_points = vdt->num_active_points;
    vdt->active_loc = NEWP_REALLOC_OBJ_ARRAY (vdt->active_loc,
                                              vdt->tot_active_points,
                                              CkitPoint3);
    CHECK_MEM (vdt->active_loc);
    vdt->active_pt = NEWP_REALLOC_OBJ_ARRAY (vdt->active_pt,
                                             vdt->tot_active_points,
                                             vdt_pt_t *);
    CHECK_MEM (vdt->active_pt);
  }

  FOR (i, n) {
    pt = vdt_bucket_pt_at (vdt->buckets, i);
    vdt->active_pt[i]     = pt;
    vdt->active_loc[i]    = pt->loc;
  }
#endif

  vdt->search_stat.sum_found += vdt->num_active_points;
  vdt->search_stat.max_num_found =
    max (vdt->search_stat.max_num_found, vdt->num_active_points);

  return vdt->num_active_points;
}

vdt_lp_result_t
vdt_get_tet_from_lp (vdt_t *vdt,
                     CkitPoint3 *sample_point,
                     int ptid[4])
{
  int nnz, i, r, result;
  lprec *lp;
  double lambda;
  vdt_lp_result_t retval = VDT_INFEASIBLE;
  
  if (vdt->pdebug) {
    char buf[256];
    sprintf (buf, "LP with %d unknowns", vdt->num_active_points);
    logit (buf);
  }
  vdt->del_stat.lp_stat.sum_unknowns += vdt->num_active_points;
  vdt->del_stat.lp_stat.sum_evals++;
  vdt->del_stat.lp_stat.max_num_unknowns
    = max (vdt->del_stat.lp_stat.max_num_unknowns, vdt->num_active_points);
  
  lp  = vdt_lp_make (vdt->active_loc, vdt->num_active_points, sample_point);
  result = solve (lp);
  switch (result) {
  case OPTIMAL:
    nnz = 0;
    r   = lp->rows + 1;
    for (i = 0; i < vdt->num_active_points; i++) {
      lambda = (double) lp->best_solution[r + i];
      if (lambda != 0 && nnz < 4) {
        ptid[nnz] = i; nnz++;
      }
      if (nnz > 4) {
        retval = VDT_TOO_MANY_POINTS;
        goto exit;
      }
    }
    if (nnz == 4) {
      retval = VDT_OPTIMAL;
    } else {
      retval = VDT_LOWER_DIM_OPTIMAL;
    }
    break;
  case UNBOUNDED:
    retval = VDT_UNBOUNDED;
    break;
  case INFEASIBLE:
  default:
    retval = VDT_INFEASIBLE;
    break;
  }
  
 exit:
  delete_lp (lp);

  return retval;
}






static int
triangulate (vdt_t *vdt)
{
  int result;
  dict_iterator di;
  ckit_timer_t *timer = ckit_timer_new ();
  void *reg_id;

  ckit_timer_restart (timer);
  logit ("Processing facets");

  if (!vdt_af_prepare (vdt)) return FALSE;

  vdt_report_time (timer, "Processing of constraining facets");

  ckit_timer_restart (timer);
  logit ("Advancing front volume triangulation");
  
  /* Increment for progress reporting */
  vdt->incr_ntets = (vdt->num_input_pts + vdt->num_gen_pts) / 5000 * 1000;
  if      (vdt->incr_ntets > 50000) vdt->incr_ntets = 50000;
  else if (vdt->incr_ntets < 10000) vdt->incr_ntets = 10000;
  vdt->ntets = 0;
  
  di = dict_get_iterator ();
  dict_reset_iterator (di);
  while ((reg_id = dict_iter_objs (vdt->regd, di))) {
    vdt_af_initial_af (vdt, reg_id);
    result = vdt_af_run (vdt, reg_id);
  }
  
  vdt_report_time (timer, "Volume triangulation");

  ckit_timer_free (timer);
  
  return result;
}




static double
eval_mesh_size (void *data, CkitPoint3 *loc)
{
  vdt_t *vdt = (vdt_t *)data;

  if (vdt->eval_mesh_size_func) { /* mesh size function defined */
    return (vdt->eval_mesh_size_func) (vdt->eval_mesh_size_func_data, loc);
  } else { /* mesh size function not defined; use defaults */
    if (vdt->eval_mesh_size_from_pt) {
      double mssum;
      double wgtsum = 0;
      double wgt;
      G3D_box_t box;
      CkitPoint3 d;
      int i, n;
      PTP pt;
      
      G3D_INIT_BOX (box, (*loc));
      G3D_INFLATE_BOX (box, vdt->max_mesh_size);
      
      n = vdt_pts_in_box (vdt, &box, VDT_ALL_PTS);
      
      if (n >= 1) {
        wgtsum = 0;
        mssum  = 0;
        FOR (i, n) {
          pt = vdt->active_pt[i];
          if (pt->mesh_size > 0) {
            VECMAC_3_VEC (d, =, (*loc), -, pt->loc);
            wgt = VECMAC_DOT (d, d);
            if (wgt > 0) wgt = 1 / wgt;
            else         {
              return pt->mesh_size;
            }
            mssum += pt->mesh_size * wgt;
            wgtsum += wgt;
          }
        }
        return mssum / wgtsum;
      } else {
        return vdt->def_mesh_size;
      }
    } else {
      return vdt->def_mesh_size;
    }
  } /* if */
}

#if VDT_USE_HIERLOBS
static void
build_bsf_ss (vdt_t *vdt)
{
  vdt_hierlobs_fill (vdt);
}
#else
static void
build_bsf_ss (vdt_t *vdt)
{
  BSFP bsf;
  G3D_box_t box;
  
  vdt->bsflobs = lobs_3d_new_instance (NULL, 100); CHECK_MEM (vdt->bsflobs);
  
  bsf = NULL;
  while ((bsf = (BSFP) get_list_next (vdt->bsfl, bsf))) {
    G3D_INIT_BOX (box, bsf->pt[0]->loc);
    G3D_UPDT_BOX (box, bsf->pt[1]->loc);
    G3D_UPDT_BOX (box, bsf->pt[2]->loc);
    lobs_3d_add_object (vdt->bsflobs, bsf,
                        box.lower.x, box.lower.y, box.lower.z,
                        box.upper.x, box.upper.y, box.upper.z);
  }
}
#endif


static void
gen_near_surf_pts (vdt_t *vdt)
{
  BSFP bsf;
#define EMPTY_EQUATORIAL_SPHERE 0
#if EMPTY_EQUATORIAL_SPHERE
  double csradius, offset;
  CkitPoint3 n, cc, il;
#else
  double iel, ih;
  CkitPoint3 n, fc, il;
#endif
  
  bsf = NULL;
  while ((bsf = (BSFP) get_list_next (vdt->bsfl, bsf))) {
#if EMPTY_EQUATORIAL_SPHERE
    if (vdt_bsf_circumsphere (bsf, &cc)) {
      VECMAC_3_VEC (n, =, bsf->pt[0]->loc, -, cc);
      csradius = VECMAC_NORM (n);
      offset   = 1.1 * csradius;
      vdt_bsf_get_normal (bsf, &n);
      if (bsf->reg_id[1] != VDT_INVALID_REG_ID) {
        VECMAC_3_VEC (il, =, cc, +, offset * n);
        vdt_gen_pt (vdt, &il, csradius, 0.5 * csradius, 0.5 * csradius);
      }
      if (bsf->reg_id[0] != VDT_INVALID_REG_ID) {
        VECMAC_3_VEC (il, =, cc, -, offset * n);
        vdt_gen_pt (vdt, &il, csradius, 0.5 * csradius, 0.5 * csradius);
      }
    }
#else
    iel = vdt_bsf_ideal_edge_length (bsf);
    ih  = sqrt (2.0/3.0) * iel;
    vdt_bsf_get_normal (bsf, &n);
    VECMAC_SCALAR (n,  *=, ih);
    VECMAC_3_VEC (fc,   =, bsf->pt[0]->loc, +, bsf->pt[1]->loc);
    VECMAC_2_VEC (fc,  +=, bsf->pt[2]->loc);
    VECMAC_SCALAR (fc, *=, 0.3333333333);
    if (bsf->reg_id[1] != VDT_INVALID_REG_ID) {
      VECMAC_3_VEC (il, =, fc, +, n);
      vdt_gen_pt (vdt, &il, iel, 0.5 * iel, 0.5 * iel);
    }
    if (bsf->reg_id[0] != VDT_INVALID_REG_ID) {
      VECMAC_3_VEC (il, =, fc, -, n);
      vdt_gen_pt (vdt, &il, iel, 0.5 * iel, 0.5 * iel);
    }
#endif
  }
}


#if VDT_USE_HIERLOBS
static void
insert_near_surf_pts (vdt_t *vdt, LIST nearsurfpts)
{
  BSFP bsf;
  lobs_3d_front_end lobs;
  BSFP projbsf;
  G3D_plane_t plane;
  G3D_tri_t tri;
  G3D_box_t box;
  PTP pt;
  G3D_loc_t xs, ab, proj;
  double lambda;
  double mindist, d;

  pt = NULL;
  while ((pt = (PTP) get_list_next (nearsurfpts, pt))) {
    G3D_INIT_BOX (box, pt->loc);
    G3D_INFLATE_BOX (box, pt->mesh_size);
    mindist = FLT_MAX;
    projbsf = NULL;
    lobs_3d_search (vdt->hierlobs,
                    box.lower.x, box.lower.y, box.lower.z,
                    box.upper.x, box.upper.y, box.upper.z);
    lobs = lobs_3d_first_overlapping_object (vdt->hierlobs);
    while (lobs) {
      lobs_3d_search (lobs,
                      box.lower.x, box.lower.y, box.lower.z,
                      box.upper.x, box.upper.y, box.upper.z);
      bsf = lobs_3d_first_overlapping_object (lobs);
      while (bsf) {
        G3D_plane_consts_3 (&plane,
                            &bsf->pt[0]->loc,
                            &bsf->pt[1]->loc,
                            &bsf->pt[2]->loc);
        tri.loc[0] = bsf->pt[0]->loc;
        tri.loc[1] = bsf->pt[1]->loc;
        tri.loc[2] = bsf->pt[2]->loc;
        VECMAC_3_VEC (ab, =, pt->loc, +, plane.n);
        if (G3D_intersect_tri_and_line (&xs, &lambda, 
                                        &pt->loc, &ab, &tri, 0.001)) {
          d = fabs (G3D_signed_dist_from_plane (&plane, &pt->loc));
          if (d < mindist) {
            proj    = xs;
            projbsf = bsf;
            mindist = d;
          }
        }
        bsf = lobs_3d_next_overlapping_object (lobs);
      }
      lobs = lobs_3d_next_overlapping_object (vdt->hierlobs);
    }
    if (projbsf) {
      vdt_bsf_inject_pt (projbsf, &proj, vdt);
    }
  } /* while */
}
#else
static void
insert_near_surf_pts (vdt_t *vdt, LIST nearsurfpts)
{
  BSFP bsf;
  BSFP projbsf;
  G3D_plane_t plane;
  G3D_tri_t tri;
  G3D_box_t box;
  PTP pt;
  G3D_loc_t xs, ab, proj;
  double lambda;
  double mindist, d;

  pt = NULL;
  while ((pt = (PTP) get_list_next (nearsurfpts, pt))) {
    G3D_INIT_BOX (box, pt->loc);
    G3D_INFLATE_BOX (box, pt->mesh_size);
    lobs_3d_search (vdt->bsflobs,
                    box.lower.x, box.lower.y, box.lower.z,
                    box.upper.x, box.upper.y, box.upper.z);
    mindist = FLT_MAX;
    projbsf = NULL;
    bsf = lobs_3d_first_overlapping_object (vdt->bsflobs);
    while (bsf) {
      G3D_plane_consts_3 (&plane,
                          &bsf->pt[0]->loc, &bsf->pt[1]->loc, &bsf->pt[2]->loc);
      tri.loc[0] = bsf->pt[0]->loc;
      tri.loc[1] = bsf->pt[1]->loc;
      tri.loc[2] = bsf->pt[2]->loc;
      VECMAC_3_VEC (ab, =, pt->loc, +, plane.n);
      if (G3D_intersect_tri_and_line (&xs, &lambda, 
                                      &pt->loc, &ab, &tri, 0.001)) {
        d = fabs (G3D_signed_dist_from_plane (&plane, &pt->loc));
        if (d < mindist) {
          proj    = xs;
          projbsf = bsf;
          mindist = d;
        }
      }
      bsf = lobs_3d_next_overlapping_object (vdt->bsflobs);
    }
    if (projbsf) {
      vdt_bsf_inject_pt (projbsf, &proj, vdt);
    }
  } /* while */
}
#endif

static void
insert_pts_with_check (vdt_t *vdt)
{
  PTP pt;
  int result;
  LIST nearsurfpts = make_list ();
  
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pt_to_insert_as_if_gen, pt))) {
    result = vdt_gen_pt (vdt, &pt->loc, pt->mesh_size,
                         0.25 * pt->mesh_size, 0.1 * pt->mesh_size);
    switch (result) {
    case VDT_PT_GEN_FAILED:
      errexit ("Failed in vdt_gen_pt");
      break;
    case VDT_PT_GEN_OUTSIDE_OF_GENSPACE:
      errexit ("Outside of genspace?");
      break;
    case VDT_PT_GEN_TOO_CLOSE_TO_BDRY:
      add_to_tail (nearsurfpts, pt);
      break;
    case VDT_PT_GEN_TOO_CLOSE_TO_ANOTHER:
    case VDT_PT_GEN_ACCEPTED:
    default:
      /* VOID */
      break;
    }
  }
  if (list_length (nearsurfpts) > 0) {
    insert_near_surf_pts (vdt, nearsurfpts);
  }
  free_list (nearsurfpts, NOT_NODES);
}

#if VDT_USE_HIERLOBS
static int
closest_proj (PTP pt, BSFP *projbsf, CkitPoint3 *proj,
              double critdist, vdt_t *vdt)
{
  lobs_3d_front_end lobs;
  BSFP bsf;
  G3D_plane_t plane;
  G3D_tri_t tri;
  G3D_box_t box;
  G3D_loc_t xs, ab;
  CkitPoint3 v;
  double lambda;
  double mindist2, d2;
  
  G3D_INIT_BOX (box, pt->loc);
  G3D_INFLATE_BOX (box, critdist);
  
  mindist2 = critdist * critdist;
  *projbsf = NULL;
  lobs_3d_search (vdt->hierlobs,
                  box.lower.x, box.lower.y, box.lower.z,
                  box.upper.x, box.upper.y, box.upper.z);
  lobs = lobs_3d_first_overlapping_object (vdt->hierlobs);
  while (lobs) {
    lobs_3d_search (lobs,
                    box.lower.x, box.lower.y, box.lower.z,
                    box.upper.x, box.upper.y, box.upper.z);
    bsf = lobs_3d_first_overlapping_object (lobs);
    while (bsf) {
      /* Do not consider facets referencing pt */
      if (pt != bsf->pt[0] && pt != bsf->pt[1] && pt != bsf->pt[2]) {
        G3D_plane_consts_3 (&plane,
                            &bsf->pt[0]->loc,
                            &bsf->pt[1]->loc,
                            &bsf->pt[2]->loc);
        tri.loc[0] = bsf->pt[0]->loc;
        tri.loc[1] = bsf->pt[1]->loc;
        tri.loc[2] = bsf->pt[2]->loc;
        VECMAC_3_VEC (ab, =, pt->loc, +, plane.n);
        if (G3D_intersect_tri_and_line (&xs, &lambda, 
                                        &pt->loc, &ab, &tri, 0.001)) {
          VECMAC_3_VEC (v, =, xs, -, pt->loc);
          d2 = VECMAC_DOT (v, v);
          if (d2 < mindist2) {
            *proj     = xs;
            *projbsf  = bsf;
            mindist2  = d2;
          }
        }
      }
      bsf = lobs_3d_next_overlapping_object (lobs);
    }
    lobs = lobs_3d_next_overlapping_object (vdt->hierlobs);
  }

  return (*projbsf != NULL);
}
#else
static int
closest_proj (PTP pt, BSFP *projbsf, CkitPoint3 *proj,
              double critdist, vdt_t *vdt)
{
  BSFP bsf;
  G3D_plane_t plane;
  G3D_tri_t tri;
  G3D_box_t box;
  G3D_loc_t xs, ab;
  CkitPoint3 v;
  double lambda;
  double mindist2, d2;
  
  G3D_INIT_BOX (box, pt->loc);
  G3D_INFLATE_BOX (box, critdist);
  
  *projbsf = NULL;
  lobs_3d_search (vdt->bsflobs,
                  box.lower.x, box.lower.y, box.lower.z,
                  box.upper.x, box.upper.y, box.upper.z);

  mindist2  = critdist * critdist;
  *projbsf = NULL;
  bsf = lobs_3d_first_overlapping_object (vdt->bsflobs);
  while (bsf) {
    /* Do not consider facets referencing pt */
    if (pt != bsf->pt[0] && pt != bsf->pt[1] && pt != bsf->pt[2]) {
      G3D_plane_consts_3 (&plane,
                          &bsf->pt[0]->loc,
                          &bsf->pt[1]->loc,
                          &bsf->pt[2]->loc);
      tri.loc[0] = bsf->pt[0]->loc;
      tri.loc[1] = bsf->pt[1]->loc;
      tri.loc[2] = bsf->pt[2]->loc;
      VECMAC_3_VEC (ab, =, pt->loc, +, plane.n);
      if (G3D_intersect_tri_and_line (&xs, &lambda, 
                                      &pt->loc, &ab, &tri, 0.001)) {
        VECMAC_3_VEC (v, =, xs, -, pt->loc);
        d2 = VECMAC_DOT (v, v);
        if (d2 < mindist2) {
          *proj     = xs;
          *projbsf  = bsf;
          mindist2  = d2;
        }
      }
    }
    bsf = lobs_3d_next_overlapping_object (vdt->bsflobs);
  }

  return (*projbsf != NULL);
}
#endif

static int
no_pts_in_box (vdt_t *vdt, double boxsize, PTP self)
{
  int i, n;
  G3D_box_t box;
  G3D_INIT_BOX (box, self->loc);
  G3D_INFLATE_BOX (box, boxsize);
  n = vdt_pts_in_box (vdt, &box, 2);
  FOR (i, n) {
    if (vdt->active_pt[i] != self) return FALSE;
  }
  return TRUE;
}

static void
inject_surf_proj (vdt_t *vdt)
{
  PTP pt;
  BSFP projbsf;
  CkitPoint3 proj;
  double critdist;
  

  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    critdist = 0.1 * pt->mesh_size;
    if (closest_proj (pt, &projbsf, &proj, critdist, vdt)) {
      if (no_pts_in_box (vdt, critdist, pt)) {
        vdt_bsf_inject_pt (projbsf, &proj, vdt);
        pt = (PTP) get_list_prev (vdt->pointl, pt);
      }
    }
  }
}

static void
gen_pts (vdt_t *vdt)
{
  CkitPoint3 center;
  double minosize, size;
  double cshift;

  if (!(
        (list_length (vdt->pt_to_insert_as_if_gen) > 0) ||
        vdt->gen_near_surf_pts ||
        vdt->gen_interior_pts  ||
        vdt->inject_surf_proj))
    return; /* nothing to do here */

  /* Set up point-generation space */
  vdt->gen_space = vdt->bbox;
  G3D_INFLATE_BOX (vdt->gen_space, 0.1 * vdt->min_mesh_size);

  /* Build structure for boundary facet proximity searching */
  if (vdt->near_surf_test != VDT_NEAR_SURF_TEST_IGNORE) 
    build_bsf_ss (vdt);

  
  /* Inject projections of points close to surfaces */
  if (vdt->inject_surf_proj)
    inject_surf_proj (vdt);

  /* Generate points input by user for which proximity check
     was requested */
  if (list_length (vdt->pt_to_insert_as_if_gen) > 0)
    insert_pts_with_check (vdt);
  
  /* Generate points by offsets from boundary facets */
  if (vdt->gen_near_surf_pts)
    gen_near_surf_pts (vdt); 
  
  /* Generate points by octree insertion */
  if (vdt->gen_interior_pts) { 
    /* compute octree center */
    VECMAC_3_VEC (center, =,
                  0.5 * vdt->bbox.upper, + , 0.5 * vdt->bbox.lower);
    cshift = (((int) (vdt->max_mesh_size / vdt->min_mesh_size) + 1)
              * vdt->min_mesh_size);
    VECMAC_SCALAR (center, -=, cshift); /* shift to have one cell outside */
    /* compute length of octree side */
    minosize = G3D_BOX_RANGE (vdt->bbox, x);
    minosize = max (minosize, G3D_BOX_RANGE (vdt->bbox, y));
    minosize = max (minosize, G3D_BOX_RANGE (vdt->bbox, z));
    minosize += 2 * cshift;
    size = vdt->min_mesh_size;
    while (size < minosize) size += minosize;
    /* build octree */
    vdt->ot = vdt_ot_new (&center, size);
    vdt_ot_refine (vdt->ot, eval_mesh_size, vdt);
    /* generate points */
    vdt_ot_gen_pts (vdt->ot, vdt_gen_pt, vdt);
  }

  /* Clean up */
#if VDT_USE_HIERLOBS
  if (vdt->hierlobs)
    vdt_hierlobs_clean (vdt); 
#else
  if (vdt->bsflobs)
    lobs_3d_free_instance (vdt->bsflobs);
#endif
}


static int
preprocess (vdt_t *vdt)
{
  ckit_timer_t *timer = ckit_timer_new ();

  vdt->num_input_pts   = list_length (vdt->pointl);

  logit ("Preprocessing");

  gen_pts (vdt);

  vdt->num_gen_pts     = list_length (vdt->pointl) - vdt->num_input_pts;
  vdt->num_input_bsfs  = list_length (vdt->bsfl);
  vdt->input_max_pt_id = vdt->max_pt_id;
  
  vdt_report_time (timer, "Preprocessing");

  ckit_timer_free (timer);
  return TRUE;
}


static void
reset_locs (vdt_t *vdt)
{
  PTP pt;
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    pt->loc = pt->origloc;
  }
}



static void
invalidate_tet (TETP tet, vdt_t *vdt)
{
  int i;
  tet->reg_id = VDT_INVALID_REG_ID;
  FOR (i, 4) { tet->pt[i]->num_conn_tets--; }
#if GIFACE
  vdt_tet_generate_graphics (tet, vdt); 
#endif
}


static void
dump_tet_wrl (vdt_t *vdt)
{
  vdt_tet_t *tet;

  printf ("#VRML V1.0 ascii\n");
  /* draw the output tets  */
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) {
#if 0
      printf ("# %d    %d %d %d %d\n", ++n,
              (int) tet->pt[0]->id, (int) tet->pt[1]->id,
              (int) tet->pt[2]->id, (int) tet->pt[3]->id
              );
#endif
      printf ("Coordinate3{point[%g %g %g,%g %g %g,%g %g %g,%g %g %g]}\n",
              tet->pt[0]->origloc.x,
              tet->pt[0]->origloc.y,
              tet->pt[0]->origloc.z,
              tet->pt[1]->origloc.x,
              tet->pt[1]->origloc.y,
              tet->pt[1]->origloc.z,
              tet->pt[2]->origloc.x,
              tet->pt[2]->origloc.y,
              tet->pt[2]->origloc.z,
              tet->pt[3]->origloc.x,
              tet->pt[3]->origloc.y,
              tet->pt[3]->origloc.z);
      printf ("IndexedFaceSet{coordIndex[0,2,1,-1,0,1,3,-1,1,2,3,-1,2,0,3]}\n");
    }
  }
}

int
vdt_count_tets (vdt_t *vdt)
{
  TETP tet;
  int count = 0;
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) count++;
  }
  return count;
}

int
vdt_count_points (vdt_t *vdt, int *connected, int *unconnected)
{
  PTP pt;
  int count = 0;
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    if (pt->num_conn_tets > 0) count++;
  }
  *connected   = count;
  *unconnected = list_length (vdt->pointl) - (*connected);
  return TRUE;
}

#define ELIXIR_FILE_HEADER  "# ELIXIR file (C) 1994 Petr Krysl; version "
#define ELIXIR_FILE_VERSION 1
#ifndef EG_TETRAHEDRON
#define EG_TETRAHEDRON 12
#endif
#ifndef EG_TRIANGLE3D
#define EG_TRIANGLE3D 7
#endif

static void
egf_graphic_header (int timeid, int serial, int type, int layer)
{
  printf ("%lu %lu %d %d %d %d\n", timeid, serial, type, layer, TRUE, FALSE);
}

static void
dump_tet_egf (vdt_t *vdt)
{
  vdt_tet_t *tet;

  printf (ELIXIR_FILE_HEADER "%d\n", ELIXIR_FILE_VERSION);

  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) {
      egf_graphic_header (0, (int) tet, EG_TETRAHEDRON, VDT_TET_LAYER);
      printf (
              "%f %f %f %f %f %f %f %f %f %f %f %f\n%d %d %d %f\n%s\n%s\n",
              tet->pt[0]->origloc.x, tet->pt[0]->origloc.y, tet->pt[0]->origloc.z, 
              tet->pt[1]->origloc.x, tet->pt[1]->origloc.y, tet->pt[1]->origloc.z,
              tet->pt[2]->origloc.x, tet->pt[2]->origloc.y, tet->pt[2]->origloc.z,
              tet->pt[3]->origloc.x, tet->pt[3]->origloc.y, tet->pt[3]->origloc.z,
              TRUE, 0, TRUE, 1.0,
              vdt_color_name_from_id (tet->reg_id),
              "black");
    }
  }
}


static void
dump_bsf_egf (vdt_t *vdt)
{
  BSFP bsf;

  printf (ELIXIR_FILE_HEADER "%d\n", ELIXIR_FILE_VERSION);

  bsf = NULL;
  while ((bsf = (BSFP) get_list_next (vdt->bsfl, bsf))) {
    egf_graphic_header (0, (int)bsf, EG_TRIANGLE3D, VDT_DELAUNAY_BSF_LAYER);
    printf ("%f %f %f %f %f %f %f %f %f\n%d %d %d %f\n%s\n%s\n",
            bsf->pt[0]->origloc.x, bsf->pt[0]->origloc.y, bsf->pt[0]->origloc.z, 
            bsf->pt[1]->origloc.x, bsf->pt[1]->origloc.y, bsf->pt[1]->origloc.z, 
            bsf->pt[2]->origloc.x, bsf->pt[2]->origloc.y, bsf->pt[2]->origloc.z, 
            TRUE, 0, TRUE, 1.0,
            vdt_color_name_from_id (bsf->surf_id),
            "black");
  }
}


static void
dump_pt (vdt_t *vdt)
{
  PTP pt;

  printf ("! points\n");
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    if (pt->num_conn_tets > 0) {
      printf ("pt %d    %g %g %g   %g\n",
              (int) pt->id,
              pt->loc.x, pt->loc.y, pt->loc.z, pt->mesh_size);
    }
  }
}


#define TET_HAS_PT(tet, apt)                     \
  (   ((tet)->pt[0] == (apt))                        \
   || ((tet)->pt[1] == (apt))                        \
   || ((tet)->pt[2] == (apt))                        \
   || ((tet)->pt[3] == (apt)))
#define TET_HAS_3_PTS(tet, pt0, pt1, pt2) \
  (TET_HAS_PT (tet, pt0) && TET_HAS_PT (tet, pt1) && TET_HAS_PT (tet, pt2))


static void
dump_bsf (vdt_t *vdt)
{
  BSFP bsf;

  printf ("! bsfs\n");
  bsf = NULL;
  while ((bsf = (BSFP) get_list_next (vdt->bsfl, bsf))) {
    printf ("bsf %d %d %d   %d   %d %d ! %d %d\n", 
            (int) bsf->pt[0]->id,
            (int) bsf->pt[1]->id,
            (int) bsf->pt[2]->id,
            (int) bsf->surf_id,
            (int) bsf->reg_id[0],
            (int) bsf->reg_id[1],
            (bsf->tet[0] ? (int) bsf->tet[0] : 0),
            (bsf->tet[1] ? (int) bsf->tet[1] : 0)
            );
    if (bsf->tet[0] && bsf->tet[0] != &vdt->stop_tet) 
      if (bsf->tet[0]->reg_id == VDT_INVALID_REG_ID)
        errexit ("Adjacent tet has been deleted");
    if (bsf->tet[1] && bsf->tet[1] != &vdt->stop_tet)
      if (bsf->tet[1]->reg_id == VDT_INVALID_REG_ID)
        errexit ("Adjacent tet has been deleted");
  }
}


static void
dump_tet (vdt_t *vdt)
{
  vdt_tet_t *tet;

  printf ("! tetrahedra\n");
  /* draw the output tets  */
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) {
      printf ("%d    %d %d %d %d   %d\n", (int) tet,
              (int) tet->pt[0]->id, (int) tet->pt[1]->id,
              (int) tet->pt[2]->id, (int) tet->pt[3]->id,
              (int) tet->reg_id
              );
    }
  }
}

static void
dump_data (vdt_t *vdt)
{
  dump_pt (vdt);
  dump_bsf (vdt);
  dump_tet (vdt);
}


static void
dump_t3d_pt (vdt_t *vdt)
{
  PTP pt;

  printf ("! points\n");
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    if (pt->num_conn_tets > 0)
      printf ("%d %g %g %g\n",
              (int) pt->id, pt->loc.x, pt->loc.y, pt->loc.z);
  }
}


static void
dump_t3d_tet (vdt_t *vdt)
{
  vdt_tet_t *tet;
  long int n;

  /* draw the output tets  */
  printf ("! tetrahedra\n");
  n   = 0;
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) {
      printf ("%ld %d %d %d %d 4 %d\n", ++n,
              (int) tet->pt[0]->id, (int) tet->pt[1]->id,
              (int) tet->pt[2]->id, (int) tet->pt[3]->id,
              (int) tet->reg_id
              );
    }
  }
}

static void
dump_t3d (vdt_t *vdt)
{
  int connected, unconnected;

  vdt_count_points (vdt, &connected, &unconnected);
  printf ("! VDT (C) 1998--2000 Petr Krysl\n");
  printf ("%d %d %d %d\n",
          connected, 0, 0, vdt_count_tets (vdt));
  printf ("3 1\n");
  dump_t3d_pt (vdt);
  dump_t3d_tet (vdt);
}



#if GIFACE
static int
refresh_view (NODE ptr, NODE elem)
{
  EVFastRedraw ((EView *)elem);
  return (int) ptr;
}
#endif

#if GIFACE
static void
refresh_views (EModel *model)
{
  EMDispatchToDependentViews (model, refresh_view, NULL);
}
#endif

#if GIFACE
int
vdt_refresh_display (vdt_t *vdt)
{
  if (vdt->gmodel)
    refresh_views (vdt->gmodel);
  return TRUE;
}
#endif

    

void
vdt_summary (vdt_t *vdt)
{
  char buf[1024];
  int i;
  int connected, unconnected;
  int ntets = vdt_count_tets (vdt);

  vdt_count_points (vdt, &connected, &unconnected);
  sprintf (buf, "Input: %d points, %d boundary facets.  Generated: %d points",
           vdt->num_input_pts, vdt->num_input_bsfs, vdt->num_gen_pts);
  logit (buf);
  sprintf (buf, "Output: %d points (%d added), %d boundary facets (%d added), %d tets",
           connected, connected + unconnected - vdt->num_input_pts - vdt->num_gen_pts,
           list_length (vdt->bsfl), list_length (vdt->bsfl) - vdt->num_input_bsfs,
           ntets);
  logit (buf);

  sprintf (buf, "Tet quality summary: ");
  logit (buf);
  sprintf (buf, "   Measure %s: min = %1.3f, max = %1.3f, rms = %1.3f, geom = %1.3f",
           (vdt->quality.quality_measure == VDT_GAMMA_QUALITY
            ? "gamma"
            : (vdt->quality.quality_measure == VDT_RADIUS_QUALITY
               ? "3*ir/cr"
               : "unknown")),
           vdt->quality.min_quality, vdt->quality.max_quality,
           vdt->quality.rms_quality, vdt->quality.geom_mean_quality);
  logit (buf);
  FOR (i, vdt->quality.nranges) {
    sprintf (buf, "   Range [%3.2f;%3.2f]: %9d",
             i*(1.0/(vdt->quality.nranges)),
             (i+1)*(1.0/(vdt->quality.nranges)),
             vdt->quality.quality_summary[i]);
    logit (buf);
  }
  sprintf (buf, "   Volume: min = %g, max = %g",
           vdt->quality.min_volume, vdt->quality.max_volume);
  logit (buf);
  if (vdt->quality.num_non_positive_vol_tets > 0) {
    sprintf (buf, "WARNING: %d non-positive volume tets found",
             vdt->quality.num_non_positive_vol_tets);
    logit (buf);
  }
  if (vdt->quality.num_small_dihed_angle_tets > 0) {
    sprintf (buf, "WARNING: %d small dihedral angle tets found",
             vdt->quality.num_small_dihed_angle_tets);
    logit (buf);
  }
  sprintf (buf, "   Dihedral angle: min = %g, max = %g, avg = %g",
           vdt->quality.min_dihed_angle * 180 / M_PI,
           vdt->quality.max_dihed_angle * 180 / M_PI,
           vdt->quality.avg_dihed_angle * 180 / M_PI);
  logit (buf);
  FOR (i, vdt->quality.nranges) {
    sprintf (buf, "   Range [%5.1f;%5.1f]: %9d",
             i*(180.0 / (vdt->quality.nranges)),
             (i+1)*(180.0 / (vdt->quality.nranges)),
             vdt->quality.angle_quality_summary[i]);
    logit (buf);
  }
}


vdt_pt_t *
vdt_find_pt (vdt_t *vdt, void *id)
{
  if (!vdt->pointd) return NULL;
  return (vdt_pt_t *) dict_find_obj_with_handle_name (vdt->pointd, id);
}


#if GIFACE
void
vdt_recolor_active_ff (vdt_t *vdt, vdt_ff_t *ff)
{
  vdt_ff_recolor_active (ff);
  vdt_refresh_display (vdt);
  ESIEventLoop (vdt->pdebug, "Activated new face");
}
#endif


void
vdt_compress_active_set (vdt_t *vdt)
{
  int i, n = vdt->num_active_points - 1;
  int nap;

  nap = i = 0;
  while (i < n) {
    if (!vdt->active_pt[i]) {
      vdt->active_pt[i]   = vdt->active_pt[i+1];
      vdt->active_loc[i]  = vdt->active_loc[i+1];
      vdt->active_pt[i+1] = NULL;
    } else {
      i++;
      nap++;
    }
  }
  vdt->num_active_points = nap;
}


int
vdt_2pt_circsph_pt_in (PTP p0, PTP p1,
                       CkitPoint3 *csc, double csr,
                       vdt_t *vdt, int closest, PTP *foundpt)
{
  CkitPoint3 d;
  double csr2 = csr * csr, bs, prevd2 = FLT_MAX, d2;
  G3D_box_t box;
  PTP pt;
  int i;
  PTP bpt = NULL;

  G3D_INIT_BOX (box, (*csc));
  bs = 1.001 * sqrt (csr2);
  G3D_INFLATE_BOX (box, bs);

  if (vdt_pts_in_box (vdt, &box, VDT_ALL_PTS) < 2)
    errexit ("suspicious number of points");

  FOR (i, vdt->num_active_points) {
    pt = vdt->active_pt[i];
    if (pt != p0 && pt != p1) {
      VECMAC_3_VEC (d, =, pt->loc, -, (*csc));
      d2 = VECMAC_DOT (d, d);
      if (d2 < csr2 && d2 < prevd2) {
        bpt = pt; prevd2 = d2;
        if (!closest) break; /* if any pt will do, return immediately */
      }
    }
  }

  *foundpt = bpt;
  return TRUE;
}


int
vdt_3pt_circsph_pt_in (PTP p0, PTP p1, PTP p2,
                       CkitPoint3 *csc, double csr,
                       vdt_t *vdt, int closest, PTP *foundpt)
{
  CkitPoint3 d;
  double csr2 = csr * csr, bs, prevd2 = FLT_MAX, d2;
  G3D_box_t box;
  PTP pt;
  int i;
  PTP bpt = NULL;

  G3D_INIT_BOX (box, (*csc));
  bs = 1.001 * sqrt (csr2);
  G3D_INFLATE_BOX (box, bs);

  if (vdt_pts_in_box (vdt, &box, VDT_ALL_PTS) < 3)
    errexit ("suspicious number of points");

  FOR (i, vdt->num_active_points) {
    pt = vdt->active_pt[i];
    if (pt != p0 && pt != p1 && pt != p2) {
      VECMAC_3_VEC (d, =, pt->loc, -, (*csc));
      d2 = VECMAC_DOT (d, d);
      if (d2 < csr2 && d2 < prevd2) {
        bpt = pt; prevd2 = d2;
        if (!closest) break; /* if any pt will do, return immediately */
      }
    }
  }

  *foundpt = bpt;
  return TRUE;
}


static int
pt_in_wrapped_box (PTP p0, PTP p1, PTP p2, PTP p3,
                   CkitPoint3 *csc, double csr,
                   double bsize, vdt_t *vdt)
{
  int i, n;
  CkitPoint3 d;
  G3D_box_t box;
  PTP pt;
  double d2, csr2 = csr * csr;

  {
    CkitPoint3 c;
    VECMAC_3_VEC (c, =, p0->loc, +, p1->loc);
    VECMAC_2_VEC (c, +=, p2->loc);
    VECMAC_2_VEC (c, +=, p3->loc);
    G3D_INIT_BOX (box, c);
    G3D_INFLATE_BOX (box, bsize);
  }
  if ((n = vdt_pts_in_box (vdt, &box, VDT_ALL_PTS)) < 4) {
    return FALSE;
  }

  n = vdt->num_active_points;
  FOR (i, n) {
    pt = vdt->active_pt[i];
    if (pt != p0 && pt != p1 && pt != p2 && pt != p3) {
      VECMAC_3_VEC (d, =, (*csc), -, pt->loc);
      d2 = VECMAC_DOT (d, d);
      if (d2 < csr2) {
        return TRUE; 
      }
    }
  }
  return FALSE;
}

int
vdt_4pt_circsph_pt_in (PTP p0, PTP p1, PTP p2, PTP p3,
                       CkitPoint3 *csc, double csr,
                       vdt_t *vdt, int max_found,
                       int limit,
                       PTP *apt)
{
  CkitPoint3 c = *csc, d;
  double csr2 = csr * csr, bs, d2;
  G3D_box_t box;
  PTP pt;
  int i, n;
  int restricted = (max_found < VDT_ALL_PTS);
  int prevn = 0;
  int outcome = TRUE;

  {/* Optimization for the case of removed circumcenter with big search box */
    double lel;
    PTP tpt[4];
    tpt[0] = p0; tpt[1] = p1; tpt[2] = p2; tpt[3] = p3;
    lel = vdt_tet_longest_edge_len (tpt);
    if (csr > lel) {
      if (pt_in_wrapped_box (p0, p1, p2, p3, csc, csr, lel, vdt)) {
        outcome = TRUE;
        goto exit;
      }
    }
  }
  
  *apt = NULL;

  G3D_INIT_BOX (box, c);
  bs = 1.001 * sqrt (csr2);
  G3D_INFLATE_BOX (box, bs);

 retry:
  
  if ((n = vdt_pts_in_box (vdt, &box, max_found)) < 4) {
    warn ("Not enough points");
    return FALSE;
  }

  n = vdt->num_active_points;
  FOR (i, n) {
    pt = vdt->active_pt[i];
    if (pt != p0 && pt != p1 && pt != p2 && pt != p3) {
      VECMAC_3_VEC (d, =, c, -, pt->loc);
      d2 = VECMAC_DOT (d, d);
      if (d2 < csr2) {
        *apt = pt;
        outcome = TRUE;
        goto exit;
      }
    }
  }

  /* If the number of found objects was restricted */
  if (restricted) {
    if (prevn == n) { outcome = FALSE; goto exit; }
    if (n > max_found && limit < max_found) { outcome = TRUE; goto exit; }
    prevn      = n;
    max_found *= 2;
    goto retry;
  }

  exit:
  return outcome;
}



int
vdt_3_pt_circumsphere (PTP p0, PTP p1, PTP p2, CkitPoint3 *cscenter)
{
  return G3D_tri_circumsphere (&p0->loc, &p1->loc, &p2->loc, cscenter);
}

int
vdt_4_pt_circumsphere (PTP p0, PTP p1, PTP p2, PTP p3, CkitPoint3 *center)
{
  return G3D_tet_circumsphere (&p0->loc, &p1->loc, &p2->loc, &p3->loc, center);
}

int
vdt_4_pt_insphere (PTP p0, PTP p1, PTP p2, PTP p3, CkitPoint3 *center)
{
  return G3D_tet_insphere (&p0->loc, &p1->loc, &p2->loc, &p3->loc, center);
}


int
vdt_4_loc_circumsphere (CkitPoint3 points[4], CkitPoint3 *center)
{
  return G3D_tet_circumsphere (&points[0], &points[1], &points[2], &points[3],
                               center);
}


int
vdt_4_loc_insphere (CkitPoint3 points[4], CkitPoint3 *center)
{
  return G3D_tet_insphere (&points[0], &points[1], &points[2], &points[3], center);
}



#if GIFACE
static void
tet_qual_go (TETP tet,
             int range,
             double quality,
             double mindiha,
             double vol,
             vdt_t *vdt)
{
  WCRec pts[4];
  int i;
  EFringeTable ft = EMGetAssocFringeTable (vdt->gmodel);
  GraphicObj *go;

  FOR (i, 4) { pts[i] = tet->pt[i]->loc; }
  go = CreateTetra (pts);
  EMNoDrawAddGraphicsToModel (vdt->gmodel, go);
  EASValsSetColor (ColorFringeValueToColorHSV (ft, 1 - quality));
  EASValsSetShrink (0.97);
  EASValsSetFillStyle (FILL_SOLID);
  if (vol <= 0)
    EASValsSetLayer (VDT_NONPOS_VOL_SLIVER_LAYER);
  else if (mindiha < vdt->quality.sliver_dihed_angle)
    EASValsSetLayer (VDT_SMALL_DIHED_SLIVER_LAYER);
  else
    EASValsSetLayer (VDT_QUAL_LAYER_BASE + range);
  EASValsSetEdgeColor (vdt_color_to_pixel ("white"));
  EASValsSetEdgeFlag (TRUE);
  EGWithMaskChangeAttributes (FILL_MASK | COLOR_MASK |
                              EDGE_COLOR_MASK | EDGE_FLAG_MASK |
                              SHRINK_MASK | LAYER_MASK, go);
}
#endif

static int
have_num_pts (BSFP bsf, PTP pt[3])
{
  int n = 0, i;
  PTP apt;
  FOR (i, 3) {
    apt = bsf->pt[i];
    if (apt == pt[0] || apt == pt[1] || apt == pt[2]) n++;
  }
  return n;
}

static BSFP 
bsf_on_tet_face (TETP tet, int face)
{
  PTP pt[3];
  int i;
  LIST l;
  BSFP bsf;
  
  FOR (i, 3) {
    pt[i] = tet->pt[g3dtet_vertex_on_face[face][i]];
  }
  
  FOR (i, 3) {
    l = pt[i]->bsfl;
    if (l) {
      loop_through_list (l, bsf, BSFP) {
        if (have_num_pts (bsf, pt) == 3) return bsf;
      }
    }
  }
  
  return NULL;
}

static int
is_sliver (TETP tet, vdt_t *vdt)
{
  double sliverangle = vdt->quality.sliver_dihed_angle;
  double maxa, mina, angles[6];

  vdt_tet_angle_quality (tet, &mina, &maxa, angles);
  return ! (mina > sliverangle && maxa < M_PI - sliverangle);
}

static int
is_surface_sliver  (TETP tet, vdt_t *vdt)
{
  int n, i;
  BSFP bsfs[4];

  n = 0;
  FOR (i, 4) {
    bsfs[i] = NULL;
    if (tet->adjtet[i] == &vdt->stop_tet) {
      bsfs[i] = bsf_on_tet_face (tet, i); n++;
    }
  }
  if (n == 2) {
    BSFP bsf[2] = {NULL, NULL};
    n = 0;
    FOR (i, 4) { if (bsfs[i]) bsf[n++] = bsfs[i]; }
    if (bsf[0] && bsf[1]
        /* On the same surface ... */
        && (bsf[0]->surf_id == bsf[1]->surf_id)
        /* ... and one-sided */
        && (   bsf[0]->reg_id[0] == VDT_INVALID_REG_ID
            || bsf[0]->reg_id[1] == VDT_INVALID_REG_ID)
        && (   bsf[1]->reg_id[0] == VDT_INVALID_REG_ID
            || bsf[1]->reg_id[1] == VDT_INVALID_REG_ID)
        ) {
      return TRUE;
    } else {
      return FALSE;
    }
  }
  return FALSE;
}




static int
share_face (TETP t0, TETP t1)
{
  PTP pt;
  int i, had = 0;
  FOR (i, 4) {
    pt = t1->pt[i];
    if (TET_HAS_PT (t0, pt)) had++;
  }
  return (had == 3);
}



static int
is_removable_interior (TETP tet, int intface[2], int outface[2], vdt_t *vdt)
{
  int ifc[6][2] = {{0,1}, {0,2}, {0,3}, {1,2}, {1,3}, {2,3}};
  int ofc[6][2] = {{2,3}, {1,3}, {1,2}, {0,3}, {0,2}, {0,1}};
  int p;
  TETP it0;
  TETP it1;
  TETP ot0;
  TETP ot1;
  
  FOR (p, 6) {
    intface[0] = ifc[p][0];
    intface[1] = ifc[p][1];
    outface[0] = ofc[p][0];
    outface[1] = ofc[p][1];
    it0 = tet->adjtet[intface[0]];
    it1 = tet->adjtet[intface[1]];
    ot0 = tet->adjtet[outface[0]];
    ot1 = tet->adjtet[outface[1]];
    if (it0 != &vdt->stop_tet && it1 != &vdt->stop_tet) {
      if (share_face (it0, it1)) {
        if (it0->reg_id == it1->reg_id &&
            ot0->reg_id == ot1->reg_id &&
            it0->reg_id == ot0->reg_id)
          return TRUE;
      }
    }
  }
  intface[0] = intface[1] = -1;
  outface[0] = outface[1] = -1;
  return FALSE;
}


static PTP 
pt_not_in_tet (TETP othertet, TETP tet)
{
  PTP pt;
  int i;
  FOR (i, 4) {
    pt = othertet->pt[i];
    if (!TET_HAS_PT (tet, pt)) return pt;
  }
  return NULL;
}

static void
find_adjt (TETP tet, TETP otet0, TETP otet1, TETP adjt[6], vdt_t *vdt)
{
  int n = 0, i, j;
  TETP at;
  TETP atets[3];
#define HAVETHISTET(adjt, tet)                  \
  (adjt[0] == tet || adjt[1] == tet ||          \
   adjt[2] == tet || adjt[3] == tet ||          \
   adjt[4] == tet || adjt[5] == tet)
  
  FOR (j, 6) { adjt[j] = NULL; }

  atets[0] = tet;
  atets[1] = otet0;
  atets[2] = otet1;
  
  n = 0;
  FOR (j, 3) {
    FOR (i, 4) {
      at = atets[j]->adjtet[i];
      if (at && at != &vdt->stop_tet &&
          at != tet && at != otet0 && at != otet1) {
        if (!HAVETHISTET (adjt, at)) {
          assert (n < 6);
          adjt[n++] = at;
        }
      }
    }
  }
}

static TETP
find_adjtet (TETP adjt[6], PTP p0, PTP p1, PTP p2, vdt_t *vdt)
{
  int i;
  FOR (i, 6) {
    if (adjt[i]) {
      if (TET_HAS_3_PTS (adjt[i], p0, p1, p2)) {
        return adjt[i];
      }
    }
  }
  return &vdt->stop_tet;
}



static void
upd_adjtet (TETP tet, TETP at)
{
  int i;
  PTP pt[3];
  FOR (i, 4) {
    pt[0] = tet->pt[g3dtet_vertex_on_face[i][0]];
    pt[1] = tet->pt[g3dtet_vertex_on_face[i][1]];
    pt[2] = tet->pt[g3dtet_vertex_on_face[i][2]];
    if (TET_HAS_3_PTS (at, pt[0], pt[1], pt[2])) {
      tet->adjtet[i] = at;
      return;
    }
  }
}


static void
reconn_tet (TETP tet, PTP p0, PTP p1, PTP p2, PTP p3)
{
  CkitPoint3 a, b, n={0,0,0}, d;
  VECMAC_3_VEC (a, =, p1->loc, -, p0->loc);
  VECMAC_3_VEC (b, =, p2->loc, -, p0->loc);
  VECMAC_CROSS (n, =, a, x, b);
  VECMAC_3_VEC (d, =, p3->loc, -, p0->loc);
  if (VECMAC_DOT (d, n) < 0) {/* swap two points to get correct orientation */
    vdt_pt_t *apt = p1; p1 = p2; p2 = apt;
  }
  tet->pt[0] = p0; tet->pt[1] = p1; tet->pt[2] = p2; tet->pt[3] = p3;
}

static int
remove_interior_sliver (TETP tet, int intface[2], int outface[2], vdt_t *vdt)
{
  TETP adjt[6];
  PTP notintet;
  TETP recontet[2];
#define PTID(tet, n) (int) (tet)->pt[n]->id
   
  /* preconditions */
  assert (intface[0] >= 0 && intface[0] < 4);
  assert (intface[1] >= 0 && intface[1] < 4);
  assert (outface[0] >= 0 && outface[0] < 4);
  assert (outface[1] >= 0 && outface[1] < 4);

  recontet[0] = tet->adjtet[intface[0]];
  recontet[1] = tet->adjtet[intface[1]];

  /* find all tets adjacent to the three tets being modified */
  find_adjt (tet, recontet[0], recontet[1], adjt, vdt);

  /* the point not used by the sliver */
  notintet = pt_not_in_tet (recontet[0], tet);
  assert (notintet);

  /* delete the sliver */
  invalidate_tet (tet, vdt);

  /* reconnect the two remaining tets (note, outface is oriented
     ccw in the 
   */
  reconn_tet (recontet[0],
              tet->pt[g3dtet_vertex_on_face[outface[0]][0]],
              tet->pt[g3dtet_vertex_on_face[outface[0]][2]],
              tet->pt[g3dtet_vertex_on_face[outface[0]][1]],
              notintet);
  reconn_tet (recontet[1],
              tet->pt[g3dtet_vertex_on_face[outface[1]][0]],
              tet->pt[g3dtet_vertex_on_face[outface[1]][2]],
              tet->pt[g3dtet_vertex_on_face[outface[1]][1]],
              notintet);

  /* recompute the adjacencies of the two tets */
#define FACEPTS(tet, face)                              \
          tet->pt[g3dtet_vertex_on_face[face][0]],      \
          tet->pt[g3dtet_vertex_on_face[face][1]],      \
          tet->pt[g3dtet_vertex_on_face[face][2]] 

  recontet[0]->adjtet[0] = find_adjtet (adjt, FACEPTS (recontet[0], 0), vdt);
  upd_adjtet (recontet[0]->adjtet[0], recontet[0]);
  recontet[0]->adjtet[1] = find_adjtet (adjt, FACEPTS (recontet[0], 1), vdt);
  upd_adjtet (recontet[0]->adjtet[1], recontet[0]);
  recontet[0]->adjtet[2] = find_adjtet (adjt, FACEPTS (recontet[0], 2), vdt);
  upd_adjtet (recontet[0]->adjtet[2], recontet[0]);
  recontet[0]->adjtet[3] = find_adjtet (adjt, FACEPTS (recontet[0], 3), vdt);
  upd_adjtet (recontet[0]->adjtet[3], recontet[0]);

  recontet[1]->adjtet[0] = find_adjtet (adjt, FACEPTS (recontet[1], 0), vdt);
  upd_adjtet (recontet[1]->adjtet[0], recontet[1]);
  recontet[1]->adjtet[1] = find_adjtet (adjt, FACEPTS (recontet[1], 1), vdt);
  upd_adjtet (recontet[1]->adjtet[1], recontet[1]);
  recontet[1]->adjtet[2] = find_adjtet (adjt, FACEPTS (recontet[1], 2), vdt);
  upd_adjtet (recontet[1]->adjtet[2], recontet[1]);
  recontet[1]->adjtet[3] = find_adjtet (adjt, FACEPTS (recontet[1], 3), vdt);
  upd_adjtet (recontet[1]->adjtet[3], recontet[1]);

  vdt_tet_update_topo (recontet[0]);
  vdt_tet_update_topo (recontet[1]);

  /* recompute the tet quality */
  recontet[0]->quality = vdt_tet_quality (recontet[0]);
  recontet[1]->quality = vdt_tet_quality (recontet[1]);

#if GIFACE
  if (vdt->gmodel) {
    vdt_tet_generate_graphics (recontet[0], vdt);
    vdt_tet_generate_graphics (recontet[1], vdt);
  }
#endif

  return TRUE;
}


static void
repl_adjtet (TETP tet, TETP at, TETP byat)
{
  int i;
  FOR (i, 4) {
    if (tet->adjtet[i] == at) {
      tet->adjtet[i] = byat;
      return;
    }
  }
}


static void
find_bsf_faces  (TETP tet, vdt_t *vdt, int bsfface[4])
{
  int i;
  FOR (i, 4) {
    bsfface[i] = FALSE;
    if (tet->adjtet[i] == &vdt->stop_tet) {
      bsfface[i] = (bsf_on_tet_face (tet, i) != NULL);
    }
  }
}


static int
remove_surface_sliver (TETP tet, vdt_t *vdt)
{
  int i, n;
  BSFP bsf;
  int isbsfface[4];
  int bsfface[2];
  int intface[2];
  
  /* mark tet as not included in the final triangulation */
  invalidate_tet (tet, vdt);

  /* update the bsf/tet connectivities */
  find_bsf_faces (tet, vdt, isbsfface);
  n = 0; FOR (i, 4) { if (isbsfface[i])  bsfface[n++] = i; assert (n < 3); }
  n = 0; FOR (i, 4) { if (!isbsfface[i]) intface[n++] = i; assert (n < 3); }
  FOR (i, 2) { 
    bsf = bsf_on_tet_face (tet, bsfface[i]);
    assert (bsf);
    vdt_bsf_disconnect (bsf);
    vdt_bsf_connect (bsf,
                     tet->pt[g3dtet_vertex_on_face[intface[i]][0]],
                     tet->pt[g3dtet_vertex_on_face[intface[i]][2]],
                     tet->pt[g3dtet_vertex_on_face[intface[i]][1]],
                     vdt);
#   if GIFACE
    bsf->is_delaunay = TRUE;    /* pretend for the sake of visualization;
                                   it is no longer true, though  */
    vdt_bsf_generate_graphics (bsf, vdt);
#   endif
  }

  /* replace tet by stop_tet in the adjacencies */
  FOR (i, 4) {
    if (tet->adjtet[i] && tet->adjtet[i] != &vdt->stop_tet) {
      repl_adjtet (tet->adjtet[i], tet, &vdt->stop_tet);
      vdt_tet_update_topo (tet->adjtet[i]);
    }
  }
  
  return TRUE;
}


static int
tet_inc_to_pt (TETP tet, PTP pt)
{
  if (tet->pt[0] == pt) return TRUE;
  if (tet->pt[1] == pt) return TRUE;
  if (tet->pt[2] == pt) return TRUE;
  if (tet->pt[3] == pt) return TRUE;
  return FALSE;
}



static void
collect_tet_inc_to_pt (TETP starttet, vdt_t *vdt, dict_dictionary d)
{
  TETP tet;
  TETP atet;
  int f;
  STACK s = make_stack ();
  
  push_stack (s, starttet);
  while (!stack_empty (s)) {
    tet = (TETP) pop_stack (s);
    FOR (f, 4) {
      atet = tet->adjtet[f];
      if (atet && atet != &vdt->stop_tet) {
        if (tet_inc_to_pt (atet, starttet->pt[0]) ||
            tet_inc_to_pt (atet, starttet->pt[1]) ||
            tet_inc_to_pt (atet, starttet->pt[2]) ||
            tet_inc_to_pt (atet, starttet->pt[3])
            ) {
          if (!dict_find_obj_with_handle_name (d, atet)) { /* not in list */
            dict_add_obj_handle_name (d, atet, atet);
            push_stack (s, atet);
          }   
        }
      }
    }
  }
  free_stack (s, NOT_NODES);
}


static int
all_connected_tets_with_positive_vol (dict_dictionary d)
{
  TETP tet;
  int result = TRUE;
  dict_iterator diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while (tet = (TETP) dict_iter_objs (d, diter)) {
    if (vdt_tet_volume (tet) <= 0) {
      result = FALSE;
      goto exit;
    }
  }
 exit:
  dict_del_iterator (diter);
  return result;
}


static double
min_dihedral (TETP tet)
{
  double maxa, mina, angles[6];
  vdt_tet_angle_quality (tet, &mina, &maxa, angles);
  return mina;
}

#if 0
static double
min_dihedral_in_group (dict_dictionary d)
{
  TETP tet;
  double mindihedral = FLT_MAX;
  double mindi;
  dict_iterator diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while (tet = (TETP) dict_iter_objs (d, diter)) {
    mindi       = min_dihedral (tet);
    mindihedral = min (mindihedral, mindi);
  }
  dict_del_iterator (diter);
  return mindihedral;
}
#endif

#if 1 /* random-movement expansion */
static void
expand_interior_sliver (TETP tet, vdt_t *vdt)
{
  int i, p, maxi = 20;
  PTP pt;
  CkitPoint3 oloc;
  TETP atet;
  double sm = 0.5;
  double mindia, maxmin_dihedral;
  PTP tomove;
  CkitPoint3 whereto;
  dict_dictionary d = dict_create_dictionary ();
  dict_iterator di  = dict_get_iterator ();

  dict_add_obj_handle_name (d, tet, tet);
  collect_tet_inc_to_pt (tet, vdt, d);

  tomove       = NULL;
  maxmin_dihedral = min_dihedral (tet);

  FOR (p, 4) {
    pt = tet->pt[p];
    if (pt->bsfl == NULL) {  /* interior pt */
      FOR (i, maxi) {
        oloc = pt->loc;
        vdt_perturb_loc (sm, pt->mesh_size, &pt->loc);
        if (vdt_tet_volume (tet) > 0) {
          if (all_connected_tets_with_positive_vol (d)) {
            mindia = min_dihedral (tet);
            if (mindia > maxmin_dihedral) {
              tomove          = pt;
              maxmin_dihedral = mindia;
              whereto         = pt->loc;
            }
          }
        }
        pt->loc = oloc;
      } /* FOR (i, maxi) */
    } /* if (pt->bsfl */
  } /* FOR (p, 4) */

  if (tomove) {
    tomove->origloc = tomove->loc = whereto;
  } else {
  }

  di = dict_get_iterator ();
  dict_reset_iterator (di);
  while ((atet = dict_iter_objs (d, di))) {
    atet->quality = vdt_tet_quality (atet);
  }
  dict_del_iterator (di);
  dict_destroy_dictionary (d);
}
#else  /* connectivity-based expansion */
static void
expand_interior_sliver (TETP tet, vdt_t *vdt)
{
  int f, p;
  PTP pt;
  PTP apt;
  CkitPoint3 oloc;
  TETP atet;
  dict_iterator di;
  double sm = 0.3;
  double mindia, maxmin_dihedral;
  PTP tomove;
  CkitPoint3 whereto;
  dict_dictionary d = dict_create_dictionary ();

  dict_add_obj_handle_name (d, tet, tet);
  collect_tet_inc_to_pt (tet, vdt, d);

  fprintf (stderr, "Expanding tet={%d,%d,%d,%d}",
                 (int) tet->pt[0]->id, (int) tet->pt[1]->id,
                 (int) tet->pt[2]->id, (int) tet->pt[3]->id);
  fprintf (stderr, "min dih before = %g, ",
           (180 / M_PI) * min_dihedral (tet));
  
  tomove       = NULL;
  maxmin_dihedral = min_dihedral (tet);
  FOR (f, 4) {
    if (tet->adjtet[f] && tet->adjtet[f] != &vdt->stop_tet) {
      FOR (p, 3) {
        pt = tet->pt[g3dtet_vertex_on_face[f][p]];
        if (pt->bsfl == NULL) {  /* interior pt */
          apt = pt_not_in_tet (tet->adjtet[f], tet);
          oloc = pt->loc;
          VECMAC_3_VEC (pt->loc, =, sm * pt->loc, +, (1 - sm) * apt->loc);
          if (vdt_tet_volume (tet) > 0) {
            if (all_connected_tets_with_positive_vol (d)) {
              mindia = min_dihedral (tet);
              if (mindia > maxmin_dihedral) {
                tomove          = pt;
                maxmin_dihedral = mindia;
                whereto         = pt->loc;
              }
            }
          }
          pt->loc = oloc;
        }
      }
    }
  }

  if (tomove) {
    tomove->origloc = tomove->loc = whereto;
    fprintf (stderr, "after = %g\n", (180 / M_PI) * min_dihedral (tet));
  } else {
    fprintf (stderr, "NO CHANGE\n");
  }
  di = dict_get_iterator ();
  dict_reset_iterator (di);
  while ((atet = dict_iter_objs (d, di))) {
    atet->quality = vdt_tet_quality (atet);
  }
  dict_del_iterator (di);
  dict_destroy_dictionary (d);
}
#endif /* connectivity-based expansion */


int
vdt_quality (vdt_t *vdt)
{
  int i, j, k;
  TETP tet;
  char buf[256];
  double vol, mina, maxa, angles[6], avga;
  int anychange;
  ckit_timer_t *timer = ckit_timer_new ();
  LIST slivers = make_list ();
  double nangles;
  double ntets;
  
  CHECK_MEM (slivers);
  
  ckit_timer_restart (timer);
  logit ("Removing slivers");

#if GIFACE
  if (vdt->giface) {
    EFringeTable ft = EMGetAssocFringeTable (vdt->gmodel);
    ColorSetupFringeTableByMinMax (ft, 0, 1);
  }
#endif

  if (vdt->quality.nranges <= 0)
    vdt->quality.nranges = 5;
  vdt->quality.quality_summary
    = NEWP_REALLOC_OBJ_ARRAY (vdt->quality.quality_summary,
                              vdt->quality.nranges, int);
  CHECK_MEM (vdt->quality.quality_summary);
  vdt->quality.angle_quality_summary
    = NEWP_REALLOC_OBJ_ARRAY (vdt->quality.angle_quality_summary,
                              vdt->quality.nranges, int);
  CHECK_MEM (vdt->quality.angle_quality_summary);
  
  FOR (i, vdt->quality.nranges) {
    vdt->quality.quality_summary[i] = 0;
    vdt->quality.angle_quality_summary[i] = 0;
  }

  /* compute tet quality */
  tet = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    tet->quality = vdt_tet_quality (tet);
    if (is_sliver (tet, vdt)) {
      add_to_tail (slivers, tet);
    }
  }

  /* delete surface slivers */
  if (!vdt->keep_slivers) {
    do {
      anychange = FALSE;
      tet = NULL;
      while ((tet = (TETP) get_list_next (slivers, tet))) {
        if (tet->reg_id != VDT_INVALID_REG_ID) {
          if (is_surface_sliver (tet, vdt)) {
            anychange = remove_surface_sliver (tet, vdt); 
          }
        }
      }
    } while (anychange);
  }
  
  /* delete interior slivers */
  if (!vdt->keep_slivers) {
    do {
      anychange = FALSE;
      tet = NULL;
      while ((tet = (TETP) get_list_next (slivers, tet))) {
        if (tet->reg_id != VDT_INVALID_REG_ID) {
          int intface[2], outface[2];
          if (is_removable_interior (tet, intface, outface, vdt)) { 
#if 1
            anychange = remove_interior_sliver (tet, intface, outface, vdt);
#endif
          }
        }
      }
    } while (anychange);
  }
  
  /* expand unremovable interior slivers */
  if (!vdt->keep_slivers) { 
    tet = NULL;
    while ((tet = (TETP) get_list_next (slivers, tet))) {
      if (tet->reg_id != VDT_INVALID_REG_ID) {
        expand_interior_sliver (tet, vdt);
      }
    }
  }
  
  /* compute summary; create graphics */
  vdt->quality.num_non_positive_vol_tets  = 0;
  vdt->quality.num_small_dihed_angle_tets = 0;
  vdt->quality.min_quality                = 1;
  vdt->quality.max_quality                = 0;
  vdt->quality.min_volume                 = FLT_MAX;
  vdt->quality.max_volume                 = 0; 
  vdt->quality.min_dihed_angle            = M_PI;
  vdt->quality.max_dihed_angle            = 0;
  vdt->quality.avg_dihed_angle            = 0;
  vdt->quality.geom_mean_quality          = 1;
  vdt->quality.rms_quality                = 0;
  nangles = 0;
  ntets   = vdt_count_tets (vdt);
  tet     = NULL;
  while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
    if (tet->reg_id != VDT_INVALID_REG_ID) {
      if ((vol = vdt_tet_volume (tet)) <= 0) {
        char buf[512];
        sprintf (buf, "** Non-positive volume: tet={%d,%d,%d,%d}",
                 (int) tet->pt[0]->id, (int) tet->pt[1]->id,
                 (int) tet->pt[2]->id, (int) tet->pt[3]->id);
        logit (buf);
        vdt->quality.num_non_positive_vol_tets++;
      }
      vdt->quality.min_volume = min (vdt->quality.min_volume, vol);
      vdt->quality.max_volume = max (vdt->quality.max_volume, vol);
      vdt_tet_angle_quality (tet, &mina, &maxa, angles);
      vdt->quality.min_dihed_angle
        = min (vdt->quality.min_dihed_angle, mina);
      vdt->quality.max_dihed_angle
        = max (vdt->quality.max_dihed_angle, maxa);
      FOR (k, 6) {
        vdt->quality.avg_dihed_angle += angles[k]; nangles++;
        avga = angles[k] / M_PI;
        if      (avga >= 1) j = vdt->quality.nranges-1;
        else if (avga > 0)  j = (int)(avga * vdt->quality.nranges);
        else                j = 0;
        vdt->quality.angle_quality_summary[j]++;
      }
      if (mina < vdt->quality.sliver_dihed_angle) {
        sprintf (buf, "** Small dihedral angle (%g deg): tet={%d,%d,%d,%d}",
                 (180 / M_PI) * mina,
                 (int) tet->pt[0]->id, (int) tet->pt[1]->id,
                 (int) tet->pt[2]->id, (int) tet->pt[3]->id);
        logit (buf);
        vdt->quality.num_small_dihed_angle_tets++;
      }
      vdt->quality.min_quality = min (vdt->quality.min_quality, tet->quality);
      vdt->quality.max_quality = max (vdt->quality.max_quality, tet->quality);
      if      (tet->quality >= 1) j = vdt->quality.nranges-1;
      else if (tet->quality > 0)  j = (int)(tet->quality *
                                            vdt->quality.nranges);
      else                        j = 0;
      vdt->quality.quality_summary[j]++;
      vdt->quality.geom_mean_quality *= pow (tet->quality, 1 / ntets);
      vdt->quality.rms_quality       += tet->quality * tet->quality;
#if GIFACE
      if (vdt->giface) { tet_qual_go (tet, j, tet->quality, mina, vol, vdt); }
#endif
    }
  }
  vdt->quality.rms_quality       = sqrt (vdt->quality.rms_quality / ntets);
  vdt->quality.avg_dihed_angle   /= (nangles > 0 ? nangles : 1);

  free_list (slivers, NOT_NODES);

  vdt_report_time (timer, "Quality improvement");

  ckit_timer_free (timer);
  
  return TRUE;
}


int
vdt_delete (vdt_t *vdt)
{
#define FREES(string) { if (string) NEWP_FREE_OBJ (string); }
#define FREELA(list) { if (list) free_list (list, AND_NODES); }
#define FREELN(list) { if (list) free_list (list, NOT_NODES); }
  FREES (vdt->input);
  FREES (vdt->input_version);
  FREES (vdt->input_true_name);
  FREELN (vdt->pt_as_read);
  FREELN (vdt->bsf_as_read);
  FREELA (vdt->pt_to_insert_as_if_gen);
  if (vdt->pointd) dict_destroy_dictionary (vdt->pointd);
  FREELA (vdt->pointl);
  FREELA (vdt->tetl);
  FREELA (vdt->bsfl);
  FREELA (vdt->fflist);
  FREELN (vdt->sliver_ccl);
#if GIFACE
  if (vdt->gmodel) {
    EMDestroyAllGraphics (vdt->gmodel);
  }
#endif
  if (vdt->active_pt) NEWP_FREE_OBJ (vdt->active_pt);
  if (vdt->active_loc) NEWP_FREE_OBJ (vdt->active_loc);
#if VDT_PIRAT
  if (vdt->pt_pirat) pirat_destroy_pirat (vdt->pt_pirat);
#endif
#if VDT_BUCKETS
  if (vdt->buckets) vdt_bucket_free (vdt->buckets);
#endif
  NEWP_FREE_OBJ (vdt);
  return TRUE;
}


TETP
vdt_next_tet (void *ptid[4], void **reg_id,
              void *adjtet[4],
              void *prevtet, 
              vdt_t *vdt)
{
  TETP tet;
  int i;

  tet = prevtet;
  while (tet = (TETP) get_list_next (vdt->tetl, tet))
    if (tet->reg_id != VDT_INVALID_REG_ID)
      break;

  if (tet) {
    *reg_id = tet->reg_id;
    FOR (i, 4) {
      ptid[i] = tet->pt[i]->id;
    }
    FOR (i, 4) {
      if (tet->adjtet[i] == &vdt->stop_tet) adjtet[i] = NULL;
      else                                  adjtet[i] = tet->adjtet[i];
    }
    return tet;
  } else
    return NULL;
}


PTP
vdt_next_pt (void **id,
             double *x, double *y, double *z,
             double *mesh_size,
             void *prevpt, 
             vdt_t *vdt)
{
  PTP pt = NULL;

  pt = prevpt;
  while (pt = (PTP) get_list_next (vdt->pointl, pt))
    if (pt->num_conn_tets > 0)  /* skip unconnected points */
      break;

  if (pt) {
    *id        = pt->id;
    *x         = pt->origloc.x;
    *y         = pt->origloc.y;
    *z         = pt->origloc.z;
    *mesh_size = pt->mesh_size;
    return pt;
  } else
    return NULL;
}


/*
   Get the next boundary surface facet.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
BSFP
vdt_next_bsf (void **id, /* the identifier of the facet */
              void *ptid[3],    /* identifiers of the points */
              void **surf_id,   /* identifier of the surface */
              void *reg_id[2], /* identifiers of regions on each side
                              of the facet */
              void *tet[2],     /* pointers to tetrahedra on each side
                              of the facet as returned by VDT_next_tet(); 
                              Note that tet[1] is on the
                              side to which the normal points,
                              tet[0] is on the other side.  */
              void *prev,/* previous boundary facet; opaque pointers as returned
                            by this function.  Pass NULL to get the
                            first facet
                            */
              vdt_t *vdt        /*  mesher handle  */
              )
{
  BSFP bsf = prev;
  bsf = (BSFP) get_list_next (vdt->bsfl, bsf);
  if (bsf) {
    *id       = bsf;
    ptid[0]   = bsf->pt[0]->id;
    ptid[1]   = bsf->pt[1]->id;
    ptid[2]   = bsf->pt[2]->id;
    *surf_id  = bsf->surf_id;
    reg_id[0] = bsf->reg_id[0];
    reg_id[1] = bsf->reg_id[1];
    tet[0]    = bsf->tet[0];
    tet[1]    = bsf->tet[1];
  } 
  return bsf;
}
     
int
vdt_set_option (vdt_t *vdt,
                vdt_option_t *option
                )
{
  if (!option) {
    vdt_set_last_error_msg (vdt, "Option arg == NULL");
    return FALSE;
  }

  if (option->option == NULL)
    return FALSE;

  if (strcmp (option->option, VDT_OPTION_DEF_MESH_SIZE) == 0) {
    vdt->def_mesh_size = option->data.dv;
    if (vdt->def_mesh_size <= 0) {
      errexit ("Default mesh size set to non-positive number");
    }
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_LP_LIMIT_SIZE) == 0) {
    vdt->lp_size_limit = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_KEEP_SLIVERS) == 0) {
    vdt->keep_slivers  = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_TET_GEOM) == 0) {
    vdt->dump_tet_wrl = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_TET_EGF) == 0) {
    vdt->dump_tet_egf = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_BSF_EGF) == 0) {
    vdt->dump_bsf_egf = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_TET_WRL) == 0) {
    vdt->dump_tet_wrl = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_T3D) == 0) {
    vdt->dump_t3d = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_DUMP_DATA) == 0) {
    vdt->dump_data = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_PDEBUG) == 0) {
    vdt->pdebug = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_MAXDEBUG) == 0) {
    vdt->maxdebug = option->data.iv;
    return TRUE;
  }
#if GIFACE
  if (strcmp (option->option, VDT_OPTION_GDEBUG) == 0) {
    vdt->gdebug = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_GIFACE) == 0) {
    vdt->giface = option->data.iv;
    return TRUE;
  }
#endif
  if (strcmp (option->option, VDT_OPTION_BDRY_PT_SHIFT_MULT) == 0) {
    vdt->bdry_pt_shift_mult = option->data.dv;
    if (fabs (vdt->bdry_pt_shift_mult) >= 1) {
      errexit ("Shift multiplier too big");
    }
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_GEN_PT_SHIFT_MULT) == 0) {
    vdt->gen_pt_shift_mult = option->data.dv;
    if (fabs (vdt->gen_pt_shift_mult) >= 1) {
      errexit ("Shift multiplier too big");
    }
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_ALLOW_FLIPS) == 0) {
    vdt->allow_flips = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_GEN_INTERIOR_PTS) == 0) {
    vdt->gen_interior_pts = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_GEN_NEAR_SURF_PTS) == 0) {
    vdt->gen_near_surf_pts = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_NEAR_SURF_TEST) == 0) {
    vdt->near_surf_test = option->data.iv;
    if (!(vdt->near_surf_test == VDT_NEAR_SURF_TEST_IGNORE ||
          vdt->near_surf_test == VDT_NEAR_SURF_TEST_DIST   ||
          vdt->near_surf_test == VDT_NEAR_SURF_TEST_CIRCUMSPHERE)) {
      errexit ("Wrong value for near surf test");
    }
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_NRANGES) == 0) {
    vdt->quality.nranges = option->data.iv;
    if (vdt->quality.nranges <= 0) {
      errexit ("Number of quality ranges must be a positive number");
    }
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_PDEBUG) == 0) {
    vdt->pdebug = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_MAXDEBUG) == 0) {
    vdt->maxdebug = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_ORIENT_INITIAL_AF) == 0) {
    vdt->orient_initial_af = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_TEST_MODE) == 0) {
    vdt->test_mode = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_INJECT_SURF_PROJ) == 0) {
    vdt->inject_surf_proj = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_EVAL_MESH_SIZE_FROM_PT) == 0) {
    vdt->eval_mesh_size_from_pt = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_EVAL_MESH_SIZE_FUNC) == 0) {
    vdt->eval_mesh_size_func = (vdt_eval_mesh_size_func) option->data.fv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_CHECK_INTERRUPT_FUNC) == 0) {
    vdt->check_interrupt.func = (vdt_check_interrupt_func) option->data.fv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_EVAL_MESH_SIZE_FUNC_DATA) == 0) {
    vdt->eval_mesh_size_func_data = option->data.pv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_REPORT_OPTS) == 0) {
    vdt->report_opts = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_INPUT) == 0) {
    NEWP_DUPL_STRING (vdt->input, (char *) option->data.pv,
                      { errexit ("Not enough memory"); });
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_INPUT_VERSION) == 0) {
    NEWP_DUPL_STRING (vdt->input_version, (char *) option->data.pv,
                      { errexit ("Not enough memory"); });
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_INPUT_TRUE_NAME) == 0) {
    NEWP_DUPL_STRING (vdt->input_true_name, (char *) option->data.pv,
                      { errexit ("Not enough memory"); });
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_BUCKET_SIZE) == 0) {
    vdt->bucket_size = option->data.dv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_SLIVER_DIHED_ANGLE) == 0) {
    vdt->quality.sliver_dihed_angle = option->data.dv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_RESET_LOCS) == 0) {
    vdt->reset_locs = option->data.iv;
    return TRUE;
  }
  if (strcmp (option->option, VDT_OPTION_RETRI_TO_DESLIVER) == 0) {
    vdt->retri_to_desliver = option->data.iv;
    return TRUE;
  }
  vdt_set_last_error_msg (vdt, "Option not recognized");
  return FALSE;
}


void
vdt_report_opts (vdt_t *vdt)
{
  char buf[1024];
  logit ("Options:");
  sprintf (buf, "   Input                     = %s",
           (vdt->input ? vdt->input : "?"));
  logit (buf);
  sprintf (buf, "   Input version             = %s",
           (vdt->input_version ? vdt->input_version : "?"));
  logit (buf);
  sprintf (buf, "   Input true name           = %s",
           (vdt->input_true_name ? vdt->input_true_name : "?"));
  logit (buf);
  sprintf (buf, "   Default mesh size         = %g", vdt->def_mesh_size);
  logit (buf);
  sprintf (buf, "   Min mesh size             = %g", vdt->min_mesh_size);
  logit (buf);
  sprintf (buf, "   Max mesh size             = %g", vdt->max_mesh_size);
  logit (buf);
#if VDT_BUCKETS
  sprintf (buf, "   Bucket size               = %g", vdt->bucket_size);
  logit (buf);
#endif
  sprintf (buf, "   Near surface test         = %s",
           (vdt->near_surf_test == VDT_NEAR_SURF_TEST_IGNORE
            ? "ignore"
            : (vdt->near_surf_test == VDT_NEAR_SURF_TEST_DIST
               ? "distance"
               : (vdt->near_surf_test == VDT_NEAR_SURF_TEST_CIRCUMSPHERE
                  ? "empty circumsphere"
                  : "unknown")
               )
            )
           );
  logit (buf);
  sprintf (buf, "   LP limit size             = %d", vdt->lp_size_limit);
  logit (buf);
  sprintf (buf, "   Bdry pt shift mult        = %g", vdt->bdry_pt_shift_mult);
  logit (buf);
  sprintf (buf, "   Generated pt shift mult   = %g", vdt->gen_pt_shift_mult);
  logit (buf);
  sprintf (buf, "   Generate interior pts     = %d", vdt->gen_interior_pts);
  logit (buf);
  sprintf (buf, "   Generate near-surface pts = %d", vdt->gen_near_surf_pts);
  logit (buf);
  sprintf (buf, "   Orient initial AF         = %d", vdt->orient_initial_af);
  logit (buf);
  sprintf (buf, "   Sliver dihed. angle (deg) = %g",
           vdt->quality.sliver_dihed_angle * 180 / M_PI);
  logit (buf);
  sprintf (buf, "   Keep slivers              = %d", vdt->keep_slivers);
  logit (buf);
  sprintf (buf, "   Dumpt tet as WRL          = %d", vdt->dump_tet_wrl);
  logit (buf);
  sprintf (buf, "   Dumpt tet as EGF          = %d", vdt->dump_tet_egf);
  logit (buf);
  sprintf (buf, "   Dumpt bsf as EGF          = %d", vdt->dump_bsf_egf);
  logit (buf);
  sprintf (buf, "   Dumpt tet as GEOM         = %d", vdt->dump_tet_geom);
  logit (buf);
  sprintf (buf, "   Dumpt tet as T3D          = %d", vdt->dump_t3d);
  logit (buf);
}

void
vdt_derive_mesh_size (vdt_t *vdt)
{
  vdt_pt_t *pt;
  vdt_bsf_t *bsf;
  double aes;
  int ne;
  CkitPoint3 v;

  vdt->max_mesh_size = 0;
  vdt->min_mesh_size = FLT_MAX;

  /* try to derive the default mesh from the mesh size stored at
   the points */
  aes = 0; ne = 0;
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    if (pt->mesh_size > 0) {
      aes += pt->mesh_size; ne++;
      vdt->min_mesh_size = min (vdt->min_mesh_size, pt->mesh_size);
      vdt->max_mesh_size = max (vdt->max_mesh_size, pt->mesh_size);
    }
  }
  if (ne > 0) {
    vdt->def_mesh_size = aes / ne;
  }

  if (vdt->def_mesh_size <= 0) {
    /* compute the average edge size */
    aes = 0; ne = 0;
    bsf = NULL;
    while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
      VECMAC_3_VEC (v, =, bsf->pt[0]->loc, -, bsf->pt[2]->loc);
      aes += VECMAC_NORM (v); ne++;
      VECMAC_3_VEC (v, =, bsf->pt[1]->loc, -, bsf->pt[0]->loc);
      aes += VECMAC_NORM (v); ne++;
      VECMAC_3_VEC (v, =, bsf->pt[2]->loc, -, bsf->pt[1]->loc);
      aes += VECMAC_NORM (v); ne++;
    }
    vdt->def_mesh_size = aes / ne;
  }
  
  /* check if any point needs the default mesh size */
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pointl, pt))) {
    if (pt->mesh_size <= 0) {
      pt->mesh_size = vdt->def_mesh_size;
    }
  }

  if (vdt->min_mesh_size > vdt->def_mesh_size)
    vdt->min_mesh_size = vdt->def_mesh_size;
  if (vdt->max_mesh_size < vdt->def_mesh_size)
    vdt->max_mesh_size = vdt->def_mesh_size;
}


/*
   Triangulate the interior.
 */
int
vdt_triangulate (vdt_t *vdt)
{
  ckit_timer_t *timer = ckit_timer_new ();

  ckit_timer_restart (timer);
  
  vdt_derive_mesh_size (vdt);
  
  if (vdt->report_opts)
    vdt_report_opts (vdt);
  
  preprocess (vdt);

#if GIFACE
  vdt->gmodel = ESIModel ();
  if (vdt->giface) {
    vdt_generate_graphics (vdt);
    ESIHandleCmd ("FIT");
    ESIEventLoop ((!vdt->test_mode), "Ready to triangulate");
    ESIEventLoop (FALSE, "Triangulation in progress ...                    ");
  }
#endif

  triangulate (vdt);

#if GIFACE
  if (vdt->giface) {
    vdt_generate_graphics (vdt);
    ESIEventLoop ((!vdt->test_mode), "Before quality improvement");
  }
#endif

  vdt_quality (vdt);

  if (vdt->reset_locs) reset_locs (vdt);

  vdt_report_time (timer, "Triangulation");

#if GIFACE
  if (vdt->giface) {
    vdt_generate_graphics (vdt);
  }
#endif

  ckit_timer_free (timer);
  
  return TRUE;
}

int
vdt_dump (vdt_t *vdt)
{
  if (vdt->dump_tet_wrl) dump_tet_wrl (vdt);
  if (vdt->dump_tet_egf) dump_tet_egf (vdt);
  if (vdt->dump_bsf_egf) dump_bsf_egf (vdt);
  if (vdt->dump_data)    dump_data (vdt);
  if (vdt->dump_t3d)     dump_t3d (vdt);
  return TRUE;
}

int
vdt_get_quality (vdt_t *vdt,
                 int *num_non_positive_vol_tets,
                 int *num_small_dihed_angle_tets,
                 int *quality_measure,
                 double *min_quality,
                 double *max_quality,
                 double *geom_mean_quality,
                 double *rms_quality,
                 double *min_volume,
                 double *max_volume,
                 double *min_dihed_angle,
                 double *max_dihed_angle,
                 double *avg_dihed_angle
                 )
{
  *num_non_positive_vol_tets = vdt->quality.num_non_positive_vol_tets;
  *num_small_dihed_angle_tets= vdt->quality.num_small_dihed_angle_tets;
  *quality_measure           = vdt->quality.quality_measure;
  *min_quality               = vdt->quality.min_quality;
  *max_quality               = vdt->quality.max_quality;
  *geom_mean_quality         = vdt->quality.geom_mean_quality;
  *rms_quality               = vdt->quality.rms_quality;
  *min_volume                = vdt->quality.min_volume;
  *max_volume                = vdt->quality.max_volume;
  *min_dihed_angle           = vdt->quality.min_dihed_angle;
  *max_dihed_angle           = vdt->quality.max_dihed_angle;
  *avg_dihed_angle           = vdt->quality.avg_dihed_angle;
  return TRUE;
}
