#ifndef VDT_H
#  define VDT_H


/* 
      Variational Delaunay Triangulation
             (C) 1998 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdt.h,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "listP.h"
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h" 
#define WANT_VECTP_ALIASES
#include "vectP.h" /* to be able to use vector library routines. */
#include "randP.h" /* to be able to use rand library routines. */
#include "newP.h"              /* Ckit library */
#include "ioP.h"              /* Ckit library */
#include "listP.h"              /* Ckit library */
#include "dictP.h"              /* Ckit library */
#include "getoptP.h"              /* Ckit library */
#include "tokensP.h"              /* Ckit library */
#include "timeP.h"              /* Ckit library */
#include "blpoolP.h"              /* Ckit library */

#include "g3d.h"
#include "lobs3d.h"

#undef REAL
#define REAL double
#include "lpkit.h"

#ifndef VDT_BUCKETS
#  define VDT_PIRAT   1
#  define VDT_BUCKETS 0
#endif

#ifndef VDT_HIER_LOBS
#  define VDT_USE_HIERLOBS 1
#else
#  define VDT_USE_HIERLOBS VDT_HIER_LOBS
#endif

#if VDT_PIRAT
#include "pirat.h"
#endif

typedef struct vdt_buckets_t *vdt_buckets;


#if GIFACE
#  include "Esimple.h"
#endif

#define VDT_CUSTOM_SIGNAL_HANDLER 1

#define VDT_REPRODUCIBLE_RANDOM 1

struct vdt_tet_t;
struct vdt_pt_t;


typedef struct vdt_bsf_t {
  struct vdt_pt_t       *pt[3];    /* corner points */
  void                  *surf_id; 
  struct vdt_pt_t       *fourth_pt;
  CkitPoint3             tetcsc;
  double                 tetcsr;
  void                  *reg_id[2];
  struct vdt_tet_t      *tet[2];
#if GIFACE
  GraphicObj            *go;
#endif  
  int                    surf_mod_pass;
  char                   is_delaunay;
}              vdt_bsf_t;


typedef struct vdt_pt_t {
  void                  *id;
  CkitPoint3             loc;
  CkitPoint3             origloc;
  LIST_OF (vdt_ff_t *)   ffl;
  LIST_OF (vdt_bsf_t *)  bsfl;
  double                 mesh_size;
  int                    num_conn_tets;
#if GIFACE
  GraphicObj            *go, *atgo;
#endif
}              vdt_pt_t;

#define PT_IS_ON_BDRY(pt) (pt->bsfl)

typedef struct vdt_tet_t {
  vdt_pt_t              *pt[4];
  struct vdt_tet_t      *adjtet[4];
  void                  *reg_id;
  float                  quality;
#if GIFACE
  GraphicObj            *go;
#endif
}              vdt_tet_t;

typedef struct vdt_ff_t {
  vdt_tet_t             *ontet;
  vdt_pt_t              *pt[3];
#if GIFACE
  GraphicObj            *go;
#endif
  struct vdt_bsf_t      *onbsf;
  unsigned char          face;  /* 0...3 */
  unsigned char          is_active; /* boolean */
}              vdt_ff_t;

typedef double (*vdt_mesh_size_func) (void *data, CkitPoint3 *loc);

#define VDT_PT_GEN_FAILED                0
#define VDT_PT_GEN_ACCEPTED              1
#define VDT_PT_GEN_TOO_CLOSE_TO_BDRY     2
#define VDT_PT_GEN_TOO_CLOSE_TO_ANOTHER  3
#define VDT_PT_GEN_OUTSIDE_OF_GENSPACE   4 

typedef int (*vdt_gen_pt_func) (void *data, CkitPoint3 *loc,
                                double mesh_size,
                                double min_dist,
                                double bdry_min_dist);

typedef struct vdt_ot_t {
  struct vdt_ot_leaf_t  *root;
  CkitPoint3             center;
  double                 size;
  vdt_mesh_size_func     mesh_size;
  void                  *mesh_size_func_data;
  vdt_gen_pt_func        gen_pt;
  void                  *gen_pt_func_data;
}              vdt_ot_t;


#define VDT_ALL_PTS INT_MAX
#define VDT_SOFT_LIMIT   VDT_ALL_PTS
#define VDT_HARD_LIMIT   0

typedef struct vdt_search_stat_t {
  int          num_searches;
  double       sum_found;
  int          max_num_found;
}              vdt_search_stat_t;

typedef struct vdt_lp_stat_t {
  double       sum_unknowns;
  int          sum_evals;
  int          max_num_unknowns;
}              vdt_lp_stat_t;

typedef struct vdt_delaunayzation_stat_t {
  vdt_lp_stat_t   lp_stat;
  int             num_bsf_updates;
}              vdt_delaunayzation_stat_t;

#define VDT_GAMMA_QUALITY  0
#define VDT_RADIUS_QUALITY 1


typedef struct vdt_tet_quality_t {
  double                 sliver_dihed_angle;
  double                 sliver_quality;
  int                    quality_measure;
  int                    nranges;
  int                   *quality_summary;
  int                   *angle_quality_summary;
  int                    num_non_positive_vol_tets;
  int                    num_small_dihed_angle_tets;
  double                 min_quality;
  double                 max_quality;
  double                 geom_mean_quality;
  double                 rms_quality;
  double                 min_volume;
  double                 max_volume;
  double                 min_dihed_angle;
  double                 max_dihed_angle;
  double                 avg_dihed_angle;
}              vdt_tet_quality_t;

typedef double (*vdt_eval_mesh_size_func) (void *data, CkitPoint3 *loc);
typedef int (*vdt_check_interrupt_func) (struct vdt_t *);

typedef struct vdt_t {
  char                      *input;
  char                      *input_version;
  char                      *input_true_name;
  LIST                       pt_as_read;
  LIST                       bsf_as_read;
  LIST                       pt_to_insert_as_if_gen;
  dict_dictionary            pointd;
  LIST                       pointl;
  dict_dictionary            regd;
  void                      *max_pt_id;
  void                      *input_max_pt_id;
  G3D_box_t                  bbox;
  G3D_box_t                  gen_space;
  LIST                       tetl;
  LIST                       bsfl; 
  double                     def_mesh_size;
  double                     min_mesh_size;
  double                     max_mesh_size;
  struct vdt_pt_t          **active_pt;
  CkitPoint3                *active_loc;
  int                        tot_active_points;
  int                        num_active_points;
#if VDT_PIRAT
  pirat_front_end_t         *pt_pirat;
#endif
  double                     bucket_size;
#if VDT_BUCKETS
  vdt_buckets                buckets;
#endif
  vdt_search_stat_t          search_stat;
  LIST                       fflist;
  vdt_tet_t                  stop_tet;
  vdt_pt_t                  *new_pt;
  int                        num_input_pts;
  int                        num_gen_pts;
  int                        num_input_bsfs;
  int                        surf_mod_pass;
  vdt_ot_t                  *ot;
  int                        near_surf_test;
#if VDT_USE_HIERLOBS
  lobs_3d_front_end          hierlobs;
#else
  lobs_3d_front_end          bsflobs;
#endif
  vdt_delaunayzation_stat_t  del_stat;
  int                        lp_size_limit;
  int                        shift_method;
  double                     bdry_pt_shift_mult;
  double                     gen_pt_shift_mult;
  int                        gen_interior_pts;
  int                        gen_near_surf_pts;
  int                        allow_flips;
  int                        pdebug;
  int                        maxdebug;
  int                        test_mode;
#if GIFACE
  int                        gdebug;
  int                        giface;
#endif
#if GIFACE
  EModel                    *gmodel;
#endif
  vdt_tet_quality_t          quality;
  int                        keep_slivers;
  int                        dump_tet_wrl;
  int                        dump_tet_geom;
  int                        dump_t3d;
  int                        dump_tet_egf;
  int                        dump_bsf_egf;
  int                        dump_data;
  struct {
    int max_ctet;
    int ctet;
    int (*func) (struct vdt_t *);
  }                          check_interrupt;
  int                        report_opts;
  int                        reset_locs;
  char                      *last_error_msg;
  int                        retri_to_desliver;
  LIST                       sliver_ccl;
  int                        inject_surf_proj;
  int                        incr_ntets;
  int                        ntets;
  int                        eval_mesh_size_from_pt;
  vdt_eval_mesh_size_func    eval_mesh_size_func;
  void                      *eval_mesh_size_func_data;
  int                        orient_initial_af;
}              vdt_t;

typedef enum vdt_lp_result_t {
  VDT_OPTIMAL           = OPTIMAL,
  VDT_UNBOUNDED         = UNBOUNDED,
  VDT_INFEASIBLE        = INFEASIBLE,
  VDT_LOWER_DIM_OPTIMAL = -1,
  VDT_TOO_MANY_POINTS   = -2
}            vdt_lp_result_t;

struct vdt_spatch_edge_t;

typedef struct vdt_spatch_face_t {
  struct vdt_spatch_edge_t   *e[3];
  vdt_bsf_t                  *bsf;
}              vdt_spatch_face_t;

typedef struct vdt_spatch_edge_t {
  vdt_pt_t                     *pt[2];
  LIST_OF (vdt_spatch_face_t *) fl;
}              vdt_spatch_edge_t;

typedef struct vdt_spatch_t {
  blpool_p                       eblp;
  blpool_p                       fblp;
  LIST_OF (vdt_spatch_edge_t *)  el;
  LIST_OF (vdt_spatch_face_t *)  fl;
}              vdt_spatch_t;



#undef PTP
#define PTP vdt_pt_t *
#undef BSFP
#define BSFP vdt_bsf_t *
#undef FFP
#define FFP vdt_ff_t *
#undef TETP
#define TETP vdt_tet_t *
#undef SPP
#define SPP vdt_spatch_t *
#undef SPEP
#define SPEP vdt_spatch_edge_t *
#undef SPFP
#define SPFP vdt_spatch_face_t *
#define OT vdt_ot_t
#define OTP vdt_ot_t *
#define LEAF vdt_ot_leaf_t 
#define LEAFP vdt_ot_leaf_t *

#define VDT_MODULE_ID(MODULE) vdt_##MODULE##_module_id ()
#define DEF_VDT_MODULE_ID(MODULE)                       \
  const char *                                          \
  vdt_##MODULE##_module_id (void) { return vcid; }

#endif
