#!/bin/bash


default_dirs="./ ./RCS/ @VDT_DIR@/inputs/ @VDT_DIR@/inputs/RCS/"
if [ -d ~/VDT/inputs ] ; then 
   dd="$default_dirs $(/bin/ls -d ~/VDT/inputs/)/"
   default_dirs="$dd"
fi
if [ -d ~/VDT/inputs/RCS ] ; then 
   dd="$default_dirs $(/bin/ls -d ~/VDT/inputs/RCS/)/"
   default_dirs="$dd"
fi

suffixes='.vin .gz .vin.gz ,v ,v.gz .vin,v .vin,v.gz'

function usage_and_exit {
cat <<EOF
          (V)ariational (D)elaunay (T)riangulation        
          VDT Copyright (C) 1998--2000, Petr Krysl           

Usage: $(basename $0) -i inp [ optional vdt arguments ] [ -vdt_exec b ] 
   Note: The option -i must be present together with the file name inp
   -vdt_exec b
      Run with vdt binary 'b'
   -vdt_dir d
      Assume the VDT distribution is under directory d.

This is a front-end to the program vdt.  It is able to figure out how
to obtain a vdt input file by (a) looking for it in various default
directories, and (b) testing various suffixes (see below).  In this
way, this script is able to figure out how to uncompress or check out
an RCS file, and makes a local copy in the working directory (this
copy is deleted when the script finishes).

The input file can be specified either by giving the full path
(e.g. /home/cairo/pkrysl/VDT/inputs/tire), or by giving just the name
(tire or tire.vin) and relying on the ability of the script to find it
in the default directories.

Directories that are searched: 
  specified directory when full path or relative path was given
`for d in $(echo $default_dirs); do echo "  $d"; done`

Suffixes: none and $(for n in $suffixes; do echo "\`$n', \c"; done)

EOF
exit 1
}

# find -i option and collect all remaining arguments in args

args=""
input='???'
vdt_exec=vdt
vdt_dir="@VDT_DIR@"
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
    -help)
        usage_and_exit
        ;;
    -i)
        if [ $# -ge 1 ] ; then
            input=$1; shift
        else
            usage_and_exit
        fi
        ;;
    -vdt_exec)
        if [ $# -ge 1 ] ; then
            vdt_exec=$1; shift
        else
            usage_and_exit
        fi
        ;;
    -vdt_dir)
        if [ $# -ge 1 ] ; then
            vdt_dir=$1; shift
        else
            usage_and_exit
        fi
        ;;
    *)
        ta="$args $a"
        args="$ta"
        ;;
    esac
done

if [ x"$input" = x'???' ] ; then
    usage_and_exit
fi

remove=f

function local_file {
    n_=$1
    echo ./$(basename ${n_})
}

function cleanup {
    if [ $remove = t ] ; then 
       rm -f $1
    fi
}

function has_suffix {
    name=$1; shift
    suffix=$1
    if [ ${name%%${suffix}} = ${name} ] ; then 
        echo f
    else
        echo t
    fi
}

specdir=$(dirname $input)/
t=$(basename $input)
input=$t
if [ x"$vdt_dir" = x ] ; then
   vdt_dirs="./inputs/"
else
   vdt_dirs="$vdt_dir/inputs/ $vdt_dir/inputs/RCS/"
fi

for d in "$specdir" $(echo $vdt_dirs) $(echo $default_dirs)
do  
   if [ -d $d ] ; then
      for s in "" $(echo $suffixes)
      do
        n=${d}${input}${s}
        if [ -f $n ] ; then 
           if [ $(has_suffix $n ".vin") = t ] ; then
              input=$n
              $vdt_exec -i $input -true_name $n $(echo $args); exit=$?;
              exit $exit
           fi
   
           if [ $(has_suffix $n ".vin,v") = t ] ; then 
              input=$(local_file ${n%%,v})
              co -q -p $n > $input
              remove=t; trap "cleanup SIGINT $input" 2 9 15
              $vdt_exec -i $input -true_name $n $(echo $args); exit=$?;
              cleanup $input; exit $exit
           fi
   
           if [ $(has_suffix $n ".vin,v.gz") = t ] ; then
              rcsfile=$(local_file ${n%%.gz})
              gunzip -qcf $n > $rcsfile
              input=$(local_file ${n%%,v.gz})
              co -q -p $rcsfile > $input
              rm -f $rcsfile
              remove=t; trap "cleanup SIGINT $input" 2 9 15
              $vdt_exec -i $input -true_name $n $(echo $args); exit=$?;
              cleanup $input; exit $exit
           fi
   
           if [ $(has_suffix $n ".gz") = t -o $(has_suffix $n ".vin.gz") = t ] ; then
              input=$(local_file ${n%%${s}})
              gunzip -qcf $n > $input
              remove=t; trap "cleanup SIGINT $input" 2 9 15
              $vdt_exec -i $input -true_name $n $(echo $args); exit=$?;
              cleanup $input; exit $exit
           fi
   
           if [ $(has_suffix $n "") = t ] ; then
              input=$n
              $vdt_exec -i $input -true_name $n $(echo $args); exit=$?;
              cleanup $input; exit $exit
           fi
        fi
      done
  fi
done

echo "*** Input $input does not exist?!"
exit 1

