/* 	$Id: vdtadlib.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $	 */

#ifndef lint
static char vcid[] = "$Id: vdtadlib.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
#endif /* lint */

/* #include <time.h> */
#include <adlib3d.h>
#include <mesh3d.h>

#include <vdtpub.h>
#include <dictP.h>
#undef REAL 
#include <smesh.h>
#include <smesh.pt>

#define VDT_PRINT       0
#define VDT_GEN_PTS     1  
#define sliver_dihed_angle (5 * M_PI / 180)

static OctNode *
find_on (dict_dictionary nmap, int id)
{
  return (OctNode *) dict_find_obj_with_int_name (nmap, id);
}

static void
update_nmap (dict_dictionary nmap, int id, OctNode *on)
{
  OctNode *fon;
  fon = find_on (nmap, id);
  if (!fon) {
    dict_add_obj_int_name (nmap, on, id);
  } else {
    if (fon != on) {
      fprintf (stderr, "Two octnodes with the same id\n");
      exit (1);
    }
  }
}



static int NodeNumber = 0;

static void
descend_updnmap (OctNode *T, dict_dictionary nmap)
{
  if (T != NULL) {
    if (!T->Gray) {
      T->Number = NodeNumber++;
      update_nmap (nmap, T->Number, T);
    }
    descend_updnmap (T->O1, nmap);
    descend_updnmap (T->O2, nmap);
    descend_updnmap (T->O3, nmap);
    descend_updnmap (T->O4, nmap);
    descend_updnmap (T->O5, nmap);
    descend_updnmap (T->O6, nmap);
    descend_updnmap (T->O7, nmap);
    descend_updnmap (T->O8, nmap);
  }
  return;
}





static void
remove_edge_midnodes (dict_dictionary nmap)
{
  GEdge *GE;
  EdgeElem *EE;
  int ge;
  OctNode *on;
  
  fprintf (stderr, "Before %d\n", dict_count (nmap));
  for (ge = 0; ge < nedges; ge++) {
    GE = GEdges[ge];
    for (EE = GE->FirstElem; EE ; EE = EE->RLink) {
      on = find_on (nmap, EE->ElVertex[1]->Number);
      dict_delete_obj_int_name (nmap, EE->ElVertex[1]->Number);
    }
  }
  fprintf (stderr, "After %d\n", dict_count (nmap));
}


/*
   Create a dictionary mapping node identifier (on->Number) to an octant node (on).
 */
static dict_dictionary
create_nmap ()
{
  dict_dictionary nmap;
  
  nmap = dict_create_dictionary ();

  NodeNumber = 1;
  descend_updnmap (OT, nmap);

  return nmap;
}

static void
define_bsfs (Body *B, VDT vdt) 
{
  Shell *SH;
  Face  *F;
  GFace *GF;
  FaceElem *FE;
  SBody *SB;
  
  for (SB=B->SBodies;SB;SB=SB->Link) {
    for (SH=SB->Shells;SH;SH=SH->Link) {
      for (F=SH->Faces;F;F=F->Link) {
        GF = GFaces[F->GlobalFace];
        for (FE=GF->FaceElements;FE;FE=FE->Link) {
          void *ptid[3];
          void *reg_id[2];
          ptid[0]   = (void *) FE->n[0]->Number;
          ptid[1]   = (void *) FE->n[2]->Number;
          ptid[2]   = (void *) FE->n[1]->Number;
          reg_id[0] = SB;
          reg_id[1] = 0;
          if (!VDT_add_bsf (ptid, GF, reg_id, vdt)) {
            fprintf (stderr, "Failed in VDT_add_bsf\n");
            exit (1);
          }
#if VDT_PRINT
          printf ("bsf %d %d %d %d\n", ptid[0], ptid[1], ptid[2], (int) GF);
#endif
        } 
      } 
    } 
  } 
}

typedef double (*eval_mesh_size_func) (double []);

typedef struct eval_mesh_size_t {
  eval_mesh_size_func msf;
}              eval_mesh_size_t;

static double
eval_mesh_size (void *data, CkitPoint3 *loc)
{
  eval_mesh_size_t *p = (eval_mesh_size_t *) data;
  return (*p->msf) ((double *)loc);
}

static eval_mesh_size_t msfp;

#define SET_VDT_OPT(vdt, op)                            \
   if (!VDT_set_option (vdt, &op)) {                    \
     fprintf (stderr, "Failed to set option \"%s\"",    \
              (op.option ? op.option : "NULL?"));       \
     exit (1);                                          \
   }


static void
define_pts (Body *B, VDT vdt, dict_dictionary nmap) 
{
  OctNode *on;
  dict_iterator di = dict_get_iterator ();
  vdt_option_t op;

  op.option  = VDT_OPTION_EVAL_MESH_SIZE_FUNC;
  op.data.pv = eval_mesh_size;
  SET_VDT_OPT (vdt, op);

  msfp.msf   = B->SBodies->HFunction;
  op.option  = VDT_OPTION_EVAL_MESH_SIZE_FUNC_DATA;
  op.data.pv = &msfp;
  SET_VDT_OPT (vdt, op);

  dict_reset_iterator (di);
  while ((on = (OctNode *) dict_iter_objs (nmap, di))) {
    double mesh_size;
    mesh_size = B->SBodies->HFunction (on->x); /* RAW a hack; need to find
                                                  subdody enclosing on!
                                                */
    if (!VDT_add_pt ((void *) on->Number, on->x[0], on->x[1], on->x[2],
                     4 * mesh_size, vdt)) {
      fprintf (stderr, "Failed in vdt_add_pt\n");
      exit (1);
    }
#if VDT_PRINT
    printf ("pt %d %g %g %g %g\n", on->Number, on->x[0], on->x[1], on->x[2],
            4 * mesh_size);
#endif
  }
  dict_del_iterator (di);
} 


static int
count_nodes (VDT vdt)
{
  void *id;
  double x, y, z;
  double ms;
  void *pt;
  int count = 0;
  pt = NULL;
  while ((pt = VDT_next_pt (&id, &x, &y, &z, &ms, pt, vdt)))
    count++;
  return count;
}

static void
redefine_pts (Body *B, VDT vdt, dict_dictionary nmap) 
{
  void *id;
  double x[3];
  double ms;
  void *pt;
  OctNode *on;
  int n;
  int numnodes;
  
  nodes = count_nodes (vdt);
  n     = nodes * spatial_dimension * sizeof (double);
  coordinates = malloc (n);
  if (!coordinates) {
    fprintf (stderr, "Failed in realloc\n");
    exit (1);
  }

  numnodes = 0;
  pt = NULL;
  while ((pt = VDT_next_pt (&id, &x[0], &x[1], &x[2], &ms, pt, vdt))) {
#if VDT_PRINT
    printf ("! output pt %d %g %g %g\n",
            (int) id, x[0], x[1], x[2]);
#endif
    on = find_on (nmap, (int) id);
    if (!on) {
      on = Create_PNode ((int) id, x, 0., white);
      PR_Insert (&on, &OT);	
      update_nmap (nmap, (int) id, on);
    }
    numnodes++;
  }
  printf ("# of nodes: %d\n", numnodes);
} 


static double
min_dih_angle (VDT vdt)
{
  int num_non_positive_vol_tets;
  int num_small_dihed_angle_tets;
  int quality_measure;
  double
    min_quality, max_quality, geom_mean_quality,
    rms_quality, min_volume, max_volume, min_dihed_angle, max_dihed_angle,
    avg_dihed_angle;
  
  VDT_get_quality (vdt,
                   &num_non_positive_vol_tets,
                   &num_small_dihed_angle_tets,
                   &quality_measure,
                   &min_quality,
                   &max_quality,
                   &geom_mean_quality,
                   &rms_quality,
                   &min_volume,
                   &max_volume,
                   &min_dihed_angle,
                   &max_dihed_angle,
                   &avg_dihed_angle);

  return min_dihed_angle;
}



static VDT
re_triangulate (VDT vdt)
{
  void *id;
  double x, y, z;
  void *pt;
  void *bsf;
  void *ptid[3];
  void *surfid;
  void *tetid[2];
  void *regid[2];
  VDT newvdt;
  void *tet;
  void *tetptid[4];
  void *adjtet[4];
  double ms;
  char buf[1024];
  double min_angle, max_angle, angles[6];
  int result;
  vdt_option_t op;

  newvdt = VDT_new_mesher ();
  if (!newvdt) {
    fprintf (stderr, "Could not create a new mesher\n");
    exit (1);
  }

  op.option  = VDT_OPTION_GEN_INTERIOR_PTS;
  op.data.iv = FALSE; /* got all the points I need */
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_GEN_NEAR_SURF_PTS;
  op.data.iv = FALSE; /* got all the points I need */
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_DEF_MESH_SIZE;
  op.data.dv = VDT_def_mesh_size (vdt);
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_EVAL_MESH_SIZE_FUNC;
  op.data.pv = (void *) eval_mesh_size;
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_NEAR_SURF_TEST;
  op.data.iv = VDT_NEAR_SURF_TEST_DIST;
  SET_VDT_OPT (newvdt, op);

  pt = NULL;
  while ((pt = VDT_next_pt (&id, &x, &y, &z, &ms, pt, vdt))) {
    if (!VDT_add_pt (id, x, y, z, ms, newvdt)) {
      fprintf (stderr, "%s", VDT_last_error_msg (newvdt));
      exit (1);
    }
  }

  bsf = NULL;
  while ((bsf = VDT_next_bsf (&id, ptid, &surfid, regid, tetid, bsf, vdt))) {
    if (!VDT_add_bsf (ptid, surfid, regid, newvdt)) {
      fprintf (stderr, "%s", VDT_last_error_msg (newvdt));
      exit (1);
    }
   }

  tet = NULL;
  while ((tet = VDT_next_tet (tetptid, &regid[0], adjtet, tet, vdt))) {
    VDT_tet_get_angle_quality (tet, &min_angle, &max_angle, angles);
    if (min_angle < sliver_dihed_angle) {
      if (VDT_tet_get_circumcenter (tet, vdt, &x, &y, &z)) {
        if (!VDT_add_pt_with_check (id, x, y, z,
                                    VDT_def_mesh_size (vdt), newvdt)) {
          fprintf (stderr, "%s", VDT_last_error_msg (newvdt));
          exit (1);
        }
      }
    }
  }
  
  VDT_delete_mesher (vdt);

  result = VDT_triangulate (newvdt);
  if (!result) {
    fprintf (stderr, "%s", VDT_last_error_msg (newvdt));
    exit (1);
  } 
 
  VDT_summary (newvdt);

  return newvdt;
}



static VDT 
triangulate (Body *B, VDT vdt,
             dict_dictionary nmap, dict_dictionary tmap)
{
  void *tet;
  void *tetptid[4];
  void *reg_id;
  void *adjtet[4];
  TetraElem *TE = NULL, *adjTE;
  OctNode *n1, *n2, *n3, *n4;
  int i;
  SBody *SB;
  int count = 10;
#define CHKON(on) if (!on) {fprintf (stderr, "No such octnode\n"); exit (1);}
  
  if (!VDT_triangulate (vdt)) {
    fprintf (stderr, "Failed in VDT_triangulate\n");
    exit (1);
  }

  VDT_summary (vdt);

  while (min_dih_angle (vdt) < sliver_dihed_angle &&
         count-- > 0) {
    vdt = re_triangulate (vdt);
  }

  redefine_pts (B, vdt, nmap);

  /* clean sub-bodies of elements */
  for (SB=B->SBodies;SB;SB=SB->Link) {
    SB->TetraElements = TE = NULL;
    SB->NTetra = 0;
    /* load tets */
    tet = NULL;
    while ((tet = VDT_next_tet (tetptid, &reg_id, adjtet, tet, vdt))) {
      if (SB == (SBody *)reg_id) { /* handling this subbody at this point */
        n1 = find_on (nmap, (int) tetptid[0]); CHKON (n1);
        n2 = find_on (nmap, (int) tetptid[1]); CHKON (n2);
        n3 = find_on (nmap, (int) tetptid[2]); CHKON (n3);
        n4 = find_on (nmap, (int) tetptid[3]); CHKON (n4);
        TE = AddTetraElem (n1,n2,n3,n4,
                           adjtet[0], adjtet[1], adjtet[2], adjtet[3], TE, SB);
        if (!SB->TetraElements) SB->TetraElements = TE;
        dict_add_obj_handle_name (tmap, TE, tet);
#if 0
        fprintf (stderr, "added tet %p {%d,%d,%d,%d}\n",
                 tet,
                 TE->ElVertex[0]->Number, 
                 TE->ElVertex[1]->Number, 
                 TE->ElVertex[2]->Number, 
                 TE->ElVertex[3]->Number);
#endif
      } /* if */
    } /* while */
  } /* for */
  
  /* complete tet adjacencies */
  for (SB=B->SBodies;SB;SB=SB->Link) {
    TE = SB->TetraElements;
    do {
      for (i = 0; i < 4; i++) {
        adjTE = dict_find_obj_with_handle_name (tmap, TE->Adj[i]);
        TE->Adj[i] = adjTE;
      }
    }  while ((TE = TE->Link));
  }

  return vdt;
}


typedef struct ge_orient_data_t {
  OctNode     *fn, *ln;
  CkitPoint3   tangent;
  GFace       *gfonleft;
  EdgeElem    *FirstEE;
  EdgeElem    *LastEE;
}              ge_orient_data_t;

static void
clean_topo (Body *B, dict_dictionary gemap)
{
  Shell *SH;
  Face  *F;
  GFace *GF;
  GEdge *GE;
  FaceElem *FE;
  ElemSides *side;
  EdgeElem *EE;
  Loop *L;
  LoopEdge *LE;
  ge_orient_data_t *geod;
  SBody *SB;

  for (SB=B->SBodies;SB;SB=SB->Link) {
    for (SH=SB->Shells;SH;SH=SH->Link) {
      for (F=SH->Faces;F;F=F->Link) {
        GF = GFaces[F->GlobalFace];
        for (FE=GF->FaceElements;FE;FE=FE->Link) {
          free (FE);
        }
        GF->FaceElements = NULL;
        for (side=GFaces[F->GlobalFace]->Side;side;side=side->Last) {
          free (side);
        }
        GF->Side = NULL;
        for (L = GF->Loops; L ; L = L->Link) {
          for (LE = L->Edges; LE ; LE = LE->Link) {
            GE = GEdges[LE->GlobalEdge];
            if (GE->FirstElem) {
              geod = NEWP_OBJ (ge_orient_data_t);
              geod->fn        = GE->FirstElem->ElVertex[0];
              geod->ln        = GE->LastElem->ElVertex[2];
              geod->FirstEE   = GE->FirstElem;
              geod->LastEE    = GE->LastElem;
              EE = GE->FirstElem;
              VECMAC_3_SCALAR (geod->tangent, =,
                               EE->ElVertex[2]->x[0] - EE->ElVertex[0]->x[0],
                               EE->ElVertex[2]->x[1] - EE->ElVertex[0]->x[1],
                               EE->ElVertex[2]->x[2] - EE->ElVertex[0]->x[2]);
              geod->gfonleft = GFaces[GE->AdjGlobalFaces[0]];
              dict_add_obj_handle_name (gemap, geod, GE);
              GE->FirstElem = NULL;
              GE->LastElem  = NULL;
            }
          }
        }
      }
    }
  }
}  

static void
setup_winged_edge (Body *B)
{
  Shell *SH;
  Face  *F;
  GFace *GF;
  SBody *SB;

  for (SB=B->SBodies;SB;SB=SB->Link) {
    for (SH=SB->Shells;SH;SH=SH->Link) {
      for (F=SH->Faces;F;F=F->Link) {
        GF = GFaces[F->GlobalFace];
        WingedEdge (GF);
      }
    }
  }
}

static void
pr_t (smesh_triangle_p t)
{
  int vid[3];
  int j, k;
  smesh_vertex_p v[6];
  smesh_edge_p e[3];

  smesh_t_get_v (t, v);
  for (j = 0; j < 3; j++) vid[j] = smesh_v_get_id (v[j]);
  fprintf (stderr, "smesh t = {%d,%d,%d}\n", vid[0], vid[1], vid[2]);
  smesh_t_get_e (t, e);
  for (k = 0; k < 3; k++) {
    smesh_e_get_v (e[k], v);
    for (j = 0; j < 2; j++) vid[j] = smesh_v_get_id (v[j]);
    fprintf (stderr, "smesh e[%d] = {%d,%d}\n", k, vid[0], vid[1]);
  }
}

static void
get_adj (smesh_triangle_p t, smesh_triangle_p adjt[3])
{
  smesh_edge_p e[3];
  smesh_triangle_p ot;
  int i;
  int ni[3] = {2,0,1};
  
  smesh_t_get_e (t, e); 
  for (i = 0; i < 3; i++) {
    adjt[ni[i]] = NULL;
    ot = NULL;
    while ((ot = smesh_e_get_next_t (e[i], ot))) {
      if (ot != t) {
        if (adjt[ni[i]]) {
          fprintf (stderr, "Non-manifold edge\n");
          exit (1);
        }
        adjt[ni[i]] = ot;
      }
    }
  }
}

static int
ge_uses_nodes (ge_orient_data_t *geod, OctNode *n0, OctNode *n1)
{
  int had[2] = {0, 0};
  EdgeElem *EE;

  for (EE = geod->FirstEE; EE; EE = EE->RLink) {
    if (EE->ElVertex[0] == n0) had[0]++;
    if (EE->ElVertex[2] == n0) had[0]++;
    if (EE->ElVertex[0] == n1) had[1]++;
    if (EE->ElVertex[2] == n1) had[1]++;
  }
  /* kludge: this is not a sufficient condition */
  return (had[0] > 0 || had[1] > 0);
}

static int
has_a_double (GEdge *iGE)
{
  int i;
  GEdge *GE = NULL;
  GFace *gf[2], *igf[2];
  
  igf[0] = GFaces[iGE->AdjGlobalFaces[0]];
  igf[1] = GFaces[iGE->AdjGlobalFaces[0]];
  
  for (i=0;i<nedges;i++) {
    gf[0] = GFaces[GE->AdjGlobalFaces[0]];
    gf[1] = GFaces[GE->AdjGlobalFaces[0]];
    if (   (igf[0] == gf[0] && igf[1] == gf[1])
        || (igf[0] == gf[1] && igf[1] == gf[0])
        )
      return TRUE;
  }
  return FALSE;
}



static int
find_ges (GFace *gf0, GFace *gf1, OctNode *n0, OctNode *n1,
          dict_dictionary gemap, GEdge *ge[])
{
  int i;
  GEdge *GE;
  ge_orient_data_t *geod;
  int n;

  n = 0;
  for (i=0;i<nedges;i++) {
    GE = GEdges[i];
    geod = dict_find_obj_with_handle_name (gemap, GE);
    if (geod) {
      if ((GFaces[GE->AdjGlobalFaces[0]] == gf0 &&
           GFaces[GE->AdjGlobalFaces[1]] == gf1)
          ||
          (GFaces[GE->AdjGlobalFaces[0]] == gf1 &&
           GFaces[GE->AdjGlobalFaces[1]] == gf0)) {
        /* check if the face on the left is correct */
        if (GFaces[GE->AdjGlobalFaces[0]] == geod->gfonleft) {
          ge[n++] = GE;
        }
      }
    }
  }
  return n;
}


static EdgeElem *
first_ee (GEdge *GE, ge_orient_data_t *geod)
{
  EdgeElem *EE;
  GFace *GF;
  FaceElem *FE;
  CkitPoint3 v;
  OctNode *f, *s;

  for (EE = GE->FirstElem; EE; EE = EE->RLink) {
    if (EE->ElVertex[0] == geod->fn || EE->ElVertex[2] == geod->fn) {
      if (EE->ElVertex[0] != geod->fn) {
        OctNode *ton = EE->ElVertex[0];
        EE->ElVertex[0] = EE->ElVertex[2];
        EE->ElVertex[2] = ton;
      }
      f = EE->ElVertex[0];
      s = EE->ElVertex[2];
      VECMAC_3_SCALAR (v, =,
                       s->x[0] - f->x[0],
                       s->x[1] - f->x[1],
                       s->x[2] - f->x[2]);
      if (VECMAC_DOT (v, geod->tangent) > 0) {
        FE = (FaceElem *) EE->AFE[0].FE;
        GF = (GFace *) FE->PQ3D;
        if (GF != GFaces[GE->AdjGlobalFaces[0]]) {
          TypeAdj ta = EE->AFE[0]; EE->AFE[0] = EE->AFE[1]; EE->AFE[1] = ta;
        }
        return EE;
      }
    }
  }
  fprintf (stderr, "No first edge\n");
  exit (1);
  return NULL;
}

static EdgeElem *
next_ee (GEdge *GE, EdgeElem *prevEE, ge_orient_data_t *geod)
{
  EdgeElem *EE;
  GFace *GF;
  FaceElem *FE;
  
  for (EE = GE->FirstElem ; EE; EE = EE->RLink) {
    if (EE != prevEE
        && (   EE->ElVertex[0] == prevEE->ElVertex[2]
            || EE->ElVertex[2] == prevEE->ElVertex[2])) {
      if (EE->ElVertex[2] == prevEE->ElVertex[2]) {
        OctNode *ton = EE->ElVertex[0];
        EE->ElVertex[0] = EE->ElVertex[2];
        EE->ElVertex[2] = ton;
      }
      if (EE->ElVertex[0] == geod->fn)/* the starting node on a cyclic edge */
        return NULL;
      FE = (FaceElem *) EE->AFE[0].FE;
      GF = (GFace *) FE->PQ3D;
      if (GF != GFaces[GE->AdjGlobalFaces[0] - 1]) {
        TypeAdj ta = EE->AFE[0]; EE->AFE[0] = EE->AFE[1]; EE->AFE[1] = ta;
      }
      return EE;
    }
  }
  return NULL;
}

static OctNode *
other_on (EdgeElem *EE, OctNode *on)
{
  return (EE->ElVertex[0] == on ? EE->ElVertex[2] : EE->ElVertex[0]);
}

static void
copy_ee (EdgeElem *src, EdgeElem *dest)
{
  int i;
  if (!src->ElVertex[0]) {
    fprintf (stderr, "NULL vertex\n");
    exit (1);
  }
  if (!src->ElVertex[2]) {
    fprintf (stderr, "NULL vertex\n");
    exit (1);
  }
  dest->ElVertex[0] = src->ElVertex[0];
  dest->ElVertex[2] = src->ElVertex[2];
  dest->AFE = NEWP_OBJ_ARRAY (src->naf, TypeAdj); 
  if (!dest->AFE) { fprintf (stderr, "Not enough memory\n"); exit (1); }
  for (i = 0; i < src->naf; i++) {
    dest->AFE[i]      = src->AFE[i];
  }
}

static OctNode *
mid_node (OctNode *n0, OctNode *n2)
{
  double x[3];

  x[0] = (n0->x[0] + n2->x[0]) / 2;
  x[1] = (n0->x[1] + n2->x[1]) / 2;
  x[2] = (n0->x[2] + n2->x[2]) / 2;
  return Create_PNode (0, x, (n0->Size + n2->Size) / 2, white);  
}

static void
pr_ge (GEdge *GE)
{
  EdgeElem *EE;
  fprintf (stderr, "GE %x\n", GE);
  for (EE=GE->FirstElem; EE; EE=EE->RLink)
    fprintf (stderr, "EE %x {%d,%d}\n",
             EE, EE->ElVertex[0]->Number, EE->ElVertex[2]->Number);
}

static void
clean_ge (GEdge *GE)
{
  EdgeElem *EE;
  for (EE = GE->FirstElem; EE ; EE = EE->RLink) {
    free (EE);
  }
}

static void 
reorder_ge (GEdge *GE, dict_dictionary gemap)
{
  EdgeElem *EE, *prevEE;
  ge_orient_data_t *geod;
  EdgeElem *new_EE, *new_EE_list;

  geod = dict_find_obj_with_handle_name (gemap, GE);
  if (!geod) {
    return; /* not yet seen by clean_topo() */
  }

  prevEE = NULL;
  EE  = first_ee (GE, geod);
  if (!EE) {
    fprintf (stderr, "No first edgeelem\n");
    exit (1);
  }
  new_EE_list = new_EE = AddEdgeElem (NULL);
  new_EE->ElVertex[1] = mid_node (EE->ElVertex[0], EE->ElVertex[2]);
  copy_ee (EE, new_EE);
  while (EE) {
    prevEE = EE;
    EE = next_ee (GE, prevEE, geod);
    if (EE) {
      new_EE = AddEdgeElem (new_EE);
      new_EE->ElVertex[1] = mid_node (EE->ElVertex[0], EE->ElVertex[2]);
      copy_ee (EE, new_EE);
    }
  }
  clean_ge (GE);
  GE->FirstElem = new_EE_list;
  GE->LastElem  = new_EE;
}

static void
reorder_ges (dict_dictionary gemap)
{
  int i;
  GEdge *GE;

  for (i=0;i<nedges;i++) {
    GE = GEdges[i];
    if (GE->FirstElem) {
      reorder_ge (GE, gemap);
    } else {
      clean_ge (GE);
    }
  }
}

static void
order_on (FaceElem *FE, OctNode *on[2])
{
#define SWAPON(on) { OctNode *tn = on[0]; on[0] = on[1]; on[1] = tn; }
  /* the FE on the left is assumed to use the edge vertices in reverse order */
  if        (FE->n[0] == on[1]) {
    if (FE->n[1] != on[0]) { SWAPON (on); }
  } else if (FE->n[1] == on[1]) {
    if (FE->n[2] != on[0]) { SWAPON (on); }
  } else if (FE->n[2] == on[1]) {
    if (FE->n[0] != on[0]) { SWAPON (on); }
  } else {
    fprintf (stderr, "Suspicious FE/on\n");
    exit (1);
  }
}

static void
find_edges (smesh_triangulation_p smesh,
            dict_dictionary nmap, dict_dictionary gemap)
{
  smesh_edge_p e;
  smesh_triangle_p t;
  smesh_triangle_p adjt[2];
  smesh_vertex_p v[6];
  int n, m, ne;
  FaceElem *FE[2];
  GFace *GF[2];
  GEdge **GES;
  EdgeElem *EE;
  OctNode *on[2];
  int j, vid[6];
  int which;
#define FEN(FE, an) (FE)->n[an]->Number
#define FEHAS(FE, n0, n1, n2)                                              \
  (   ((FEN (FE, 0) == n0) || (FEN (FE, 1) == n0) || (FEN (FE, 2) == n0))  \
   && ((FEN (FE, 0) == n1) || (FEN (FE, 1) == n1) || (FEN (FE, 2) == n1))  \
   && ((FEN (FE, 0) == n2) || (FEN (FE, 1) == n2) || (FEN (FE, 2) == n2)))

  GES = NEWP_OBJ_ARRAY (nedges, GEdge *);
  if (!GES) { fprintf (stderr, "Not enough memory\n"); exit (1); }

  e = NULL;
  while ((e = smesh_next_e (smesh, e))) {
    n = 0;
    t = NULL;
    while ((t = smesh_e_get_next_t (e, t))) {
      adjt[n++] = t;
      if (n > 2) {
        fprintf (stderr, "More than two triangles at an edge\n");
        exit (1);
      }
    }
    FE[0] = (FaceElem *) smesh_t_get_ap (adjt[0]);
    FE[1] = (FaceElem *) smesh_t_get_ap (adjt[1]);
    GF[0] = (GFace *) FE[0]->PQ3D;
    GF[1] = (GFace *) FE[1]->PQ3D;
    if (GF[0] != GF[1]) {
      smesh_e_get_v (e, v);
      for (j = 0; j < 2; j++) {
        vid[j] = smesh_v_get_id (v[j]);
        on[j] = find_on (nmap, vid[j]);
        if (!on[j]) {
          fprintf (stderr, "Octnode not found\n");
          exit (1);
        }
      }
      /* find the global edge */
      ne = find_ges (GF[0], GF[1], on[0], on[1], gemap, GES);
      for (m = 0; m < ne; m++) {
        if (!GES[m]) {
          fprintf (stderr, "Global edge not found\n");
          exit (1);
        }
        /* swap adjacent faces if not in agreement with global faces */
        if (GFaces[GES[m]->AdjGlobalFaces[0] - 1] != GF[0]) {
          { GFace *agf = GF[0]; GF[0] = GF[1]; GF[1] = agf; }
          { FaceElem *afe = FE[0]; FE[0] = FE[1]; FE[1] = afe; }
          { smesh_triangle_p at = adjt[0]; adjt[0] = adjt[1]; adjt[1] = at; }
        }
        EE = AddEdgeElem (GES[m]->LastElem);
        if (!GES[m]->FirstElem) GES[m]->FirstElem = EE;
        GES[m]->LastElem = EE;
        order_on (FE[0], on);
        EE->ElVertex[0] = on[0];
        EE->ElVertex[2] = on[1];
        EE->naf         = GES[m]->NAdjF;
        if (EE->naf != 2) { fprintf (stderr, "Non-manifold topology; cannot handle properly\n"); exit (1); }
        EE->AFE = NEWP_OBJ_ARRAY (EE->naf, TypeAdj); 
        if (!EE->AFE) { fprintf (stderr, "Not enough memory\n"); exit (1); }
        for (j = 0; j < 2; j++) {
          smesh_t_get_v (adjt[j], v);
          vid[0] = smesh_v_get_id (v[0]);
          vid[1] = smesh_v_get_id (v[1]);
          vid[2] = smesh_v_get_id (v[2]);
          EE->AFE[j].FE    = FE[j];
          if      (   (   EE->ElVertex[0]->Number == vid[0]
                       && EE->ElVertex[2]->Number == vid[1])
                   || (   EE->ElVertex[0]->Number == vid[1]
                       && EE->ElVertex[2]->Number == vid[0]))
            which = 0;
          else if (   (   EE->ElVertex[0]->Number == vid[1]
                       && EE->ElVertex[2]->Number == vid[2])
                   || (   EE->ElVertex[0]->Number == vid[2]
                       && EE->ElVertex[2]->Number == vid[1]))
            which = 1;
          else
            which = 2;
          EE->AFE[j].which = which;
        } /* for (j = 0; j < 2; j++) */
      } /* for (m = 0; m < ne; m++) */
    } /* if (GF[0] != GF[1]) */
  } /* while ((e = smesh_next_e (smesh, e))) */

  reorder_ges (gemap);
}

static int
which_adj_tet (FaceElem *FE, TetraElem *TE)
{
  int vertex_on_face[4 /* face # */][3 /* vertex index */] = {
/* 0 */ { 0, 2, 1 },
/* 1 */ { 0, 1, 3 },
/* 2 */ { 1, 2, 3 },
/* 3 */ { 2, 0, 3 }
      };
  int which;
#define TET_FACE_USES_PT(te, face, p)                   \
 (   p == te->ElVertex[vertex_on_face[face][0]]         \
  || p == te->ElVertex[vertex_on_face[face][1]]         \
  || p == te->ElVertex[vertex_on_face[face][2]])
#define TET_FACE_USES_PTS(te, face, p0, p1, p2) \
  (   TET_FACE_USES_PT(te, face, p0)            \
   && TET_FACE_USES_PT(te, face, p1)            \
   && TET_FACE_USES_PT(te, face, p2))
  OctNode *n0 = FE->n[0];
  OctNode *n1 = FE->n[1];
  OctNode *n2 = FE->n[2];
  if (!TE) {
    fprintf (stderr, "NULL tet\n"); exit (1);
  }
  for (which = 0; which < 4; which++) {
    if     (TET_FACE_USES_PTS (TE, which, n0, n1, n2))
      return which;
  }
  fprintf (stderr, "Wrong tet-face correspondance\n");
  return -1;
}

static void
redefine_surf_discret (Body *B, VDT vdt,
                       dict_dictionary nmap, dict_dictionary tmap)
{
  void *bsf;
  void *id;
  void *ptid[3];
  CkitPoint3 locs[3];
  OctNode *on[3];
  int i;
  GFace *GF;
  void *surf_id;
  smesh_triangulation_p smesh;
  smesh_triangle_p t;
  smesh_triangle_p adjt[3];
  int vid[3];
  FaceElem *FE;
  FaceElem *FFE;
  dict_dictionary gemap;
  void *tet[2];
  void *reg_id[2];

  smesh = smesh_new_triangulation (NULL, 0.00001);
  if (!smesh) {
    fprintf (stderr, "Failed to create smesh\n");
    exit (1);
  }
  
  /* clean the old face and side elements  */
  gemap = dict_create_dictionary ();
  clean_topo (B, gemap);

  /* build a connectivity model of the surfaces */
  bsf = NULL;
  while ((bsf = VDT_next_bsf (&id, ptid, &surf_id, reg_id, tet, bsf, vdt))) {
#if VDT_PRINT
    printf ("! output bsf %d %d %d %d\n",
            (int) ptid[0],
            (int) ptid[1],
            (int) ptid[2],
            (int) surf_id); 
#endif
    GF = (GFace *)surf_id;
    FE = GF->FaceElements;
    /* swap two points to change to adlib orientation */
    { void *tid = ptid[1]; ptid[1] = ptid[2]; ptid[2] = tid; }
    for (i = 0; i < 3; i++) {
      on[i] = find_on (nmap, (int) ptid[i]);
      if (!on[i]) {
        fprintf (stderr, "Octnode not found\n");
        exit (1);
      }
      vid[i] = on[i]->Number;
      VECMAC_3_SCALAR (locs[i], =, on[i]->x[0], on[i]->x[1], on[i]->x[2]);
      if (!smesh_add_v (smesh, vid[i], &locs[i])) {
        fprintf (stderr, "Failed to add node to smesh\n");
        exit (1);
      }
    } 
    FE = AddFaceElem (on[0], on[1], on[2], NULL, NULL, NULL, NULL);
    if (tet[0]) {
      FE->ATE[0].TE = dict_find_obj_with_handle_name (tmap, tet[0]);
      FE->ATE[0].which = which_adj_tet (FE, FE->ATE[0].TE);
    }
    if (tet[1]) {
      FE->ATE[1].TE = dict_find_obj_with_handle_name (tmap, tet[1]);
      FE->ATE[1].which = which_adj_tet (FE, FE->ATE[1].TE);
    } 
    FFE = GF->FaceElements;
    if (FFE) {FFE->Rink = FE;}
    FE->Link  = FFE;
    GF->FaceElements = FE;
    if (!smesh_add_t_with_ap (smesh, vid, FE)) {
      fprintf (stderr, "Failed to add bsf to smesh\n");
      exit (1);
    }
    FE->PQ3D = GF;
  }
  
  /* extract face elements with the connectivities */
  t = NULL;
  while ((t = smesh_next_t (smesh, t))) {
    get_adj (t, adjt);
    FE = (FaceElem *) smesh_t_get_ap (t);
    for (i = 0; i < 3; i++) {
      FE->a[i] = (FaceElem *) smesh_t_get_ap (adjt[i]);
      if (!FE->a[i]) {
        fprintf (stderr, "No adjacent FE\n");
        exit (1);
      }
    }
  }

  /* extract the edge elements */
  find_edges (smesh, nmap, gemap);

  /* winged edge */
  setup_winged_edge (B);
  
  smesh_destroy_triangulation (smesh);
  {
    dict_iterator diter;
    EdgeElem *EE;
    ge_orient_data_t *geod;
    diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((geod = dict_iter_objs (gemap, diter))) {
      for (EE = geod->FirstEE; EE; EE = EE->RLink) {
        free (EE->ElVertex[1]);
        free (EE);
      }
      free (geod);
    }
    dict_del_iterator (diter);
  }
  dict_destroy_dictionary (gemap);
}


void
MeshBRepVDT (void) 
{
  Body *B;
  VDT vdt;
  dict_dictionary nmap;
  dict_dictionary tmap;
  vdt_option_t vdt_option;

  CompleteGEdgesInfo ();
  
  for (B=BRep;B;B=B->Link) {
    
    vdt = VDT_new_mesher ();
    if (!vdt) {
      fprintf (stderr, "Failed in VDT_new_mesher\n");
      exit (1);
    }  
#if 0
    vdt_option.option = VDT_OPTION_DUMP_DATA;
    vdt_option.data.iv = TRUE;
    SET_VDT_OPT (vdt, vdt_option);
#endif
#if 1
    vdt_option.option = VDT_OPTION_REPORT_OPTS;
    vdt_option.data.iv = TRUE; 
    SET_VDT_OPT (vdt, vdt_option);
#endif
#if VDT_GEN_PTS 
    vdt_option.option = VDT_OPTION_GEN_INTERIOR_PTS;
    vdt_option.data.iv = TRUE; 
    SET_VDT_OPT (vdt, vdt_option);
#endif
    vdt_option.option = VDT_OPTION_BDRY_PT_SHIFT_MULT;
    vdt_option.data.dv = 0.001;
    SET_VDT_OPT (vdt, vdt_option);
    vdt_option.option = VDT_OPTION_GEN_PT_SHIFT_MULT;
    vdt_option.data.dv = 0.0001;
    SET_VDT_OPT (vdt, vdt_option);
    
    nmap = create_nmap ();
    
    define_pts (B, vdt, nmap); 
    
    define_bsfs (B, vdt);
    
    tmap = dict_create_dictionary ();
    vdt = triangulate (B, vdt, nmap, tmap);
    
    redefine_surf_discret (B, vdt, nmap, tmap);
    
    VDT_delete_mesher (vdt);
    
    dict_destroy_dictionary (nmap);
    dict_destroy_dictionary (tmap);
  }
} 


