/* 
      Variational Delaunay Triangulation
             (C) 1998-2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtaf.c,v 1.2 2004/04/23 19:11:16 cvs Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdtaf.c,v 1.2 2004/04/23 19:11:16 cvs Exp $";
DEF_VDT_MODULE_ID (vdtaf) 
 

#define GET_CIRCUMSPHERE(tetpts, csc) vdt_4_loc_circumsphere (tetpts, csc)


#if GIFACE
#  define ANNOUNCE_ACTION(action)               \
    if (vdt->gdebug && vdt->pdebug) {           \
      char buf[128];                            \
      if (vdt->pdebug) logit (action);          \
      sprintf (buf, "   Bsf pt={%d,%d,%d}",     \
               (int) bsf->pt[0]->id,            \
               (int) bsf->pt[1]->id,            \
               (int) bsf->pt[2]->id);           \
      logit (buf);                              \
      vdt_generate_graphics (vdt);              \
      EMHiliteGraphics (vdt->gmodel, bsf->go);  \
      vdt_refresh_display (vdt);                \
      ESIEventLoop (vdt->pdebug, action);       \
    }
#  define SHOW_ACTION_RESULT                    \
  if (vdt->gdebug && vdt->pdebug) {             \
    vdt_generate_graphics (vdt);                \
    EMUnhiliteGraphics (vdt->gmodel, bsf->go);  \
    vdt_refresh_display (vdt);                  \
  }
#else
#  define ANNOUNCE_ACTION(action)
#  define SHOW_ACTION_RESULT
#endif



static vdt_pt_t *
pick_pt (vdt_t *vdt,
         int any_side,
         CkitPoint3 tetloc[4],
         CkitPoint3 *normal,
         CkitPoint3 *cscenter,
         double *csrad)
{
  int i, nap = vdt->num_active_points, si;
  int picked;
  CkitPoint3 pickedcsc;
  double pickedcsrad2;
  CkitPoint3 csc, d, n = *normal;
  double csrad2, d2;
  
  picked = -1;

  /* initialize the circumsphere center and radius */
  si = nap;
  FOR (i, nap) {
    if (vdt->active_pt[i]) { 
      VECMAC_3_VEC (d, =, vdt->active_loc[i], -, tetloc[0]);
      if (any_side || VECMAC_DOT (d, n) > 0) {
        VECMAC_2_VEC (tetloc[3], =, vdt->active_loc[i]);
        if (GET_CIRCUMSPHERE (tetloc, &csc)) {
          VECMAC_3_VEC (d, =, csc, -, tetloc[0]);
          pickedcsrad2 = csrad2 = VECMAC_DOT (d, d);
          pickedcsc    = csc;
          picked       = i;
          si           = i+1;
          break;
        }
      }
    }
  }

  for (i = si; i < nap; i++) {
    if (vdt->active_pt[i]) { 
      VECMAC_3_VEC (d, =, vdt->active_loc[i], -, tetloc[0]);
      if (any_side || VECMAC_DOT (d, n) > 0) {
        VECMAC_2_VEC (tetloc[3], =, vdt->active_loc[i]);
        VECMAC_3_VEC (d, =, csc, -, vdt->active_loc[i]);
        d2 = VECMAC_DOT (d, d);
        if (d2 < csrad2) { /* inside circumsphere? */
          if (GET_CIRCUMSPHERE (tetloc, &csc)) {
            VECMAC_3_VEC (d, =, csc, -, tetloc[0]);
            csrad2       = VECMAC_DOT (d, d);
            pickedcsc    = csc;
            pickedcsrad2 = csrad2;
            picked       = i;
          }
        } /* inside circumsphere? */
      } /* if (VECMAC_DOT (d, n) > 0) */
    } /* if (active */
  } /* FOR */

  if (picked >= 0) {
    *cscenter = pickedcsc;
    *csrad    = sqrt (pickedcsrad2);
    return vdt->active_pt[picked];
  } else {
    return NULL;
  }
}


static int
find_active_points (vdt_t *vdt, G3D_box_t *box, vdt_pt_t *facept[3])
{
  int i, nap;
  vdt_pt_t *pt;
  int true_nap;

  if (vdt_pts_in_box (vdt, box, VDT_ALL_PTS) < 4)
    return vdt->num_active_points - 3; /* removed the face points */

  true_nap = nap = vdt->num_active_points;
  FOR (i, nap) {
    pt = vdt->active_pt[i];
    /* deactivate the face points  */
    if (pt == facept[0] || pt == facept[1] || pt == facept[2]) {
      vdt->active_pt[i] = NULL;
      true_nap--;
    }
  }
  return true_nap;
}


vdt_pt_t *
vdt_af_find_fourth (vdt_pt_t *facept[3], vdt_t *vdt, int any_side)
{
  CkitPoint3 cscenter, a, b, n = {0,0,0}, fc, apptc;
  double csrad, bsize;
  int i;
  CkitPoint3 tetloc[4];
  G3D_box_t box;
  double mesh_size;
  vdt_pt_t *picked, *prevpicked, *prevprevpicked;
  int ntries = 0;
  
  /* init tet locations; the fourth later */
  FOR (i, 3) { VECMAC_2_VEC (tetloc[i], =, facept[i]->loc); }
  /* face normal  */
  VECMAC_3_VEC (a, =, tetloc[1], -, tetloc[0]);
  VECMAC_3_VEC (b, =, tetloc[2], -, tetloc[0]);
  VECMAC_CROSS (n, =, a, x, b);
  VECMAC_NORMALIZE (n);
  /* face circumcenter */
#if 0
  if (!vdt_3_pt_circumsphere (facept[0], facept[1], facept[2], &fc)) {
    warn ("No circumsphere for facet?");
    return NULL;
  }
#else
  VECMAC_3_VEC (fc, =, tetloc[0], +, tetloc[1]);
  VECMAC_2_VEC (fc, +=, tetloc[2]);
  VECMAC_SCALAR (fc, *=, 1.0/3.0);
#endif
  /* mesh size */
  mesh_size = vdt->def_mesh_size;
  mesh_size = max (mesh_size, facept[0]->mesh_size);
  mesh_size = max (mesh_size, facept[1]->mesh_size);
  mesh_size = max (mesh_size, facept[2]->mesh_size);
  /* approximate tet center */
  VECMAC_3_VEC (apptc, =, fc, +, 0.5 * mesh_size * n);

  /* start selection with the probable search extent */
  prevpicked = NULL; prevprevpicked = NULL;
  cscenter = apptc;
  csrad    = mesh_size;
 
 retry_for_another_box:
  if (++ntries > 100) {
    int i;
    warn ("Too many tries");
    printf ("! ff %d %d %d\n",
            (int) facept[0]->id,
            (int) facept[1]->id,
            (int) facept[2]->id);
    FOR (i, 3) { 
      printf ("! pt %d: %g %g %g\n", i,
              facept[i]->loc.x, facept[i]->loc.y, facept[i]->loc.z);
    }
#   if GIFACE
    vdt_generate_graphics (vdt);
    ESIEventLoop (TRUE, "too many tries");
#   endif
    ntries=0;
  }

  G3D_INIT_BOX (box, cscenter);
  bsize = 1.01 * csrad;
  G3D_INFLATE_BOX (box, bsize);
  
  if (find_active_points (vdt, &box, facept) < 1) {
    csrad *= 2;
    goto retry_for_another_box;
  }
  
  picked = pick_pt (vdt, any_side, tetloc, &n, &cscenter, &csrad);

  if (!picked) {
    /* No suitable point found; increase box, repeat search */
    csrad *= 2;
    goto retry_for_another_box;
  } else {
    if (picked == prevpicked)
      return picked;            /* got it */
    else {
      /* The circumsphere might have changed in pick_pt (); recompute
         to make sure you've got the correct answer
       */
      if (prevprevpicked == picked) /* detected a cycle */
        return NULL;
      else {
        prevprevpicked = prevpicked;
        prevpicked     = picked;
      }
      goto retry_for_another_box;
    }
  }
}





void
vdt_af_add_ff (vdt_t *vdt, FFP ff)
{
  add_to_head (vdt->fflist, ff);
#if GIFACE
  if (vdt->gdebug)
    vdt_ff_generate_graphics (ff, vdt);
#endif
}


static void
print_search_stat (vdt_t *vdt)
{
  char buf[256];
  logit        ("   Search statistics");
  sprintf (buf, "     Number of searches: %d", vdt->search_stat.num_searches);
  logit (buf);
  if (vdt->search_stat.num_searches > 0) {
    sprintf (buf, "     Average number of found points: %d", 
             (int) (vdt->search_stat.sum_found
                    / vdt->search_stat.num_searches));
    logit (buf);
    sprintf (buf, "     Max num of found points:        %d",
             vdt->search_stat.max_num_found);
    logit (buf);
  }
}



int
vdt_af_run (vdt_t *vdt, void *reg_id)
{
  vdt_tet_t *tet;
  int f;
  vdt_pt_t *tetpts[4];
  FFP ff, *nff;
 
  vdt->search_stat.num_searches  = 0;
  vdt->search_stat.sum_found     = 0;
  vdt->search_stat.max_num_found = 0;
  
  while (!list_empty (vdt->fflist)) {
    ff = (FFP ) get_from_head (vdt->fflist);
    if (ff->is_active) {
#if GIFACE
      if (vdt->gdebug)
        vdt_recolor_active_ff (vdt, ff);
#endif
      if (vdt->pdebug) {
        char buf[256];
        sprintf (buf, "Building on ff, pt = {%d, %d, %d}",
                 (int) ff->pt[0]->id, (int) ff->pt[1]->id, (int) ff->pt[2]->id);
        logit (buf);
      }
      tetpts[0] = ff->pt[0];
      tetpts[1] = ff->pt[1];
      tetpts[2] = ff->pt[2];
      tetpts[3] = vdt_af_find_fourth (tetpts, vdt, FALSE);
      if (!tetpts[3]) { /* failed to make a tet */
        if (!ff->ontet) errexit ("Wrong data");
        if (!ff->ontet->adjtet[ff->face]) errexit ("Wrong data");
        ff->ontet->adjtet[ff->face] = &vdt->stop_tet;
      } else { /* success */
        tet = vdt_tet_add (tetpts, reg_id, vdt); 
        FOR (f, 4) {          
          nff = vdt_tet_update_ffl (tet, f, vdt);
          if (nff->is_active) 
            vdt_af_add_ff (vdt, nff);
        }
        vdt_tet_update_topo (tet);
        if (vdt->ntets / vdt->incr_ntets * vdt->incr_ntets == vdt->ntets) {
          char buf[128];
          sprintf (buf, "   %9.0d tets", vdt->ntets);
          logit (buf);
        }
      } /* if (!tet->adjtet[f]) */
    }
    vdt_ff_delete (ff, vdt);
#if VDT_CUSTOM_SIGNAL_HANDLER
    {
      if (++(vdt->check_interrupt.ctet) > 1000) {
        if (vdt->check_interrupt.func) (*vdt->check_interrupt.func) (vdt);
        vdt->check_interrupt.ctet = 0;
      }
    }
#endif
  } 
  
  return TRUE;
}


static int
count_non_delaunay_bsfs (vdt_t *vdt)
{
  vdt_bsf_t *bsf;
  int count = 0;
  
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    if (!bsf->is_delaunay)
      count++;
  }
  return count;
}

static void
print_del_stat (vdt_t *vdt)
{
  char buf[256];
  logit        ("   Delaunayhood evaluation statistics");
  sprintf (buf, "     Number of Delaunayhood updates: %d",
           vdt->del_stat.num_bsf_updates);
  logit (buf);
  logit        ("     LP statistics");
  sprintf (buf, "       Solved %d LP's",
           (int) (vdt->del_stat.lp_stat.sum_evals));
  logit (buf);
  if (vdt->del_stat.lp_stat.sum_evals > 0) {
    sprintf (buf, "       Average num of unknowns: %d",
             (int) (vdt->del_stat.lp_stat.sum_unknowns
                    / vdt->del_stat.lp_stat.sum_evals));
    logit (buf);
    sprintf (buf, "       Max num of unknowns:     %d",
             vdt->del_stat.lp_stat.max_num_unknowns);
    logit (buf);
  }
}




static int
init_delaunay_prop (vdt_t *vdt)
{
  vdt_bsf_t *bsf;
  PTP tetpt[4];
  int n, i;
  char buf[256];
  ckit_timer_t *timer = ckit_timer_new ();

  ckit_timer_restart (timer);
  logit ("Computing Delaunay properties of surface facets");
  
  vdt->del_stat.num_bsf_updates          = 0;
  vdt->del_stat.lp_stat.sum_unknowns     = 0;
  vdt->del_stat.lp_stat.sum_evals        = 0;
  vdt->del_stat.lp_stat.max_num_unknowns = 0;

  vdt->search_stat.num_searches  = 0;
  vdt->search_stat.sum_found     = 0;
  vdt->search_stat.max_num_found = 0;

  n   = max (list_length (vdt->bsfl) / 10, 10000);
  i   = 0;
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    bsf->is_delaunay = FALSE;
    bsf->tetcsr      = 0;
    vdt_bsf_update_delaunay_prop (bsf, vdt, tetpt);
    if (++i % n == 0) {
      sprintf (buf, "   %d facets", i);
      logit (buf);
    }
  }
  
  vdt_report_time (timer, "Computation of Delaunay properties");

  print_search_stat (vdt);
  print_del_stat (vdt);
  
#if GIFACE
    if (vdt->giface) {
      vdt_generate_graphics (vdt);
    }
#endif

  ckit_timer_free (timer);

  return TRUE;
}

static int
update_delaunay_prop (vdt_t *vdt)
{
  vdt_bsf_t *bsf;
  CkitPoint3 d;
  PTP tetpt[4];
  double d2;

  if (vdt->new_pt) {
    bsf = NULL;
    while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
      if (bsf->tetcsr == 0) { /* A new bsf, or ... */
        vdt_bsf_update_delaunay_prop (bsf, vdt, tetpt); 
      } else {
        VECMAC_3_VEC (d, =, vdt->new_pt->loc, -, bsf->tetcsc);
        d2 = VECMAC_DOT (d, d);
        if (d2 <= bsf->tetcsr*bsf->tetcsr) { /* bsf invalidated by new point */
          vdt_bsf_update_delaunay_prop (bsf, vdt, tetpt); 
        }
      }
    }
  }

  return TRUE;
}

int
vdt_af_update_delaunay_prop (vdt_t *vdt)
{
  return update_delaunay_prop (vdt);
}


static int
num_intersections (vdt_t *vdt, CkitPoint3 *start, CkitPoint3 *dir, double vltol)
{
  FFP ff;
  int num_intersect = 0;
  G3D_tri_t tri;
  CkitPoint3 s;
  double lambda;
  double dlambda = vltol / VECMAC_NORM ((*dir));
#define EPS 0.0001
  
  ff = NULL;
  while ((ff = (FFP ) get_list_next (vdt->fflist, ff))) {
      /* need to check for intersections */
    tri.loc[0] = ff->pt[0]->loc;
    tri.loc[1] = ff->pt[1]->loc;
    tri.loc[2] = ff->pt[2]->loc;
    if (G3D_intersect_tri_and_line (&s, &lambda, start, dir, &tri, EPS)) {
      if (lambda >= 0 || (lambda < 0 && fabs (lambda) < dlambda)) 
        num_intersect++;
    }
  }

  return num_intersect;
}

static int
get_num_intersections (vdt_t *vdt, CkitPoint3 *start, CkitPoint3 *direction)
{
  CkitPoint3 dir = *direction;
  VECMAC_NORMALIZE (dir);
  return num_intersections (vdt, start, &dir, 0.0001 * vdt->def_mesh_size);
}

static FFP 
random_ff (vdt_t *vdt, int segnum)
{
  int randn = rand_0_to_1 () * 67;
  int i;
  FFP ff = NULL;
  int forever = TRUE;
  
  i = -1;
  while (forever) {
    loop_through_list (vdt->fflist, ff, FFP) {
      if (ff->is_active == segnum)
        if (++i == randn) return ff;
    }
  }
  return ff;
}

static void
centroid (FFP ff, CkitPoint3 *centroid)
{
  int j;
  VECMAC_SCALAR ((*centroid), =, 0);
  FOR (j, 3) {
    VECMAC_2_VEC ((*centroid), +=, 0.333333 * (ff->pt[j]->loc));
  }
}

static void
lin_n (CkitPoint3 *p0, CkitPoint3 *p1, CkitPoint3 *p2, CkitPoint3 *n)
{
  CkitPoint3 a, b;
  VECMAC_3_VEC (a, =, (*p1), -, (*p0));
  VECMAC_3_VEC (b, =, (*p2), -, (*p0));
  VECMAC_CROSS ((*n), =, a, x, b);
  if (VECMAC_NORM ((*n)) <= 0)
    errexit ("Zero normal");    
  VECMAC_NORMALIZE ((*n));
}

static void
compute_direction (FFP ff, CkitPoint3 *l, CkitPoint3 *direction)
{
  double randu = 0.1 * rand_0_to_1 (), randv = 0.1 * rand_0_to_1 ();
  CkitPoint3 n;
  G3D_plane_t plane;
  lin_n (&ff->pt[0]->loc, &ff->pt[1]->loc, &ff->pt[2]->loc, &n);
  G3D_plane_consts_p_and_n (&plane, l, &n);
  *direction = plane.n;
  VECMAC_2_VEC ((*direction), +=, randu * plane.u);
  VECMAC_2_VEC ((*direction), +=, randv * plane.v);
}

static void
start_and_direction (vdt_t *vdt, int segnum,
                     CkitPoint3 *start, CkitPoint3 *direction)
{
  FFP ff = random_ff (vdt, segnum);
  centroid (ff, start);
  compute_direction (ff, start, direction);
}

typedef void *PTID;

typedef struct ff_edge_t {
  PTID  anchor_pt;
  PTID  other_pt;
  FFP   ff[2];
  int   is_manifold;
}              ff_edge_t;

typedef struct ff_edge_t *FFEP;


static PTID
anchor_id (PTID p0, PTID p1)
{
  return (p0 < p1? p0: p1);
}

static PTID
other_id (PTID p0, PTID p1)
{
  return (p0 < p1? p1: p0);
}

static FFEP 
find_edge_in_list (LIST el, PTID p)
{
  FFEP e = NULL;
  while (e = (FFEP)get_list_next (el, e)) {
    if (e->other_pt == p) return e;
  }
  return NULL;
}

static FFEP
find_edge (dict_dictionary eld, PTID p0, PTID p1)
{
  LIST el = NULL;
  if (p0 < p1) el = dict_find_obj_with_handle_name (eld, p0);
  else         el = dict_find_obj_with_handle_name (eld, p1);
  if (!el) return NULL;
  else     return find_edge_in_list (el, other_id (p0, p1));
}

static FFEP
new_edge (PTID p0, PTID p1, FFP ff)
{
  FFEP e;
  e = NEWP_OBJ (ff_edge_t); CHECK_MEM (e);
  e->is_manifold = TRUE;        /* assumption */
  e->ff[0] = ff;                /* created by this ff */
  e->ff[1] = NULL;              /* this should get set later as we collect
                                 the info */
  e->anchor_pt = anchor_id (p0, p1);
  e->other_pt = other_id (p0, p1);
  return e;
}

static void
edge_attach_ff (FFEP e, FFP ff)
{
  if (e->ff[0] == ff || e->ff[1] == ff) return; /* already have this ff */
  if (e->ff[1] == NULL) {
    e->ff[1] = ff;
  } else {
    e->is_manifold = FALSE;     /* can't store: too many ff's along
                                   this edge; edge is non-manifold */
  }
}

static FFP
other_ff (FFEP e, FFP ff)
{
  if (!e->is_manifold) return NULL;
  if      (e->ff[0] == ff) return e->ff[1];
  else if (e->ff[1] == ff) return e->ff[0];
  else                     return NULL;
}

static void
destroy_edges (dict_dictionary eld)
{
  LIST el;
  dict_iterator di = dict_get_iterator ();
  CHECK_MEM (di);
  dict_reset_iterator (di);
  while (el = (LIST) dict_iter_objs (eld, di)) {
    free_list (el, AND_NODES);
  }
  dict_del_iterator (di);
  dict_destroy_dictionary (eld);
}

static void
add_edge (dict_dictionary eld, FFEP e, PTID pid)
{
  LIST el;
  el = dict_find_obj_with_handle_name (eld, pid);
  if (!el) {
    el = make_list (); CHECK_MEM (el);
    dict_add_obj_handle_name (eld, el, pid);
  }
  add_to_head (el, e);
}

static void
merge_edges (FFP ff, dict_dictionary eld)
{
  FFEP e;
#define MERGEE(n0, n1)                                                  \
  {                                                                     \
    e = find_edge (eld, ff->pt[n0]->id, ff->pt[n1]->id);                \
    if (!e) {                                                           \
      e = new_edge (ff->pt[n0]->id, ff->pt[n1]->id, ff);                \
      add_edge (eld, e, anchor_id (ff->pt[n0]->id, ff->pt[n1]->id));    \
    } else {                                                            \
      edge_attach_ff (e, ff);                                           \
    }                                                                   \
  }
  MERGEE (0, 1);
  MERGEE (1, 2);
  MERGEE (2, 0);
}

static dict_dictionary
create_edges (vdt_t *vdt)
{
  FFP ff;
  dict_dictionary ed = dict_create_dictionary ();
  CHECK_MEM (ed);
  ff = NULL;
  while (ff = (FFP) get_list_next (vdt->fflist, ff)) {
    merge_edges (ff, ed);
  }
  return ed;
}

#define NOSEG 0

static void
toggle_ff (FFP ff)
{
  PTP apt = ff->pt[1]; ff->pt[1] = ff->pt[2]; ff->pt[2] = apt;
}

static int
toggle_af_segment (vdt_t *vdt, int segnum)
{
  FFP ff;
  ff = NULL;
  while ((ff = (FFP ) get_list_next (vdt->fflist, ff))) {
    if (ff->is_active == segnum) toggle_ff (ff);
  }
  return TRUE;
}

static int
orientation_is_compatible (FFP ff, FFP aff)
{
#define SAMEPT(a, b) (ff->pt[a] == aff->pt[b])
#define COMPATEDGE(ffpta, ffptb, affpta, affptb) \
  (SAMEPT(ffpta, affptb) && SAMEPT(ffptb, affpta))

  if (COMPATEDGE (0, 1, 0, 1)) return TRUE;
  if (COMPATEDGE (0, 1, 1, 2)) return TRUE;
  if (COMPATEDGE (0, 1, 2, 0)) return TRUE;

  if (COMPATEDGE (1, 2, 0, 1)) return TRUE;
  if (COMPATEDGE (1, 2, 1, 2)) return TRUE;
  if (COMPATEDGE (1, 2, 2, 0)) return TRUE;

  if (COMPATEDGE (2, 0, 0, 1)) return TRUE;
  if (COMPATEDGE (2, 0, 1, 2)) return TRUE;
  if (COMPATEDGE (2, 0, 2, 0)) return TRUE;

  return FALSE;
}

static void
orient_ff (FFP off, FFP ff)
{
  if (!orientation_is_compatible (off, ff)) {
    toggle_ff (off);
  }
}

static FFP
ff_on_segment (vdt_t *vdt, int segnum)
{
  FFP ff;
  ff = NULL;
  while ((ff = (FFP ) get_list_next (vdt->fflist, ff))) {
    if (ff->is_active  == segnum) return ff;
  }
  return NULL;
}

static int                      /* returns the number of segments */
gather_and_reorient_segments (vdt_t *vdt)
{
  STACK S = NULL;
  FFP ff;  FFP off;
  FFEP e;
  int i;
  dict_dictionary ed;
  int segnum = NOSEG;
  
  loop_through_list (vdt->fflist, ff, FFP) {ff->is_active = NOSEG;}
  
  ed = create_edges (vdt);
  S  = make_stack ();
  
  ff = (FFP ) get_list_head (vdt->fflist);
  
 next:
  segnum++;
  { char buf[132]; sprintf (buf, "      Collecting segment = %d", segnum);
    logit (buf);
  }
  push_stack (S, ff);
  while (!stack_empty (S)) {
    ff = (FFP ) pop_stack (S);
    ff->is_active = segnum;
    FOR (i, 3) {
      e = find_edge (ed, ff->pt[i]->id, ff->pt[(i+1)%3]->id);
      if (e->is_manifold) {
        off = other_ff (e, ff);
        if (!off) { errexit ("Sheet front (edge with just one facet)!?"); }
        if (off->is_active == NOSEG) {
          orient_ff (off, ff);
          push_stack (S, off);
        } /* if not oriented yet */
      } /* if manifold */
    } /* FOR */
  } /* while */

  if (ff = ff_on_segment (vdt, NOSEG)) {
    goto next;
  }
  
  destroy_edges (ed);
  free_stack (S, NOT_NODES);

  return segnum;
}

static void
activate_ffs (vdt_t *vdt)
{
  FFP ff;
  loop_through_list (vdt->fflist, ff, FFP) {
    if (ff->is_active  != FALSE) ff->is_active = TRUE;
  }
}

static void
orient_af (vdt_t *vdt)
{
#define NRETRY 7
#define NTESTS 3
  int num_intersect[NRETRY];
  int i, k;
  CkitPoint3 start;
  CkitPoint3 direction;
  int result, prev_result;
  int nsegs, segnum;
#define IS_EVEN(num) (num / 2 * 2 == num)
  ckit_timer_t *timer = ckit_timer_new ();

  ckit_timer_restart (timer);
  logit ("     Orienting the initial AF");
  
  nsegs = gather_and_reorient_segments (vdt);

  for (segnum = 1; segnum <= nsegs; segnum++) {

    { char buf[132]; sprintf (buf, "      Orienting segment = %d", segnum);
      logit (buf);
    }
    
    FOR (k, NTESTS) { 
      FOR (i, NRETRY) {
        start_and_direction (vdt, segnum, &start, &direction);
        num_intersect[i] = get_num_intersections (vdt, &start, &direction);
      }
      prev_result = IS_EVEN (num_intersect[0]);
      for (i = 1; i < NRETRY; i++) {
        result = IS_EVEN (num_intersect[i]);
        if (prev_result != result)
          goto next_test; /* inconsistent results; try another test */
        prev_result = result;
      }
      if (!IS_EVEN (num_intersect[0]))
        toggle_af_segment (vdt, segnum);
      goto next_seg;
    next_test: ;
    }
    
  next_seg:;
  }

  activate_ffs (vdt);

  vdt_report_time (timer, "     Orienting AF");
  ckit_timer_free (timer);
}

static void
deactivate_dupl_ffs (vdt_t *vdt)
{
  PTP pt;
  FFP ff;
  pt = NULL;
  while (pt = (PTP) get_list_next (vdt->pointl, pt)) {
    vdt_pt_deactivate_dupl_ff (pt);
  }
  /* now delete deactivated ff's from the global list */
 again:
  ff = NULL;
  while (ff = (FFP) get_list_next (vdt->fflist, ff)) {
    if (!ff->is_active) {
      delete_list_node (vdt->fflist, ff);
      NEWP_FREE_OBJ (ff);
      goto again;
    }
  }
}

static int
initial_af (vdt_t *vdt, void *reg_id)
{
  FFP ff;
  vdt_bsf_t *bsf;
  int side;
  char printbuf[128];

  sprintf (printbuf, "   Region %ld", (long int) reg_id);
  logit (printbuf);
  
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    if (!bsf->is_delaunay) {errexit ("non-delaunay bsf");}
    if (bsf->reg_id[0] == bsf->reg_id[1]) {
      /* VOID */
      /* bsf's internal to a single region are *not* added to the front */
    } else {
      FOR (side, 2) {
        if (bsf->reg_id[side] == reg_id) {
          ff = vdt_bsf_make_ff (bsf, vdt, side);
          vdt_af_add_ff (vdt, ff);
        }
      }
    }
  }

  deactivate_dupl_ffs (vdt);

  if (vdt->orient_initial_af) orient_af (vdt);
  
  return TRUE;
}

int
vdt_af_initial_af (vdt_t *vdt, void *reg_id)
{
  return initial_af (vdt, reg_id);
}

static int
flip (vdt_bsf_t *bsf, vdt_t *vdt)
{
  vdt_bsf_t *nbsf;
  int result = FALSE;
  
#define STOP_FOR_FLIP 1
#if STOP_FOR_FLIP
  ANNOUNCE_ACTION (">>>>>>> Diagonal flip");
#endif
  nbsf = vdt_bsf_adj_for_flip (bsf);
  if (nbsf) {
    vdt_bsf_flip_diagonal (bsf, nbsf, vdt);
    result = (bsf->is_delaunay || nbsf->is_delaunay);
  }
#if STOP_FOR_FLIP
  SHOW_ACTION_RESULT;
#endif
  return result;
}




#if 0
#define HAVEPT(bsf, ptid) \
  (ptid==(int) bsf->pt[0]->id||ptid==(int) bsf->pt[1]->id||ptid==(int) bsf->pt[2]->id)
#define BSFPTSARE(bsf, n0, n1, n2)                              \
  (HAVEPT (bsf, n0) && HAVEPT (bsf, n1) && HAVEPT (bsf, n2))
#endif
  
static int
split (BSFP bsf, vdt_t *vdt)
{
  int result;
  LIST al = NULL;
  al = vdt_bsf_adj_patch (bsf);
  al = vdt_bsf_reduce_adj_patch_for_split (bsf, al);
  ANNOUNCE_ACTION (">>>>>>> Edge split");
  result = vdt_bsf_edge_split (bsf, al, vdt);
  free_list (al, NOT_NODES);
  SHOW_ACTION_RESULT;
  return result;
}

static int
map_to_l (double aggq, int maxl)
{
  double inv_aggq = 1 / aggq;
  if (inv_aggq > 1) return (maxl - 1);
  else              return (int) (inv_aggq * maxl);
}

static LIST
build_non_delaunay_list (vdt_t *vdt)
{
  BSFP bsf;
  double sigma, tau;
  double aggq;
  double maxaggq = (6 * 6 + 3 * 3);
#define MAXL 20
  LIST l[MAXL];
  LIST ndl;
  int i, j;

  ndl = make_list ();
  if (!ndl) return NULL;
  
  FOR (i, MAXL) { l[i] = make_list (); }
  
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    if (!bsf->is_delaunay) {
      vdt_tri_quality (bsf->pt[0], bsf->pt[1], bsf->pt[2], &sigma, &tau);
      aggq = (sigma * sigma + tau * tau) / maxaggq;
      j = map_to_l (aggq, MAXL);
      add_to_tail (l[j], bsf);
    }
  }

  FOR (i, MAXL) {
    if (list_length (l[i]) > 0) {
      append_list (ndl, l[i]);
    }
  }

  return ndl;
}


static int
f_pass (vdt_t *vdt)
{
  vdt_bsf_t *bsf;
  int nch = 0;
  LIST il;

  vdt->surf_mod_pass++;

  if (vdt->pdebug) {
    char buf[128];
    sprintf (buf, "Flipping pass #%d", vdt->surf_mod_pass);
    logit (buf);
  }

  if (vdt->new_pt) {
    il = vdt->new_pt->bsfl;
  } else {
    il = vdt->bsfl;
  }
  
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (il, bsf))) {
    if (!bsf->is_delaunay) {
      if (bsf->surf_mod_pass < vdt->surf_mod_pass)
        nch += flip (bsf, vdt); 
    }
  }

  return nch;
}



static int
d_pass (vdt_t *vdt)
{
  vdt_bsf_t *bsf;
  LIST non_delaunay;
  int nd;

  vdt->surf_mod_pass++;

  if (vdt->pdebug) {
    char buf[128];
    sprintf (buf, "Splitting pass #%d", vdt->surf_mod_pass);
    logit (buf);
  }

  non_delaunay = build_non_delaunay_list (vdt);
  nd = list_length (non_delaunay);

  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (non_delaunay, bsf))) {
    vdt->new_pt = NULL;
    if (!bsf->is_delaunay && bsf->surf_mod_pass < vdt->surf_mod_pass) {
      split (bsf, vdt); 
      update_delaunay_prop (vdt);
      if (vdt->allow_flips) f_pass (vdt);
    }
  }
    
  free_list (non_delaunay, NOT_NODES);

#if GIFACE
  if (vdt->gdebug) {
    char buf[128];
    vdt_generate_graphics (vdt);
    sprintf (buf, "After delaunayzation pass #%d", vdt->surf_mod_pass);
    ESIEventLoop (TRUE, buf);
  }
#endif

  return nd;
}

static void
check_delaunay_props (vdt_t *vdt)
{
  BSFP bsf;
  PTP apt;
  bsf = NULL;
  while ((bsf = (vdt_bsf_t *) get_list_next (vdt->bsfl, bsf))) {
    if (bsf->is_delaunay && bsf->fourth_pt) {
      if (vdt_4pt_circsph_pt_in (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                 bsf->fourth_pt, &bsf->tetcsc, bsf->tetcsr,
                                 vdt, VDT_ALL_PTS, VDT_SOFT_LIMIT, &apt))
        errexit ("Non-delaunay bsf");
    } else {
      if (!vdt_3pt_circsph_pt_in (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                  &bsf->tetcsc, bsf->tetcsr,
                                  vdt, FALSE, &apt))
        errexit ("Non-delaunay bsf");
      else {
        if (apt)
          errexit ("Non-delaunay bsf");
      }
    }
  }
} 
    

static int
make_delaunay (vdt_t *vdt)
{
  int nd = 0;
  char buf[256];
  ckit_timer_t *timer = ckit_timer_new ();

  ckit_timer_restart (timer);
  logit ("Enforcing empty circumsphere of boundary facets");
   
  vdt->del_stat.num_bsf_updates          = 0;
  vdt->del_stat.lp_stat.sum_unknowns     = 0;
  vdt->del_stat.lp_stat.sum_evals        = 0;
  vdt->del_stat.lp_stat.max_num_unknowns = 0;
  
  vdt->search_stat.num_searches  = 0;
  vdt->search_stat.sum_found     = 0;
  vdt->search_stat.max_num_found = 0;
  
  vdt->surf_mod_pass = 0;

  nd = count_non_delaunay_bsfs (vdt);

 restart:

  do {
    sprintf (buf, "   %d non-Delaunay facets", nd);
    logit (buf);
    vdt->new_pt = NULL;
    if (vdt->allow_flips) f_pass (vdt);
  } while ((nd = d_pass (vdt)) > 0);

  if (count_non_delaunay_bsfs (vdt) > 0) {
    goto restart;
  }

  vdt_report_time (timer, "Boundary enforcement");
  
  print_search_stat (vdt);
  print_del_stat (vdt);
  
  if (vdt->maxdebug) {
    check_delaunay_props (vdt);
  }

  ckit_timer_free (timer);
  return TRUE;
}

int
vdt_af_prepare (vdt_t *vdt)
{
  init_delaunay_prop (vdt);
  
#if GIFACE
  if (vdt->giface) {
    vdt_generate_graphics (vdt);
    ESIEventLoop ((!vdt->test_mode), "Before surface modification");
  }
#endif
  
  if (count_non_delaunay_bsfs (vdt) >= 1)
    make_delaunay (vdt);

#if GIFACE
  if (vdt->giface) {
    vdt_generate_graphics (vdt);
    ESIEventLoop ((!vdt->test_mode), "After surface modification");
  }
#endif
  
  return TRUE;
}



static void
normal (CkitPoint3 *p0, CkitPoint3 *p1, CkitPoint3 *p2, CkitPoint3 *n)
{
  CkitPoint3 a, b;
  VECMAC_3_VEC (a, =, (*p1), -, (*p0));
  VECMAC_3_VEC (b, =, (*p2), -, (*p0));
  VECMAC_CROSS ((*n), =, a, x, b);
  if (VECMAC_NORM ((*n)) <= 0)
    errexit ("Zero normal");    
  VECMAC_NORMALIZE ((*n));
}

vdt_lp_result_t
vdt_af_tet_at_bsf (vdt_t *vdt, PTP bsfpt[3], PTP tetpt[4],
                   CkitPoint3 *tetcsc, double *tetcsr)
{
  double bsfcd  = vdt_bsf_char_dim (bsfpt);
  double dist   = bsfcd / 1000;
  CkitPoint3 n, sp, fc;
  int nap, prevnap;
  vdt_lp_result_t result;
  double boxsize, csrad;
  CkitPoint3 csc, d, tetloc[4], approxtc;
  G3D_box_t box;
  int ptid[4];
  int i;
  PTP apt;
  PTP prevapt = NULL;
  int num_retries;
  int max_search = (vdt->lp_size_limit > 0
                    ? vdt->lp_size_limit
                    : VDT_ALL_PTS);
#define AEQ_(X, Y) (fabs (X - Y) < MACHEPS)
#define LOCEQ(P, Q) (AEQ_ (P.x, Q.x) &&  AEQ_ (P.y, Q.y) && AEQ_ (P.z, Q.z))
#define TET_USES_PT(tetpt, p) \
 (p == tetpt[0] || p == tetpt[1] || p == tetpt[2] || p == tetpt[3])
#define TET_USES_ALL_FACE_PTS(tetpt, bsfpt)     \
  (   TET_USES_PT (tetpt, bsfpt[0])             \
   && TET_USES_PT (tetpt, bsfpt[1])             \
   && TET_USES_PT (tetpt, bsfpt[2])             \
   )

  /* normal at the barycenter */
  normal (&bsfpt[0]->loc, &bsfpt[1]->loc, &bsfpt[2]->loc, &n);
  /* center */
  VECMAC_3_VEC (fc, =, bsfpt[0]->loc, +, bsfpt[1]->loc);
  VECMAC_2_VEC (fc, +=, bsfpt[2]->loc);
  VECMAC_SCALAR (fc, *=, 1.0/3.0);
  /* sampling point */
  VECMAC_3_VEC (sp, =, fc, -, dist * n);

  /* approx tet barycenter */
  boxsize = 2 * bsfcd;
  VECMAC_3_VEC (approxtc, =, fc, -, 0.7 * boxsize / 2 * n);
  csc = approxtc;

  G3D_INIT_BOX (box, csc);
  G3D_INFLATE_BOX (box, boxsize);

  while ((nap = vdt_pts_in_box (vdt, &box, max_search)) < 4) {
    G3D_INFLATE_BOX (box, boxsize);
  }
  
  num_retries = 0;

 retry_for_another_active_set:
  num_retries++;
  prevnap = nap;

  if (   vdt->lp_size_limit > 0
      && vdt->num_active_points >= vdt->lp_size_limit)
    return VDT_INFEASIBLE;

  result = vdt_get_tet_from_lp (vdt, &sp, ptid);
  switch (result) {
  case VDT_UNBOUNDED:
    return VDT_UNBOUNDED;
  case VDT_INFEASIBLE:
    if (vdt->pdebug) { logit ("Infeasible LP"); }
    G3D_INFLATE_BOX (box, boxsize);
    nap = vdt_pts_in_box (vdt, &box, max_search);
    if (nap == prevnap || num_retries > 2) {
      return VDT_INFEASIBLE;
    }
    goto retry_for_another_active_set;
  case VDT_TOO_MANY_POINTS:
    errexit ("Too many points in LP problem solution");
    break;
  case VDT_LOWER_DIM_OPTIMAL:
    if (vdt->pdebug) { logit ("Lower-dimensional LP"); }
    dist *= 2;
    VECMAC_3_VEC (sp, =, fc, -, dist * n);
    goto retry_for_another_active_set;
  case VDT_OPTIMAL:
    FOR (i, 4) { tetpt[i] = vdt->active_pt[ptid[i]]; }
    FOR (i, 4) { tetloc[i] = tetpt[i]->loc; }
    if (!vdt_4_loc_circumsphere (tetloc, &csc))
      return VDT_INFEASIBLE;
    VECMAC_3_VEC (d, =, csc, -, tetpt[0]->loc);
    csrad = VECMAC_NORM (d);
    /* Check if the tet is Delaunay by the circumsphere test; if not
       it is an indication that the LP solver did not get all the points ...
       */
    if (!vdt_4pt_circsph_pt_in (tetpt[0], tetpt[1], tetpt[2], tetpt[3],
                                &csc, csrad,
                                vdt, max_search, VDT_SOFT_LIMIT,
                                &apt) && apt == NULL) {
      /* Found tet enclosing the sampling point; finished */
      *tetcsc = csc;
      *tetcsr = csrad;
      return VDT_OPTIMAL;
    } else {
      /* ... therefore, increase the box, and ... */
      if (G3D_PT_INSIDE_BOX (box, csc)) {
        G3D_INIT_BOX (box, csc);
        boxsize += csrad;
      } else {
        G3D_INIT_BOX (box, approxtc);
        boxsize += bsfcd;
      }
      G3D_INFLATE_BOX (box, boxsize);
      if (vdt_pts_in_box (vdt, &box, max_search) == 0)
        return VDT_INFEASIBLE;
      if (prevapt == apt) { /* Stuck in a cycle: jiggle the sampling point */
        double r[3];
        int j;
        FOR (j, 2) { r[j] = vdt_rand_0_to_1 (j); }
        r[2] = 1 - r[0] - r[1];
        VECMAC_SCALAR (sp, =, 0);
        VECMAC_2_VEC (sp, +=, r[0] * bsfpt[0]->loc);
        VECMAC_2_VEC (sp, +=, r[1] * bsfpt[1]->loc);
        VECMAC_2_VEC (sp, +=, r[2] * bsfpt[2]->loc);
        dist *= 2;
        VECMAC_2_VEC (sp, -=, dist * n);
      }
      prevapt = apt;
      goto retry_for_another_active_set;
    }
  }
  
  return VDT_INFEASIBLE;
}

