
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtbsf.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"


static char vcid[] = "$Id: vdtbsf.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtbsf) 
 

#if 1
#define HAVEPT(bsf, ptid) \
  (ptid==(int) bsf->pt[0]->id||ptid==(int) bsf->pt[1]->id||ptid==(int) bsf->pt[2]->id)
#define BSFPTSARE(bsf, n0, n1, n2)                              \
  (HAVEPT (bsf, n0) && HAVEPT (bsf, n1) && HAVEPT (bsf, n2))
#endif


static int etov[3][2] = {
  {2,0}, {0,1}, {1,2}
};


static void
lin_n (CkitPoint3 *p0, CkitPoint3 *p1, CkitPoint3 *p2, CkitPoint3 *n)
{
  CkitPoint3 a, b;
  VECMAC_3_VEC (a, =, (*p1), -, (*p0));
  VECMAC_3_VEC (b, =, (*p2), -, (*p0));
  VECMAC_CROSS ((*n), =, a, x, b);
  if (VECMAC_NORM ((*n)) <= 0)
    errexit ("Zero normal");    
  VECMAC_NORMALIZE ((*n));
}


BSFP 
vdt_bsf_add (PTP pt[3],    /* corner points;
                                    vertex 0 is shared by edges 0 and 1,
                                    vertex 1 is shared by edges 1 and 2,
                                    vertex 2 is shared by edges 2 and 0,
                                    */
             void *surf_id,       /* id of parent surface */
             void *reg_id[2],     /* region id; reg_id[1] is on the
                                   side to which the normal points,
                                   reg_id[0] is on the other side.
                                 */
             vdt_t *vdt /* triangulation */
             )
{
  BSFP bsf;
  bsf = vdt_bsf_new (pt, surf_id, reg_id);
  vdt_bsf_connect (bsf, bsf->pt[0], bsf->pt[1], bsf->pt[2], vdt);
  if (bsf) vdt_attach_bsf (vdt, bsf);
  return bsf;
}


BSFP 
vdt_bsf_new (PTP pt[3],    /* corner points;
                                    vertex 0 is shared by edges 0 and 1,
                                    vertex 1 is shared by edges 1 and 2,
                                    vertex 2 is shared by edges 2 and 0,
                                    */
             void *surf_id,       /* id of parent surface */
             void *reg_id[2]     /* region id; reg_id[1] is on the
                                   side to which the normal points,
                                   reg_id[0] is on the other side.
                                 */
             )
{
  BSFP bsf;
  CkitPoint3 zerov = {0,0,0};

  bsf = NEWP_OBJ (vdt_bsf_t); CHECK_MEM (bsf);

  bsf->pt[0]         = pt[0];
  bsf->pt[1]         = pt[1];
  bsf->pt[2]         = pt[2];

  bsf->surf_id       = surf_id;
  bsf->reg_id[0]     = reg_id[0];
  bsf->reg_id[1]     = reg_id[1];
  bsf->tet[0]        = NULL;
  bsf->tet[1]        = NULL;
  
  bsf->is_delaunay   = FALSE;
  bsf->fourth_pt     = NULL;
  
  bsf->tetcsc        = zerov;
  bsf->tetcsr        = 0;
  bsf->surf_mod_pass = 0;
  
#if GIFACE
  bsf->go = NULL;
#endif
  
  return bsf;
}

#define TET_USES_PT(tetpt, p) \
 (p == tetpt[0] || p == tetpt[1] || p == tetpt[2] || p == tetpt[3])

static  int
tet_uses_bsf_pts (PTP tetpt[4], vdt_bsf_t *bsf)
{
  return (   TET_USES_PT (tetpt, bsf->pt[0])
          && TET_USES_PT (tetpt, bsf->pt[1])
          && TET_USES_PT (tetpt, bsf->pt[2])
          );
}


static  PTP
the_fourth (PTP tetpt[4], vdt_bsf_t *bsf)
{
  int i, f = -1;
  FOR (i, 4) {
    if (!(tetpt[i] == bsf->pt[0] ||
          tetpt[i] == bsf->pt[1] ||
          tetpt[i] == bsf->pt[2]))
      f = i;
  }
  if (f >= 0 && f < 4) return tetpt[f];
  else                 return NULL;
}


static void
csrad (BSFP bsf)
{
  CkitPoint3 d;
  VECMAC_3_VEC (d, =, bsf->pt[0]->loc, -, bsf->tetcsc);
  bsf->tetcsr = VECMAC_NORM (d);
}

int
vdt_bsf_can_blow_bubble (BSFP bsf, vdt_t *vdt, PTP *bubblept)
{
  CkitPoint3 d;
  double csr2, bs, prevd2 = FLT_MAX, d2;
  G3D_box_t box;
  PTP pt;
  LIST inpts;
  int i;
  int can_do_it = FALSE;
  PTP bpt = NULL;
  PTP p0 = bsf->pt[0];
  PTP p1 = bsf->pt[1];
  PTP p2 = bsf->pt[2];

  *bubblept = NULL;
  
  if (!vdt_3_pt_circumsphere (p0, p1, p2, &bsf->tetcsc)) {
    warn ("No circumsphere for facet?");
    return FALSE;
  }
  csrad (bsf); csr2 = bsf->tetcsr * bsf->tetcsr;
  
  G3D_INIT_BOX (box, bsf->tetcsc);
  bs = 1.001 * bsf->tetcsr;
  G3D_INFLATE_BOX (box, bs);

  inpts = make_list ();
  
  if (vdt_pts_in_box (vdt, &box, VDT_ALL_PTS) < 3)
    errexit ("suspicious number of points");

  FOR (i, vdt->num_active_points) {
    pt = vdt->active_pt[i];
    if (pt != p0 && pt != p1 && pt != p2) {
      VECMAC_3_VEC (d, =, pt->loc, -, bsf->tetcsc);
      d2 = VECMAC_DOT (d, d);
      if (d2 < csr2 && d2 < prevd2) {
        bpt = pt; prevd2 = d2;
        add_to_tail (inpts, bpt);
      }
    }
  }

  if (list_length (inpts) < 1) {
    can_do_it = TRUE;
    goto exit;
  } else if (list_length (inpts) < 3) {
    PTP apt;
    bpt = NULL;
    while ((bpt = (PTP) get_list_next (inpts, bpt))) {
      if (!vdt_4_pt_circumsphere (p0, p1, p2, bpt, &bsf->tetcsc)) {
        can_do_it = FALSE;
        goto exit;
      }
      csrad (bsf);
      if (!vdt_4pt_circsph_pt_in (p0, p1, p2, bpt, &bsf->tetcsc, bsf->tetcsr,
                                  vdt, VDT_ALL_PTS, VDT_SOFT_LIMIT, &apt)) {
        *bubblept = bpt;
        can_do_it = TRUE;
        goto exit;
      }
    } /* while */
  }
   
 exit:
  if (inpts) free_list (inpts, NOT_NODES);
  
  return can_do_it;
}

int 
vdt_bsf_update_delaunay_prop (BSFP bsf, vdt_t *vdt, PTP tetpt[4])
{
  PTP apt;
  vdt_lp_result_t result;
  char buf[128];

  vdt->del_stat.num_bsf_updates++;
  
  if (bsf->is_delaunay && bsf->fourth_pt) {
    if (!vdt_4pt_circsph_pt_in (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                bsf->fourth_pt, &bsf->tetcsc, bsf->tetcsr,
                                vdt, VDT_ALL_PTS, VDT_SOFT_LIMIT, &apt)) {
      bsf->is_delaunay = TRUE;
      return TRUE;
    }
  }

  if (!bsf->is_delaunay) {
    if (!vdt_3_pt_circumsphere (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                &bsf->tetcsc)) {
      warn ("No circumsphere for facet?");
      return FALSE;
    }
    csrad (bsf);
    if (vdt_3pt_circsph_pt_in (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                               &bsf->tetcsc, bsf->tetcsr,
                               vdt, TRUE, &apt)) {
      if (!apt) { /* No point within the circumsphere; return immediately */
        bsf->fourth_pt   = NULL;
        bsf->is_delaunay = TRUE;
        return TRUE;
      } else  {
        bsf->fourth_pt = apt;
        if (!vdt_4_pt_circumsphere (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                    bsf->fourth_pt, &bsf->tetcsc)) {
          bsf->is_delaunay = FALSE;
          bsf->tetcsr      = 0; /* cannot be used anyway */
          bsf->fourth_pt   = NULL;
          return TRUE;
        }
        csrad (bsf);
        if (!vdt_4pt_circsph_pt_in (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                    bsf->fourth_pt, &bsf->tetcsc, bsf->tetcsr,
                                    vdt, 200, VDT_HARD_LIMIT, &apt)) {
          bsf->is_delaunay = TRUE;
          return TRUE;
        } else {
          bsf->fourth_pt   = NULL;
          bsf->is_delaunay = FALSE;
        }
      } /* else */
    } /* if (vdt_3pt_circsph_pt_in */
  }
  
  /* Need to check by solving the LP; otherwise can return immediately;
   */
  bsf->fourth_pt           = NULL;
  bsf->is_delaunay         = FALSE;

  if (vdt->lp_size_limit <= 0) return TRUE;

  result = vdt_af_tet_at_bsf (vdt, bsf->pt, tetpt,
                              &bsf->tetcsc, &bsf->tetcsr);

  switch (result) {
  case  VDT_OPTIMAL:
    if (tet_uses_bsf_pts (tetpt, bsf)) {
      bsf->fourth_pt   = the_fourth (tetpt, bsf);
      if (!bsf->fourth_pt) errexit ("Wrong fourth");
      bsf->is_delaunay = TRUE;
    }
    break;
  case VDT_LOWER_DIM_OPTIMAL:
    break;
  case VDT_INFEASIBLE:
    break;
  default:
    sprintf (buf, "vdt_af_tet_at_bsf failed: pt={%d,%d,%d}",
                 bsf->pt[0]->id, bsf->pt[1]->id, bsf->pt[2]->id);
    warn (buf);
    return FALSE;
  }

  return TRUE;
}


#if GIFACE
/*
   Generate graphics
 */
void
vdt_bsf_generate_graphics (BSFP bsf, vdt_t *vdt)
{
  WCRec pts[6];

  if (!vdt->gmodel) return;

  if (!bsf->go) {
    pts[0] = bsf->pt[0]->loc;
    pts[1] = bsf->pt[1]->loc;
    pts[2] = bsf->pt[2]->loc;
    bsf->go = CreateTriangle3D (pts);
    
    EASValsSetColor (vdt_color_from_id (bsf->surf_id));
    EASValsSetShrink (0.9);
    EASValsSetFillStyle (FILL_SOLID);
    EASValsSetLayer (VDT_DELAUNAY_BSF_LAYER);
    EGWithMaskChangeAttributes (FILL_MASK | COLOR_MASK |
                                SHRINK_MASK | LAYER_MASK, bsf->go);
    EMNoDrawAddGraphicsToModel (vdt->gmodel, bsf->go);
  } else {
    if (bsf->is_delaunay) 
      EASValsSetLayer (VDT_DELAUNAY_BSF_LAYER);
    else
      EASValsSetLayer (VDT_NON_DELAUNAY_BSF_LAYER);
    EGWithMaskChangeAttributes (LAYER_MASK, bsf->go);
    pts[0] = bsf->pt[0]->loc;
    pts[1] = bsf->pt[1]->loc;
    pts[2] = bsf->pt[2]->loc;
    EGModifyGraphicsGeometry (bsf->go, pts);
  }
}
#endif


/*
   Get normal at a corner.
 */
int
vdt_bsf_get_normal (BSFP bsf, CkitPoint3 *n)
{
  lin_n (&bsf->pt[0]->loc, &bsf->pt[1]->loc, &bsf->pt[2]->loc, n);
  return TRUE;
}

int
vdt_bsf_circumsphere (BSFP bsf, CkitPoint3 *cscenter)
{
  return G3D_tri_circumsphere (&bsf->pt[0]->loc,
                               &bsf->pt[1]->loc,
                               &bsf->pt[2]->loc,
                               cscenter);
}


vdt_ff_t *
vdt_bsf_make_ff (BSFP bsf, vdt_t *vdt, int side)
{
  vdt_ff_t *ff;
  PTP pt[3];
  
  if        (bsf->pt[0]->id < bsf->pt[1]->id &&
             bsf->pt[0]->id < bsf->pt[2]->id) {
    pt[0] = bsf->pt[0];
    pt[1] = bsf->pt[1];
    pt[2] = bsf->pt[2];
  } else if (bsf->pt[1]->id < bsf->pt[0]->id &&
             bsf->pt[1]->id < bsf->pt[2]->id) {
    pt[0] = bsf->pt[1];
    pt[1] = bsf->pt[2];
    pt[2] = bsf->pt[0];
  } else {
    pt[0] = bsf->pt[2];
    pt[1] = bsf->pt[0];
    pt[2] = bsf->pt[1];
  }

  if (side == 0) { /* toggle orientation for the other side of bsf */
    PTP apt = pt[1]; pt[1] = pt[2]; pt[2] = apt;
  }
  
  ff = vdt_ff_add (&vdt->stop_tet, 0, pt[0], pt[1], pt[2], vdt, bsf);
  if (!pt[0]->ffl) pt[0]->ffl = make_list ();
  add_to_tail (pt[0]->ffl, ff);

  return ff;
}


static int
has_pt (BSFP bsf, PTP pt)
{
  return (bsf->pt[0] == pt || bsf->pt[1] == pt || bsf->pt[2] == pt);
}

static int
shares_edge (BSFP bsf, BSFP absf)
{
  return (   (has_pt (bsf, absf->pt[0]) && has_pt (bsf, absf->pt[1]))
          || (has_pt (bsf, absf->pt[1]) && has_pt (bsf, absf->pt[2]))
          || (has_pt (bsf, absf->pt[2]) && has_pt (bsf, absf->pt[0])));
}


static PTP
the_third (BSFP bsf, PTP pt0, PTP pt1)
{
  if (bsf->pt[0] != pt0 && bsf->pt[0] != pt1) return bsf->pt[0];
  if (bsf->pt[1] != pt0 && bsf->pt[1] != pt1) return bsf->pt[1];
  if (bsf->pt[2] != pt0 && bsf->pt[2] != pt1) return bsf->pt[2];
  return NULL; /* shouldn't get here, though; assuming pt IS used by bsf */
}

static int
swap_edge_pt (BSFP bsf, PTP pt0, PTP pt1)
{
  if        (pt0 == bsf->pt[0]) {
    if (pt1 == bsf->pt[1]) return FALSE;
    else                   return TRUE;
  } else if (pt0 == bsf->pt[1]) {
    if (pt1 == bsf->pt[2]) return FALSE;
    else                   return TRUE;
  } else if (pt0 == bsf->pt[2]) {
    if (pt1 == bsf->pt[0]) return FALSE;
    else                   return TRUE;
  } else {
    errexit ("pt is not used by bsf");
    return FALSE;
  }
}


static void
order_quad_pts (BSFP bsf, BSFP obsf, PTP pt[4])
{
  int i;
  PTP shared[3];
  PTP third_bsf;
  PTP third_obsf;
  
  i = 0;
  if (has_pt (obsf, bsf->pt[0])) shared[i++] = bsf->pt[0];
  if (has_pt (obsf, bsf->pt[1])) shared[i++] = bsf->pt[1];
  if (has_pt (obsf, bsf->pt[2])) shared[i++] = bsf->pt[2];
  if (i > 2) {
    errexit ("bsf's share three pts");
  }

  third_bsf  = the_third (bsf, shared[0], shared[1]);
  third_obsf = the_third (obsf, shared[0], shared[1]);
  
  if (swap_edge_pt (bsf, shared[0], shared[1])) {
    pt[0] = third_bsf;
    pt[1] = shared[1];
    pt[2] = third_obsf;
    pt[3] = shared[0];
  } else {
    pt[0] = third_bsf;
    pt[1] = shared[0];
    pt[2] = third_obsf;
    pt[3] = shared[1];
  }
}


static int
quad_is_convex (BSFP bsf, BSFP absf, double min_area_ratio)
{
  PTP pt[4];
  PTP third;
  double uc[4], vc[4];
  double a[4];
  double totarea, minarea;
  G3D_plane_t plane;
  int i;
  CkitPoint3 *locs[4];
  
  if        (has_pt (absf, bsf->pt[0]) && has_pt (absf, bsf->pt[1])) {
    third = the_third (absf, bsf->pt[0], bsf->pt[1]);
    pt[0] = bsf->pt[0];
    pt[1] = third;
    pt[2] = bsf->pt[1];
    pt[3] = bsf->pt[2];
  } else if (has_pt (absf, bsf->pt[1]) && has_pt (absf, bsf->pt[2])) {
    third = the_third (absf, bsf->pt[1], bsf->pt[2]);
    pt[0] = bsf->pt[0];
    pt[1] = bsf->pt[1];
    pt[2] = third;
    pt[3] = bsf->pt[2];
  } else if (has_pt (absf, bsf->pt[2]) && has_pt (absf, bsf->pt[0])) {
    third = the_third (absf, bsf->pt[2], bsf->pt[0]);
    pt[0] = bsf->pt[0];
    pt[1] = bsf->pt[1];
    pt[2] = bsf->pt[2];
    pt[3] = third;
  } else
    return FALSE;

  FOR (i, 4) { locs[i] = &pt[i]->loc; }
  
  if (!G3D_plane_consts_4 (&plane, locs[0], locs[1], locs[2], locs[3])) {
    return FALSE;
  }
    
  G3D_plane_project_points (&plane, locs, 4, uc, vc);

  a[0] = G3D_tri2d_area (uc[0], vc[0], uc[1], vc[1], uc[2], vc[2]);
  a[1] = G3D_tri2d_area (uc[0], vc[0], uc[2], vc[2], uc[3], vc[3]);
  a[2] = G3D_tri2d_area (uc[1], vc[1], uc[2], vc[2], uc[3], vc[3]);
  a[3] = G3D_tri2d_area (uc[1], vc[1], uc[3], vc[3], uc[0], vc[0]);

  /* All the areas need to be positive... */
  if (a[0] >= 0 && a[1] >= 0 && a[2] >= 0 && a[3] >= 0) {
    /* ... and the areas of the new triangles need to be larger
       than some ratio of the total area of the quadrilateral
     */
    totarea = 0.5 * (a[0] + a[1] + a[2] + a[3]);
    minarea = min (FLT_MAX, a[2]);
    minarea = min (minarea, a[3]);
    return (minarea >= min_area_ratio * totarea);
  } else {
    return FALSE;
  }
}



static BSFP 
longest_diagonal_pick_one_bsf (BSFP bsf, LIST onsl)
{
  BSFP obsf;
  BSFP pbsf = NULL;
  double d, ld = 0;
  CkitPoint3 v;
  
  loop_through_list (onsl, obsf, BSFP) {
    if        (has_pt (obsf, bsf->pt[0]) && has_pt (obsf, bsf->pt[1])) {
      VECMAC_3_VEC (v, =, bsf->pt[1]->loc, -, bsf->pt[0]->loc);
    } else if (has_pt (obsf, bsf->pt[1]) && has_pt (obsf, bsf->pt[2])) {
      VECMAC_3_VEC (v, =, bsf->pt[2]->loc, -, bsf->pt[1]->loc);
    } else {
      VECMAC_3_VEC (v, =, bsf->pt[0]->loc, -, bsf->pt[2]->loc);
    }
    d = VECMAC_DOT (v, v);
    if (d > ld) {
      pbsf = obsf;
      ld   = d;
    }
  }
  return pbsf;
}



static int
angles_acceptable (BSFP bsf, BSFP obsf)
{
  PTP pt[4];
  CkitPoint3 normals[4];
  double cosa, cosnewa;
  order_quad_pts (bsf, obsf, pt);
  lin_n (&bsf->pt[0]->loc, &bsf->pt[1]->loc, &bsf->pt[2]->loc, &normals[0]);
  lin_n (&obsf->pt[0]->loc, &obsf->pt[1]->loc, &obsf->pt[2]->loc, &normals[1]);
  lin_n (&pt[0]->loc, &pt[1]->loc, &pt[2]->loc, &normals[2]);
  lin_n (&pt[2]->loc, &pt[3]->loc, &pt[0]->loc, &normals[3]);
  cosa    = VECMAC_DOT (normals[0], normals[1]);
  cosnewa = VECMAC_DOT (normals[2], normals[3]);
  return (cosnewa > 0.9 * (cosa + 0.05));
}


static int
for_flip_across_edge (BSFP bsf, BSFP ae[3])
{
  int edge, nshared;
  LIST l;
  BSFP absf;

  FOR (edge, 3) {
    ae[edge] = NULL;
    l = bsf->pt[etov[edge][0]]->bsfl;
    nshared = 0;
    absf = NULL;
    while ((absf = (BSFP) get_list_next (l, absf))) {
      if (absf != bsf) { /* not self */
        if (   has_pt (absf, bsf->pt[etov[edge][0]])
            && has_pt (absf, bsf->pt[etov[edge][1]])) { /* shares edge */
          nshared++;
          if ((bsf->surf_id != absf->surf_id) || /* distinct surfaces along this edge */
              nshared > 1 /* too many facets on the same surface
                             meet along this edge */
              ) {
            
            ae[edge] = NULL;
            goto next_edge; /* next edge */
          } else {
            if (quad_is_convex (bsf, absf, 0.01)) {
              if (angles_acceptable (bsf, absf)) {
                ae[edge] = absf;
              }
            }
          } /* if */
        }
      } /* not self */
    } /* while */
  next_edge:;
  }
  return TRUE;
}

BSFP 
vdt_bsf_adj_for_flip (BSFP bsf)
{
  BSFP aebsf[3];
  BSFP obsf = NULL;
  BSFP nbsf;
  double d, ld = 0;
  CkitPoint3 v;
  PTP p0;
  PTP p1;
  int i;

  obsf = NULL;
  for_flip_across_edge (bsf, aebsf);
  /* Pick the one across the longest edge of bsf */
  FOR (i, 3) {
    nbsf = aebsf[i];
    if (nbsf) {
      if        (   has_pt (nbsf, bsf->pt[0])
                 && has_pt (nbsf, bsf->pt[1])) {
        p0 = bsf->pt[0]; p1 = bsf->pt[1];
      } else if (   has_pt (nbsf, bsf->pt[1])
                 && has_pt (nbsf, bsf->pt[2])) {
        p0 = bsf->pt[1]; p1 = bsf->pt[2];
      } else {
        p0 = bsf->pt[2]; p1 = bsf->pt[0];
      }
      VECMAC_3_VEC (v, =, p1->loc, -, p0->loc);
      d = VECMAC_DOT (v, v);
      if (d > ld) {
        obsf = nbsf;
        ld   = d;
      }
    } /* if (nbsf) */
  } /* FOR */
  
  return obsf;
}

BSFP 
vdt_bsf_longest_edge_adj (BSFP bsf)
{
  BSFP nbsf;
  BSFP absf;
  LIST l;
  LIST a;
  int i;

  a = make_list ();
  FOR (i, 3) {
    l = bsf->pt[i]->bsfl;
    loop_through_list (l, absf, BSFP) {
      if (
          (absf != bsf) &&         /* not self */
          shares_edge (bsf, absf)   /* shares edge; not only a point */
          ) {
        add_to_tail (a, absf);
      }
    }
  }

  nbsf = longest_diagonal_pick_one_bsf (bsf, a);
  free_list (a, NOT_NODES);
  return nbsf;
}


static void
common_edge_pts (BSFP bsf, BSFP obsf, PTP cept[2])
{
  int i = 0;
  if (has_pt (obsf, bsf->pt[0])) cept[i++] = bsf->pt[0];
  if (has_pt (obsf, bsf->pt[1])) cept[i++] = bsf->pt[1];
  if (has_pt (obsf, bsf->pt[2])) cept[i++] = bsf->pt[2];
  if (i > 2) {
    char buf[256];
    sprintf (buf, "Bsf pt={%d,%d,%d} shares three points with another bsf",
             (int) bsf->pt[0]->id,
             (int) bsf->pt[1]->id,
             (int) bsf->pt[2]->id);
    errexit (buf);
  }
}


LIST
vdt_bsf_reduce_adj_patch_for_split (BSFP bsf, LIST al)
{
  BSFP nbsf = NULL;
  double d, ld = 0;
  CkitPoint3 v;
  PTP p0;
  PTP p1;
  PTP cept[2];
  LIST newal = make_list ();
  
  /* find the longest edge */
  loop_through_list (al, nbsf, BSFP) {
    if (nbsf != bsf) { /* do not add self */
      p0 = p1 = NULL;
      if        (   has_pt (nbsf, bsf->pt[0])
                 && has_pt (nbsf, bsf->pt[1])) {
        p0 = bsf->pt[0]; p1 = bsf->pt[1];
      } else if (   has_pt (nbsf, bsf->pt[1])
                 && has_pt (nbsf, bsf->pt[2])) {
        p0 = bsf->pt[1]; p1 = bsf->pt[2];
      } else {
        p0 = bsf->pt[2]; p1 = bsf->pt[0];
      }
      if (p0 && p1) {
        VECMAC_3_VEC (v, =, p1->loc, -, p0->loc);
        d = VECMAC_DOT (v, v);
        if (d > ld) {
          cept[0] = p0;
          cept[1] = p1;
          ld      = d;
        }
      }
    }
  }

  /* collect all bsfs sharing the edge */
  loop_through_list (al, nbsf, BSFP) {
    if (nbsf != bsf             /* not self */
        && has_pt (nbsf, cept[0]) && has_pt (nbsf, cept[1]) /* shares selected edge */
        ) {
      add_to_tail (newal, nbsf);
    }
  }

  free_list (al, NOT_NODES);
  
  return newal;
}


LIST
vdt_bsf_reduce_adj_patch_along_edge (BSFP bsf, LIST al, PTP p0, PTP p1)
{
  BSFP nbsf = NULL;
  LIST newal = make_list ();
  
  /* collect all bsfs sharing the edge */
  loop_through_list (al, nbsf, BSFP) {
    if (nbsf != bsf                               /* not self */
        && has_pt (nbsf, p0) && has_pt (nbsf, p1) /* shares given edge */
        ) {
      add_to_tail (newal, nbsf);
    }
  }

  free_list (al, NOT_NODES);
  
  return newal;
}



int
vdt_bsf_flip_diagonal (BSFP bsf, BSFP obsf, vdt_t *vdt)
{
  PTP pt[4];
  int result = TRUE;
  double sigma[2], tau[2];
  double newsigma[2], newtau[2];
  double max_sigma = 6, max_tau = 3;
  double q, newq;
  PTP tetpt[4];
  
  order_quad_pts (bsf, obsf, pt);

  vdt_tri_quality (pt[0], pt[1], pt[3], &sigma[0], &tau[0]);
  vdt_tri_quality (pt[1], pt[2], pt[3], &sigma[1], &tau[1]);
  
  vdt_tri_quality (pt[0], pt[1], pt[2], &newsigma[0], &newtau[0]);
  vdt_tri_quality (pt[2], pt[3], pt[0], &newsigma[1], &newtau[1]);
  
  q    = (  (sigma[0] + sigma[1]) / max_sigma
          + (tau[0] + tau[1]) / max_tau);
  newq = (  (newsigma[0] + newsigma[1]) / max_sigma
          + (newtau[0] + newtau[1]) / max_tau);

  if (newq < 100 * q) { /* eliminate flips leading to very bad triangles */
    vdt_bsf_disconnect (bsf);
    vdt_bsf_disconnect (obsf);
    vdt_bsf_connect (bsf, pt[0], pt[1], pt[2], vdt);
    vdt_bsf_connect (obsf, pt[2], pt[3], pt[0], vdt);
    vdt_bsf_update_delaunay_prop (bsf, vdt, tetpt);
    vdt_bsf_update_delaunay_prop (obsf, vdt, tetpt);
    if ( /*no Delaunay-hood improvement*/
        !(bsf->is_delaunay || obsf->is_delaunay) &&
        newq > q /* and even worse quality triangles */
        ) {
      /* Undo */
      vdt_bsf_disconnect (bsf);
      vdt_bsf_disconnect (obsf);
      vdt_bsf_connect (bsf, pt[0], pt[1], pt[3], vdt);
      vdt_bsf_connect (obsf, pt[2], pt[3], pt[1], vdt);
      vdt_bsf_update_delaunay_prop (bsf, vdt, tetpt);
      vdt_bsf_update_delaunay_prop (obsf, vdt, tetpt);
    }
  } 
  
  obsf->surf_mod_pass = bsf->surf_mod_pass = vdt->surf_mod_pass;
  return result;
}

void
vdt_bsf_disconnect (BSFP bsf)
{
  int i;
  FOR (i, 3) { vdt_pt_detach_bsf (bsf->pt[i], bsf); }
}

void
vdt_bsf_connect (BSFP bsf, PTP p0, PTP p1, PTP p2, vdt_t *vdt)
{
  int i;
  
  if (vdt->pdebug) {
    char buf[256];
    if (bsf->pt[0] && bsf->pt[1] && bsf->pt[2]) {
      sprintf (buf, "Reconnected bsf pt={%d,%d,%d} to pt={%d,%d,%d}",
               (int) bsf->pt[0]->id, (int) bsf->pt[1]->id, (int) bsf->pt[2]->id,
               (int) p0->id, (int) p1->id, (int) p2->id);
    } else {
      sprintf (buf, "Connected bsf pt={%d,%d,%d}",
               (int) p0->id, (int) p1->id, (int) p2->id);
    }
    logit (buf);
  }
  bsf->pt[0] = p0;
  bsf->pt[1] = p1;
  bsf->pt[2] = p2;

  FOR (i, 3) { vdt_pt_attach_bsf (bsf->pt[i], bsf); }
  /* Initialize the Delaunay-hood properties */
  bsf->is_delaunay         = FALSE;
  bsf->fourth_pt           = NULL;
  bsf->tetcsr              = 0;
}

LIST
vdt_bsf_adj_patch (BSFP bsf)
{
  BSFP nbsf;
  BSFP absf;
  int i;
  LIST retl = make_list ();
  LIST l;
  STACK bs = make_stack ();
  
  push_stack (bs, bsf);
  while (!stack_empty (bs)) {
    nbsf = (BSFP) pop_stack (bs);
    if (!in_list (retl, nbsf)) {
      add_to_head (retl, nbsf);
      FOR (i, 3) {
        l = nbsf->pt[i]->bsfl;
        loop_through_list (l, absf, BSFP) {
          if (
              (absf != nbsf) &&         /* not self */
              shares_edge (bsf, absf)   /* shares edge; not only a point */
              ) {
            push_stack (bs, absf);
          }
        }
      }
    }
  }
  free_stack (bs, NOT_NODES);
  
  return retl;
}



double
vdt_bsf_rms_edge_len (PTP pt[3])
{
  double rms;
  CkitPoint3 v;
  VECMAC_3_VEC (v, =, pt[1]->loc, -, pt[0]->loc);
  rms = VECMAC_DOT (v, v); 
  VECMAC_3_VEC (v, =, pt[2]->loc, -, pt[1]->loc);
  rms += VECMAC_DOT (v, v); 
  VECMAC_3_VEC (v, =, pt[0]->loc, -, pt[2]->loc);
  rms += VECMAC_DOT (v, v); 
  return sqrt (rms / 3);
}

double
vdt_bsf_char_dim (PTP pt[3])
{
  return vdt_bsf_rms_edge_len (pt);
}



double
vdt_bsf_max_edge_len (PTP pt[3])
{
  double el, mel;
  CkitPoint3 v;
  VECMAC_3_VEC (v, =, pt[1]->loc, -, pt[0]->loc);
  mel = el = VECMAC_NORM (v);
  VECMAC_3_VEC (v, =, pt[2]->loc, -, pt[1]->loc);
  el = VECMAC_NORM (v); mel = max (el, mel);
  VECMAC_3_VEC (v, =, pt[0]->loc, -, pt[2]->loc);
  el = VECMAC_NORM (v); mel = max (el, mel);
  return mel;
}





static int
common_edge_split (BSFP bsf, LIST al, vdt_t *vdt)
{
  PTP newpt;
  BSFP newbsf;
  BSFP obsf;
  PTP pts[3];
  CkitPoint3 loc;
  double h;
  PTP bpt[3];
  PTP apt;
#define SWAP_PTS(bsf, pt0, pt1)                 \
  {                                             \
    if (swap_edge_pt (bsf, pt0, pt1)) {         \
      apt = pt0; pt0 = pt1; pt1 = apt;          \
    }                                           \
  }
#define DEF_PTS(bsf, bpt)                       \
  {                                             \
    SWAP_PTS(bsf, bpt[0], bpt[1]);              \
    bpt[2] = the_third (bsf, bpt[0], bpt[1]);   \
  }
  
  common_edge_pts (bsf, (BSFP) get_list_head (al), bpt);
  
  VECMAC_3_VEC (loc,  =, 0.5 * bpt[0]->origloc, +, 0.5 * bpt[1]->origloc);
  h      = 0.5 * (bpt[0]->mesh_size + bpt[1]->mesh_size);

  vdt->max_pt_id = (void *)((char *)vdt->max_pt_id + 1); 
  newpt  = vdt_pt_new (vdt->max_pt_id, &loc, h);
  if (!newpt) { errexit ("Failed to create point"); }
  vdt_perturb_loc (vdt->bdry_pt_shift_mult, h, &newpt->loc);
  vdt_attach_pt (vdt, newpt);
  
  vdt_bsf_disconnect (bsf); bsf->surf_mod_pass = vdt->surf_mod_pass;
  DEF_PTS (bsf, bpt);
  vdt_bsf_connect (bsf, newpt, bpt[1], bpt[2], vdt);
  pts[0] = bpt[0]; pts[1] = newpt; pts[2] = bpt[2];
  newbsf = vdt_bsf_add (pts, bsf->surf_id, bsf->reg_id, vdt);
  newbsf->surf_mod_pass = vdt->surf_mod_pass;
  
  obsf = NULL;
  while ((obsf = (BSFP) get_list_next (al, obsf))) {
    vdt_bsf_disconnect (obsf);
    vdt_bsf_disconnect (obsf); obsf->surf_mod_pass = vdt->surf_mod_pass;
    DEF_PTS (obsf, bpt);
    vdt_bsf_connect (obsf, newpt, bpt[1], bpt[2], vdt);
    pts[0] = bpt[0]; pts[1] = newpt; pts[2] = bpt[2];
    newbsf = vdt_bsf_add (pts, obsf->surf_id, obsf->reg_id, vdt);
    newbsf->surf_mod_pass = vdt->surf_mod_pass;
  }
  
  return TRUE;
}



int
vdt_bsf_edge_split (BSFP bsf, LIST al, vdt_t *vdt)
{
  return common_edge_split (bsf, al, vdt);
}




void
vdt_bsf_conn_tet (BSFP bsf, TETP tet)
{
  if      (tet->reg_id == bsf->reg_id[0]) bsf->tet[0] = tet;
  else if (tet->reg_id == bsf->reg_id[1]) bsf->tet[1] = tet;
  else { errexit ("Tet in wrong region"); }
}


double
vdt_bsf_ideal_edge_length (BSFP bsf)
{
  double x[3], y[3], a;
  G3D_tri_to_tri2d (&bsf->pt[0]->loc, &bsf->pt[1]->loc, &bsf->pt[2]->loc,
                    x, y);
  a = G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
  return sqrt (4.0 / sqrt (3.0) * a);
}



int
vdt_bsf_1_to_3_split (BSFP bsf, CkitPoint3 *newloc, vdt_t *vdt)
{
  BSFP obsf[2];
  double h;
  PTP pts[3];
  PTP newpt;
  PTP *bpt = bsf->pt;
  
  h = (bpt[0]->mesh_size + bpt[1]->mesh_size + bpt[2]->mesh_size) / 3;

  vdt->max_pt_id = (void *)((char *)vdt->max_pt_id + 1); 
  newpt  = vdt_pt_new (vdt->max_pt_id, newloc, h);
  if (!newpt) { errexit ("Failed to create point"); }
  vdt_perturb_loc (vdt->bdry_pt_shift_mult, h, &newpt->loc);
  vdt_attach_pt (vdt, newpt);

  pts[0]  = bsf->pt[1]; pts[1] = bsf->pt[2]; pts[2] = newpt;
  obsf[0] = vdt_bsf_add (pts, bsf->surf_id, bsf->reg_id, vdt);

  pts[0]  = bsf->pt[2]; pts[1] = bsf->pt[0]; pts[2] = newpt;
  obsf[1] = vdt_bsf_add (pts, bsf->surf_id, bsf->reg_id, vdt);

  vdt_bsf_disconnect (bsf);
  vdt_bsf_connect (bsf, bsf->pt[0], bsf->pt[1], newpt, vdt);

  obsf[0]->surf_mod_pass = obsf[1]->surf_mod_pass = bsf->surf_mod_pass =
    vdt->surf_mod_pass;

  return TRUE;
}

int
vdt_bsf_inject_pt (BSFP bsf, CkitPoint3 *newloc, vdt_t *vdt)
{
  PTP *bpt = bsf->pt;
  double x[3], y[3];
  double nx, ny;
  double L1, L2, L3;
  double Lmin = 0.05;
  LIST al;

  G3D_tri_to_tri2d (&bpt[0]->loc, &bpt[1]->loc, &bpt[2]->loc, x, y);
  G3D_tri_2dpt (&bpt[0]->loc, &bpt[1]->loc, &bpt[2]->loc,
                newloc, &nx, &ny);
  if (!G3D_tri_inv_map (nx, ny, x[0], y[0], x[1], y[1], x[2], y[2],
                        &L1, &L2, &L3)) {
    return FALSE;
  }
  if        (L1 < Lmin || L1 > 1 - Lmin) {
    al = vdt_bsf_adj_patch (bsf);
    al = vdt_bsf_reduce_adj_patch_along_edge (bsf, al, bpt[1], bpt[2]);
    return vdt_bsf_edge_split (bsf, al, vdt);
  } else if (L2 < Lmin || L2 > 1 - Lmin) {
    al = vdt_bsf_adj_patch (bsf);
    al = vdt_bsf_reduce_adj_patch_along_edge (bsf, al, bpt[0], bpt[2]);
    return vdt_bsf_edge_split (bsf, al, vdt);
  } else if (L3 < Lmin || L3 > 1 - Lmin) {
    al = vdt_bsf_adj_patch (bsf);
    al = vdt_bsf_reduce_adj_patch_along_edge (bsf, al, bpt[0], bpt[1]);
    return vdt_bsf_edge_split (bsf, al, vdt);
  } else {
    return vdt_bsf_1_to_3_split (bsf, newloc, vdt);
  }
}
