 
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtdriv.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#ifndef lint
static char vcid[] = "$Id: vdtdriv.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
#endif /* lint */

#include "listP.h"
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h" 
#define WANT_VECTP_ALIASES
#include "vectP.h" /* to be able to use vector library routines. */
#include "randP.h" /* to be able to use rand library routines. */
#include "newP.h"              /* Ckit library */
#include "ioP.h"              /* Ckit library */
#include "listP.h"              /* Ckit library */
#include "dictP.h"              /* Ckit library */
#include "getoptP.h"              /* Ckit library */
#include "tokensP.h"              /* Ckit library */
#include "timeP.h"              /* Ckit library */
#include "blpoolP.h"              /* Ckit library */

#include "g3d.h"
#include "lobs3d.h"

#include "vdtpub.h"


#if GIFACE
#  include "Esimple.h"
#endif

static  char            *input_file = NULL;
static  char            *input_version = NULL;
static  char            *input_true_name = NULL;
static  VDT              vdt = NULL;
static  int              pdebug = FALSE;
static  int              maxdebug = FALSE;
#if GIFACE
static  int              gdebug = FALSE;
static  int              giface = FALSE;
#endif
static  int              test_mode = FALSE;
static  int              auto_mode = FALSE;
#if VDT_CUSTOM_SIGNAL_HANDLER
static  int              no_intr = FALSE;
#endif
static  int              keep_slivers = FALSE;
static  double           bdry_pt_shift_mult = VDT_BDRY_PT_DEFAULT_SHIFT_MULT;
static  double           gen_pt_shift_mult = VDT_GEN_PT_DEFAULT_SHIFT_MULT;
static  int              gen_interior_pts = FALSE;
static  int              gen_near_surf_pts = FALSE;
static  int              lp_size_limit = 100;
static  int              near_surf_test = VDT_NEAR_SURF_TEST_DIST;
static  int              dump_tet_wrl = FALSE;
static  int              dump_tet_egf = FALSE;
static  int              dump_bsf_egf = FALSE;
static  int              dump_data = FALSE;
static  int              dump_t3d = FALSE;
static  int              orient_initial_af = FALSE;
static  int              report_opts = FALSE;
static  int              reset_locs = FALSE;
static  int              allow_flips = TRUE;
static  int              nranges = 5;
static  double           bucket_size = 0;
static  int              retri_to_desliver = FALSE;
static  int              inject_surf_proj = FALSE;
static  int              eval_mesh_size_from_pt = FALSE;
static  double           sliver_dihed_angle = VDT_SLIVER_DIHED_ANGLE_DEG;

static  int              cmd_line_lp_size_limit = FALSE;
static  int              cmd_line_gen_near_surf_pts = FALSE;
static  int              cmd_line_bdry_pt_shift_mult = FALSE;
static  int              cmd_line_gen_pt_shift_mult = FALSE;


static tokens_parser parser = NULL;


static int
errexit_ (char *file, int line, char *msg)
{
  fprintf (stderr, "\n*** ERROR: %s, l.%d: %s\n", file, line, msg);
  exit (1);
  return FALSE;
}
#define errexit(msg) errexit_ (__FILE__, __LINE__, msg)

static int
usage (int argc, char **argv)
{
  char *option, *usage;

  fprintf (stderr, "usage: %s -i file + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option != NULL ? option : "?"),
             (usage  != NULL ? usage  : "?"));
  }
  exit (0);
  return argc;
}

static void
check_presence_of_input (void)
{
  char suffix[] = ".vin";
  char prefix[] = "../inputs/";
  char *base, *testname;
  int il, pl, sl;
  
  if (!ckit_io_file_exists (input_file)) {
    il = strlen (input_file);
    pl = strlen (prefix);
    sl = strlen (suffix);
    base = NEWP_OBJ_ARRAY (il+1, char);
    ckit_io_strip_path (base, input_file);

    if (!ckit_io_file_has_suffix (input_file, suffix)) {
      testname = NEWP_OBJ_ARRAY (il+sl+1, char);
      sprintf (testname, "%s%s", input_file, suffix);
      if (ckit_io_file_exists (testname)) {
        NEWP_FREE_OBJ (base);
        NEWP_FREE_OBJ (input_file);
        input_file = testname;
        return;
      }
      NEWP_FREE_OBJ (testname);
    }

    if (!ckit_io_file_has_suffix (base, suffix)) {
      testname = NEWP_OBJ_ARRAY (il+sl+1, char);
      sprintf (testname, "%s%s", base, suffix);
      if (ckit_io_file_exists (testname)) {
        NEWP_FREE_OBJ (base);
        NEWP_FREE_OBJ (input_file);
        input_file = testname;
        return;
      }
      NEWP_FREE_OBJ (testname);
    }

    testname = NEWP_OBJ_ARRAY (pl+il+1, char);
    sprintf (testname, "%s%s", prefix, base);
    if (ckit_io_file_exists (testname)) {
      NEWP_FREE_OBJ (base);
      NEWP_FREE_OBJ (input_file);
      input_file = testname;
      return;
    }
    NEWP_FREE_OBJ (testname);
    
    if (!ckit_io_file_has_suffix (base, suffix)) {
      testname = NEWP_OBJ_ARRAY (pl+sl+il+1, char);
      sprintf (testname, "%s%s%s", prefix, base, suffix);
      if (ckit_io_file_exists (testname)) {
        NEWP_FREE_OBJ (base);
        NEWP_FREE_OBJ (input_file);
        input_file = testname;
        return;
      }
      NEWP_FREE_OBJ (testname);
    }
    
    NEWP_FREE_OBJ (base);
    
    errexit ("Input file not found");
  } else {                      /* file exists */
    return;
  }
}

static void 
process_options (int argc, char **argv)
{
  int help = FALSE;

#define DOSTRINGZ(s) #s
#define STRINGZ(s) DOSTRINGZ(s)
#define SET_BOOL(name)                                          \
  {                                                             \
    if (report_opts) { fprintf (stderr, "Set: %s\n", #name); }  \
    name = TRUE;                                                \
  }
#define SET_WITH_ARG(name, format)                                        \
  {                                                                       \
    if (report_opts) { fprintf (stderr, "Set: %s = %" #format "\n",       \
                                STRINGZ(name), name); }                   \
  }
  
  GET_OPT_BOOL (argc, argv, "-report_opts", SET_BOOL(report_opts),
                "prints options as they are being set");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL(help),
                "prints usage for each option recognized");
  
#if GIFACE
  GET_OPT_BOOL (argc, argv, "-gdebug", SET_BOOL(gdebug),
                "debug graphically");
#endif
  
  GET_OPT_BOOL (argc, argv, "-maxdebug", SET_BOOL(maxdebug),
                "debug by checking certain (expensive) invariants");

  GET_OPT_BOOL (argc, argv, "-pdebug", SET_BOOL(pdebug),
                "debug printouts");
  
  GET_OPT_BOOL (argc, argv, "-test_mode", SET_BOOL(test_mode),
                "run in test mode");
  
  GET_OPT_BOOL (argc, argv, "-auto_mode", SET_BOOL(auto_mode),
                "run in automatic mode");
  
  GET_OPT_BOOL (argc, argv, "-keep_slivers", SET_BOOL(keep_slivers),
                "keep slivers");
  
#if VDT_CUSTOM_SIGNAL_HANDLER
  GET_OPT_BOOL (argc, argv, "-no_intr", SET_BOOL(no_intr),
                "run in no-interrupt mode");
#endif
  
  GET_OPT_BOOL (argc, argv, "-gen_interior_pts", SET_BOOL(gen_interior_pts),
                "generate interior points");
  
  GET_OPT_BOOL (argc, argv, "-gen_near_surf_pts", SET_BOOL(gen_near_surf_pts),
                "generate near-surface points");
  
  GET_OPT_BOOL (argc, argv, "-retri_to_desliver", SET_BOOL(retri_to_desliver),
                "remove slivers by re-triangulation");
  
  GET_OPT_BOOL (argc, argv, "-inject_surf_proj", SET_BOOL(inject_surf_proj),
                "inject projections of points close to surfaces");
  
  GET_OPT_BOOL (argc, argv, "-eval_mesh_size_from_pt",
                SET_BOOL(eval_mesh_size_from_pt),
                "evaluate mesh size from mesh sizes associated with points");
  
  GET_OPT_BOOL (argc, argv, "-dump_tet_wrl", SET_BOOL(dump_tet_wrl),
                "dump tet as VRML");
  
  GET_OPT_BOOL (argc, argv, "-orient_initial_af", SET_BOOL(orient_initial_af),
                "orient initial advancing front");
  
  GET_OPT_BOOL (argc, argv, "-dump_tet_egf", SET_BOOL(dump_tet_egf),
                "dump tet as EGF file");
  
  GET_OPT_BOOL (argc, argv, "-dump_bsf_egf", SET_BOOL(dump_bsf_egf),
                "dump boundary facets as EGF file");
  
  GET_OPT_BOOL (argc, argv, "-dump_data", SET_BOOL(dump_data),
                "dump data in the VDT input format");
  
  GET_OPT_BOOL (argc, argv, "-dump_t3d", SET_BOOL(dump_t3d),
                "dump mesh in the T3D output format");
  
#if GIFACE
  GET_OPT_BOOL (argc, argv, "-giface", SET_BOOL(giface),
                "graphical interface");
#endif
  
#if VDT_BUCKETS
  bucket_size = 0;
  GET_OPT_ARG (argc, argv, "-bucket_size", double, bucket_size,
               { SET_WITH_ARG (bucket_size, g); },
               "bucket size for bucket-based search (>0)");
#endif

  allow_flips  = TRUE;
  GET_OPT_ARG (argc, argv, "-allow_flips", int, allow_flips,
               { SET_WITH_ARG (allow_flips, d); },
               "allow flips (0=FALSE; TRUE otherwise)");
  
  sliver_dihed_angle = VDT_SLIVER_DIHED_ANGLE_DEG * M_PI / 180;
  GET_OPT_ARG (argc, argv, "-sliver_dihed_angle", double, sliver_dihed_angle,
               { SET_WITH_ARG (sliver_dihed_angle, g);
                 sliver_dihed_angle *= M_PI / 180; },
               "sliver dihedral angle in degrees (>0)");
  
  bdry_pt_shift_mult  = VDT_BDRY_PT_DEFAULT_SHIFT_MULT;
  GET_OPT_ARG (argc, argv, "-bdry_pt_shift_mult", double, bdry_pt_shift_mult,
               { SET_WITH_ARG (bdry_pt_shift_mult, g);
                 cmd_line_bdry_pt_shift_mult = TRUE; },
               "bdry pt shift multiplier (>0)");
  
  gen_pt_shift_mult  = VDT_GEN_PT_DEFAULT_SHIFT_MULT;
  GET_OPT_ARG (argc, argv, "-gen_pt_shift_mult", double, gen_pt_shift_mult,
               { SET_WITH_ARG (gen_pt_shift_mult, g);
                 cmd_line_gen_pt_shift_mult = TRUE; },
               "generated pt shift multiplier (>0)");
  
  lp_size_limit = 100;
  GET_OPT_ARG (argc, argv, "-lp_size_limit", int, lp_size_limit,
               { SET_WITH_ARG (lp_size_limit, d);
                 cmd_line_lp_size_limit = TRUE; },
               "limit size of LP");

  near_surf_test = VDT_NEAR_SURF_TEST_DIST;
  GET_OPT_ARG (argc, argv, "-near_surf_test", int, near_surf_test,
               { SET_WITH_ARG (near_surf_test, d); },
                "near-surface test (0=ignore,1=dist,2=circumsphere)");
  
  nranges = 5;
  GET_OPT_ARG (argc, argv, "-nranges", int, nranges,
               SET_WITH_ARG (nranges, d),
               "number of quality ranges");
  
  reset_locs = TRUE;
  GET_OPT_ARG (argc, argv, "-reset_locs", int, reset_locs,
               SET_WITH_ARG (reset_locs, d),
               "reset locations of points (un-randomize) at the end");
  
  GET_OPT_ARG (argc, argv, "-i", char, input_file,
               SET_WITH_ARG (input_file, s),
               "input file");
  
  GET_OPT_ARG (argc, argv, "-true_name", char, input_true_name,
               SET_WITH_ARG (input_true_name, s),
               "input file true name");
  
  if (help)
    usage (argc, argv);

  if (!input_file)
    usage (argc, argv);

  check_presence_of_input ();

#if GIFACE
  if (gdebug && !giface)
    giface = TRUE;
#endif
}

#define SET_VDT_OPT(vdt, op)                            \
   if (!VDT_set_option (vdt, &op)) {                    \
     char buf[1024];                                    \
     sprintf (buf, "Failed to set option \"%s\"",       \
              (op.option ? op.option : "NULL?"));       \
     errexit (buf);                                     \
   }

static int
invalid_input_ (tokens_parser parser, char *msg)
{
  char buf[1024];
  sprintf (buf, "%s (l.%d: %s)", msg,
           tokens_curr_line_num (parser),
           tokens_curr_line_contents (parser));
  errexit (buf);
  return 0;
}
#define invalid_input(parser, msg) invalid_input_ (parser, msg)


static int
handle_unknown_command (void *data)
{
  char buf[512];

  sprintf(buf, "Unknown command: %s", tokens_curr_line_contents (parser));
  errexit(buf);
  data = NULL;
  return (int) (data);
} 


static int had_one_pt = FALSE;

static int
handle_pt (void *data)
{
  VDT vdt = (VDT) data;
  int id;
  double x, y, z;
  double mesh_size;
  
  if        (tokens_total_of_tokens (parser) < 5) {
    invalid_input (parser, "Expected four tokens");
  }

  id = tokens_token_as_int (parser, 2);
  if (id <= 0) errexit ("non-positive pt id");

  x  = tokens_token_as_double (parser, 3);
  y  = tokens_token_as_double (parser, 4);
  z  = tokens_token_as_double (parser, 5);

  mesh_size = VDT_def_mesh_size (vdt);
  if (tokens_total_of_tokens (parser) >= 6) {
    mesh_size = tokens_token_as_double (parser, 6);
  }

  if (!VDT_add_pt ((void *)id, x, y, z, mesh_size, vdt)) {
    char buf[1024];
    sprintf (buf, "%s", VDT_last_error_msg (vdt));
    errexit (buf);
  }
  
  had_one_pt = TRUE;
  
  return TRUE;
} 



static int
handle_bsf (void *data)
{
  VDT vdt = (VDT) data;
  void *pt[3];
  void *surf_id;
  void *reg_id[2];

  if        (tokens_total_of_tokens (parser) < 5) {
    invalid_input (parser, "Expected 5, 6, or 7 tokens");
  }

  pt[0] = (void *) tokens_token_as_int (parser, 2);
  pt[1] = (void *) tokens_token_as_int (parser, 3);
  pt[2] = (void *) tokens_token_as_int (parser, 4);

  surf_id   = (void *) tokens_token_as_int    (parser, 5);
  reg_id[0] = VDT_DEFAULT_REG_ID; 
  if (tokens_total_of_tokens (parser) >= 6) {
    reg_id[0] = (void *) tokens_token_as_int (parser, 6);
  }
  reg_id[1] = VDT_INVALID_REG_ID;
  if (tokens_total_of_tokens (parser) >= 7) {
    reg_id[1] = (void *) tokens_token_as_int (parser, 7);
  }
  
  if (!VDT_add_bsf (pt, surf_id, reg_id, vdt)) {
    char buf[1024];
    sprintf (buf, "%s", VDT_last_error_msg (vdt));
    errexit (buf);
  }
  
  return TRUE;
} 


static int
handle_mesh_size (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;
  
  if        (tokens_total_of_tokens (parser) < 2) {
    invalid_input (parser, "Expected two tokens");
  }

  op.option = VDT_OPTION_DEF_MESH_SIZE;
  op.data.dv = tokens_token_as_double (parser, 2);
  SET_VDT_OPT (vdt, op);

  return TRUE;
} 


static int
handle_bdry_pt_shift_mult (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;
  
  if (had_one_pt) {
    invalid_input (parser, "bdry_pt_shift_mult must precede pt input");
  }
  
  if        (tokens_total_of_tokens (parser) < 2) {
    invalid_input (parser, "Expected two tokens");
  }

  if (!cmd_line_bdry_pt_shift_mult) {
    op.option = VDT_OPTION_BDRY_PT_SHIFT_MULT;
    op.data.dv = tokens_token_as_double (parser, 2);
    SET_VDT_OPT (vdt, op);
  }

  return TRUE;
} 


static int
handle_gen_pt_shift_mult (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;
  
  if        (tokens_total_of_tokens (parser) < 2) {
    invalid_input (parser, "Expected two tokens");
  }

  if (!cmd_line_gen_pt_shift_mult) {
    op.option = VDT_OPTION_GEN_PT_SHIFT_MULT;
    op.data.dv = tokens_token_as_double (parser, 2);
    SET_VDT_OPT (vdt, op);
  }

  return TRUE;
} 


static int
handle_gen_interior_pts (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;

  op.option  = VDT_OPTION_GEN_INTERIOR_PTS;
  op.data.iv = TRUE;
  SET_VDT_OPT (vdt, op);

  return TRUE;
}
 
static int
handle_lp_size_limit (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;

  if        (tokens_total_of_tokens (parser) < 2) {
    invalid_input (parser, "Expected two tokens");
  }

  if (!cmd_line_lp_size_limit) {
    op.option = VDT_OPTION_LP_LIMIT_SIZE;
    op.data.iv = tokens_token_as_int (parser, 2);
    SET_VDT_OPT (vdt, op);
  }

  return TRUE;
}

static int
handle_gen_near_surf_pts (void *data)
{
  VDT vdt = (VDT) data;
  int flag = FALSE;
  vdt_option_t op;

  if        (tokens_total_of_tokens (parser) < 2) {
    flag = TRUE;
  } else {
    flag = tokens_token_as_int (parser, 2);
  }
  
  if (!cmd_line_gen_near_surf_pts) {
    op.option  = VDT_OPTION_GEN_NEAR_SURF_PTS;
    op.data.iv = flag;
    SET_VDT_OPT (vdt, op);
  }

  return TRUE;
}

static int
handle_input_version (void *data)
{
  VDT vdt = (VDT) data;
  vdt_option_t op;

  if        (tokens_total_of_tokens (parser) < 2) {
    invalid_input (parser, "Expected two tokens");
  }
  if (input_version)
    errexit ("Resetting input version");

  NEWP_DUPL_STRING (input_version, tokens_token_as_string (parser, 2),
                    { errexit ("failed to duplicate string"); });

  op.option  = VDT_OPTION_INPUT_VERSION;
  op.data.pv = input_version;
  SET_VDT_OPT (vdt, op);
  
  return TRUE;
}
 


static int
read_input (VDT vdt, char *file)
{
  tokens_command_t commands[] = {
    { "pt",                   handle_pt                  },
    { "bsf",                  handle_bsf                 },
    { "bdry_pt_shift_mult",   handle_bdry_pt_shift_mult  },
    { "gen_pt_shift_mult",    handle_gen_pt_shift_mult   },
    { "mesh_size",            handle_mesh_size           },
    { "gen_interior_pts",     handle_gen_interior_pts    },
    { "lp_size_limit",        handle_lp_size_limit       },
    { "gen_near_surf_pts",    handle_gen_near_surf_pts   },
    { "input_version",        handle_input_version       },
    { "*",                    handle_unknown_command     },
    { NULL, NULL }  
  };

  parser = tokens_new_parser ();
  if (!parser) {
    errexit ("No parser");
  }
  
  if (!tokens_open_file (parser, file)) {
    errexit ("File could not be opened");
  }

  while (tokens_next_line (parser)) {
    if (!tokens_process_command (parser, vdt, commands)) {
      char buf[512];
      sprintf (buf, "failed to process: %s",
               tokens_curr_line_contents (parser));
      errexit (buf);
    }
  }

  tokens_close_curr_file (parser);

  return TRUE;
}



static void
remove_suffix (char *str, char *suffix)
{
  if (ckit_io_file_has_suffix (str, suffix)) {
    int i, n = strlen (str);
    for (i = n-1; i >= 0; i--)
      if (str[i] == '.') {
        str[i] = '\0';
        break;
      }
  }
}

static void
print_module_ids (void)
{
  VDT_log_module_ids ();
}

static void
print_test_tag (VDT vdt, double time)
{
  int connected, unconnected;
  char buf[512];
  char line[1024];
  char suffix[] = ".vin";
  ckit_io_strip_path (buf, input_file);
  remove_suffix (buf, suffix);
  VDT_count_points (vdt, &connected, &unconnected);
  sprintf (line, "VDT test @ %s @ %g @ %d @ %d @ %d @ %d @ %d @\n",
          buf,
          time,
          VDT_count_input_points (vdt),
          VDT_count_input_bsfs (vdt),
          connected,
          VDT_count_bsfs (vdt),
          VDT_count_tets (vdt)
          );
  fprintf (stderr, "# %s\n", line);
}



#if VDT_CUSTOM_SIGNAL_HANDLER
#include <signal.h>

static void (*old_signal_handler) (int sig);

#define FAILED_SIGNAL()                                                 \
  {                                                                     \
    char buf[512];                                                      \
    sprintf (buf, "Failed to install signal handler (%s, %d)\n",        \
            __FILE__, __LINE__);                                        \
    errexit (buf);                                                      \
  }
#define SET_HANDLER(sig, handler)                               \
  { if ((old_signal_handler = signal(sig, handler)) == SIG_ERR) \
      FAILED_SIGNAL();                                          \
  }




typedef void (*action_proc) (VDT vdt);

typedef struct menu_choice_t {
  char            option;
  int             only_with_giface;
  char           *descr;
  action_proc     action;
}              menu_choice_t;


static void
stop_choice (VDT vdt)
{
  exit (EXIT_SUCCESS);
}


#if GIFACE
static void
ginteract_choice (VDT vdt)
{
  char buf[256];
  if (vdt)
    VDT_generate_graphics (vdt);
  sprintf (buf, "Stopped by an interrupt.  Press ^P to continue.\n");
  ESIEventLoop (TRUE, buf);
}
#endif


#if GIFACE
#if 0
static void
tet_wrl_choice (VDT vdt)
{
  FILE *fp;
  char buf[L_tmpnam];

  if (!vdt) {
    errexit ("No triangulation defined yet");
    return;
  }
  
  tmpnam(buf);
  fp = fopen(buf, "w");
  if (!fp) {
    errexit ("Could not open temporary file");
  } else {
    vdt_tet_t *tet;
    
    vdt_generate_graphics (vdt);

    EXVRMLStart (vdt->gmodel, fp);
    tet = NULL;
    while ((tet = (TETP) get_list_next (vdt->tetl, tet))) {
      EXVRMLWriteGeom (fp, vdt->gmodel, tet->go);
    }
    EXVRMLClose (fp);
    fclose(fp);
    { char s[512]; sprintf (s, "Wrote \"%s\"", buf);
      fprintf (stderr, "%s\n", s);
    }
  }
}
#endif
#endif


static void
abort_choice (VDT vdt)
{
  exit (EXIT_FAILURE);
}





static menu_choice_t choices[] = {
  { 's',  FALSE, "stop",                            stop_choice },
  { 'a',  FALSE, "abort (no cleanup)",              abort_choice },
  { 'c',  FALSE, "continue",                        NULL },
#if GIFACE
  { 'X',   TRUE, "graphic interface",               ginteract_choice },
#endif
#if GIFACE
#if 0
  { 't',   TRUE, "write tet wrl file",              tet_wrl_choice },
#endif
#endif
  { '\0', FALSE, NULL,                              NULL }
};



static int received_an_interrupt = FALSE;

static void 
signal_handler (int sig)
{
#define DO_GLUE(A, B)  A##B
#define GLUE(A, B)     DO_GLUE(A, B)
#define DO_STRNGZ(S)   #S
#define STRNGZ(S)      DO_STRNGZ(S) 
#define SIGNLNM(S)     GLUE(STRNGZ(SIG),STRNGZ(S))
#define PRERR(NAME)    fprintf(stderr,                                  \
                               "vdt_signal_handler: signal %s\n",       \
                               SIGNLNM(NAME))
#define CASE(NAME)                                                      \
   case GLUE(SIG,NAME):                                                 \
     PRERR(NAME);                                                       \
     SET_HANDLER(SIGINT, SIG_DFL);                                      \
     if (isatty (2)) {                                                  \
       fprintf(stderr, "      ^c will interrupt the program\n");        \
       fflush(stderr);                                                  \
       { int i = 1; while (i) {;} }                                     \
     } else {                                                           \
       errexit ("Caught signal " SIGNLNM(NAME));                        \
     }                                                                  \
     break

  switch (sig) {
  CASE(FPE);
  CASE(SEGV);
  CASE(BUS);
  CASE(ILL);
  case SIGINT:
    {
      received_an_interrupt = TRUE;
      SET_HANDLER(SIGINT, &signal_handler);
    }
    break;
  default:
    fprintf(stderr, "VDT signal handler: signal %d\n", sig);
    fflush(stderr);
    exit(EXIT_FAILURE);
    break;
  }

  return;
}



static int
menu (VDT vdt)
{
  int i, c;
  int forever = TRUE;

  while (forever) {
    fprintf(stderr, "Interrupt.\n");
    i = 0;
    while (choices[i].option != '\0') {
      if ((!choices[i].only_with_giface)
#         if GIFACE
	  || giface
#         endif
	  )
	fprintf(stderr, "   %c   %s\n", choices[i].option, choices[i].descr);
      i++;
    }
    fprintf(stderr,   "         Your choice? ");
    while (forever) {
      c = fgetc(stdin);
      i = 0;
      while (choices[i].option != '\0') {
        if (choices[i].option == c
	    && ((!choices[i].only_with_giface)
#               if GIFACE
		|| giface
#               endif
		))
          return i;
        i++;
      }
    }
  }
  return 0;
}


static int
check_interrupt (VDT vdt)
{
  if (!received_an_interrupt) {
    return TRUE;
  } else {
    int c;
    signal(SIGINT, SIG_IGN);
    c = menu (vdt);
    if (choices[c].action) choices[c].action (vdt);
    SET_HANDLER(SIGINT, &signal_handler);
    received_an_interrupt = FALSE;
    return TRUE;
  }
}

#endif /* VDT_CUSTOM_SIGNAL_HANDLER */


#if GIFACE
static void
dump_image (char *input_file)
{
  char buf[1024];
  char img[1024];
  char suffix[] = ".vin";
  ckit_io_strip_path (buf, input_file);
  remove_suffix (buf, suffix);
  sprintf (img, "image %s.jpg", buf);
  ESIHandleCmd (img);
}
#endif

static void
set_vdt_opts (VDT vdt)
{
  vdt_option_t op;

  op.option  = VDT_OPTION_INPUT;
  op.data.pv = input_file;
  SET_VDT_OPT (vdt, op);

  if (input_version) {
    op.option  = VDT_OPTION_INPUT_VERSION;
    op.data.pv = input_version;
    SET_VDT_OPT (vdt, op);
  }

  if (input_true_name) {
    op.option  = VDT_OPTION_INPUT_TRUE_NAME;
    op.data.pv = input_true_name;
    SET_VDT_OPT (vdt, op);
  }

  op.option  = VDT_OPTION_BUCKET_SIZE;
  op.data.dv = bucket_size;
  SET_VDT_OPT (vdt, op);
  
  op.option  = VDT_OPTION_SLIVER_DIHED_ANGLE;
  op.data.dv = sliver_dihed_angle;
  SET_VDT_OPT (vdt, op);
  
  op.option  = VDT_OPTION_BDRY_PT_SHIFT_MULT;
  op.data.dv = bdry_pt_shift_mult;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_GEN_PT_SHIFT_MULT;
  op.data.dv = gen_pt_shift_mult;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_LP_LIMIT_SIZE;
  op.data.iv = lp_size_limit;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_NEAR_SURF_TEST;
  op.data.iv = near_surf_test;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_NRANGES;
  op.data.iv = nranges;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_PDEBUG;
  op.data.iv = pdebug;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_MAXDEBUG;
  op.data.iv = maxdebug;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_KEEP_SLIVERS;
  op.data.iv = keep_slivers;
  SET_VDT_OPT (vdt, op);

#if GIFACE
  op.option  = VDT_OPTION_GDEBUG;
  op.data.iv = gdebug;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_GIFACE;
  op.data.iv = giface;
  SET_VDT_OPT (vdt, op);
#endif

  op.option  = VDT_OPTION_ALLOW_FLIPS;
  op.data.iv = allow_flips;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_GEN_INTERIOR_PTS;
  op.data.iv = gen_interior_pts;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_GEN_NEAR_SURF_PTS;
  op.data.iv = gen_near_surf_pts;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_RETRI_TO_DESLIVER;
  op.data.iv = retri_to_desliver;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_INJECT_SURF_PROJ;
  op.data.iv = inject_surf_proj;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_EVAL_MESH_SIZE_FROM_PT;
  op.data.iv = eval_mesh_size_from_pt;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_DUMP_TET_WRL;
  op.data.iv = dump_tet_wrl;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_DUMP_TET_EGF;
  op.data.iv = dump_tet_egf;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_DUMP_BSF_EGF;
  op.data.iv = dump_bsf_egf;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_DUMP_DATA;
  op.data.iv = dump_data;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_DUMP_T3D;
  op.data.iv = dump_t3d;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_ORIENT_INITIAL_AF;
  op.data.iv = orient_initial_af;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_REPORT_OPTS;
  op.data.iv = report_opts;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_RESET_LOCS;
  op.data.iv = reset_locs;
  SET_VDT_OPT (vdt, op);

  op.option  = VDT_OPTION_TEST_MODE;
  op.data.iv = test_mode;
  SET_VDT_OPT (vdt, op);
}


static VDT
re_triangulate (VDT vdt)
{
  void *id;
  double x, y, z;
  void *pt;
  void *bsf;
  void *ptid[3];
  void *surfid;
  void *tetid[2];
  void *regid[2];
  VDT newvdt;
  void *tet;
  void *tetptid[4];
  void *adjtet[4];
  double ms;
  char buf[1024];
  double min_angle, max_angle, angles[6];
  int result;
  vdt_option_t op;

  newvdt = VDT_new_mesher ();
  if (!newvdt) errexit ("Could not create a new mesher");
  set_vdt_opts (newvdt);

  op.option  = VDT_OPTION_GEN_INTERIOR_PTS;
  op.data.iv = FALSE; /* got all the points I need */
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_GEN_NEAR_SURF_PTS;
  op.data.iv = FALSE; /* got all the points I need */
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_DEF_MESH_SIZE;
  op.data.dv = VDT_def_mesh_size (vdt);
  SET_VDT_OPT (newvdt, op);

  op.option  = VDT_OPTION_NEAR_SURF_TEST;
  /* better check that the circumcenters do not fall
     on some bounding surface */
  if (near_surf_test == VDT_NEAR_SURF_TEST_IGNORE)
    near_surf_test = VDT_NEAR_SURF_TEST_DIST;
  op.data.iv = near_surf_test; 
                         
  SET_VDT_OPT (newvdt, op);

  pt = NULL;
  while ((pt = VDT_next_pt (&id, &x, &y, &z, &ms, pt, vdt))) {
    if (!VDT_add_pt (id, x, y, z, ms, newvdt)) {
      sprintf (buf, "%s", VDT_last_error_msg (newvdt));
      errexit (buf);
    }
  }

  bsf = NULL;
  while ((bsf = VDT_next_bsf (&id, ptid, &surfid, regid, tetid, bsf, vdt))) {
    if (!VDT_add_bsf (ptid, surfid, regid, newvdt)) {
      sprintf (buf, "%s", VDT_last_error_msg (newvdt));
      errexit (buf);
    }
  }

  tet = NULL;
  while ((tet = VDT_next_tet (tetptid, &regid[0], adjtet, tet, vdt))) {
    VDT_tet_get_angle_quality (tet, &min_angle, &max_angle, angles);
    if (min_angle < sliver_dihed_angle) {
      if (VDT_tet_get_circumcenter (tet, vdt, &x, &y, &z)) {
        if (!VDT_add_pt_with_check (id, x, y, z,
                                    VDT_def_mesh_size (vdt), newvdt)) {
          sprintf (buf, "%s", VDT_last_error_msg (newvdt));
          errexit (buf);
        }
      }
    }
  }
  
  VDT_delete_mesher (vdt);

  result = VDT_triangulate (newvdt);
  if (!result) {
    char buf[1024];
    sprintf (buf, "%s", VDT_last_error_msg (newvdt));
    errexit (buf);
  } 
 
  VDT_summary (newvdt);

  return newvdt;
}


static void
init_graphics_settings (void)
{
#if GIFACE
#define LON  TRUE
#define LOFF FALSE
  if (giface) {
    if (auto_mode) 
      ESIHandleCmd ("NEWFRAME name vdt x 500 y 1 width 192 height 192 wire");
    else
      ESIHandleCmd ("NEWFRAME name vdt x 500 y 1 width 320 height 320 wire");
    {
      int i;
      for (i = 0; i < 1000; i++) {
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()),
                                    VDT_QUAL_LAYER_BASE + i, LOFF);
      }
    }
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_DELETED_FF_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_PT_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_ADD_PT_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_PT_W_CHECK_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_DELAUNAY_BSF_LAYER, LON);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_PT_NUM_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_FF_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_NONPOS_VOL_SLIVER_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_SMALL_DIHED_SLIVER_LAYER, LOFF);
    EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_REMOVED_SLIVER_LAYER, LOFF);

    if (test_mode) {
      ESIHandleCmd ("fit");
      ESIHandleCmd ("render shade");
    }
  }
#endif
}

static void
init_exc_handler (void)
{
#if VDT_CUSTOM_SIGNAL_HANDLER
  if (no_intr) {
    int check_interrupt (VDT vdt);
    vdt_option_t op;
    op.option  = VDT_OPTION_CHECK_INTERRUPT_FUNC;
    op.data.fv = (void (*)()) check_interrupt;
    SET_VDT_OPT(vdt, op);
    SET_HANDLER (SIGINT, &signal_handler);
  }
#endif  
}

static void
final_graphic_interaction (void)
{
#if GIFACE
  int forever = TRUE;
  if (giface) {
    while (forever) {
      if (test_mode) {
        int beenhere = 0;
        ESIHandleCmd ("view axes off");
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_DELAUNAY_BSF_LAYER, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_NON_DELAUNAY_BSF_LAYER, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_TET_LAYER, LON);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_QUAL_LAYER_0, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_QUAL_LAYER_1, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_QUAL_LAYER_2, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_QUAL_LAYER_3, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_QUAL_LAYER_4, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_NONPOS_VOL_SLIVER_LAYER, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_SMALL_DIHED_SLIVER_LAYER, LOFF);
        EVSetLayerOnOffNoRedisplay (EMAnyViewP (ESIModel ()), VDT_REMOVED_SLIVER_LAYER, LOFF);
        ESIHandleCmd ("render shade");
        ESIHandleCmd ("fit");
        do {
          ESIEventLoop (!auto_mode, "Triangulation done; press Ctrl-p to exit             ");
        } while (auto_mode && ++beenhere < 100000);
        dump_image (input_file);
        break;
      } else {
        ESIEventLoop (TRUE, "Triangulation done; press ^x to exit                 ");
      }
    }
  }
#endif
}


static void
graphic_interaction (void)
{
#if GIFACE
  if (giface) {
    ESIEventLoop (TRUE, "Before re-triangulation; press ^x to exit, ^p to continue");
  }
#endif
}

static double
min_dih_angle (VDT vdt)
{
  int num_non_positive_vol_tets;
  int num_small_dihed_angle_tets;
  int quality_measure;
  double
    min_quality, max_quality, geom_mean_quality,
    rms_quality, min_volume, max_volume, min_dihed_angle, max_dihed_angle,
    avg_dihed_angle;
  
  VDT_get_quality (vdt,
                   &num_non_positive_vol_tets,
                   &num_small_dihed_angle_tets,
                   &quality_measure,
                   &min_quality,
                   &max_quality,
                   &geom_mean_quality,
                   &rms_quality,
                   &min_volume,
                   &max_volume,
                   &min_dihed_angle,
                   &max_dihed_angle,
                   &avg_dihed_angle);

  return min_dihed_angle;
}

static void
triangulate (VDT vdt)
{
  
  if (!VDT_triangulate (vdt)) {
    char buf[1024];
    sprintf (buf, "%s", VDT_last_error_msg (vdt));
    errexit (buf);
  } 
  VDT_summary (vdt);
   
  if (retri_to_desliver) {
    int count = 10;
    while (min_dih_angle (vdt) < sliver_dihed_angle &&
           count-- > 0) {
      graphic_interaction ();
      vdt = re_triangulate (vdt);
    }
  }

  VDT_dump (vdt);
}

static int
run_vdt (char *file)
{
  ckit_timer_t *timer = ckit_timer_new ();

  fprintf (stderr, "# Program started [%s]\n", ckit_get_date_and_time ());
  ckit_timer_restart (timer);

  vdt = VDT_new_mesher ();
  if (!vdt) errexit ("Could not create context");
  set_vdt_opts (vdt);
    
  read_input (vdt, file);
  
  init_graphics_settings ();
  init_exc_handler ();

  triangulate (vdt);

  if (test_mode) {
    print_module_ids ();
    print_test_tag (vdt, ckit_timer_current (timer));
  }

  final_graphic_interaction ();

  fprintf (stderr, "# Program finished [%s]\n", ckit_get_date_and_time ());
  ckit_timer_free (timer);  
  return TRUE;
}




int
main (int argc, char **argv)
{
#if GIFACE
  unsigned long m = ESI_GRAPHIC_EDITOR_MASK;
#endif

  process_options (argc, argv);

#if GIFACE
  if (giface) {
    ESIBuildInterface (m, argc, argv);
    ESIPopup ();
    ESIHandleCmd ("EDGECOLOR white");
    ESIHandleCmd ("EDGEFLAG off");
    ESIHandleCmd ("COLOR yellow");
    ESIHandleCmd ("MSIZE 10");
    ESIHandleCmd ("MTYPE 1");
    ESIHandleCmd ("FONT \"9x15\"");
    ESIHandleCmd ("SELCRIT inside");
  }
#endif
  
  fprintf (stderr, "   ======================================================\n");
  fprintf (stderr, "         (V)ariational (D)elaunay (T)riangulation        \n");
  fprintf (stderr, "         VDT Copyright (C) 1998--2000, Petr Krysl        \n");
  fprintf (stderr, "   ======================================================\n");

  return (run_vdt (input_file) ? 0 : 1);
}

#if GIFACE
#if 0
static int
center_view (NODE ptr, NODE elem)
{
  CkitPoint3 *loc = (CkitPoint3 *)ptr;
  EView *vp = (EView *)elem;
  EVSetOriginVC (vp, loc);
  EVSetVRP (vp, loc);
  EVFastRedraw (vp);
  return TRUE;
}
#endif

static void
find_entity_cb (Widget w, XtPointer ptr, XtPointer call_data)
{
#if 1
  ERptErrMessage(1, 2, "Not enabled", ERROR_GRADE);
#else
  int kp = 0;
  void *id;
  PTP pt;
  
  if        ((kp = TypeInIsKeywordInLine (1, "PT")) > 0)  {
    id = (void *) TypeInGetTokenAsInt (kp+1);
    pt = vdt_find_pt (vdt, id);
    if (!pt) {
      goto errexit;
    } else {
      EMDispatchToDependentViews (vdt->gmodel, center_view,
                                  (caddr_t)&pt->loc);
    }
  } else if ((kp = TypeInIsKeywordInLine (1, "TET")) > 0)  {
  } else if ((kp = TypeInIsKeywordInLine (1, "BSF")) > 0)  {
  }

  return;
  
 errexit:
  ERptErrMessage(1, 2, "No such entity", ERROR_GRADE);
#endif
}
#endif

#if GIFACE
static void
toggle_layer_cb (Widget w, XtPointer ptr, XtPointer call_data)
{
  Boolean s;
  Arg al[1];
  int ac;
  char buf[132];
  int layer = (int)ptr;
  
  ac = 0;
  XtSetArg(al[ac], XtNstate, &s); ac++;
  XtGetValues(w, al, ac);
  if (s)
    sprintf(buf, "LAYER ON %d", layer);
  else
    sprintf(buf, "LAYER OFF %d", layer);
  ESIHandleCmd(buf);
}
#endif



#if GIFACE
void
ESICustomize (Widget parent_pane)
{
  int argn;
  Arg args[13];
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("pt_visibility",
               "points",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_PT_LAYER);

  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("pt_w_check_visibility",
               "points w check",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_PT_W_CHECK_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("add_pt_visibility",
               "added points",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_ADD_PT_LAYER);

  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("pt_num_visibility",
               "point numbers",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_PT_NUM_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, TRUE); argn++;
  ESIAddButton("bsf_visibility",
               "delaunay facets",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb,
               (XtPointer)VDT_DELAUNAY_BSF_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, TRUE); argn++;
  ESIAddButton("bsf_non_visibility",
               "non-delaunay facets",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb,
               (XtPointer)VDT_NON_DELAUNAY_BSF_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, TRUE); argn++;
  ESIAddButton("tet_visibility",
               "tets",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_TET_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_qual_visibility",
               "tet quality [0.0;0.2]",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_QUAL_LAYER_0);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_qual_visibility",
               "tet quality [0.2;0.4]",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_QUAL_LAYER_1);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_qual_visibility",
               "tet quality [0.4;0.6]",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_QUAL_LAYER_2);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_qual_visibility",
               "tet quality [0.6;0.8]",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_QUAL_LAYER_3);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_qual_visibility",
               "tet quality [0.8;1.0]",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_QUAL_LAYER_4);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("nonpos_vol_sliver_visibility",
               "non-pos vol slivers",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_NONPOS_VOL_SLIVER_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("dihed_sliver_visibility",
               "small dihedral slivers",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_SMALL_DIHED_SLIVER_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("tet_sliver_visibility",
               "removed slivers",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_REMOVED_SLIVER_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("ff_visibility",
               "front faces",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_FF_LAYER);
  
  argn = 0;
  XtSetArg (args[argn], XtNstate, FALSE); argn++;
  ESIAddButton("octree_visibility",
               "octree",
               toggleWidgetClass, parent_pane,
               args, argn, toggle_layer_cb, (XtPointer)VDT_OCTREE_LAYER);
  
  argn = 0;
  ESIAddPopupDialog("find_entity", "find entity",
                    "entity + id [PT, TET, BSF] >",
                    "PT 1", parent_pane, args, argn,
                    find_entity_cb, NULL, ESIDialogValueAny,
                    NULL);
}
#endif


