
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtff.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdtff.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtff) 
 
/* RAW: the code marked is not definitive; don't delete this line */

/*
   Add a front face.
 */
vdt_ff_t *
vdt_ff_add (vdt_tet_t *tet, int face,
            vdt_pt_t *pt, vdt_pt_t *opt0, vdt_pt_t *opt1,
            vdt_t *vdt, BSFP bsf)
{
  vdt_ff_t *ff;

  ff = NEWP_OBJ (vdt_ff_t); CHECK_MEM (ff);

  ff->ontet     = tet;
  assert (ff->ontet);
  assert (face >= 0 && face < 4);
  ff->face      = face;
  ff->is_active = TRUE;
  
  ff->pt[0]     = pt;
  ff->pt[1]     = opt0;
  ff->pt[2]     = opt1;

  ff->onbsf     = bsf;
  
  if (vdt->pdebug) {
    if (pt == opt0 || pt == opt1 || opt0 == opt1) {
      char buf[64];
      sprintf (buf, "Invalid ff pt = { %d %d %d }",
               (int) pt->id, (int) opt0->id, (int) opt1->id);
      warn (buf);
    }
  }
  
#if GIFACE
  ff->go        = NULL;
#endif

  return ff;
}

void
vdt_ff_delete (vdt_ff_t *ff, vdt_t *vdt)
{
#if GIFACE
  if (vdt->gmodel && ff->go) {
    EASValsSetLayer (VDT_DELETED_FF_LAYER);
    EGWithMaskChangeAttributes (LAYER_MASK, ff->go);
  }
#endif
  NEWP_FREE_OBJ (ff);
}


#if GIFACE
void
vdt_ff_generate_graphics (vdt_ff_t *ff, vdt_t *vdt)
{
  if (vdt->gmodel && !ff->go) {
    WCRec pts[3];
    int i;
    FOR (i, 3) { pts[i] = ff->pt[i]->loc; }
    ff->go = CreateTriangle3D (pts);
    EMNoDrawAddGraphicsToModel (vdt->gmodel, ff->go);
    EASValsSetColor (vdt_color_to_pixel ("slateblue"));
    EASValsSetShrink (0.8);
    EASValsSetFillStyle (FILL_SOLID);
    EASValsSetLayer (VDT_FF_LAYER);
    EGWithMaskChangeAttributes (FILL_MASK | COLOR_MASK |
                                SHRINK_MASK | LAYER_MASK, ff->go);
  }
}

void
vdt_ff_recolor_active (vdt_ff_t *ff)
{
  if (ff->go) {
    EASValsSetColor (vdt_color_to_pixel ("yellow"));
    EGWithMaskChangeAttributes (COLOR_MASK, ff->go);
  }
}
#endif

