#!/bin/ksh

default_dirs="./ ./RCS/ @VDT_DIR@/inputs/ @VDT_DIR@/inputs/RCS/"
suffixes='.vin .gz .vin.gz ,v ,v.gz .vin,v .vin,v.gz'

function usage_and_exit {
cat <<EOF
          (V)ariational (D)elaunay (T)riangulation        
          VDT Copyright (C) 1998--2000, Petr Krysl           

Usage: $(basename $0) [-i] inp [ -vdt_dir d ]
   -vdt_dir d
      Assume the VDT distribution is under directory d.

Use this script to get a local copy of a VDT input file 
stored in some default directory in one of the standard forms.
This script is able to figure out how to uncompress or check out
an RCS file.

The input file can be specified either by giving the full path
(e.g. /home/cairo/pkrysl/VDT/inputs/tire), or by giving just the name
(tire or tire.vin) and relying on the ability of the script to find it
in the default directories.

Directories that are searched: 
  specified directory when full path or relative path was given
`for d in $(echo $default_dirs); do echo "  $d"; done`

Suffixes: none and $suffixes

EOF
exit 1
}

# find -i option and collect all remaining arguments in args

input='???'
vdt_exec=vdt
vdt_dir="@VDT_DIR@"
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
    -i)
        if [ $# -ge 1 ] ; then
            input=$1; shift
        else
            usage_and_exit
        fi
        ;;
    -vdt_dir)
        if [ $# -ge 1 ] ; then
            vdt_dir=$1; shift
        else
            usage_and_exit
        fi
        ;;
     -help)
        usage_and_exit
        ;;
    *)
        input=$a
        ;;
    esac
done

if [ x"$input" = x'???' ] ; then
    usage_and_exit
fi



function local_file {
    n_=$1
    echo ./$(basename ${n_})
}

function cleanup {
   echo Killed
}

function has_suffix {
    name=$1; shift
    suffix=$1
    if [ ${name%%${suffix}} = ${name} ] ; then 
        echo f
    else
        echo t
    fi
}

specdir=$(dirname $input)/
t=$(basename $input)
input=$t
if [ x"$vdt_dir" = x ] ; then
   vdt_dirs="./inputs/"
else
   vdt_dirs="$vdt_dir/inputs/ $vdt_dir/inputs/RCS/"
fi

for d in "$specdir" $(echo $vdt_dirs) $(echo $default_dirs)
do 
   for s in "" $(echo $suffixes)
   do
     n=${d}${input}${s}
     if [ -f $n ] ; then 
        if [ $(has_suffix $n ".vin") = t ] ; then
           echo Found $n
           input=$n
           exit 0
        fi

        if [ $(has_suffix $n ".vin,v") = t ] ; then 
           input=$(local_file ${n%%,v})
           echo Checking out $n into $input
           co -q -p $n > $input
           exit 0
        fi

        if [ $(has_suffix $n ".vin,v.gz") = t ] ; then
           rcsfile=$(local_file ${n%%.gz})
           echo De-compressing $n into $rcsfile
           gunzip -qcf $n > $rcsfile
           input=$(local_file ${n%%,v.gz})
           echo Checking out $rcsfile into $input
           co -q -p $rcsfile > $input
           echo Removing temporary $rcsfile
           rm -f $rcsfile
           exit 0
        fi

        if [ $(has_suffix $n ".gz") = t -o $(has_suffix $n ".vin.gz") = t ] ; then
           input=$(local_file ${n%%${s}})
           echo De-compressing $n into $input
           gunzip -qcf $n > $input
           exit 0
        fi

        if [ $(has_suffix $n "") = t ] ; then
           echo Found $n
           input=$n
           exit 0
        fi
     fi
   done
done

if [ ! -f "$input" ] ; then
   echo "*** Input $input does not exist?!"
   exit 1
fi

