/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdthierlobs.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdthierlobs.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdthierlobs) 
 
#include "vdthierlobs.h"

#if VDT_USE_HIERLOBS

static void
error_notify (char *file, int line, char *explanation)
{
  char *buf;
  buf = NEWP_OBJ_ARRAY (strlen (file) + strlen (explanation) + 32, char);
  if (buf) {
    sprintf (buf, "%s (l.%d): %s", file, line, explanation);
    errexit (buf);
    NEWP_FREE_OBJ (buf);
  } else {
    errexit ("Error in LOBS");
  }
}
static double
csrad (CkitPoint3 *l0, CkitPoint3 *l1)
{
  CkitPoint3 d;
  VECMAC_3_VEC (d, =, (*l0), -, (*l1));
  return VECMAC_NORM (d);
}

static void
descend (lobs_3d_front_end toplobs,
         G3D_box_t *leaf,
         LIST pl,
         int max_bo_per_leaf,
         int near_surf_test)
{
  BSFP bsf;
  lobs_3d_front_end lobs;
  int i;

  if (list_length (pl) <= max_bo_per_leaf) {
    G3D_box_t b;

    /* create leaf lobs and fill it; expand its box to enclose
       all bsfs
     */
    lobs = lobs_3d_new_instance (error_notify, 10); CHECK_MEM (lobs);
 
    bsf = NULL;
    while ((bsf = (BSFP) get_list_next (pl, bsf))) {
      switch (near_surf_test) {
      case VDT_NEAR_SURF_TEST_CIRCUMSPHERE:
        {
          CkitPoint3 csc;
          double csr;
          if (!vdt_3_pt_circumsphere (bsf->pt[0], bsf->pt[1], bsf->pt[2],
                                      &csc)) {
            errexit ("No circumsphere for facet?");
          }
          csr = csrad (&bsf->pt[0]->loc, &csc);
          G3D_INIT_BOX (b, csc);
          G3D_INFLATE_BOX (b, csr);
        }
        break;
      case VDT_NEAR_SURF_TEST_DIST:
      default:
        G3D_INIT_BOX (b, bsf->pt[0]->loc);
        G3D_UPDT_BOX (b, bsf->pt[1]->loc);
        G3D_UPDT_BOX (b, bsf->pt[2]->loc);
        break;
      }
      lobs_3d_add_object (lobs, bsf,
                          b.lower.x, b.lower.y, b.lower.z,
                          b.upper.x, b.upper.y, b.upper.z);
      G3D_UPDT_BOX ((*leaf), b.lower);
      G3D_UPDT_BOX ((*leaf), b.upper);
    } /* while */

    /* add the leaf lobs to the top lobs */
    lobs_3d_add_object (toplobs, lobs,
                        leaf->lower.x, leaf->lower.y, leaf->lower.z,
                        leaf->upper.x, leaf->upper.y, leaf->upper.z);
  } else {
    G3D_box_t desc[8], bz0, bz1, by0, by1;
    LIST dl[8] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

    /* compute boxes of the descendants in the octree */
    G3D_SPLIT_ACROSS_DIR (bz0, bz1, (*leaf), z);
    G3D_SPLIT_ACROSS_DIR (by0, by1, bz0, y);
    G3D_SPLIT_ACROSS_DIR (desc[0], desc[1], by0, x);
    G3D_SPLIT_ACROSS_DIR (desc[2], desc[3], by1, x);
    G3D_SPLIT_ACROSS_DIR (by0, by1, bz1, y);
    G3D_SPLIT_ACROSS_DIR (desc[4], desc[5], by0, x);
    G3D_SPLIT_ACROSS_DIR (desc[6], desc[7], by1, x);
    
    /* split leaf list into descendant lists */
    bsf = NULL;
    while ((bsf = (BSFP) get_list_next (pl, bsf))) {
      FOR (i, 8) {
        if        (   G3D_PT_INSIDE_BOX (desc[i], bsf->pt[0]->loc)
                   || G3D_PT_INSIDE_BOX (desc[i], bsf->pt[1]->loc)
                   || G3D_PT_INSIDE_BOX (desc[i], bsf->pt[2]->loc)
                   ) {
          if (dl[i] == NULL) {
            dl[i] = make_list (); CHECK_MEM (dl[i]);
          }
          add_to_tail (dl[i], bsf);
          break;
        }
      }
    } /* while */

    /* recursive descent */
    FOR (i, 8) {
      if (dl[i]) {
        descend (toplobs, &desc[i], dl[i], max_bo_per_leaf, near_surf_test);
        free_list (dl[i], NOT_NODES);
      }
    } /* FOR */
  } /* else */
}

void
vdt_hierlobs_fill (vdt_t *vdt)
{
  vdt->hierlobs = lobs_3d_new_instance (error_notify, 2);
  CHECK_MEM (vdt->hierlobs);
  descend (vdt->hierlobs, &vdt->bbox, vdt->bsfl, 200, vdt->near_surf_test);
#if 0
  lobs_3d_print (vdt->hierlobs, stderr); 
#endif
}

void
vdt_hierlobs_clean (vdt_t *vdt)
{
  lobs_3d_front_end lobs;
  G3D_box_t box;
  CkitPoint3 mininf = {-FLT_MAX,-FLT_MAX,-FLT_MAX}; 
  CkitPoint3 pluinf = {+FLT_MAX,+FLT_MAX,+FLT_MAX}; 
  G3D_INIT_BOX (box, mininf); G3D_UPDT_BOX (box, pluinf);
  lobs_3d_search (vdt->hierlobs,
                  box.lower.x, box.lower.y, box.lower.z,
                  box.upper.x, box.upper.y, box.upper.z);
  lobs = lobs_3d_first_overlapping_object (vdt->hierlobs);
  while (lobs) {
    lobs_3d_free_instance (lobs);
    lobs = lobs_3d_next_overlapping_object (vdt->hierlobs);
  }
  lobs_3d_free_instance (vdt->hierlobs);
}

#endif /* VDT_USE_HIERLOBS */
