
#include "vdtmaster.h"

static char vcid[] = "$Id: vdtoctree.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtoctree) 
 

typedef struct vdt_ot_leaf_t {
  union {
    struct vdt_ot_leaf_t      *leaf[8];
    double                     mesh_size;
  }                                          data;
  char                                       is_terminal;
}              vdt_ot_leaf_t;




static LEAFP
new_leaf (void)
{
  LEAFP leaf;
  int i;
  leaf = NEWP_OBJ (LEAF); CHECK_MEM (leaf);
  leaf->is_terminal = TRUE;
  FOR (i, 8) leaf->data.leaf[i] = NULL;

  return leaf;
}

static int
refine (OTP ot, LEAFP leaf, CkitPoint3 *center, double size)
{
  double ms = (*ot->mesh_size) (ot->mesh_size_func_data, center);
  
  if (size > 2.5 * ms) {
    if (leaf->is_terminal) {
      int i;
      CkitPoint3 c;
      double qsize = size / 4, hsize = size / 2;

      FOR (i, 8) {
        VECMAC_2_VEC (c, =, (*center));
        c.x += qsize * g3dhex_vertex_param_coord[i][0];
        c.y += qsize * g3dhex_vertex_param_coord[i][1];
        c.z += qsize * g3dhex_vertex_param_coord[i][2];
        leaf->data.leaf[i] = new_leaf ();
        refine (ot, leaf->data.leaf[i], &c, hsize);
      }
      leaf->is_terminal = FALSE;
    }
  } else {
    if (leaf->is_terminal) {
      leaf->data.mesh_size = ms;
    } else {
      errexit ("should not get here");
    }
  }
  return TRUE;
}

int
vdt_ot_refine (OTP ot,
               vdt_mesh_size_func mesh_size,
               void *mesh_size_func_data)
{
  ot->mesh_size = mesh_size;
  ot->mesh_size_func_data = mesh_size_func_data;
  return refine (ot, ot->root, &ot->center, ot->size);
}



OTP
vdt_ot_new (CkitPoint3 *center,
            double size
            )
{
  OTP ot;
  ot = NEWP_OBJ (OT); CHECK_MEM (ot);
  ot->center              = *center;
  ot->size                = size;
  ot->root                = new_leaf ();
  ot->mesh_size           = NULL;
  ot->mesh_size_func_data = NULL;
  ot->gen_pt              = NULL;
  ot->gen_pt_func_data    = NULL;
  return ot;
}



#if GIFACE
static void
generate_graphics (EModel *gmodel, CkitPoint3 *center, double size)
{
  GraphicObj *go;
  WCRec pts[2];
  double hsize = size / 2;
  VECMAC_2_VEC (pts[0], =, (*center));
  VECMAC_SCALAR (pts[0], -=, hsize); 
  VECMAC_2_VEC (pts[1], =, (*center));
  VECMAC_SCALAR (pts[1], +=, hsize); 
  go = CreateRAPrism (pts);
  EASValsSetColor (vdt_color_to_pixel ("white"));
  EASValsSetLayer (VDT_OCTREE_LAYER);
  EGWithMaskChangeAttributes (COLOR_MASK | LAYER_MASK, go);
  EMNoDrawAddGraphicsToModel (gmodel, go);
}

static int
leaf_generate_graphics (OTP ot, EModel *gmodel,
                        LEAFP leaf, CkitPoint3 *center, double size)
{
  if (leaf->is_terminal) {
    generate_graphics (gmodel, center, size);
  } else {
    int i;
    CkitPoint3 c;
    double qsize = size / 4, hsize = size / 2;
    
    FOR (i, 8) {
      VECMAC_2_VEC (c, =, (*center));
      c.x += qsize * g3dhex_vertex_param_coord[i][0];
      c.y += qsize * g3dhex_vertex_param_coord[i][1];
      c.z += qsize * g3dhex_vertex_param_coord[i][2];
      leaf_generate_graphics (ot, gmodel, leaf->data.leaf[i], &c, hsize);
    }
  }
  return TRUE;
}

int
vdt_ot_generate_graphics (OTP ot, EModel *gmodel)
{
  return leaf_generate_graphics (ot, gmodel, ot->root, &ot->center, ot->size);
}
#endif



static int
gen_pts (OTP ot, LEAFP leaf, CkitPoint3 *center, double size)
{
  if (leaf->is_terminal) {
    CkitPoint3 p;
    double ms = leaf->data.mesh_size;
    double nmd = 0.95 * ms;     /* distance to closest point */
    double bmd = 0.87 * ms;     /* distance to closest facet */
    double hsize = size / 2;
#define GEN(C) (*ot->gen_pt) (ot->gen_pt_func_data, C, ms, nmd, bmd)

    /* center of the leaf */
    GEN (center);

    /* centers of the `plus'-side edges */
    VECMAC_3_SCALAR (p, =, center->x + hsize, center->y, center->z + hsize);
    GEN (&p);
    VECMAC_3_SCALAR (p, =, center->x, center->y + hsize, center->z + hsize);
    GEN (&p);
    VECMAC_3_SCALAR (p, =, center->x + hsize, center->y + hsize, center->z);
    GEN (&p);

    return TRUE;
  } else {
    int i;
    CkitPoint3 c;
    double qsize = size / 4, hsize = size / 2;
    
    FOR (i, 8) {
      VECMAC_2_VEC (c, =, (*center));
      c.x += qsize * g3dhex_vertex_param_coord[i][0];
      c.y += qsize * g3dhex_vertex_param_coord[i][1];
      c.z += qsize * g3dhex_vertex_param_coord[i][2];
      gen_pts (ot, leaf->data.leaf[i], &c, hsize);
    }
    return TRUE;
  }
}

int
vdt_ot_gen_pts (OTP ot, vdt_gen_pt_func gen_pt, void *gen_pt_func_data)
{
  ot->gen_pt           = gen_pt;
  ot->gen_pt_func_data = gen_pt_func_data;
  return gen_pts (ot, ot->root, &ot->center, ot->size);
}
