
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtpatch.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdtpatch.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtpatch) 
 
static SPEP
find_e (SPP sp, PTP p, PTP q)
{
  SPEP e;
  loop_through_list (sp->el, e, SPEP) {
    if (   (e->pt[0] == p && e->pt[1] == q)
        || (e->pt[1] == p && e->pt[0] == q)
        )
      return e;
  }
  return NULL;
}

static SPEP
add_e (SPP sp, PTP p, PTP q)
{
  SPEP e;
  e = find_e (sp, p, q);
  if (!e) {
    e = blpool_get_mem (sp->eblp); CHECK_MEM (e);
    e->pt[0] = p;
    e->pt[1] = q;
    e->fl    = make_list ();
    add_to_tail (sp->el, e);
  }
  return e;
}

static void
attach_f (SPEP e, SPFP f)
{
  add_to_tail (e->fl, f);
}

static void
add_bsf (SPP sp, BSFP bsf)
{
  SPFP f;

  f = blpool_get_mem (sp->fblp); CHECK_MEM (f);
  f->bsf = bsf;
  f->e[0] = add_e (sp, bsf->pt[2], bsf->pt[0]);
  attach_f (f->e[0], f);
  f->e[1] = add_e (sp, bsf->pt[0], bsf->pt[1]);
  attach_f (f->e[1], f);
  f->e[2] = add_e (sp, bsf->pt[1], bsf->pt[2]);
  attach_f (f->e[2], f);
  add_to_tail (sp->fl , f);
}

SPP
vdt_sp_new (void)
{
  SPP sp;
  sp = NEWP_OBJ (vdt_spatch_t); CHECK_MEM (sp);
  sp->eblp = blpool_new (sizeof (vdt_spatch_edge_t)); CHECK_MEM (sp->eblp);
  sp->fblp = blpool_new (sizeof (vdt_spatch_face_t)); CHECK_MEM (sp->fblp);
  sp->el = make_list ();
  sp->fl = make_list ();
  return sp;
}

static int
e_is_on_perimeter (SPEP e)
{
  return (list_length (e->fl) == 1);
}

static int
which_e (SPFP f, SPEP e)
{
  if      (f->e[0] == e) return 0;
  else if (f->e[1] == e) return 1;
  else if (f->e[2] == e) return 2;
  else { errexit ("wrong f/e relationship"); return -1; }
}

   
static PTP
other_pt (SPEP e, PTP pt)
{
  return (e->pt[0] == pt ? e->pt[1] : e->pt[0]);
}

static PTP
start_perimeter_pt (SPEP e)
{
  SPFP f;
  int en;
  
  f = (SPFP) get_list_head (e->fl);
  if (!f) return NULL;
  en = which_e (f, e);
  switch (en) {
  case 0:
    return f->bsf->pt[2];
  case 1:
    return f->bsf->pt[0];
  case 2: default:
    return f->bsf->pt[1];
  }
}

static SPEP
first_perimeter_e (SPP sp)
{
  SPEP oe;
  
  /* first edge */
  loop_through_list (sp->el, oe, SPEP) {
    if (e_is_on_perimeter (oe))
      return oe;
  }
  return NULL;
}

static SPEP
next_perimeter_e (SPP sp, PTP pt)
{
  SPEP oe = NULL; 

  /* continuation edge */
  oe = NULL;
  while ((oe = (SPEP) get_list_next (sp->el, oe))) {
    if (e_is_on_perimeter (oe)) {
      if (start_perimeter_pt (oe) == pt)
        return oe;
    }
  }
  return NULL;
}


int
vdt_sp_recycle (SPP sp)
{
  SPEP e;
  loop_through_list (sp->el, e, SPEP) {
    free_list (e->fl, NOT_NODES);
  }
  flush_list (sp->el, NOT_NODES);
  flush_list (sp->fl, NOT_NODES);
  blpool_recycle (sp->eblp);
  blpool_recycle (sp->fblp);
  return TRUE;
}

int
vdt_sp_add_bsf (SPP sp, BSFP bsf)
{
  add_bsf (sp, bsf);
  return TRUE;
}

static int
count_perimeter_e (SPP sp)
{
  SPEP e;
  int count = 0;

  loop_through_list (sp->el, e, SPEP) {
    if (e_is_on_perimeter (e)) count++;
  }
  return count;
}

LIST
vdt_sp_perimeter_pt_list (SPP sp)
{
  LIST ptl = make_list ();
  SPEP e;
  PTP spt;
  PTP pt;
  PTP ppt;
  int pec = count_perimeter_e (sp);
  int en;

  e = first_perimeter_e (sp);
  if (!e) return NULL;
  spt = start_perimeter_pt (e);
  if (!spt) return NULL;

  ppt = spt;
  add_to_tail (ptl, spt);
  en  = 0;
  while (++en < pec) {
    pt = other_pt (e, ppt);
    add_to_tail (ptl, pt);
    e  = next_perimeter_e (sp, pt);
    ppt = pt;
  }
  return ptl;
}
