#ifndef VDTPUB_H
#define VDTPUB_H


/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtpub.h,v 1.2 2004/04/23 19:11:16 cvs Exp $
   */

/* This is the mesher handle type. */
typedef struct vdt_t *VDT;

/* Default region ID */
#define VDT_DEFAULT_REG_ID ((void *)1)
/* ID indicating an invalid region (meaning `no region specified') */
#define VDT_INVALID_REG_ID ((void *)0)

/* Default values of shift multipliers (boundary and interior point shift) */
#define VDT_BDRY_PT_DEFAULT_SHIFT_MULT 0.0001
#define VDT_GEN_PT_DEFAULT_SHIFT_MULT  0.01

/* Default dihedral angle of a `sliver' tet */
#define VDT_SLIVER_DIHED_ANGLE_DEG (5)
#define VDT_SLIVER_DIHED_ANGLE (VDT_SLIVER_DIHED_ANGLE_DEG * M_PI / 180)

/* Values for the near-surface test option */
#define VDT_NEAR_SURF_TEST_IGNORE       0
#define VDT_NEAR_SURF_TEST_DIST         1
#define VDT_NEAR_SURF_TEST_CIRCUMSPHERE 2

/* This is the type for setting options */
typedef struct vdt_option_t {
  const char *option;           /* option name; the VDT_OPTION_* names
                                   below may be passed in this field
                                   for example as
                                   o->option = VDT_OPTION_DEF_MESH_SIZE;
                                */
  union {
    int                                      iv;
    double                                   dv;
    void                                    *pv;
    void                                (*fv)();
  }            data;            /* value for the option; assign to
                                   appropriate field
                                 */
}              vdt_option_t;

/* Option names */
#define VDT_OPTION_DEF_MESH_SIZE          "def_mesh_size"
#define VDT_OPTION_LP_LIMIT_SIZE          "lp_limit_size"
#define VDT_OPTION_KEEP_SLIVERS           "keep_slivers"
#define VDT_OPTION_DUMP_TET_GEOM          "dump_tet_geom"
#define VDT_OPTION_DUMP_TET_WRL           "dump_tet_wrl"
#define VDT_OPTION_DUMP_T3D               "dump_t3d"
#define VDT_OPTION_DUMP_DATA              "dump_data"
#define VDT_OPTION_GEN_PT_SHIFT_MULT      "gen_pt_shift_mult"
#define VDT_OPTION_BDRY_PT_SHIFT_MULT     "bdry_pt_shift_mult"
#define VDT_OPTION_GEN_INTERIOR_PTS       "gen_interior_pts"
#define VDT_OPTION_GEN_NEAR_SURF_PTS      "gen_near_surf_pts"
#define VDT_OPTION_PDEBUG                 "pdebug"
#define VDT_OPTION_MAXDEBUG               "maxdebug"
#define VDT_OPTION_TEST_MODE              "test_mode"
#define VDT_OPTION_REPORT_OPTS            "report_opts"
#define VDT_OPTION_DUMP_TET_EGF           "dump_tet_egf"
#define VDT_OPTION_DUMP_BSF_EGF           "dump_bsf_egf"
#define VDT_OPTION_RESET_LOCS             "reset_locs"
#define VDT_OPTION_NEAR_SURF_TEST         "near_surf_test"
#define VDT_OPTION_ORIENT_INITIAL_AF      "orient_initial_af"
#define VDT_OPTION_NRANGES                "nranges"
#define VDT_OPTION_GDEBUG                 "gdebug"
#define VDT_OPTION_GIFACE                 "giface"
#define VDT_OPTION_BUCKET_SIZE            "bucket_size"
#define VDT_OPTION_INPUT                  "input"
#define VDT_OPTION_INPUT_VERSION          "input_version"
#define VDT_OPTION_INPUT_TRUE_NAME        "input_true_name"
#define VDT_OPTION_RETRI_TO_DESLIVER      "retri_to_desliver"
#define VDT_OPTION_SLIVER_DIHED_ANGLE     "sliver_dihed_angle"
#define VDT_OPTION_INJECT_SURF_PROJ       "inject_surf_proj"
#define VDT_OPTION_EVAL_MESH_SIZE_FROM_PT "eval_mesh_size_from_pt"
#define VDT_OPTION_EVAL_MESH_SIZE_FUNC    "eval_mesh_size_func"
#define VDT_OPTION_EVAL_MESH_SIZE_FUNC_DATA "eval_mesh_size_func_data"
#define VDT_OPTION_ALLOW_FLIPS            "allow_flips"
#define VDT_OPTION_CHECK_INTERRUPT_FUNC   "check_interrupt_func"

/* Layer number for graphics version.  */
#define VDT_PT_LAYER                   0 /* points */
#define VDT_ADD_PT_LAYER               1 /* points added during Delaunazation */
#define VDT_PT_NUM_LAYER               2 /* point numbers */
#define VDT_DELAUNAY_BSF_LAYER         3 /* Delaunay facets */
#define VDT_NON_DELAUNAY_BSF_LAYER     4 /* non-Delaunay facets */
#define VDT_TET_LAYER                 11 /* tetrahedra */
#define VDT_FF_LAYER                  12 /* advancing front facets */
#define VDT_DELETED_FF_LAYER          13 /* deleted front facets */
#define VDT_OCTREE_LAYER              14 /* mesh-size control octree */
#define VDT_PT_W_CHECK_LAYER          15 /* points */
#define VDT_QUAL_LAYER_BASE          100 /* tetrahedron quality  */
#define VDT_QUAL_LAYER_0             100 /*    range 0 */
#define VDT_QUAL_LAYER_1             101 /*    range 1 */
#define VDT_QUAL_LAYER_2             102 /*    range 2 */
#define VDT_QUAL_LAYER_3             103 /*    range 3 */
#define VDT_QUAL_LAYER_4             104 /*    range 4 */
#define VDT_NONPOS_VOL_SLIVER_LAYER  313 /* slivers w/ non-positive volume */
#define VDT_SMALL_DIHED_SLIVER_LAYER 314 /* ordinary slivers */
#define VDT_REMOVED_SLIVER_LAYER     611 /* removed slivers */

/*
   Create a new VDT mesher.
 */
VDT /* mesher handle or NULL if not enough memory */
VDT_new_mesher (void);

/*
   Delete mesher.
 */
int /* TRUE if ok */
VDT_delete_mesher (VDT vdt /* mesher handle */
                   );

/*
   Add a point.
   A point is added to the mesher input data.
 */
int /* FALSE if (vdt==NULL), or if
       (id<=0), or if there was not enough memory */
VDT_add_pt (void *id, /* unique id (>0) */
            double x, /* x-coordinate of the point */
            double y, /* y-coordinate of the point */
            double z, /* z-coordinate of the point */
            double mesh_size, /* mesh size associated with the
                                 point (can be supplied as 0; in that
                                 case its value is ignored)
                               */
            VDT vdt /* mesher handle */
            );

/*
   Add a point.
   A point is added to the mesher input data, however, the point is
   first checked if it is not too close to other points or to
   the boundary surfaces.
 */
int /* FALSE if (vdt==NULL), or if
       (id==0), or if there was not enough memory */
VDT_add_pt_with_check (void *id, /* unique id (>0) */
                       double x, /* x-coordinate of the point */
                       double y, /* y-coordinate of the point */
                       double z, /* z-coordinate of the point */
                       double mesh_size, /* mesh size associated with the
                                            point (can be supplied as 0; in that
                                            case its value is ignored)
                                            */
                       VDT vdt /* mesher handle */
                       );

/*
   Add a boundary facet.
   A boundary facet is added to the mesher input data.
   The facet is assumed to be oriented so that its points
   are traversed counterclockwise when looking against the normal.
   The normal points outside (in other words, it is the outer normal).
 */
int /* FALSE if inconsistent data have been supplied
       (points not found, surf_id<=0,
       (reg_id[0]<=0), or if (vdt==NULL)),
       or if there was not enough memory
     */
VDT_add_bsf (void *ptid[3], /* identifiers of points; the points
                             need to be defined first by vdt_add_pt().
                             The order in which points are defined
                             determines the normal so that the points
                             are counterclockwise when looking
                             against the normal.  The normal must be
                             either the outer normal (for the outer
                             surfaces), or can be set arbitrarily
                             (for inner, interface surfaces).
                           */
             void *surf_id, /* identifier of the surface the
                             facet is on; positive number
                           */
             void *reg_id[2], /* identifiers of regions on each side
                              of the facet (positive numbers) if there
                              is a region on the given side of the facet;
                              -1 (minus one) otherwise;
                              Note that reg_id[1] is on the
                              side to which the normal points,
                              reg_id[0] is on the other side.
                              If the facet is incident on one region
                              only, it must be the zeroth element of reg_id
                            */
             VDT vdt /* mesher handle */
             );



/*
   Triangulate the interior.
 */
int /* FALSE if the triangulation failed for some reason */
VDT_triangulate (VDT vdt        /* mesher handle */
                 );

void *
VDT_next_tet (void *tetptid[4],
              void **reg_id,
              void *adjtet[4],
              void *prevtet,
              VDT vdt
              );

/*
   Get next point.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_pt (void **id, /* the identifier of the point; a unique number >0 */
             double *x, /* x-coordinate */
             double *y, /* y-coordinate */
             double *z, /* z-coordinate */
             double *mesh_size, /* mesh size */
             void *prev,/* previous point; opaque pointers as returned
                                  by this function.  Pass NULL to get the
                                  first point.
                                  */
             VDT vdt            /*  mesher handle  */
             ) ;

/*
   Get the next boundary surface facet.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_bsf (void **id, /* the identifier of the facet */
              void *ptid[3],    /* identifiers of the points */
              void **surf_id,   /* identifier of the surface */
              void *reg_id[2], /* identifiers of regions on each side
                              of the facet */
              void *tet[2],     /* pointers to tetrahedra on each side
                              of the facet as returned by VDT_next_tet(); 
                              Note that tet[1] is on the
                              side to which the normal points,
                              tet[0] is on the other side.  */
              void *prev,/* previous boundary facet; opaque pointers as returned
                            by this function.  Pass NULL to get the
                            first facet
                            */
              VDT vdt        /*  mesher handle  */
              );

/*
   Produce the triangulation summary.
 */
void
VDT_summary (VDT vdt        /* mesher handle */
                 );


/*
   Get the next boundary surface facet.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_bsf (void **id, /* the identifier of the facet */
              void *ptid[3],    /* identifiers of the points */
              void **surf_id,   /* identifier of the surface */
              void *reg_id[2], /* identifiers of regions on each side
                              of the facet */
              void *tet[2],     /* pointers to tetrahedra on each side
                              of the facet as returned by VDT_next_tet(); 
                              Note that tet[1] is on the
                              side to which the normal points,
                              tet[0] is on the other side.  */
              void *prev,/* previous boundary facet; opaque pointers as returned
                            by this function.  Pass NULL to get the
                            first facet
                            */
              VDT vdt        /*  mesher handle  */
              );

/*
   Set option.
   The option is passed by setting the option number
   (as defined in the vdtpub.h header) and the option value.
 */
int /* TRUE if successful; FALSE otherwise */
VDT_set_option (VDT vdt,         /* mesher handle */
                vdt_option_t *option /* option */
                );


/*
   Get default mesh size.
 */
double
VDT_def_mesh_size (VDT vdt);

/*
   Log module ids.
 */
void
VDT_log_module_ids (void);

/*
   Count points (both connected by tetrahedra and those unconnected).
 */
int
VDT_count_points (VDT vdt, int *connected, int *unconnected);


/*
   Count input points.
 */
int
VDT_count_input_points (VDT vdt);

/*
   Count input boundary facets.
 */
int
VDT_count_input_bsfs (VDT vdt);

/*
   Count boundary facets.
 */
int
VDT_count_bsfs (VDT vdt);

/*
   Count tetrahedra.
 */
int
VDT_count_tets (VDT vdt);

/*
   Access message pertaining to last error raised by
   the public VDT interface.
 */
const char *
VDT_last_error_msg (VDT vdt);

/*
   Get mesh quality.
 */
int
VDT_get_quality (VDT vdt,
                 int *num_non_positive_vol_tets,
                 int *num_small_dihed_angle_tets,
                 int *quality_measure,
                 double *min_quality,
                 double *max_quality,
                 double *geom_mean_quality,
                 double *rms_quality,
                 double *min_volume,
                 double *max_volume,
                 double *min_dihed_angle,
                 double *max_dihed_angle,
                 double *avg_dihed_angle
                 );

/*
   Compute tet dihedral angle quality.
 */
double
VDT_tet_get_angle_quality (void *obj,
                           double *min_angle,
                           double *max_angle,
                           double angles[6]
                           );

/*
   Get tet circumcenter.
 */
int /* TRUE if successful; FALSE otherwise, in which case
       the data in *cx, ..., *cz are not valid.
     */
VDT_tet_get_circumcenter (void *obj, VDT vdt,
                          double *cx, double *cy, double *cz);

/*
  Dump whatever data were requested by options.
 */
int
VDT_dump (VDT vdt);

#if GIFACE
/*
   Generate graphics.
 */
void
VDT_generate_graphics (VDT vdt);
#endif

#endif
