#!/sbin/ksh

#  failure for:
#  bumps3 bridge3
#  arch crack1 part2  manycubes mosk slab1

args=
start_at=1

#
#  usage function
#
function usage_and_exit {
cat <<EOF
Usage: $(basename $0) + optional arguments
   -start_at n
        start testing with input # n
EOF
   exit 0
}

#
#  get the additional arguments
#
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
      -start_at)
         if [ $# -ge 1 ] ; then
            start_at=$1; shift
         else
            usage_and_exit
         fi
         ;;
      *)
         aargs=$args
         args="$a $aargs"
         ;;
   esac
done

inexpensive="aholecyl aholecyl1 aholecyl2 aholecyl3 awheel awheel1 axle bcube blockhole blockhole1 blockstack bracket bracket1 brekeke bumps2 cage cage1 cage2 clover cog0 cogwheel cogwheel0 crack crankshaft cubes cyls elementaler funny funny2 funny4 funny6 holecyl1 hollowcube joint junction maze1 nofancy1 opencage part1 pierced1 pierced2 pipeline3 prism1 prism2 prism3 realmosk ring ring1 ringlet rosette sandia5 sandia7 sandia8 sandia14 sandia15 sandia20 sandia22 sandia24 sandia25 sandia26 sandia27 sandia28 sgi slab2 slitblock slitshell sphere5 spherewcutouts splitshell1 twocubes vejir vertebra wheel yawheel"

more_expensive="apierce bumps1 cog1 funny3 "

expensive="amosk apart1 archfiner1 blocks1 blocks2 brekeke1 bridge1 bridge3 cheese dumbbell dumbbell1 funny5 funny7 junction1 MANDIBLE mol mosk1 nofancy1 pierce pierced2 pipeline3 realmosk ring ring1 ringlet rosette sandia5 sandia7 sandia8 sandia14 sandia15 sandia20 sandia22 sandia24 sandia25 sandia26 sandia27 sandia28 sgi slab2 slitblock slitshell sphere5 spherewcutouts splitshell1 twocubes vejir vertebra wheel yawheel"

log=test.log
oldlog=$log.$$
intrlog=$log.intr

if [ -f $log ] ; then
        echo Moving $log into $oldlog
        mv $log $oldlog
fi

tags=.testtags
if [ -f "$oldlog" ] ; then
        rm -f $tags
        egrep 'VDT test @' $oldlog > $tags
fi

function expected_time {
        input=$1;
        l="$(egrep "\@$input\@" $tags)"
        if [ x"$l" != x ] ; then
                t=$(echo $l | cut -d@ -f3)
        else
                t="?"
        fi
        echo "$t"
}

function kill_vdt {
        echo "Killing vdt"
        echo "Moving current log \"$log\" to \"$intrlog\""
        mv $log $intrlog
        echo "Copying old log \"$oldlog\" to \"$log\""
        cp $oldlog $log
        exit 0
}

trap "kill_vdt SIGINT $file" 2 
trap "kill_vdt SIGKILL $file" 9 
trap "kill_vdt SIGTERM $file" 15

testsuite="$inexpensive $more_expensive $expensive"

echo "\nVDT (C) 1998, Petr Krysl\n"
serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
done
echo Running test for $(($serial-$start_at)) inputs    

serial=0
for n in $(echo $testsuite)
do
        serial=$(($serial+1))
        if [ $serial -ge $start_at ] ; then
          echo "Input #${serial}: $n, expected time: $(expected_time $n) seconds, started at $(date +%H:%M:%S)"
          echo "" >> $log 2>&1 
          echo "<<<<<<<<<<<<<<<< Problem #${serial}: $n >>>>>>>>>>>>>>>>" >> $log 2>&1 
          vdt -test_mode -i $n $(eval echo $args) >> $log 2>&1 
          if [ $? -ne 0 ] ; then
                  echo "***** ERROR ($(basename $0)); see the log $log"
                  exit 1
          fi
          echo "" >> $log 2>&1 
        fi
done
