
/* 
      Variational Delaunay Triangulation
 (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdttet.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"


static char vcid[] = "$Id: vdttet.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdttet) 
 
static int
has_tet_face_orientation (TETP tet, int face, BSFP bsf) 
{
  int i0, i1;
  
  i0 = g3dtet_vertex_on_face[face][0];
  i1 = g3dtet_vertex_on_face[face][1];

  if (tet->pt[i0] == bsf->pt[0]) {
    if (tet->pt[i1] == bsf->pt[1]) 
      return TRUE;
    else
      return FALSE;
  }  else if (tet->pt[i0] == bsf->pt[1]) {
    if (tet->pt[i1] == bsf->pt[2]) 
      return TRUE;
    else
      return FALSE;
  }  else if (tet->pt[i0] == bsf->pt[2]) {
    if (tet->pt[i1] == bsf->pt[0]) 
      return TRUE;
    else
      return FALSE;
  }  else  {
    errexit ("Wront tet face");
  } 
  return FALSE;
} 


#if 0
#define HAVEPT(bsf, ptid) \
 (ptid== (int) bsf->pt[0]->id||ptid== (int) bsf->pt[1]->id||ptid== (int) bsf->pt[2]->id||ptid== (int) bsf->pt[3]->id) 
#define TETPTSARE(bsf, n0, n1, n2, n3) \
 (HAVEPT (bsf, n0) && HAVEPT (bsf, n1) && HAVEPT (bsf, n2) && HAVEPT (bsf, n3)) 
#endif
 

void
vdt_tet_update_topo (TETP tet) 
{
  BSFP bsf;
  int j;
  int nbsfpts = 0;
  if (tet->pt[0]->bsfl) nbsfpts++;
  if (tet->pt[1]->bsfl) nbsfpts++;
  if (tet->pt[2]->bsfl) nbsfpts++;
  if (tet->pt[3]->bsfl) nbsfpts++;

  /* Update boundary facet -- tet connectivity */
  if (nbsfpts >= 3) {
    int face;
    LIST l;
    FOR (j, 4) {
      if (tet->pt[j]->bsfl) {
        l = tet->pt[j]->bsfl;
        bsf = NULL;
        while ((bsf = (BSFP) get_list_next (l, bsf)) ) {
          face = vdt_tet_get_face_num (tet, bsf->pt[0], bsf->pt[1], bsf->pt[2]);
          if (face != -1) {
            if (has_tet_face_orientation (tet, face, bsf)) {
              bsf->tet[0] = tet;
            }  else  {
              bsf->tet[1] = tet;
            } 
          } 
        } 
      } 
    } 
  }  
} 


/*
   Add a tet.
   Add a tet to the triangulation.
   Reorder the points if necessary to define a positive-volume tet.
 */
vdt_tet_t *
vdt_tet_add (vdt_pt_t *pt[4], /* points */
             void *reg_id,      /* region id */
             vdt_t *vdt /* triangulation */
             ) 
{
  int i;
  vdt_tet_t *tet;

  tet = NEWP_OBJ (vdt_tet_t); CHECK_MEM (tet);

  FOR (i, 4) {
    tet->pt[i]     = pt[i]; tet->pt[i]->num_conn_tets++;
    tet->adjtet[i] = NULL;
  } 
  tet->reg_id = reg_id;
  tet->quality = 1;
  
#if GIFACE
  tet->go = NULL;
#endif

  vdt_attach_tet (vdt, tet);

  return tet;
} 


#if GIFACE
/*
   Generate graphics.
 */
void
vdt_tet_generate_graphics (vdt_tet_t *tet, vdt_t *vdt) 
{
  if (vdt->gmodel) {
    if (!tet->go) {
      WCRec pts[4];
      int i;
      FOR (i, 4) { pts[i] = tet->pt[i]->loc; } 
      tet->go = CreateTetra (pts);
      EMNoDrawAddGraphicsToModel (vdt->gmodel, tet->go);
      EASValsSetColor (vdt_color_from_id (tet->reg_id));
      EASValsSetShrink (vdt->test_mode ? 1.0 : 0.85);
      EASValsSetFillStyle (FILL_SOLID);
      EASValsSetLayer (VDT_TET_LAYER);
      EASValsSetEdgeColor (vdt_color_to_pixel ("white"));
      EASValsSetEdgeFlag (TRUE);
      EGWithMaskChangeAttributes (FILL_MASK | COLOR_MASK |
                                  EDGE_COLOR_MASK | EDGE_FLAG_MASK |
                                  SHRINK_MASK | LAYER_MASK, tet->go);
    }  else  {
      WCRec pts[4];
      int i;
      FOR (i, 4) { pts[i] = tet->pt[i]->loc; } 
      EGModifyGraphicsGeometry (tet->go, pts);
      EASValsSetColor (vdt_color_from_id (tet->reg_id));
      if (tet->reg_id == VDT_INVALID_REG_ID) {
        EASValsSetLayer (VDT_REMOVED_SLIVER_LAYER);
        EGWithMaskChangeAttributes (LAYER_MASK, tet->go);
      } 
      EGWithMaskChangeAttributes (COLOR_MASK, tet->go);
    } 
  } 
} 
#endif


/*
   Update 
 */
vdt_ff_t *
vdt_tet_update_ffl (vdt_tet_t *tet, int face, vdt_t *vdt) 
{
  int i0, i1, i2;
  vdt_pt_t *pt[3];
  vdt_ff_t *ff;
  
  i0 = g3dtet_vertex_on_face[face][0];
  i1 = g3dtet_vertex_on_face[face][1];
  i2 = g3dtet_vertex_on_face[face][2];

  if (tet->pt[i0]->id < tet->pt[i1]->id &&
             tet->pt[i0]->id < tet->pt[i2]->id) {
    pt[0] = tet->pt[i0];
    pt[1] = tet->pt[i1];
    pt[2] = tet->pt[i2];
  }  else if (tet->pt[i1]->id < tet->pt[i0]->id &&
             tet->pt[i1]->id < tet->pt[i2]->id) {
    pt[0] = tet->pt[i1];
    pt[1] = tet->pt[i2];
    pt[2] = tet->pt[i0];
  }  else  {
    pt[0] = tet->pt[i2];
    pt[1] = tet->pt[i0];
    pt[2] = tet->pt[i1];
  } 

  ff = vdt_pt_update_ffl (pt[0], pt[1], pt[2], tet, face, vdt);
  assert (ff->ontet);

  if (ff->ontet == tet) {
    /* the tet owns the face; it should be added to the front */
    ff->is_active               = TRUE;
  }  else if (ff->ontet) {
    if (ff->ontet != &vdt->stop_tet) 
      ff->ontet->adjtet[ff->face] = tet;
    tet->adjtet[face]           = ff->ontet;
    ff->is_active               = FALSE;
  } 
    
  return ff;
} 

int
vdt_tet_get_face_num (vdt_tet_t *tet, vdt_pt_t *p0, vdt_pt_t *p1, vdt_pt_t *p2) 
{
  int face;
  int i0, i1, i2;
  
  FOR (face, 4) {
    i0 = g3dtet_vertex_on_face[face][0];
    i1 = g3dtet_vertex_on_face[face][1];
    i2 = g3dtet_vertex_on_face[face][2];
    if ((tet->pt[i0] == p0 || tet->pt[i0] == p1 || tet->pt[i0] == p2) 
        && (tet->pt[i1] == p0 || tet->pt[i1] == p1 || tet->pt[i1] == p2) 
        && (tet->pt[i2] == p0 || tet->pt[i2] == p1 || tet->pt[i2] == p2) 
        ) return face;
  } 
  return -1;
} 

double
vdt_tet_volume (vdt_tet_t *tet) 
{
  double x0 = tet->pt[0]->loc.x;
  double y0 = tet->pt[0]->loc.y;
  double z0 = tet->pt[0]->loc.z;
  double x1 = tet->pt[1]->loc.x;
  double y1 = tet->pt[1]->loc.y;
  double z1 = tet->pt[1]->loc.z;
  double x2 = tet->pt[2]->loc.x;
  double y2 = tet->pt[2]->loc.y;
  double z2 = tet->pt[2]->loc.z;
  double x3 = tet->pt[3]->loc.x;
  double y3 = tet->pt[3]->loc.y;
  double z3 = tet->pt[3]->loc.z;

  double a = x1-x0;
  double b = y1-y0;
  double c = z1-z0;
  double d = x2-x0;
  double e = y2-y0;
  double f = z2-z0;
  double g = x3-x0;
  double h = y3-y0;
  double i = z3-z0;
  
  double VOL = a*e*i - a*h*f - b*d*i + b*g*f + c*d*h - c*g*e;
  
  return  VOL / 6;
} 


double
vdt_tet_edge_quality (vdt_tet_t *tet) 
{	
  double l, lmin, lmax;
  CkitPoint3 v;
  int i;

  lmin = FLT_MAX;
  lmax = 0;

  FOR (i, 6) {
    VECMAC_3_VEC (v, =,
                  tet->pt[g3dtet_vertex_on_edge[i][0]]->loc, -,
                  tet->pt[g3dtet_vertex_on_edge[i][1]]->loc);
    l = VECMAC_NORM (v);
    lmin = min (lmin, l);
    lmax = max (lmax, l);
  } 
  
  return (lmin / lmax);
} 

double
vdt_tet_longest_edge_len (PTP pt[4])
{	
  double l, lmax;
  CkitPoint3 v;
  int i;

  lmax = 0;

  FOR (i, 6) {
    VECMAC_3_VEC (v, =,
                  pt[g3dtet_vertex_on_edge[i][0]]->loc, -,
                  pt[g3dtet_vertex_on_edge[i][1]]->loc);
    l = VECMAC_NORM (v);
    lmax = max (lmax, l);
  } 
  
  return lmax;
} 

double 
vdt_tet_quality (vdt_tet_t *tet)
{
  return vdt_tet_gamma_quality (tet);
}

double
vdt_tet_rms_edge_len (PTP pt[4])
{
  double l_rms_2 = 0, l_2;
  CkitPoint3 v;
  int i;

  FOR (i, 6) {
    VECMAC_3_VEC (v, =,
                  pt[g3dtet_vertex_on_edge[i][0]]->loc, -,
                  pt[g3dtet_vertex_on_edge[i][1]]->loc);
    l_2     = VECMAC_DOT (v, v);
    l_rms_2 += l_2;
  }
  return sqrt (l_rms_2 / 6);
}

double
vdt_tet_gamma_quality (vdt_tet_t *tet)
{
  double vol = vdt_tet_volume (tet);
  double l_rms = vdt_tet_rms_edge_len (tet->pt);
  return (vol / (l_rms * l_rms * l_rms) * 8.47967); 
}

double
vdt_tet_radius_quality (vdt_tet_t *tet) 
{
  CkitPoint3 a, b, n= {0,0,0} , v, c;
  double ir, or;
  PTP *pts = tet->pt;

  if (!vdt_4_pt_circumsphere (pts[0], pts[1], pts[2], pts[3], &c)) {
    vdt_4_pt_circumsphere (pts[0], pts[1], pts[2], pts[3], &c);
    warn ("No circumsphere");
    return 0;
  } 
  VECMAC_3_VEC (v, =, pts[0]->loc, -, c);
  or = VECMAC_NORM (v);
  if (!vdt_4_pt_insphere (pts[0], pts[1], pts[2], pts[3], &c)) {
    warn ("No insphere");
    return 0;
  } 
  VECMAC_3_VEC (a, =, pts[1]->loc, -, pts[0]->loc);
  VECMAC_3_VEC (b, =, pts[2]->loc, -, pts[0]->loc);
  VECMAC_CROSS (n, =, a, x, b);
  VECMAC_NORMALIZE (n);
  VECMAC_3_VEC (v, =, c, -, pts[0]->loc);
  ir = VECMAC_DOT (v, n);
  
  return (3 * ir / or);
} 



double
vdt_tet_angle_quality_pt (CkitPoint3 *pos1, CkitPoint3 *pos2,
                          CkitPoint3 *pos3, CkitPoint3 *pos4,
                          double *min_angle,
                          double *max_angle,
                          double angles[6]) 
{
  CkitPoint3 normal={0,0,0}, normal1={0,0,0}, normal2={0,0,0}, normal3={0,0,0};
  CkitPoint3 vec1, vec2, vec3;
  double norm, norm1, norm2, norm3, mina = M_PI, maxa = 0.0;
  double angle1, angle2, angle3, angle4, angle5, angle6;

  *min_angle = 0;
  *max_angle = M_PI;
  
  /* face normals */
  VECMAC_3_VEC (vec1, =, (*pos1), -, (*pos3));     
  VECMAC_3_VEC (vec2, =, (*pos2), -, (*pos3)); 
  VECMAC_CROSS (normal, =, vec1, x, vec2);
  norm = VECMAC_NORM (normal);
  if (norm <= 0) {
    warn ("Zero-area tet face");
    return 0;
  }
  VECMAC_3_VEC (vec1, =, (*pos1), -, (*pos4));
  VECMAC_3_VEC (vec2, =, (*pos2), -, (*pos4)); 
  VECMAC_3_VEC (vec3, =, (*pos3), -, (*pos4)); 
  VECMAC_CROSS (normal1, =, vec3, x, vec2); 
  norm1 = VECMAC_NORM (normal1);
  VECMAC_CROSS (normal2, =, vec1, x, vec3);        
  norm2 = VECMAC_NORM (normal2);        
  VECMAC_CROSS (normal3, =, vec2, x, vec1);        
  norm3 = VECMAC_NORM (normal3);        
  if (norm1 <= 0 || norm2 <= 0 || norm3 <= 0) {
    warn ("Zero-area tet face");
    return 0;
  }
  
  /* convert to unit normals */
  VECMAC_SCALAR (normal, /=, norm);                         
  VECMAC_SCALAR (normal1, /=, norm1);                       
  VECMAC_SCALAR (normal2, /=, norm2);                       
  VECMAC_SCALAR (normal3, /=, norm3);                       
  
  /* compute angles */
  angle1 = VECMAC_DOT (normal, normal1);    
  angle2 = VECMAC_DOT (normal, normal2);    
  angle3 = VECMAC_DOT (normal, normal3);    
  angle4 = VECMAC_DOT (normal2, normal3);   
  angle5 = VECMAC_DOT (normal3, normal1);   
  angle6 = VECMAC_DOT (normal1, normal2);

#define EVA(N)                                                          \
  {                                                                     \
    if (fabs (angle##N) > 1) angle##N = (angle##N > 0 ? +1 : -1);       \
    angle##N = acos (-angle##N);                                        \
  }
  EVA (1); EVA (2); EVA (3); EVA (4); EVA (5); EVA (6);

  /* evaluate min/max */
  mina = min (mina, angle1);                        
  mina = min (mina, angle2);                        
  mina = min (mina, angle3);                        
  mina = min (mina, angle4);                        
  mina = min (mina, angle5);                        
  mina = min (mina, angle6);                        
  
  maxa = max (maxa, angle1);                        
  maxa = max (maxa, angle2);                        
  maxa = max (maxa, angle3);                        
  maxa = max (maxa, angle4);                        
  maxa = max (maxa, angle5);                        
  maxa = max (maxa, angle6);                        

  /* output arguments */
  *min_angle = mina;
  *max_angle = maxa;
  angles[0] = angle1;
  angles[1] = angle2;
  angles[2] = angle3;
  angles[3] = angle4;
  angles[4] = angle5;
  angles[5] = angle6;
  
  return (maxa > 0 ? mina / maxa : 0);
} 


double
vdt_tet_angle_quality (TETP tet,
                       double *min_angle,
                       double *max_angle,
                       double angles[6]) 
{
  return vdt_tet_angle_quality_pt (&tet->pt[0]->loc,
                                   &tet->pt[1]->loc,
                                   &tet->pt[2]->loc,
                                   &tet->pt[3]->loc,
                                   min_angle, max_angle, angles);
}

