
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtutil.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"
#include "vdtutil.h"

static char vcid[] = "$Id: vdtutil.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtutil) 
 
void
vdt_assert_failed (char *expr, char *file, int line)
{
  char buf[4096];
  sprintf (buf, "Assertion %s failed.", expr); 
  vdt_warning (file, line, buf);
  exit (1);
}

int
vdt_error (char *file, int line, char *msg)
{
  fprintf (stderr, "\n*** ERROR: %s, l.%d: %s\n", file, line, msg);
  exit (1);
  return FALSE;
}

int
vdt_warning (char *file, int line, char *msg)
{
  char buf[1024];
  sprintf (buf, "*** WARNING: %s, l.%d: %s\n", file, line, msg);
  fprintf (stderr, "\n%s\n", buf);
  return TRUE;
}

int
vdt_log (FILE *stream, char *msg)
{
  if (msg)
    fprintf (stream, "%s\n", msg);
  return TRUE;
}

static char *ct[] = {
  "red",
  "green",
  "steelblue",
  "cyan",
  "magenta",
  "yellow",
  "grey",
  "salmon",
  "SpringGreen",
  "white",
  "CadetBlue",
  "orange",
  "IndianRed",
  "seagreen",
  "steelblue",
  "HotPink"
};

#define NUMCOLORS ((int) (sizeof(ct) / sizeof(ct[0])))

const char *
vdt_color_name_from_id (void *id)
{
  int nc = NUMCOLORS;
  int i = abs ((int) id) % nc;
  return ct[i];
}


#if GIFACE
EPixel 
vdt_color_to_pixel (char *name)
{
  BOOLEAN suc;
  static EPixel p = 0;
  static char prevname[256] = "";
  if (strcmp (name, prevname) == 0) return p;
  sprintf (prevname, "%s", name);
  p = ColorGetPixelFromString (name, &suc);
  if (!suc) return 0;
  else      return p;
}


EPixel
vdt_color_from_id (void *id)
{
  BOOLEAN suc;
  int nc = NUMCOLORS;
  int i = abs ((int) id) % nc;
  { EPixel p = ColorGetPixelFromString (ct[i], &suc);
    if (!suc) { fprintf (stderr, "no color for id=%d\n", (int) id); return 0; }
    else      return p;
  }
}


#endif


void
vdt_report_time (ckit_timer_t *timer, char *what)
{
  char buf[128];
  double t = ckit_timer_lap (timer);
  sprintf (buf, "%s took %g [seconds]", what, t);
  logit (buf);
}


void
vdt_tri_quality (PTP p0, PTP p1, PTP p2,
                 double *sigma, double *tau)
{
  double x[3], y[3];
  int good, type;
  G3D_tri_to_tri2d (&p0->loc, &p1->loc, &p2->loc, x, y);
  G3D_triangle_shape (x[0], y[0], x[1], y[1], x[2], y[2], 6, 3, sigma, tau,
                      &good, &type);
}


#if VDT_REPRODUCIBLE_RANDOM
static short int rn[] = {
'\123','\24','\201','\246','\307','\354','\13','\130',
'\12','\27','\43','\333','\364','\212','\222','\203',
'\364','\233','\34','\23','\6','\72','\342','\75',
'\157','\66','\221','\53','\320','\216','\45','\352',
'\274','\135','\176','\323','\275','\3','\67','\365',
'\131','\337','\115','\121','\3','\110','\26','\254',
'\247','\245','\71','\236','\247','\10','\56','\51',
'\304','\111','\110','\31','\115','\125','\263','\23',
'\275','\103','\215','\242','\207','\111','\155','\107',
'\363','\165','\330','\147','\222','\331','\16','\105',
'\160','\107','\341','\266','\100','\317','\66','\250',
'\237','\73','\105','\32','\172','\354','\207','\21',
'\353','\327','\167','\233','\265','\73','\236','\142',
'\64','\360','\273','\357','\307','\136','\362','\34',
'\251','\15','\323','\63','\331','\305','\212','\254',
'\127','\265','\53','\353','\222','\62','\250','\221',
'\100','\347','\252','\241','\22','\257','\122','\124',
'\235','\324','\342','\15','\364','\160','\36','\361',
'\343','\216','\355','\216','\131','\125','\147','\132',
'\226','\216','\357','\24','\50','\356','\17','\337',
'\122','\345','\14','\272','\150','\201','\374','\266',
'\50','\104','\104','\113','\153','\65','\104','\374',
'\37','\234','\372','\176','\11','\271','\201','\140',
'\224','\303','\225','\127','\106','\317','\126','\346',
'\253','\236','\307','\251','\4','\347','\255','\346',
'\157','\121','\222','\332','\114','\255','\327','\65',
'\370','\53','\175','\320','\247','\77','\210','\153',
'\71','\154','\361','\276','\175','\25','\11','\101',
'\324','\362','\270','\220','\347','\226','\44','\351',
'\242','\317','\0','\70','\22','\102','\326','\51',
'\211','\50','\144','\207','\122','\101','\335','\151',
'\311','\324','\360','\163','\0','\26','\301','\204',
'\256','\302','\331','\23','\232','\225','\135','\317',
'\240','\124','\153','\40','\353','\107','\215','\317',
'\241','\35','\314','\11','\24','\361','\176','\117',
'\104','\365','\266','\323','\372','\215','\213','\322',
'\56','\253','\313','\361','\125','\217','\233','\271',
'\206','\107','\271','\110','\54','\353','\351','\172',
'\217','\372','\274','\53','\33','\64','\370','\332',
'\223','\56','\107','\250','\100','\315','\63','\346',
'\315','\44','\145','\242','\32','\241','\344','\47',
'\214','\321','\114','\350','\325','\112','\125','\150',
'\141','\210','\214','\121','\115','\274','\236','\107',
'\334','\246','\213','\261','\164','\371','\324','\137',
'\271','\346','\373','\102','\25','\237','\45','\47',
'\231','\64','\235','\310','\346','\73','\74','\114',
'\120','\237','\35','\215','\130','\71','\21','\156',
'\152','\175','\246','\327','\213','\104','\1','\145',
'\70','\30','\113','\14','\143','\31','\54','\24',
'\67','\354','\64','\276','\143','\145','\65','\156',
'\74','\16','\12','\234','\155','\220','\135','\123',
'\256','\342','\366','\10','\74','\13','\372','\204',
'\373','\260','\224','\47','\326','\326','\26','\340',
'\142','\17','\202','\273','\313','\200','\21','\26',
'\316','\303','\340','\203','\10','\365','\152','\262',
'\33','\232','\167','\73','\102','\125','\374','\125',
'\211','\45','\73','\247','\230','\264','\135','\46',
'\77','\106','\273','\351','\207','\146','\350','\372',
'\306','\134','\220','\0','\310','\245','\206','\245',
'\47','\156','\352','\57','\75','\140','\203','\74',
'\317','\74','\0','\340','\367','\251','\54','\322',
'\373','\204','\362','\7','\55','\304','\54','\165',
'\162','\10','\120','\22','\30','\233','\321','\114',
'\125','\376','\246','\267','\274','\263','\370','\153',
'\111','\66','\366','\177','\71','\44','\316','\121',
'\350','\207','\257','\265','\215','\351','\323','\331',
'\31','\353','\207','\204','\243','\317','\103','\143',
'\361','\310','\351','\167','\157','\10','\4','\251',
'\233','\152','\127','\166','\67','\101','\240','\230',
'\356','\144','\211','\222','\367','\1','\11','\360',
'\14','\334','\315','\105','\103','\360','\147','\236',
'\232','\213','\47','\302','\354','\215','\260','\117',
'\227','\112','\53','\371','\361','\322','\320','\47',
'\336','\252','\266','\135','\213','\121','\372','\123',
'\55','\135','\117','\254','\32','\165','\370','\113',
'\322','\14','\112','\74','\116','\315','\307','\306',
'\204','\336','\332','\231','\164','\266','\364','\270',
'\256','\157','\364','\226','\325','\360','\244','\117',
'\270','\32','\167','\260','\76','\341','\207','\317',
'\305','\114','\37','\127','\135','\31','\61','\46',
'\372','\107','\150','\364','\173','\12','\255','\335',
'\244','\64','\65','\151','\67','\27','\303','\2',
'\106','\375','\35','\366','\160','\256','\225','\136',
'\343','\223','\61','\177','\73','\352','\13','\320',
'\312','\300','\141','\153','\271','\215','\213','\266',
'\323','\170','\222','\106','\161','\330','\72','\147',
'\147','\212','\45','\334','\334','\106','\33','\52',
'\174','\246','\222','\17','\120','\37','\122','\354',
'\137','\273','\325','\357','\244','\352','\306','\336',
'\225','\221','\320','\222','\320','\117','\103','\271',
'\27','\212','\353','\54','\232','\352','\175','\334',
'\140','\244','\225','\310','\347','\237','\357','\364',
'\341','\1','\165','\335','\7','\352','\44','\255',
'\163','\232','\357','\260','\252','\57','\253','\237',
'\106','\355','\170','\206','\66','\261','\374','\152',
'\330','\324','\172','\306','\342','\332','\271','\347',
'\35','\65','\242','\37','\243','\23','\16','\167',
'\103','\375','\262','\23','\264','\313','\336','\364',
'\313','\300','\115','\41','\274','\24','\72','\371',
'\343','\100','\73','\347','\106','\20','\237','\373',
'\127','\361','\375','\227','\172','\337','\277','\120',
'\41','\267','\71','\25','\364','\350','\57','\335',
'\357','\72','\212','\372','\223','\250','\107','\40',
'\335','\257','\41','\14','\230','\55','\266','\362',
'\333','\371','\331','\236','\325','\364','\305','\354',
'\317','\23','\225','\301','\177','\220','\262','\302',
'\141','\212','\305','\72','\167','\161','\105','\143',
'\25','\3','\274','\332','\11','\312','\211','\64',
'\347','\45','\176','\62','\151','\235','\53','\55',
'\327','\0','\64','\131','\61','\324','\332','\172',
'\232','\35','\116','\274','\44','\353','\151','\247',
'\22','\264','\40','\243','\213','\47','\300','\241',
'\73','\225','\215','\46','\137','\334','\347','\346',
'\357','\250','\353','\271','\114','\131','\371','\155',
'\360','\10','\271','\145','\77','\142','\162','\105',
'\351','\347','\244','\103','\75','\116','\105','\360',
'\20','\212','\154','\166','\207','\201','\200','\55',
'\53','\105','\2','\276','\254','\6','\64','\313',
'\311','\70','\375','\107','\350','\162','\134','\111',
'\205','\2','\270','\350','\124','\146','\146','\301',
'\174','\246','\266','\215','\254','\362','\70','\45',
'\76','\201','\146','\366','\170','\0','\347','\237',
'\142','\241','\211','\124','\370','\335','\211','\115',
'\16','\305','\276','\371','\323','\214','\363','\6',
'\7','\200','\127','\264','\215','\64','\342','\52',
'\324','\133','\274','\224','\171','\175','\276','\67',
'\363','\362','\52','\132','\221','\330','\314','\166',
'\237','\66','\65','\366','\145','\4','\101','\24',
'\253','\13','\71','\343','\206','\152','\337','\52',
'\255','\226','\52','\236','\114','\260','\177','\227',
'\46','\207','\304','\334','\47','\357','\303','\361',
'\323','\114','\173','\250','\255','\170','\111','\252',
'\255','\46','\152','\331','\61','\16','\266','\134',
'\314','\241','\224','\263','\261','\334','\261','\254',
'\142','\347','\172','\273','\126','\310','\311','\1',
'\134','\23','\171','\253','\71','\321','\162','\120',
'\311','\207','\11','\230','\375','\242','\344','\170',
'\45','\333','\141','\152','\230','\372','\50','\322',
'\357','\150','\57','\1','\76','\167','\70','\134',
'\133','\341','\320','\230','\17','\24','\166','\123',
'\174','\270','\121','\106','\103','\265','\203','\216',
'\367','\22','\213','\124','\211','\122','\307','\0',
'\266','\226','\332','\240','\301','\135','\330','\166',
'\13','\120','\106','\267','\136','\57','\363','\225',
'\343','\117','\336','\104','\160','\324','\12','\357',
'\260','\230','\46','\235','\264','\331','\202','\344',
'\326','\177','\121','\127','\264','\107','\136','\345',
'\323','\2','\147','\376','\105','\256','\203','\7',
'\140','\362','\56','\352','\234','\354','\174','\321',
'\276','\132','\116','\151','\250','\221','\30','\44',
'\173','\103','\117','\137','\351','\105','\21','\333',
'\355','\352','\25','\205','\70','\35','\162','\336',
'\330','\312','\117','\113','\61','\360','\313','\262',
'\4','\16','\10','\124','\135','\30','\324','\366',
'\24','\174','\350','\277','\166','\146','\223','\201',
'\32','\142','\233','\121','\273','\212','\5','\14',
'\126','\100','\141','\123','\63','\376','\117','\203',
'\57','\322','\166','\51','\316','\252','\107','\154',
'\320','\262','\350','\345','\141','\347','\330','\313',
'\164','\17','\350','\335','\274','\313','\143','\177',
'\335','\64','\335','\63','\111','\52','\257','\246',
'\252','\66','\325','\217','\5','\253','\247','\115',
'\355','\214','\147','\334','\314','\70','\250','\326',
'\141','\55','\164','\376','\265','\44','\323','\207',
'\47','\100','\123','\77','\244','\233','\303','\203',
'\304','\134','\176','\203','\32','\240','\43','\235',
'\254','\273','\150','\275','\26','\76','\252','\125',
'\14','\235','\336','\11','\204','\24','\342','\47',
'\160','\170','\21','\201','\221','\6','\51','\120',
'\336','\252','\164','\17','\42','\151','\53','\14',
'\40','\326','\37','\253','\237','\277','\244','\0',
'\316','\332','\56','\43','\11','\274','\131','\164',
'\175','\322','\16','\12','\175','\271','\161','\237',
'\233','\221','\235','\151','\236','\352','\257','\14',
'\315','\4','\332','\141','\12','\45','\314','\373',
'\372','\75','\162','\206','\210','\157','\335','\76',
'\17','\342','\161','\270','\312','\100','\133','\24',
'\10','\33','\317','\76','\265','\177','\62','\321',
'\336','\357','\260','\202','\64','\351','\225','\76',
'\303','\53','\230','\304','\355','\61','\104','\207',
'\210','\202','\230','\137','\65','\60','\331','\372',
'\44','\352','\257','\207','\357','\342','\145','\274',
'\171','\106','\355','\302','\200','\254','\371','\207',
'\131','\356','\37','\152','\13','\52','\71','\14',
'\365','\344','\215','\55','\216','\116','\154','\42',
'\166','\151','\223','\161','\150','\350','\103','\74',
'\320','\251','\244','\203','\46','\31','\115','\174',
'\214','\262','\321','\7','\241','\212','\216','\2',
'\164','\106','\362','\176','\171','\173','\170','\75',
'\164','\104','\35','\374','\205','\340','\71','\334',
'\142','\154','\230','\4','\210','\256','\175','\362',
'\160','\107','\72','\20','\324','\357','\250','\71',
'\364','\117','\124','\153','\142','\142','\2','\363',
'\106','\70','\351','\41','\217','\236','\376','\230',
'\102','\211','\313','\220','\265','\236','\130','\73',
'\120','\242','\52','\16','\31','\26','\72','\200',
'\374','\343','\22','\143','\120','\232','\335','\310',
'\51','\16','\67','\366','\367','\335','\231','\65',
'\365','\227','\345','\305','\326','\221','\375','\163',
'\122','\157','\131','\375','\21','\147','\25','\174',
'\16','\333','\351','\227','\245','\40','\375','\15',
'\223','\64','\252','\356','\306','\305','\74','\334',
'\120','\271','\264','\32','\5','\322','\152','\177',
'\324','\66','\171','\267','\363','\275','\312','\34',
'\172','\3','\114','\11','\27','\323','\367','\31',
'\107','\53','\21','\274','\214','\53','\11','\322',
'\231','\42','\204','\130','\162','\127','\261','\361',
'\140','\345','\113','\234','\11','\52','\24','\171',
'\354','\177','\334','\12','\71','\21','\322','\174',
'\40','\221','\224','\253','\137','\22','\315','\342',
'\325','\222','\154','\127','\264','\240','\7','\0',
'\26','\313','\53','\14','\232','\346','\105','\311',
'\12','\177','\47','\340','\312','\365','\5','\265',
'\114','\360','\353','\32','\242','\20','\351','\137',
'\317','\176','\160','\373','\217','\55','\64','\43',
'\164','\135','\342','\130','\105','\334','\167','\151',
'\105','\243','\277','\105','\157','\274','\367','\165',
'\301','\354','\346','\267','\21','\330','\313','\362',
'\126','\315','\51','\101','\240','\147','\121','\237',
'\72','\203','\40','\205','\205','\107','\275','\326',
'\56','\375','\63','\351','\376','\20','\223','\334',
'\50','\242','\321','\115','\214','\313','\320','\324',
'\43','\136','\67','\277','\126','\57','\253','\156',
'\164','\301','\21','\205','\275','\346','\341','\343',
'\304','\103','\344','\303','\114','\376','\41','\201',
'\252','\311','\12','\337','\20','\41','\363','\313',
'\16','\136','\163','\0','\211','\325','\363','\112',
'\70','\356','\4','\72','\303','\223','\273','\245',
'\367','\10','\102','\253','\272','\256','\337','\126',
'\120','\330','\114','\126','\245','\325','\171','\52',
'\210','\42','\200','\202','\43','\116','\50','\37',
'\113','\334','\200','\237','\104','\161','\240','\123',
'\171','\61','\223','\271','\172','\35','\121','\343',
'\325','\121','\176','\6','\261','\245','\41','\31',
'\243','\46','\271','\43','\116','\111','\4','\243',
'\216','\21','\30','\1','\63','\302','\4','\153',
'\3','\371','\326','\236','\43','\323','\10','\156',
'\176','\132','\310','\0','\376','\335','\32','\247',
'\116','\156','\207','\103','\75','\335','\116','\15',
'\55','\230','\341','\333','\371','\175','\333','\146',
'\322','\44','\75','\261','\245','\103','\375','\372',
'\327','\112','\233','\13','\14','\46','\125','\200',
'\342','\176','\36','\127','\307','\206','\217','\233',
'\47','\262','\141','\161','\56','\304','\157','\202',
'\201','\176','\143','\363','\321','\363','\122','\315',
'\212','\10','\44','\344','\25','\307','\174','\221',
'\161','\57','\171','\210','\117','\26','\366','\120',
'\241','\276','\115','\260','\161','\12','\250','\11',
'\276','\222','\345','\46','\20','\214','\3','\121',
'\15','\356','\334','\77','\66','\170','\244','\325',
'\20','\242','\43','\174','\123','\145','\330','\167',
'\276','\320','\15','\252','\343','\304','\311','\71',
'\372','\65','\7','\324','\220','\3','\354','\232',
'\36','\326','\144','\322','\302','\220','\167','\51',
'\376','\174','\243','\305','\45','\343','\77','\204',
'\227','\260','\36','\72','\373','\376','\144','\161',
'\317','\6','\331','\143','\265','\23','\136','\346',
'\243','\245','\330','\214','\77','\363','\70','\71',
'\70','\134','\214','\250','\307','\311','\324','\210',
'\144','\1','\372','\54','\115','\330','\74','\67',
'\63','\275','\371','\65','\121','\202','\360','\304',
'\103','\106','\71','\173','\33','\164','\257','\16',
'\113','\35','\77','\316','\260','\212','\65','\230',
'\322','\33','\151','\177','\301','\333','\365','\335',
'\142','\221','\123','\302','\327','\303','\211','\52',
'\21','\157','\151','\367','\31','\111','\123','\313',
'\56','\336','\302','\315','\27','\140','\101','\124',
'\133','\147','\341','\215','\251','\54','\73','\30',
'\64','\311','\22','\43','\133','\145','\271','\176',
'\356','\137','\14','\342','\162','\323','\126','\175',
'\277','\1','\235','\63','\61','\310','\242','\111',
'\135','\211','\260','\264','\147','\43','\160','\221',
'\177','\11','\353','\65','\325','\177','\164','\257',
'\36','\300','\57','\343','\56','\170','\260','\22',
'\12','\127','\334','\363','\304','\27','\247','\242',
'\111','\10','\66','\45','\216','\72','\12','\77',
'\100','\113','\262','\216','\244','\335','\314','\255',
'\366','\105','\125','\74','\4','\213','\240','\272',
'\100','\247','\176','\102','\223','\65','\207','\100',
'\307','\237','\201','\107','\331','\225','\341','\160',
'\310','\350','\216','\361','\164','\45','\160','\117',
'\27','\65','\165','\202','\203','\162','\32','\10',
'\77','\225','\243','\107','\140','\163','\134','\317',
'\201','\311','\376','\262','\151','\247','\74','\247',
'\56','\40','\305','\200','\10','\56','\237','\67',
'\147','\347','\235','\114','\324','\23','\343','\367',
'\36','\267','\107','\46','\54','\72','\64','\41',
'\271','\156','\53','\362','\35','\102','\257','\336',
'\314','\215','\343','\66','\234','\7','\44','\107',
'\167','\353','\376','\71','\107','\42','\233','\140',
'\17','\220','\206','\166','\212','\371','\367','\117',
'\122','\70','\234','\320','\220','\227','\23','\40',
'\157','\261','\115','\233','\313','\300','\134','\274',
'\250','\266','\237','\265','\257','\71','\50','\152',
'\33','\362','\125','\113','\261','\27','\312','\253',
'\35','\61','\350','\1','\73','\142','\351','\271',
'\102','\107','\230','\221','\162','\141','\70','\244',
'\165','\222','\40','\76','\110','\340','\204','\246',
'\32','\152','\346','\216','\134','\260','\65','\265',
'\156','\5','\2','\71','\116','\372','\204','\3',
'\101','\15','\224','\160','\360','\107','\344','\164',
'\330','\351','\143','\42','\271','\32','\165','\147',
'\302','\342','\375','\1','\23','\356','\65','\76',
'\6','\213','\373','\1','\157','\357','\276','\0',
'\271','\65','\331','\45','\46','\151','\60','\262',
'\247','\3','\217','\133','\234','\106','\177','\57',
'\376','\166','\75','\115','\212','\223','\104','\100',
'\122','\137','\226','\313','\40','\172','\103','\45',
'\102','\140','\221','\135','\174','\57','\346','\222',
'\343','\207','\302','\305','\144','\240','\371','\304',
'\70','\234','\152','\243','\317','\103','\274','\223',
'\362','\171','\222','\307','\354','\0','\312','\25',
'\37','\10','\227','\325','\340','\336','\65','\71',
'\75','\240','\341','\346','\303','\1','\202','\276',
'\153','\34','\70','\127','\215','\102','\132','\54',
'\100','\211','\304','\245','\115','\10','\135','\252',
'\112','\42','\342','\174','\334','\370','\101','\171',
'\60','\115','\152','\167','\264','\171','\71','\305',
'\70','\11','\2','\106','\262','\302','\267','\164',
'\275','\165','\362','\144','\2','\74','\366','\302',
'\160','\224','\174','\113','\113','\10','\204','\245',
'\343','\56','\124','\107','\217','\226','\164','\313',
'\313','\117','\262','\350','\5','\40','\262','\206',
'\260','\116','\0','\220','\254','\70','\363','\211',
'\151','\73','\43','\155','\351','\241','\121','\153',
'\75','\262','\222','\310','\274','\226','\300','\66',
'\303','\244','\143','\160','\216','\275','\315','\345',
'\53','\100','\50','\334','\36','\342','\124','\167',
'\360','\240','\263','\305','\331','\335','\13','\224',
'\212','\16','\204','\224','\221','\363','\267','\312',
'\42','\277','\150','\317','\32','\21','\65','\360',
'\112','\201','\270','\43','\307','\163','\140','\250',
'\50','\204','\103','\50','\4','\313','\176','\370',
'\201','\243','\10','\276','\243','\274','\257','\72',
'\365','\246','\43','\371','\32','\340','\27','\110',
'\167','\57','\201','\260','\330','\76','\25','\220',
'\213','\340','\213','\322','\120','\372','\32','\347',
'\310','\301','\253','\40','\365','\124','\330','\1',
'\11','\211','\240','\27','\231','\41','\222','\227',
'\367','\142','\233','\61','\173','\203','\335','\231',
'\332','\106','\23','\203','\305','\133','\140','\363',
'\234','\231','\130','\3','\106','\364','\264','\337',
'\67','\37','\24','\247','\302','\221','\262','\361',
'\6','\105','\277','\314','\64','\276','\207','\323',
'\44','\21','\306','\361','\16','\342','\114','\50',
'\122','\201','\2','\361','\136','\1','\343','\5',
'\134','\126','\160','\255','\116','\36','\21','\6',
'\41','\1','\273','\322','\114','\174','\247','\164',
'\106','\325','\1','\123','\226','\263','\372','\61',
'\301','\274','\66','\24','\102','\223','\337','\0',
'\43','\276','\20','\141','\161','\12','\54','\257',
'\311','\357','\360','\271','\114','\227','\54','\40',
'\277','\104','\273','\111','\275','\120','\44','\134',
'\63','\261','\235','\354','\253','\7','\330','\161',
'\66','\313','\340','\222','\60','\17','\107','\46',
'\26','\273','\226','\115','\364','\252','\15','\144',
'\211','\203','\166','\124','\346','\330','\46','\260',
'\207','\104','\177','\336','\331','\34','\147','\243',
'\331','\64','\42','\375','\242','\201','\23','\132',
'\271','\70','\321','\150','\150','\112','\66','\201',
'\52','\130','\155','\314','\54','\273','\11','\222',
'\361','\253','\157','\61','\154','\305','\23','\272',
'\1','\136','\173','\213','\101','\301','\56','\10',
'\130','\36','\346','\32','\253','\151','\247','\364',
'\122','\322','\157','\247','\257','\241','\220','\36',
'\63','\303','\201','\26','\77','\120','\54','\261',
'\336','\262','\201','\116','\236','\313','\330','\65',
'\6','\232','\134','\204','\14','\147','\1','\322',
'\170','\252','\256','\125','\205','\301','\216','\117',
'\177','\307','\11','\362','\43','\142','\56','\43',
'\30','\114','\274','\47','\135','\100','\362','\176',
'\131','\303','\367','\132','\33','\315','\217','\342',
'\104','\271','\14','\114','\55','\311','\203','\272',
'\232','\65','\220','\166','\155','\230','\101','\206',
'\100','\230','\213','\260','\113','\62','\103','\224',
'\146','\364','\110','\71','\224','\304','\313','\303',
'\54','\323','\141','\253','\363','\224','\341','\220',
'\13','\44','\232','\253','\363','\277','\73','\303',
'\134','\3','\21','\340','\71','\312','\27','\177',
'\214','\70','\352','\200','\324','\355','\330','\12',
'\52','\154','\333','\300','\244','\200','\112','\174',
'\73','\210','\236','\151','\370','\170','\175','\345',
'\147','\20','\116','\23','\356','\334','\351','\42',
'\250','\363','\317','\62','\357','\140','\105','\215',
'\347','\21','\14','\211','\306','\302','\106','\165',
'\232','\1','\222','\45','\332','\71','\225','\152',
'\216','\365','\260','\242','\374','\170','\303','\233',
'\203','\311','\263','\33','\271','\207','\66','\304',
'\355','\235','\326','\322','\234','\103','\120','\374',
'\200','\127','\142','\114','\337','\161','\136','\14',
'\160','\5','\175','\324','\245','\275','\127','\43',
'\42','\156','\262','\362','\45','\2','\45','\234',
'\200','\72','\301','\215','\131','\43','\14','\201',
'\204','\252','\44','\50','\321','\5','\104','\140',
'\247','\105','\147','\305','\141','\63','\237','\121',
'\34','\233','\20','\16','\350','\55','\360','\375',
'\322','\266','\50','\266','\236','\130','\307','\20',
'\135','\70','\154','\336','\133','\44','\202','\320',
'\121','\26','\341','\26','\31','\367','\252','\201',
'\147','\13','\145','\242','\204','\263','\271','\164',
'\115','\352','\272','\362','\251','\40','\335','\325',
'\362','\120','\342','\112','\31','\247','\245','\232',
'\63','\5','\247','\106','\247','\75','\104','\135',
'\150','\74','\372','\253','\255','\246','\62','\206',
'\105','\33','\372','\21','\12','\331','\52','\266',
'\342','\334','\155','\321','\117','\155','\74','\12',
'\241','\51','\246','\327','\5','\22','\102','\306',
'\245','\237','\366','\227','\26','\3','\253','\73',
'\155','\6','\333','\106','\130','\70','\136','\51',
'\143','\176','\237','\316','\70','\62','\33','\241',
'\226','\133','\264','\26','\220','\56','\332','\11',
'\254','\364','\64','\302','\200','\361','\130','\270',
'\171','\356','\62','\177','\64','\11','\355','\376',
'\221','\113','\43','\356','\200','\24','\353','\233',
'\361','\1','\143','\122','\201','\271','\342','\263',
'\214','\354','\24','\330','\4','\124','\72','\143',
'\354','\124','\253','\111','\260','\6','\265','\371',
'\150','\262','\261','\75','\253','\41','\46','\370',
'\150','\106','\256','\247','\354','\4','\176','\25',
'\2','\335','\322','\352','\127','\347','\110','\226',
'\341','\157','\74','\102','\147','\311','\165','\370',
'\230','\34','\243','\330','\313','\336','\65','\167',
'\352','\65','\5','\31','\373','\122','\32','\125',
'\161','\73','\14','\156','\236','\352','\204','\157',
'\32','\224','\230','\321','\125','\107','\303','\360',
'\133','\170','\242','\323','\7','\54','\371','\253',
'\154','\227','\335','\350','\110','\135','\21','\161',
'\125','\246','\250','\224','\164','\55','\65','\247',
'\300','\52','\66','\366','\326','\11','\174','\76',
'\310','\154','\72','\22','\11','\274','\103','\376',
'\344','\144','\304','\12','\165','\127','\72','\161',
'\323','\156','\233','\135','\157','\272','\163','\34',
'\257','\65','\231','\364','\42','\74','\150','\154',
'\262','\170','\44','\76','\346','\271','\67','\343',
'\322','\72','\370','\27','\274','\171','\64','\243',
'\172','\0','\135','\341','\345','\100','\325','\134',
'\271','\301','\354','\312','\325','\210','\307','\1',
'\221','\211','\21','\136','\231','\277','\75','\57',
'\362','\232','\0','\215','\323','\63','\105','\123',
'\173','\326','\7','\126','\145','\336','\243','\261',
'\345','\232','\55','\207','\40','\167','\246','\243',
'\107','\135','\26','\57','\245','\1','\135','\47',
'\204','\205','\331','\217','\331','\366','\267','\10',
'\332','\122','\235','\116','\150','\120','\132','\354',
'\101','\262','\110','\152','\31','\74','\320','\313',
'\26','\171','\263','\103','\130','\116','\232','\162',
'\372','\264','\267','\27','\165','\123','\127','\367',
'\130','\171','\76','\265','\135','\151','\321','\173',
'\106','\233','\172','\141','\247','\17','\213','\21',
'\212','\45','\326','\301','\200','\143','\260','\341',
'\370','\44','\17','\262','\361','\357','\33','\305',
'\154','\253','\356','\266','\224','\301','\77','\4',
'\356','\215','\154','\226','\315','\67','\300','\30',
'\272','\223','\10','\302','\141','\146','\334','\222',
'\15','\222','\305','\245','\324','\20','\73','\102',
'\254','\354','\364','\20','\100','\124','\117','\162',
'\152','\102','\271','\274','\1','\111','\5','\146',
'\26','\174','\353','\113','\32','\12','\211','\51',
'\375','\2','\301','\203','\200','\352','\263','\174',
'\43','\51','\333','\20','\65','\167','\23','\123',
'\174','\212','\326','\25','\113','\322','\10','\342',
'\223','\303','\343','\120','\217','\120','\216','\60',
'\311','\312','\63','\60','\36','\327','\136','\26',
'\304','\56','\345','\353','\375','\61','\135','\63',
'\350','\47','\147','\311','\171','\70','\232','\315',
'\67','\365','\117','\117','\313','\332','\35','\14',
'\143','\255','\51','\111','\171','\340','\365','\111',
'\241','\256','\255','\201','\130','\317','\110','\304',
'\213','\355','\0','\77','\204','\262','\23','\274',
'\147','\205','\6','\324','\136','\175','\40','\173',
'\213','\67','\75','\173','\327','\215','\352','\43',
'\233','\170','\310','\71','\263','\346','\271','\172',
'\216','\311','\111','\366','\173','\237','\353','\347',
'\41','\361','\13','\374','\11','\147','\136','\176',
'\22','\64','\270','\37','\33','\140','\13','\20',
'\136','\66','\365','\214','\66','\163','\317','\176',
'\144','\23','\311','\140','\163','\267','\35','\15',
'\44','\11','\301','\270','\162','\211','\30','\212',
'\32','\221','\103','\46','\257','\106','\251','\17',
'\145','\65','\54','\32','\53','\63','\312','\23',
'\314','\1','\73','\117','\364','\323','\146','\207',
'\32','\47','\332','\363','\257','\303','\146','\117',
'\320','\277','\140','\34','\110','\234','\263','\120',
'\105','\307','\45','\44','\302','\110','\245','\234',
'\356','\232','\334','\175','\330','\123','\44','\311',
'\277','\202','\305','\223','\375','\15','\133','\24',
'\356','\144','\364','\7','\157','\201','\155','\60',
'\22','\237','\162','\166','\371','\22','\143','\117',
'\146','\130','\56','\33','\305','\130','\362','\273',
'\320','\372','\46','\100','\54','\61','\342','\25',
'\265','\205','\260','\114','\115','\223','\70','\236',
'\64','\204','\117','\275','\215','\312','\325','\4',
'\261','\311','\154','\274','\340','\51','\61','\331',
'\256','\157','\125','\43','\64','\330','\367','\162',
'\153','\354','\46','\74','\204','\227','\170','\142',
'\245','\143','\256','\10','\167','\311','\212','\365',
'\114','\266','\352','\10','\373','\11','\34','\102',
'\31','\303','\45','\20','\16','\237','\46','\320',
'\335','\353','\16','\204','\123','\14','\170','\272',
'\175','\264','\321','\327','\55','\50','\274','\154',
'\125','\220','\375','\43','\312','\127','\114','\231',
'\16','\72','\55','\376','\104','\153','\350','\352',
'\344','\370','\66','\236','\233','\142','\71','\337',
'\141','\131','\12','\330','\76','\304','\364','\176',
'\265','\351','\341','\17','\133','\271','\347','\362',
'\6','\25','\116','\263','\157','\26','\201','\235',
'\123','\167','\314','\124','\350','\367','\166','\325',
'\332','\23','\101','\272','\372','\222','\261','\32',
'\63','\30','\257','\217','\372','\321','\150','\107',
'\244','\276','\321','\107','\170','\100','\150','\212',
'\176','\366','\226','\202','\311','\142','\314','\232',
'\265','\311','\136','\313','\275','\25','\212','\276',
'\160','\217','\341','\50','\235','\100','\64','\310',
'\370','\325','\274','\146','\55','\24','\11','\100',
'\134','\166','\134','\57','\323','\233','\217','\164',
'\246','\40','\255','\322','\63','\47','\22','\333',
'\305','\373','\212','\21','\22','\123','\200','\75',
'\232','\40','\46','\232','\215','\60','\12','\101',
'\42','\75','\36','\351','\340','\142','\253','\223',
'\310','\345','\234','\214','\55','\337','\107','\272',
'\72','\233','\77','\222','\206','\327','\65','\32',
'\123','\36','\20','\350','\271','\55','\310','\22',
'\127','\307','\66','\246','\371','\62','\224','\13',
'\300','\356','\320','\177','\314','\250','\10','\177',
'\313','\7','\211','\243','\166','\334','\170','\17',
'\345','\125','\263','\72','\150','\171','\376','\275',
'\44','\255','\272','\255','\23','\235','\262','\272',
'\316','\67','\11','\342','\141','\130','\241','\103',
'\70','\272','\65','\177','\304','\116','\23','\230',
'\242','\210','\353','\113','\36','\365','\127','\5',
'\37','\230','\165','\351','\107','\256','\334','\210',
'\316','\315','\155','\365','\27','\51','\373','\115',
'\262','\53','\115','\34','\132','\11','\145','\36',
'\215','\41','\70','\51','\126','\3','\334','\106',
'\150','\72','\164','\152','\100','\170','\107','\351',
'\7','\12','\61','\247','\362','\11','\62','\203',
'\371','\311','\23','\364','\236','\124','\47','\327',
'\60','\72','\164','\334','\156','\144','\101','\343',
'\306','\143','\304','\157','\65','\362','\32','\56',
'\60','\202','\11','\145','\151','\21','\365','\264',
'\357','\374','\263','\171','\224','\170','\5','\203',
'\163','\21','\244','\320','\102','\123','\317','\327',
'\331','\256','\236','\215','\236','\41','\214','\104',
'\331','\25','\100','\45','\277','\256','\322','\254',
'\306','\143','\67','\235','\42','\354','\24','\355',
'\211','\26','\320','\247','\123','\370','\241','\352',
'\250','\363','\45','\104','\120','\107','\341','\102',
'\66','\120','\101','\365','\125','\30','\156','\204',
'\207','\216','\376','\232','\376','\326','\66','\355',
'\5','\251','\150','\132','\317','\105','\141','\226',
'\335','\356','\342','\310','\130','\272','\76','\101',
'\332','\247','\235','\207','\354','\75','\266','\232',
'\20','\46','\77','\370','\50','\133','\210','\21',
'\26','\65','\140','\206','\355','\47','\4','\364',
'\164','\231','\130','\315','\52','\61','\315','\353',
'\176','\330','\132','\5','\256','\331','\144','\362',
'\223','\62','\370','\111','\126','\72','\221','\347',
'\171','\111','\160','\211','\220','\15','\162','\356',
'\51','\304','\54','\347','\216','\334','\66','\34',
'\221','\14','\0','\336','\5','\116','\177','\55',
'\272','\27','\26','\143','\226','\366','\201','\205',
'\157','\223','\241','\102','\241','\360','\265','\141',
'\106','\116','\5','\316','\324','\25','\163','\145',
'\167','\340','\63','\362','\243','\232','\145','\362',
'\330','\334','\347','\115','\152','\201','\70','\306',
'\175','\334','\341','\276','\7','\40','\250','\54',
'\223','\36','\252','\21','\252','\343','\307','\306',
'\76','\70','\173','\110','\213','\0','\114','\250',
'\25','\304','\137','\342','\350','\224','\270','\32',
'\272','\345','\44','\302','\152','\216','\235','\257',
'\242','\371','\166','\132','\16','\145','\246','\60',
'\307','\43','\355','\207','\230','\243','\16','\143',
'\42','\373','\232','\103','\264','\31','\215','\342',
'\173','\260','\231','\117','\325','\142','\40','\16',
'\160','\130','\131','\4','\147','\250','\231','\260',
'\241','\142','\107','\245','\336','\236','\213','\326',
'\151','\162','\50','\32','\270','\73','\217','\224',
'\211','\36','\0','\11','\375','\374','\225','\26',
'\200','\213','\77','\361','\303','\156','\326','\113',
'\221','\47','\32','\231','\140','\122','\123','\345',
'\47','\173','\116','\313','\144','\42','\167','\132',
'\326','\315','\304','\164','\343','\52','\331','\126',
'\317','\40','\212','\104','\76','\71','\202','\203',
'\113','\254','\312','\353','\326','\154','\34','\56',
'\106','\62','\104','\210','\30','\152','\73','\51',
'\272','\37','\146','\176','\60','\234','\307','\16',
'\333','\154','\316','\163','\172','\345','\311','\367',
'\64','\34','\207','\332','\12','\15','\36','\255',
'\22','\245','\15','\145','\371','\74','\216','\300',
'\103','\36','\165','\311','\276','\233','\270','\122',
'\170','\256','\144','\372','\142','\17','\20','\230',
'\37','\236','\173','\262','\160','\376','\232','\370',
'\332','\110','\10','\146','\335','\53','\124','\210',
'\363','\37','\17','\176','\203','\271','\347','\233',
'\374','\347','\137','\127','\233','\36','\23','\31',
'\311','\303','\134','\223','\135','\142','\154','\271',
'\51','\126','\272','\77','\30','\376','\152','\333',
'\261','\173','\233','\12','\163','\360','\205','\6',
'\323','\266','\125','\334','\240','\67','\214','\40',
'\37','\261','\252','\236','\23','\125','\36','\27',
'\155','\222','\342','\170','\106','\365','\26','\366',
'\24','\250','\1','\276','\175','\332','\213','\251',
'\244','\251','\347','\276','\117','\27','\233','\21',
'\347','\50','\134','\236','\270','\163','\305','\272',
'\147','\362','\11','\263','\163','\251','\171','\51',
'\22','\252','\364','\340','\347','\165','\204','\16',
'\310','\202','\101','\164','\233','\252','\121','\262',
'\52','\356','\157','\221','\202','\153','\174','\351',
'\225','\125','\311','\233','\362','\26','\45','\226',
'\340','\144','\174','\332','\172','\357','\277','\65',
'\374','\154','\57','\47','\357','\132','\75','\174',
'\134','\263','\351','\326','\266','\142','\4','\173',
'\311','\217','\143','\15','\372','\223','\115','\223',
'\205','\324','\105','\164','\272','\365','\165','\254',
'\222','\272','\246','\253','\263','\32','\345','\32',
'\212','\71','\67','\344','\33','\220','\110','\114',
'\336','\247','\224','\5','\65','\310','\277','\117',
'\331','\163','\177','\103','\245','\42','\226','\101',
'\50','\260','\236','\320','\107','\162','\62','\174',
'\260','\273','\167','\364','\135','\204','\322','\65',
'\371','\320','\364','\166','\120','\122','\25','\120',
'\221','\365','\14','\342','\322','\162','\346','\15',
'\210','\173','\155','\46','\63','\303','\171','\250',
'\254','\303','\20','\276','\170','\155','\72','\211',
'\340','\211','\66','\302','\117','\44','\132','\33',
'\47','\42','\142','\40','\77','\156','\3','\7',
'\16','\263','\343','\236','\13','\145','\372','\160',
'\357','\153','\77','\214','\3','\360','\365','\166',
'\315','\351','\177','\261','\63','\170','\11','\200',
'\155','\271','\101','\223','\267','\24','\255','\74',
'\322','\244','\176','\352','\152','\110','\223','\151',
'\272','\114','\102','\234','\357','\340','\21','\246',
'\264','\250','\336','\253','\265','\356','\130','\321',
'\337','\77','\40','\40','\372','\262','\334','\205',
'\257','\361','\112','\145','\1','\24','\145','\47',
'\64','\326','\322','\207','\301','\32','\50','\232',
'\331','\2','\163','\227','\30','\372','\136','\114',
'\364','\213','\240','\4','\107','\176','\4','\223',
'\30','\153','\311','\150','\37','\366','\175','\31',
'\316','\333','\70','\36','\325','\135','\163','\222',
'\111','\223','\176','\365','\152','\136','\201','\34',
'\331','\335','\232','\227','\364','\155','\250','\32',
'\10','\365','\116','\243','\103','\312','\47','\106',
'\267','\336','\30','\42','\217','\0','\312','\327',
'\260','\301','\151','\256','\210','\225','\351','\374',
'\143','\323','\31','\172','\262','\262','\361','\207',
'\44','\130','\5','\263','\314','\163','\306','\161',
'\342','\274','\214','\273','\324','\352','\61','\334',
'\172','\306','\175','\171','\272','\364','\50','\177',
'\320','\365','\235','\217','\374','\263','\176','\106',
'\373','\133','\225','\323','\52','\337','\246','\323',
'\107','\241','\61','\17','\332','\364','\323','\342',
'\2','\43','\334','\42','\265','\75','\11','\335',
'\134','\250','\35','\370','\110','\346','\311','\16',
'\164','\351','\211','\305','\17','\74','\10','\226',
'\304','\131','\234','\227','\77','\352','\166','\362',
'\230','\364','\161','\4','\153','\227','\173','\354',
'\345','\252','\111','\144','\317','\152','\312','\257',
'\334','\326','\161','\103','\265','\350','\154','\16',
'\211','\346','\260','\261','\321','\160','\141','\157',
'\301','\162','\256','\342','\26','\137','\103','\304',
'\15','\262','\123','\313','\205','\37','\52','\265',
'\317','\264','\346','\214','\31','\5','\221','\51',
'\62','\376','\124','\207','\200','\34','\250','\41',
'\234','\225','\20','\4','\343','\252','\204','\46',
'\7','\273','\133','\50','\316','\146','\174','\140',
'\264','\260','\134','\111','\357','\105','\305','\240',
'\324','\30','\322','\22','\173','\230','\77','\346',
'\253','\227','\157','\135','\117','\303','\71','\156',
'\0','\65','\57','\316','\25','\114','\143','\221',
'\51','\73','\264','\156','\240','\20','\340','\265',
'\351','\170','\223','\206','\132','\300','\354','\227',
'\273','\46','\367','\0','\120','\60','\374','\43',
'\207','\40','\224','\200','\244','\264','\276','\37',
'\56','\131','\175','\364','\75','\175','\370','\23',
'\332','\111','\144','\117','\162','\10','\161','\333',
'\255','\16','\72','\224','\246','\255','\321','\12',
'\15','\132','\357','\253','\322','\142','\273','\325',
'\234','\112','\41','\13','\116','\343','\45','\336',
'\170','\350','\305','\273','\210','\116','\120','\174',
'\23','\72','\220','\367','\341','\133','\142','\127',
'\274','\63','\322','\61','\57','\24','\354','\373',
'\320','\172','\211','\265','\177','\42','\212','\60',
'\372','\376','\3','\256','\313','\165','\246','\265',
'\235','\273','\327','\241','\63','\246','\347','\240',
'\170','\245','\201','\227','\144','\3','\123','\107',
'\355','\305','\270','\331','\202','\332','\325','\134',
'\234','\324','\40','\100','\50','\272','\156','\33',
'\271','\54','\256','\226','\257','\351','\250','\210',
'\214','\302','\374','\245','\237','\257','\132','\235',
'\142','\56','\215','\212','\115','\353','\107','\201',
'\1','\252','\361','\233','\366','\375','\367','\357',
'\30','\276','\231','\276','\143','\242','\353','\117',
'\143','\211','\113','\366','\214','\350','\25','\32',
'\322','\164','\116','\213','\303','\160','\43','\311',
'\24','\357','\373','\360','\206','\5','\233','\75',
'\137','\163','\131','\247','\43','\160','\210','\107',
'\133','\31','\74','\301','\50','\325','\135','\66',
'\226','\1','\236','\6','\166','\361','\123','\331',
'\136','\46','\144','\272','\177','\173','\305','\33',
'\56','\62','\272','\312','\300','\77','\274','\233',
'\335','\114','\353','\274','\371','\61','\166','\71',
'\220','\325','\213','\312','\205','\40','\166','\103',
'\105','\47','\353','\130','\257','\111','\211','\276',
'\72','\356','\324','\152','\333','\53','\277','\135',
'\157','\310','\171','\252','\154','\62','\142','\33',
'\340','\75','\316','\47','\313','\174','\220','\156',
'\366','\312','\364','\126','\142','\220','\332','\200',
'\334','\43','\103','\116','\304','\103','\325','\165',
'\364','\116','\373','\136','\124','\371','\261','\325',
'\57','\41','\142','\130','\345','\370','\342','\271',
'\22','\261','\46','\214','\146','\173','\206','\370',
'\323','\53','\255','\16','\334','\150','\150','\12',
'\15','\254','\303','\162','\226','\227','\212','\75',
'\267','\102','\57','\325','\52','\136','\370','\342',
'\117','\27','\240','\266','\145','\154','\15','\351',
'\100','\117','\231','\341','\142','\32','\112','\361',
'\150','\40','\50','\17','\316','\203','\277','\56',
'\142','\16','\353','\257','\26','\31','\353','\350',
'\351','\136','\163','\372','\5','\116','\276','\33',
'\104','\243','\111','\206','\310','\6','\174','\211',
'\337','\40','\351','\171','\110','\336','\366','\145',
'\55','\101','\102','\151','\146','\276','\52','\45',
'\12','\365','\227','\333','\120','\104','\165','\335',
'\100','\7','\45','\124','\73','\337','\66','\373',
'\340','\251','\154','\253','\261','\54','\276','\304',
'\123','\174','\143','\51','\374','\172','\244','\145',
'\142','\101','\140','\164','\312','\233','\122','\342',
'\77','\77','\15','\234','\54','\311','\246','\147',
'\355','\305','\6','\75','\300','\256','\177','\244',
'\223','\126','\172','\333','\170','\330','\213','\135',
'\307','\344','\107','\235','\262','\153','\125','\171',
'\272','\246','\244','\164','\153','\17','\104','\255',
'\350','\14','\256','\41','\2','\312','\331','\5',
'\170','\172','\55','\55','\347','\346','\214','\305',
'\204','\146','\367','\267','\136','\226','\42','\77',
'\333','\370','\216','\236','\254','\263','\312','\215',
'\334','\376','\255','\112','\360','\244','\317','\2',
'\140','\340','\244','\372','\15','\275','\126','\172',
'\311','\252','\345','\313','\142','\217','\342','\132',
'\57','\256','\162','\266','\331','\65','\64','\166',
'\31','\114','\207','\353','\330','\316','\156','\11',
'\233','\115','\320','\221','\64','\221','\134','\200',
'\311','\337','\54','\364','\331','\225','\130','\146',
'\345','\141','\302','\4','\312','\1','\222','\127',
'\104','\127','\52','\172','\210','\303','\36','\110',
'\337','\6','\237','\155','\265','\114','\2','\310',
'\307','\20','\363','\324','\151','\63','\342','\301',
'\7','\57','\301','\136','\105','\252','\351','\122',
'\154','\312','\361','\306','\333','\17','\176','\0',
'\27','\33','\1','\4','\53','\45','\203','\44',
'\132','\222','\102','\153','\273','\347','\155','\360',
'\352','\137','\210','\301','\346','\16','\311','\337',
'\356','\204','\153','\260','\41','\334','\105','\243',
'\265','\140','\222','\375','\167','\371','\124','\26',
'\104','\142','\325','\202','\302','\332','\210','\214',
'\333','\155','\132','\52','\355','\241','\272','\162',
'\144','\240','\161','\55','\254','\210','\316','\103',
'\116','\337','\225','\70','\44','\221','\147','\136',
'\203','\33','\146','\274','\40','\50','\44','\173',
'\37','\312','\77','\362','\57','\274','\324','\161',
'\21','\261','\147','\76','\325','\302','\255','\344',
'\335','\272','\34','\6','\272','\140','\147','\137',
'\115','\217','\352','\371','\161','\375','\235','\361',
'\216','\277','\305','\155','\261','\14','\225','\150',
'\1','\332','\34','\36','\325','\7','\365','\43',
'\314','\326','\321','\45','\265','\46','\335','\54',
'\73','\176','\55','\300','\12','\307','\333','\247',
'\257','\213','\42','\267','\353','\246','\355','\272',
'\354','\104','\64','\224','\252','\101','\200','\40',
'\123','\153','\313','\257','\363','\325','\26','\345',
'\17','\71','\113','\371','\40','\356','\172','\106',
'\354','\326','\351','\101','\134','\264','\242','\253',
'\202','\145','\131','\5','\62','\10','\26','\124',
'\157','\363','\165','\17','\246','\353','\55','\151',
'\236','\113','\315','\252','\111','\71','\30','\101',
'\301','\315','\73','\244','\17','\170','\12','\375',
'\372','\326','\71','\343','\56','\51','\270','\244',
'\231','\376','\54','\165','\244','\100','\56','\253',
'\25','\33','\6','\266','\245','\51','\240','\203',
'\152','\201','\313','\115','\330','\353','\233','\311',
'\132','\166','\341','\61','\107','\353','\72','\361',
'\66','\261','\244','\222','\223','\355','\237','\75',
'\143','\270','\330','\300','\236','\36','\142','\61',
'\174','\302','\1','\373','\341','\313','\152','\246',
'\310','\172','\317','\22','\254','\100','\13','\326',
'\267','\114','\335','\343','\212','\202','\41','\71',
'\5','\10','\113','\334','\11','\316','\137','\262',
'\254','\147','\140','\64','\76','\371','\346','\136',
'\370','\127','\263','\256','\143','\35','\34','\52',
'\262','\317','\165','\301','\24','\222','\320','\310',
'\26','\231','\232','\254','\214','\157','\277','\151',
'\200','\170','\136','\375','\333','\355','\272','\215',
'\231','\15','\156','\307','\155','\215','\147','\73',
'\360','\261','\101','\134','\101','\357','\361','\102',
'\345','\347','\325','\100','\36','\367','\243','\153',
'\275','\245','\100','\232','\257','\215','\301','\166',
'\335','\303','\200','\267','\121','\337','\325','\322',
'\262','\261','\325','\114','\120','\307','\304','\25',
'\271','\267','\274','\100','\76','\54','\312','\42',
'\56','\251','\107','\264','\74','\206','\144','\36',
'\375','\216','\205','\364','\236','\52','\174','\320',
'\56','\36','\203','\334','\270','\167','\237','\233',
'\166','\226','\123','\371','\120','\72','\22','\277',
'\304','\360','\62','\133','\101','\253','\125','\341',
'\203','\267','\307','\163','\16','\115','\164','\36',
'\101','\46','\354','\104','\332','\217','\20','\156',
'\343','\125','\202','\340','\105','\270','\307','\34',
'\12','\314','\55','\53','\34','\56','\134','\17',
'\333','\135','\356','\164','\233','\17','\213','\10',
'\261','\374','\166','\157','\364','\130','\366','\250',
'\345','\144','\56','\25','\217','\46','\204','\164',
'\332','\152','\203','\230','\261','\144','\222','\70',
'\164','\357','\314','\152','\232','\266','\166','\73',
'\156','\35','\261','\57','\51','\240','\226','\225',
'\360','\301','\246','\202','\174','\366','\173','\242',
'\350','\376','\247','\60','\164','\323','\220','\166',
'\215','\131','\274','\304','\57','\126','\22','\120',
'\162','\153','\155','\14','\251','\215','\367','\67',
'\201','\204','\360','\200','\357','\275','\247','\374',
'\142','\162','\255','\324','\206','\40','\50','\36',
'\251','\106','\301','\34','\366','\10','\235','\270',
'\100','\42','\221','\373','\7','\275','\372','\352',
'\234','\64','\202','\236','\115','\160','\115','\156',
'\256','\166','\3','\22','\136','\11','\262','\271',
'\53','\113','\171','\254','\206','\71','\74','\225',
'\346','\273','\374','\121','\160','\66','\257','\362',
'\303','\66','\321','\163','\301','\36','\153','\357',
'\32','\166','\241','\111','\201','\143','\174','\251',
'\72','\325','\375','\265','\114','\304','\73','\251',
'\160','\114','\313','\330','\256','\56','\3','\275',
'\134','\214','\160','\24','\207','\117','\74','\200',
'\20','\243','\256','\320','\354','\167','\132','\75',
'\113','\260','\204','\227','\207','\200','\356','\6',
'\330','\105','\3','\254','\226','\33','\263','\230',
'\145','\374','\232','\257','\165','\5','\116','\374',
'\211','\114','\60','\263','\74','\272','\271','\151',
'\162','\206','\324','\1','\227','\347','\12','\302',
'\46','\162','\170','\326','\330','\67','\62','\23',
'\32','\141','\237','\70','\72','\237','\202','\340',
'\75','\375','\302','\206','\347','\7','\372','\26',
'\36','\220','\64','\276','\102','\174','\146','\122',
'\342','\70','\20','\75','\311','\256','\171','\250',
'\17','\62','\360','\265','\321','\262','\155','\134',
'\147','\103','\314','\327','\244','\215','\134','\323',
'\216','\100','\306','\233','\154','\202','\230','\160',
'\122','\155','\106','\204','\315','\316','\162','\340',
'\225','\176','\271','\17','\141','\155','\271','\112',
'\2','\362','\367','\246','\17','\342','\270','\161',
'\344','\242','\365','\104','\340','\251','\65','\55',
'\234','\162','\372','\301','\121','\0','\242','\34',
'\101','\330','\336','\375','\175','\170','\32','\34',
'\132','\44','\244','\47','\241','\355','\73','\44',
'\317','\323','\211','\13','\175','\162','\75','\170',
'\75','\340','\57','\143','\357','\100','\60','\344',
'\230','\323','\1','\346','\100','\311','\66','\277',
'\321','\41','\272','\324','\53','\301','\357','\43',
'\64','\32','\221','\354','\17','\141','\353','\134',
'\100','\13','\373','\225','\67','\117','\345','\231',
'\121','\202','\154','\314','\302','\170','\64','\254',
'\236','\156','\170','\217','\142','\333','\213','\337',
'\237','\44','\146','\131','\10','\340','\375','\375',
'\201','\176','\172','\163','\7','\274','\212','\44',
'\202','\114','\134','\233','\137','\23','\226','\124',
'\171','\155','\206','\265','\130','\13','\135','\101',
'\210','\175','\33','\303','\107','\50','\16','\54',
'\0','\40','\160','\47','\31','\40','\25','\112',
'\364','\13','\11','\254','\202','\200','\202','\136',
'\233','\350','\76','\104','\230','\325','\356','\125',
'\175','\132','\3','\337','\243','\151','\17','\67',
'\126','\120','\221','\261','\220','\26','\320','\33',
'\320','\235','\166','\24','\161','\207','\122','\63',
'\0','\277','\300','\146','\204','\243','\35','\310',
'\310','\37','\224','\147','\217','\376','\170','\330',
'\71','\153','\66','\155','\16','\371','\13','\150',
'\347','\175','\203','\62','\121','\112','\67','\264',
'\221','\342','\113','\74','\331','\17','\275','\175',
'\26','\46','\172','\42','\7','\326','\20','\150',
'\264','\367','\217','\77','\66','\10','\373','\112',
'\56','\334','\33','\107','\311','\16','\12','\112',
'\11','\156','\353','\154','\367','\241','\310','\320',
'\266','\75','\335','\316','\303','\2','\154','\235',
'\117','\246','\26','\11','\177','\50','\252','\106',
'\331','\333','\117','\154','\26','\34','\264','\105',
'\157','\75','\104','\67','\22','\15','\373','\26',
'\250','\350','\50','\54','\207','\322','\237','\347',
'\2','\76','\367','\134','\366','\354','\225','\302',
'\376','\256','\50','\277','\217','\273','\323','\237',
'\224','\363','\245','\227','\172','\230','\375','\147',
'\243','\327','\21','\43','\61','\274','\257','\26',
'\334','\240','\104','\126','\213','\367','\160','\316',
'\247','\120','\221','\172','\221','\142','\30','\101',
'\105','\120','\33','\144','\241','\341','\67','\353',
'\140','\126','\27','\240','\140','\234','\122','\57',
'\6','\326','\56','\25','\51','\160','\206','\122',
'\255','\125','\173','\26','\366','\34','\57','\266',
'\166','\271','\13','\217','\133','\22','\303','\254',
'\231','\154','\47','\166','\316','\376','\107','\124',
'\241','\73','\375','\173','\153','\373','\21','\215',
'\330','\141','\130','\21','\205','\44','\50','\263',
'\274','\252','\171','\351','\315','\122','\346','\260',
'\65','\361','\26','\254','\5','\212','\101','\37',
'\173','\32','\274','\344','\225','\12','\216','\302',
'\374','\140','\26','\161','\104','\160','\305','\347',
'\67','\64','\140','\41','\371','\274','\23','\202',
'\276','\343','\275','\205','\145','\6','\376','\26',
'\143','\367','\372','\24','\337','\266','\354','\323',
'\162','\104','\211','\253','\130','\52','\235','\261',
'\62','\61','\356','\177','\44','\72','\37','\145',
'\265','\303','\351','\201','\52','\326','\324','\51',
'\354','\320','\347','\227','\27','\233','\112','\204',
'\1','\56','\62','\171','\227','\330','\142','\131',
'\56','\370','\232','\343','\333','\50','\100','\72',
'\13','\51','\53','\110','\236','\121','\210','\355',
'\102','\133','\171','\335','\221','\304','\117','\330',
'\3','\364','\343','\77','\136','\366','\45','\172',
'\270','\300','\125','\341','\67','\203','\237','\376',
'\215','\214','\35','\71','\11','\205','\64','\47',
'\52','\75','\225','\127','\171','\152','\64','\262',
'\77','\331','\216','\371','\357','\274','\232','\13',
'\232','\236','\52','\145','\107','\2','\270','\32',
'\60','\74','\110','\71','\112','\344','\244','\117',
'\264','\12','\260','\143','\6','\150','\337','\255',
'\45','\265','\266','\320','\73','\235','\106','\163',
'\133','\340','\360','\316','\73','\117','\202','\331',
'\207','\212','\246','\26','\143','\312','\322','\221',
'\263','\330','\102','\36','\200','\373','\304','\47',
'\103','\111','\164','\170','\305','\150','\227','\126',
'\263','\274','\222','\64','\200','\330','\173','\43',
'\125','\327','\41','\346','\346','\342','\350','\375',
'\161','\155','\134','\302','\326','\334','\332','\5',
'\54','\63','\74','\7','\315','\250','\135','\144',
'\64','\316','\363','\173','\343','\357','\247','\265',
'\146','\323','\301','\275','\44','\346','\113','\76',
'\34','\100','\74','\276','\232','\355','\367','\117',
'\207','\303','\77','\5','\364','\213','\172','\237',
'\160','\114','\45','\51','\275','\1','\261','\63',
'\121','\34','\205','\30','\102','\354','\14','\334',
'\253','\167','\307','\310','\263','\321','\351','\224',
'\135','\246','\126','\272','\61','\246','\3','\62',
'\202','\244','\67','\14','\244','\136','\61','\77',
'\224','\303','\366','\320','\71','\46','\144','\362',
'\146','\227','\255','\71','\132','\322','\233','\75',
'\311','\177','\360','\261','\134','\123','\2','\21',
'\207','\301','\41','\160','\262','\345','\167','\270',
'\21','\57','\32','\171','\166','\26','\367','\37',
'\334','\325','\310','\177','\111','\257','\60','\160',
'\35','\165','\351','\64','\252','\112','\306','\332',
'\262','\306','\265','\276','\256','\116','\371','\106',
'\272','\327','\160','\331','\210','\353','\15','\130',
'\142','\173','\306','\50','\247','\266','\111','\253',
'\17','\131','\271','\362','\67','\51','\275','\312',
'\163','\102','\11','\234','\214','\13','\22','\204',
'\163','\114','\267','\41','\7','\170','\47','\161',
'\146','\217','\171','\266','\23','\235','\35','\347',
'\123','\163','\265','\22','\300','\356','\64','\66',
'\143','\43','\42','\147','\343','\223','\61','\357',
'\25','\67','\255','\357','\243','\253','\215','\47',
'\6','\106','\75','\44','\150','\316','\225','\75',
'\335','\344','\171','\45','\47','\175','\122','\202',
'\67','\200','\163','\27','\324','\172','\342','\77',
'\211','\63','\171','\355','\65','\116','\121','\26',
'\224','\317','\227','\232','\306','\265','\342','\337',
'\353','\241','\362','\314','\276','\265','\15','\325',
'\254','\171','\166','\265','\166','\77','\324','\357',
'\260','\154','\214','\314','\7','\337','\131','\374',
'\367','\327','\62','\131','\276','\234','\165','\140',
'\322','\370','\237','\37','\4','\242','\151','\230',
'\201','\14','\141','\147','\106','\160','\273','\226',
'\327','\277','\150','\155','\366','\106','\152','\213',
'\315','\371','\54','\237','\260','\152','\154','\23',
'\366','\151','\123','\16','\370','\144','\63','\32',
'\224','\313','\143','\231','\347','\22','\222','\154',
'\11','\67','\55','\37','\352','\12','\363','\276',
'\1','\66','\312','\365','\251','\4','\145','\147',
'\167','\224','\3','\254','\210','\250','\4','\152',
'\266','\60','\77','\327','\132','\137','\227','\17',
'\367','\365','\323','\272','\300','\147','\255','\217',
'\361','\276','\12','\113','\332','\211','\25','\52',
'\25','\63','\131','\20','\266','\62','\324','\371',
'\77','\67','\112','\57','\111','\213','\21','\161',
'\103','\14','\352','\135','\51','\332','\121','\114',
'\177','\317','\122','\170','\360','\321','\224','\274',
'\360','\10','\215','\52','\26','\13','\375','\205',
'\261','\307','\316','\142','\76','\321','\75','\244',
'\77','\71','\367','\142','\227','\103','\224','\130',
'\102','\120','\201','\213','\33','\311','\372','\270',
'\204','\202','\140','\345','\114','\76','\64','\235',
'\314','\261','\144','\305','\102','\15','\267','\46',
'\215','\2','\365','\144','\265','\312','\300','\67',
'\203','\12','\311','\57','\275','\277','\121','\24',
'\371','\35','\375','\170','\216','\57','\354','\30',
'\252','\365','\154','\62','\215','\310','\226','\13',
'\314','\167','\301','\353','\13','\73','\147','\160',
'\207','\103','\347','\323','\256','\4','\260','\264',
'\6','\155','\130','\20','\177','\356','\216','\202',
'\64','\176','\242','\310','\277','\204','\316','\22',
'\202','\206','\107','\351','\20','\115','\142','\105',
'\26','\220','\37','\176','\257','\375','\210','\47',
'\300','\355','\217','\304','\10','\22','\73','\271',
'\363','\73','\244','\41','\52','\31','\15','\101',
'\55','\127','\264','\57','\123','\262','\53','\336',
'\253','\152','\177','\272','\345','\174','\107','\172',
'\211','\147','\44','\236','\131','\265','\5','\100',
'\130','\356','\65','\340','\370','\257','\32','\67',
'\343','\264','\376','\364','\27','\13','\337','\106',
'\143','\50','\240','\151','\267','\205','\264','\263',
'\341','\264','\353','\111','\40','\135','\47','\365',
'\162','\0','\60','\136','\232','\37','\376','\145',
'\6','\127','\32','\173','\304','\301','\305','\325',
'\204','\156','\262','\20','\224','\247','\30','\142',
'\174','\65','\257','\231','\24','\371','\70','\175',
'\255','\246','\332','\232','\106','\35','\200','\215',
'\242','\61','\266','\247','\357','\15','\232','\177',
'\70','\265','\231','\205','\110','\55','\275','\364',
'\121','\210','\203','\376','\330','\340','\344','\170',
'\263','\347','\304','\35','\320','\270','\72','\5',
'\66','\373','\165','\16','\136','\352','\244','\207',
'\342','\46','\26','\25','\41','\126','\230','\65',
'\362','\361','\325','\50','\113','\306','\122','\235',
'\372','\271','\61','\340','\255','\256','\331','\260',
'\322','\162','\157','\63','\14','\146','\133','\34',
'\225','\220','\51','\257','\224','\35','\45','\26',
'\353','\12','\353','\357','\12','\212','\52','\204',
'\50','\3','\236','\53','\336','\75','\53','\305',
'\220','\204','\352','\160','\300','\172','\17','\333',
'\5','\312','\74','\133','\274','\306','\162','\165',
'\221','\55','\102','\202','\317','\234','\204','\14',
'\340','\352','\107','\232','\244','\340','\14','\270',
'\213','\136','\162','\71','\140','\135','\242','\166',
'\170','\275','\255','\207','\272','\215','\376','\34',
'\41','\263','\342','\103','\227','\112','\271','\326',
'\367','\214','\375','\206','\32','\261','\70','\216',
'\121','\176','\254','\64','\14','\343','\165','\164',
'\13','\215','\237','\213','\4','\72','\122','\12',
'\361','\272','\234','\35','\305','\132','\200','\374',
'\14','\363','\213','\53','\120','\276','\0','\3',
'\261','\37','\347','\56','\211','\304','\310','\27',
'\300','\47','\246','\171','\203','\136','\126','\372',
'\177','\12','\372','\355','\140','\171','\215','\115',
'\313','\215','\114','\335','\132','\260','\253','\363',
'\314','\167','\344','\211','\266','\363','\263','\365',
'\73','\222','\42','\327','\371','\6','\156','\276',
'\47','\375','\336','\162','\143','\116','\127','\274',
'\122','\157','\110','\376','\16','\324','\143','\312',
'\353','\24','\126','\252','\144','\117','\214','\300',
'\341','\374','\16','\354','\25','\77','\364','\17',
'\73','\333','\224','\132','\242','\326','\220','\1',
'\213','\320','\62','\13','\216','\143','\150','\171',
'\257','\45','\311','\14','\360','\56','\310','\30',
'\301','\3','\335','\11','\253','\200','\54','\120',
'\343','\167','\314','\2','\105','\376','\141','\161',
'\143','\61','\340','\127','\136','\135','\366','\311',
'\72','\13','\17','\14','\222','\164','\341','\352',
'\366','\34','\215','\237','\220','\221','\237','\31',
'\154','\1','\51','\67','\5','\12','\201','\344',
'\351','\275','\373','\343','\233','\123','\65','\325',
'\13','\300','\172','\14','\357','\232','\257','\312',
'\13','\312','\330','\16','\370','\126','\356','\176',
'\106','\43','\132','\171','\301','\243','\224','\143',
'\376','\116','\220','\206','\204','\107','\72','\54',
'\21','\301','\306','\202','\227','\111','\51','\355',
'\272','\6','\333','\213','\110','\335','\120','\206',
'\275','\255','\123','\24','\316','\354','\232','\335',
'\256','\267','\105','\155','\2','\146','\240','\34',
'\20','\321','\61','\155','\16','\21','\56','\277',
'\144','\254','\46','\254','\323','\246','\307','\41',
'\236','\327','\101','\27','\104','\323','\213','\364',
'\261','\363','\120','\51','\73','\30','\277','\371',
'\241','\57','\145','\232','\12','\250','\275','\234',
'\16','\172','\322','\267','\247','\306','\122','\271',
'\6','\120','\35','\327','\53','\207','\134','\25',
'\60','\2','\10','\162','\6','\204','\2','\306',
'\235','\252','\330','\34','\375','\1','\316','\344',
'\320','\357','\371','\153','\250','\4','\227','\233',
'\302','\115','\41','\307','\21','\72','\367','\331',
'\220','\271','\300','\354','\251','\273','\315','\51',
'\250','\342','\275','\74','\36','\115','\41','\152',
'\215','\316','\266','\142','\165','\41','\35','\126',
'\161','\115','\260','\31','\332','\105','\370','\163',
'\1','\365','\136','\301','\342','\200','\302','\146',
'\142','\52','\264','\105','\215','\166','\176','\154',
'\332','\322','\27','\367','\136','\261','\340','\34',
'\267','\312','\216','\346','\330','\115','\146','\102',
'\200','\222','\206','\327','\277','\252','\353','\40',
'\362','\342','\112','\175','\301','\152','\164','\124',
'\224','\345','\211','\363','\154','\253','\15','\276',
'\165','\312','\73','\12','\277','\27','\172','\101',
'\66','\320','\371','\20','\263','\23','\366','\206',
'\75','\61','\155','\152','\104','\65','\17','\370',
'\124','\302','\253','\315','\214','\117','\306','\77',
'\134','\264','\334','\256','\366','\330','\50','\154',
'\244','\31','\17','\127','\45','\77','\104','\233',
'\220','\306','\107','\236','\37','\274','\366','\66',
'\156','\205','\54','\21','\37','\241','\30','\311',
'\54','\340','\0','\133','\354','\267','\102','\241',
'\240','\246','\42','\245','\25','\122','\1','\36',
'\124','\46','\240','\172','\57','\45','\46','\222',
'\210','\373','\255','\233','\352','\160','\52','\3',
'\170','\150','\65','\177','\51','\361','\232','\162',
'\161','\41','\374','\243','\214','\320','\17','\333',
'\373','\60','\142','\203','\265','\246','\237','\51',
'\150','\227','\46','\155','\215','\117','\6','\55',
'\64','\0','\247','\161','\304','\57','\16','\170',
'\51','\160','\30','\253','\157','\177','\362','\137',
'\223','\26','\164','\175','\342','\331','\373','\112',
'\63','\64','\326','\266','\55','\153','\143','\370',
'\367','\214','\212','\340','\141','\217','\347','\234',
'\233','\114','\16','\70','\302','\304','\54','\35',
'\143','\361','\225','\325','\42','\232','\311','\73',
'\10','\33','\346','\342','\314','\176','\217','\214',
'\323','\313','\235','\66','\273','\55','\63','\165',
'\261','\202','\346','\14','\267','\141','\200','\367',
'\53','\273','\317','\111','\235','\344','\151','\152',
'\267','\276','\102','\52','\322','\272','\151','\46',
'\231','\212','\120','\265','\15','\311','\234','\153',
'\305','\331','\5','\327','\374','\343','\320','\234',
'\142','\50','\240','\242','\252','\323','\43','\255',
'\61','\23','\345','\5','\40','\150','\327','\106',
'\274','\14','\7','\336','\143','\233','\46','\215',
'\130','\247','\250','\372','\72','\224','\107','\12',
'\214','\57','\104','\315','\232','\0','\115','\123',
'\315','\104','\256','\55','\145','\1','\104','\67',
'\201','\135','\57','\2','\251','\74','\124','\130',
'\232','\160','\135','\333','\71','\42','\137','\335',
'\105','\240','\125','\361','\2','\326','\1','\204',
'\212','\250','\257','\141','\361','\351','\70','\376',
'\14','\67','\234','\135','\132','\45','\251','\256',
'\351','\242','\167','\247','\375','\165','\121','\110',
'\270','\6','\175','\4','\16','\132','\375','\201',
'\120','\366','\144','\113','\341','\57','\255','\353',
'\206','\271','\143','\221','\43','\350','\213','\373',
'\347','\117','\164','\375','\307','\246','\205','\10',
'\130','\270','\374','\245','\7','\11','\120','\40',
'\260','\341','\370','\307','\1','\277','\43','\355',
'\272','\50','\63','\0','\255','\252','\302','\257',
'\254','\304','\6','\147','\122','\364','\60','\57',
'\105','\201','\121','\31','\55','\355','\307','\230',
'\376','\73','\136','\43','\73','\337','\70','\67',
'\241','\257','\366','\137','\124','\220','\327','\50',
'\162','\172','\124','\355','\333','\142','\127','\30',
'\342','\275','\151','\335','\203','\102','\264','\202',
'\205','\202','\11','\115','\126','\62','\327','\204',
'\122','\213','\23','\144','\31','\314','\327','\55',
'\236','\223','\257','\116','\151','\205','\342','\174',
'\264','\332','\31','\355','\211','\6','\137','\35',
'\135','\152','\114','\356','\166','\12','\120','\104',
'\332','\277','\75','\214','\370','\321','\64','\145',
'\10','\215','\304','\225','\267','\255','\43','\150',
'\45','\236','\0','\146','\272','\315','\63','\15',
'\260','\122','\46','\213','\323','\5','\370','\71',
'\212','\41','\211','\323','\314','\65','\130','\114',
'\265','\172','\355','\43','\113','\221','\350','\117',
'\270','\370','\307','\352','\65','\112','\113','\241',
'\65','\33','\3','\45','\4','\375','\15','\273',
'\327','\111','\246','\224','\142','\221','\176','\325',
'\344','\7','\235','\316','\227','\350','\301','\17',
'\35','\13','\210','\256','\1','\371','\53','\315',
'\111','\77','\54','\70','\322','\306','\314','\337',
'\225','\346','\204','\173','\370','\0','\215','\251',
'\333','\274','\142','\206','\111','\120','\165','\146',
'\321','\302','\367','\312','\53','\105','\355','\367',
'\262','\277','\17','\345','\227','\176','\107','\72',
'\52','\337','\154','\20','\237','\160','\231','\265',
'\71','\304','\265','\65','\315','\345','\337','\121',
'\214','\153','\50','\353','\57','\116','\227','\132',
'\273','\243','\17','\350','\140','\34','\20','\45',
'\203','\234','\317','\23','\202','\216','\206','\177',
'\66','\214','\242','\76','\25','\237','\137','\127',
'\4','\153','\326','\255','\35','\53','\52','\4',
'\134','\317','\100','\213','\355','\12','\57','\270',
'\216','\51','\222','\216','\317','\77','\360','\64',
'\147','\273','\25','\200','\60','\153','\60','\266',
'\107','\76','\250','\323','\55','\6','\365','\312',
'\15','\246','\200','\362','\275','\106','\215','\1',
'\45','\345','\51','\224','\157','\363','\47','\114',
'\44','\227','\5','\110','\374','\43','\15','\221',
'\301','\230','\13','\372','\113','\347','\52','\214',
'\4','\253','\145','\73','\317','\374','\15','\32',
'\227','\174','\172','\346','\156','\14','\250','\264',
'\173','\17','\310','\62','\271','\16','\237','\177',
'\242','\323','\214','\202','\166','\220','\363','\12',
'\115','\253','\31','\207','\204','\32','\242','\156',
'\33','\365','\177','\311','\151','\376','\360','\102',
'\365','\66','\155','\310','\160','\231','\227','\370',
'\135','\270','\42','\107','\167','\361','\167','\51',
'\140','\204','\146','\243','\242','\251','\302','\71',
'\30','\221','\304','\360','\234','\5','\346','\250',
'\165','\322','\137','\56','\30','\2','\222','\44',
'\112','\224','\40','\340','\166','\370','\252','\340',
'\335','\355','\5','\167','\141','\353','\132','\115',
'\300','\231','\150','\101','\201','\225','\146','\71',
'\156','\215','\47','\226','\172','\175','\176','\177',
'\246','\170','\235','\52','\107','\326','\336','\312',
'\317','\52','\207','\133','\203','\121','\102','\73',
'\144','\321','\7','\200','\206','\24','\245','\312',
'\355','\246','\102','\114','\260','\322','\2','\22',
'\35','\332','\357','\6','\150','\10','\12','\204',
'\65','\225','\72','\1','\16','\121','\223','\242',
'\236','\177','\124','\56','\314','\252','\174','\223',
'\133','\302','\72','\101','\156','\325','\277','\113',
'\364','\360','\40','\202','\102','\360','\247','\301',
'\62','\207','\54','\130','\106','\371','\115','\131',
'\113','\176','\22','\156','\144','\53','\121','\321',
'\345','\375','\344','\345','\305','\31','\334','\163',
'\143','\261','\323','\265','\12','\231','\124','\105',
'\63','\324','\34','\135','\106','\121','\101','\211',
'\334','\73','\133','\210','\20','\246','\155','\201',
'\215','\156','\255','\213','\345','\15','\154','\354',
'\123','\126','\74','\350','\263','\373','\30','\301',
'\241','\325','\23','\333','\33','\252','\143','\144',
'\157','\166','\70','\71','\62','\316','\6','\165',
'\136','\175','\171','\41','\344','\127','\55','\161',
'\62','\105','\112','\152','\66','\226','\45','\170',
'\313','\344','\120','\325','\163','\235','\172','\30',
'\17','\57','\22','\304','\50','\24','\277','\341',
'\43','\365','\236','\111','\305','\200','\234','\51',
'\12','\56','\212','\104','\66','\304','\350','\127',
'\146','\10','\46','\155','\246','\360','\265','\173',
'\131','\202','\202','\233','\221','\253','\323','\11',
'\73','\33','\336','\361','\137','\137','\216','\163',
'\175','\107','\257','\6','\366','\211','\142','\223',
'\325','\216','\102','\213','\202','\23','\15','\32',
'\271','\202','\152','\323','\40','\74','\231','\225',
'\260','\103','\223','\60','\315','\136','\364','\322',
'\62','\226','\52','\32','\155','\241','\156','\77',
'\310','\376','\61','\274','\212','\227','\156','\245',
'\152','\233','\135','\56','\220','\57','\176','\251',
'\25','\30','\332','\333','\33','\300','\45','\225',
'\305','\102','\132','\305','\210','\165','\252','\24',
'\27','\316','\166','\360','\253','\151','\27','\6',
'\55','\173','\125','\360','\103','\64','\255','\233',
'\320','\347','\276','\247','\30','\367','\223','\263',
'\347','\141','\106','\230','\353','\4','\215','\144',
'\155','\340','\330','\237','\353','\111','\373','\171',
'\106','\361','\35','\163','\247','\54','\1','\21',
'\131','\370','\176','\356','\211','\245','\206','\262',
'\140','\224','\30','\13','\175','\315','\12','\262',
'\64','\142','\323','\316','\214','\201','\231','\276',
'\215','\305','\50','\51','\156','\355','\55','\65',
'\1','\260','\200','\337','\206','\277','\131','\272',
'\254','\165','\312','\35','\274','\172','\337','\207',
'\144','\126','\113','\175','\275','\23','\343','\44',
'\366','\315','\156','\251','\36','\306','\130','\245',
'\347','\242','\77','\137','\20','\217','\322','\254',
'\375','\174','\203','\16','\302','\205','\332','\151',
'\315','\132','\272','\304','\316','\146','\334','\132',
'\310','\334','\237','\47','\256','\311','\346','\376',
'\166','\116','\230','\256','\266','\131','\254','\371',
'\210','\357','\302','\367','\104','\175','\107','\233',
'\4','\220','\11','\366','\27','\216','\275','\347',
'\53','\360','\305','\141','\330','\170','\0','\3',
'\363','\222','\112','\312','\100','\134','\41','\61',
'\125','\51','\272','\326','\116','\115','\234','\337',
'\261','\255','\140','\115','\360','\155','\376','\162',
'\132','\6','\143','\353','\30','\117','\330','\126',
'\232','\230','\75','\155','\56','\370','\376','\43',
'\335','\133','\226','\311','\235','\117','\347','\270',
'\226','\277','\375','\354','\140','\257','\52','\234',
'\367','\260','\214','\322','\56','\50','\62','\61',
'\112','\137','\100','\204','\257','\114','\367','\367',
'\27','\237','\117','\3','\15','\116','\107','\274',
'\15','\261','\136','\104','\40','\257','\44','\150',
'\115','\334','\55','\250','\110','\261','\53','\345',
'\103','\366','\172','\231','\101','\230','\160','\330',
'\262','\113','\215','\33','\216','\171','\170','\217',
'\143','\256','\245','\204','\336','\177','\5','\144',
'\33','\110','\12','\26','\12','\136','\207','\76',
'\353','\3','\165','\361','\277','\315','\104','\174',
'\116','\256','\236','\233','\320','\243','\142','\347',
'\301','\343','\236','\62','\246','\314','\135','\5',
'\12','\166','\44','\117','\330','\242','\223','\65',
'\374','\336','\174','\362','\250','\323','\55','\172',
'\354','\301','\345','\254','\327','\70','\7','\26',
'\76','\222','\361','\272','\204','\344','\251','\266',
'\6','\225','\165','\313','\44','\160','\117','\322',
'\300','\333','\264','\43','\312','\51','\340','\252',
'\244','\14','\77','\353','\107','\170','\132','\123',
'\160','\215','\331','\15','\175','\43','\345','\51',
'\35','\240','\75','\204','\10','\3','\231','\101',
'\35','\331','\154','\353','\357','\216','\272','\352',
'\41','\144','\340','\167','\146','\206','\317','\333',
'\170','\324','\215','\236','\150','\114','\71','\177',
'\275','\332','\266','\336','\71','\122','\356','\270',
'\351','\332','\13','\317','\22','\322','\202','\224',
'\24','\5','\176','\336','\271','\230','\357','\235',
'\341','\142','\315','\50','\311','\107','\122','\144',
'\356','\220','\340','\40','\174','\347','\150','\214',
'\117','\332','\367','\207','\313','\67','\352','\332',
'\361','\53','\320','\77','\202','\360','\54','\131',
'\352','\45','\375','\31','\325','\7','\117','\135',
'\213','\57','\327','\211','\7','\113','\250','\71',
'\22','\123','\15','\172','\26','\375','\160','\207',
'\166','\250','\52','\260','\161','\112','\24','\221',
'\5','\260','\107','\361','\337','\35','\342','\25',
'\310','\243','\171','\121','\40','\354','\151','\152',
'\263','\50','\24','\321','\177','\341','\105','\131',
'\40','\114','\320','\342','\46','\62','\171','\165',
'\146','\63','\20','\265','\52','\273','\316','\247',
'\236','\344','\145','\104','\175','\216','\102','\34',
'\207','\334','\336','\215','\43','\341','\27','\216',
'\112','\24','\340','\52','\362','\47','\105','\40',
'\4','\224','\352','\333','\156','\332','\50','\227',
'\237','\66','\361','\361','\275','\143','\101','\220',
'\2','\274','\2','\372','\77','\122','\60','\105',
'\56','\173','\311','\116','\114','\363','\177','\7',
'\104','\56','\121','\213','\317','\336','\332','\366',
'\332','\42','\213','\50','\242','\304','\136','\4',
'\344','\145','\140','\243','\331','\121','\240','\374',
'\117','\273','\211','\342','\12','\111','\273','\250',
'\133','\317','\146','\225','\277','\174','\62','\375',
'\327','\233','\232','\232','\170','\335','\244','\61',
'\53','\16','\303','\151','\350','\66','\112','\233',
'\322','\201','\45','\72','\36','\135','\303','\334',
'\112','\357','\214','\7','\113','\177','\137','\262',
'\72','\121','\336','\157','\147','\152','\261','\262',
'\350','\7','\274','\316','\53','\136','\343','\176',
'\205','\165','\10','\270','\60','\262','\235','\44',
'\141','\110','\272','\10','\165','\71','\176','\167',
'\44','\274','\254','\32','\13','\353','\266','\123',
'\373','\256','\62','\144','\55','\21','\76','\45',
'\215','\103','\114','\103','\12','\45','\63','\301',
'\55','\42','\302','\115','\137','\304','\373','\224',
'\13','\16','\234','\305','\60','\322','\100','\137',
'\347','\303','\153','\354','\247','\170','\357','\236',
'\33','\227','\105','\47','\331','\271','\173','\217',
'\260','\236','\174','\105','\272','\343','\22','\234',
'\306','\155','\367','\100','\150','\304','\64','\355',
'\50','\263','\375','\341','\112','\171','\105','\234',
'\161','\44','\320','\32','\11','\30','\201','\156',
'\106','\56','\314','\357','\152','\172','\52','\370',
'\233','\267','\355','\305','\55','\166','\131','\175',
'\46','\40','\126','\277','\257','\143','\15','\102',
'\127','\115','\167','\312','\170','\227','\161','\314',
'\321','\100','\121','\242','\227','\275','\224','\251',
'\214','\334','\213','\212','\311','\341','\217','\57',
'\314','\202','\241','\76','\236','\351','\373','\170',
'\134','\214','\236','\307','\341','\153','\143','\360',
'\53','\21','\72','\241','\254','\166','\354','\366',
'\123','\322','\36','\160','\304','\42','\105','\335',
'\360','\352','\53','\260','\264','\5','\151','\337',
'\24','\77','\150','\364','\21','\35','\274','\224',
'\211','\267','\154','\130','\35','\155','\350','\355',
'\220','\265','\22','\354','\64','\365','\17','\173',
'\167','\136','\136','\354','\132','\105','\64','\225',
'\207','\201','\65','\124','\131','\177','\15','\144',
'\101','\147','\136','\326','\355','\116','\315','\165',
'\163','\313','\121','\307','\252','\43','\253','\176',
'\267','\345','\276','\274','\303','\74','\206','\44',
'\165','\4','\370','\105','\177','\232','\40','\33',
'\102','\106','\232','\153','\113','\353','\316','\370',
'\246','\170','\55','\234','\201','\366','\16','\142',
'\14','\61','\132','\24','\0','\162','\104','\46',
'\45','\30','\74','\201','\3','\341','\245','\172',
'\45','\330','\232','\15','\240','\263','\157','\124',
'\331','\215','\142','\256','\74','\24','\337','\117',
'\335','\340','\11','\375','\227','\75','\165','\306',
'\40','\301','\320','\75','\112','\33','\210','\121',
'\236','\237','\72','\321','\272','\370','\202','\274',
'\123','\170','\12','\2','\331','\6','\260','\135',
'\366','\171','\257','\305','\227','\211','\360','\322',
'\71','\372','\322','\315','\367','\165','\321','\35',
'\71','\350','\244','\72','\72','\330','\235','\177',
'\351','\216','\115','\321','\245','\352','\324','\115',
'\207','\325','\152','\173','\42','\22','\177','\201',
'\145','\121','\55','\200','\276','\250','\36','\322',
'\215','\173','\263','\10','\140','\304','\210','\251',
'\361','\331','\133','\33','\147','\336','\123','\267',
'\224','\200','\240','\167','\266','\51','\375','\146',
'\132','\50','\205','\62','\170','\330','\366','\100',
'\234','\163','\166','\212','\203','\365','\370','\316',
'\100','\350','\332','\251','\133','\126','\202','\375',
'\365','\4','\371','\214','\57','\104','\333','\115',
'\336','\100','\220','\324','\101','\221','\257','\42',
'\230','\161','\77','\274','\202','\374','\64','\225',
'\212','\233','\175','\51','\370','\270','\212','\15',
'\111','\105','\240','\242','\226','\144','\217','\171',
'\331','\4','\112','\305','\162','\373','\232','\251',
'\135','\77','\54','\61','\134','\310','\163','\347',
'\271','\264','\353','\213','\255','\166','\274','\144',
'\146','\214','\212','\172','\36','\263','\207','\356',
'\24','\0','\220','\121','\237','\153','\42','\241',
'\340','\310','\75','\154','\13','\24','\117','\344',
'\336','\23','\30','\33','\106','\333','\361','\351',
'\357','\112','\65','\41','\133','\56','\207','\272',
'\136','\143','\363','\356','\274','\137','\227','\324',
'\324','\327','\163','\221','\57','\337','\162','\173',
'\234','\100','\105','\374','\127','\130','\146','\130',
'\241','\54','\135','\257','\255','\301','\11','\13',
'\362','\144','\370','\171','\343','\157','\161','\143',
'\166','\327','\325','\110','\204','\230','\203','\226',
'\247','\10','\303','\252','\200','\304','\261','\122',
'\160','\216','\365','\103','\173','\50','\143','\161',
'\141','\232','\273','\346','\133','\331','\367','\6',
'\227','\74','\71','\253','\326','\241','\246','\117',
'\273','\315','\271','\151','\106','\340','\247','\200',
'\3','\135','\275','\373','\23','\227','\153','\160',
'\77','\230','\16','\327','\300','\340','\356','\130',
'\114','\364','\360','\27','\175','\317','\251','\356',
'\213','\227','\205','\176','\24','\312','\6','\360',
'\133','\42','\327','\202','\30','\26','\26','\76',
'\357','\104','\272','\277','\43','\325','\244','\313',
'\102','\42','\343','\307','\313','\334','\171','\22',
'\2','\232','\43','\220','\177','\124','\132','\1',
'\247','\311','\35','\1','\275','\31','\263','\177',
'\261','\315','\62','\265','\4','\314','\165','\171',
'\224','\351','\132','\203','\113','\261','\226','\131',
'\171','\16','\145','\26','\130','\373','\212','\157',
'\221','\354','\214','\154','\320','\326','\126','\310',
'\214','\370','\154','\107','\221','\307','\173','\247',
'\40','\25','\52','\150','\373','\135','\166','\70',
'\67','\57','\163','\120','\107','\333','\51','\347',
'\100','\71','\111','\273','\73','\254','\375','\353',
'\311','\105','\130','\341','\264','\70','\342','\101',
'\326','\65','\242','\10','\264','\255','\106','\251',
'\253','\123','\43','\24','\233','\250','\312','\240',
'\57','\46','\355','\256','\22','\102','\303','\336',
'\75','\364','\304','\363','\250','\307','\360','\215',
'\203','\40','\54','\62','\100','\200','\35','\237',
'\11','\42','\314','\214','\232','\272','\47','\351',
'\10','\143','\34','\300','\217','\33','\117','\111',
'\261','\202','\332','\201','\154','\305','\213','\14',
'\111','\1','\107','\16','\15','\20','\170','\104',
'\36','\212','\221','\120','\262','\250','\216','\363',
'\103','\102','\30','\46','\132','\53','\1','\167',
'\325','\52','\352','\61','\162','\15','\46','\304',
'\3','\102','\371','\111','\10','\176','\133','\23',
'\40','\331','\250','\150','\277','\43','\110','\71',
'\253','\260','\226','\73','\122','\176','\344','\204',
'\375','\4','\24','\302','\7','\371','\347','\13',
'\151','\266','\247','\111','\350','\152','\253','\11',
'\225','\147','\217','\37','\330','\201','\221','\121',
'\53','\346','\35','\256','\5','\54','\333','\367',
'\327','\102','\241','\275','\224','\360','\147','\27',
'\103','\101','\327','\237','\112','\305','\247','\132',
'\204','\256','\35','\242','\171','\267','\247','\6',
'\53','\320','\41','\111','\144','\366','\242','\122',
'\163','\146','\161','\111','\37','\344','\141','\73',
'\341','\337','\2','\65','\7','\330','\251','\45',
'\272','\114','\215','\272','\306','\344','\124','\204',
'\316','\73','\114','\113','\320','\202','\265','\10',
'\136','\366','\235','\25','\276','\315','\260','\136',
'\363','\320','\25','\275','\360','\302','\260','\246',
'\1','\345','\374','\31','\45','\40','\234','\152',
'\10','\171','\252','\12','\46','\222','\212','\203',
'\353','\60','\37','\110','\347','\345','\164','\24',
'\112','\246','\275','\255','\276','\64','\347','\144',
'\62','\321','\237','\306','\133','\42','\175','\331',
'\13','\154','\364','\165','\344','\311','\351','\337',
'\205','\155','\376','\102','\4','\174','\275','\234',
'\262','\52','\170','\136','\303','\172','\354','\55',
'\155','\75','\25','\260','\255','\277','\230','\13',
'\251','\261','\110','\244','\341','\366','\125','\255',
'\363','\223','\64','\137','\157','\155','\167','\5',
'\144','\141','\260','\252','\163','\274','\101','\0',
'\301','\103','\54','\77','\346','\361','\212','\327',
'\171','\153','\16','\243','\14','\120','\45','\255',
'\304','\26','\254','\31','\6','\57','\353','\155',
'\172','\321','\203','\155','\327','\246','\200','\321',
'\67','\334','\226','\76','\352','\5','\210','\106',
'\316','\35','\261','\126','\147','\107','\31','\365',
'\345','\336','\46','\6','\344','\231','\176','\357',
'\13','\116','\250','\250','\7','\204','\116','\63',
'\234','\360','\360','\265','\350','\334','\223','\315',
'\57','\315','\361','\376','\111','\121','\355','\212',
'\305','\64','\304','\56','\237','\372','\203','\3',
'\100','\245','\115','\64','\33','\253','\261','\0',
'\354','\343','\143','\147','\375','\306','\7','\133',
'\30','\360','\100','\265','\371','\202','\154','\7',
'\16','\3','\133','\321','\151','\242','\201','\233',
'\3','\14','\121','\200','\336','\313','\306','\323',
'\264','\0','\275','\223','\127','\153','\36','\275',
'\46','\301','\322','\314','\221','\337','\201','\243',
'\241','\100','\341','\231','\313','\344','\71','\30',
'\57','\143','\250','\151','\157','\233','\74','\132',
'\312','\176','\341','\54','\360','\213','\205','\151',
'\140','\32','\356','\316','\325','\233','\142','\351',
'\334','\355','\327','\164','\134','\52','\227','\132',
'\65','\77','\260','\52','\171','\221','\70','\274',
'\140','\267','\334','\302','\11','\344','\100','\266',
'\143','\362','\301','\127','\207','\16','\200','\324',
'\55','\151','\270','\104','\42','\251','\42','\15',
'\36','\52','\116','\56','\202','\237','\50','\250',
'\266','\373','\15','\176','\23','\123','\151','\7',
'\300','\156','\272','\30','\7','\354','\67','\47',
'\40','\206','\333','\113','\17','\227','\223','\266',
'\1','\340','\265','\212','\320','\354','\115','\133',
'\62','\122','\233','\157','\7','\222','\234','\5',
'\110','\62','\150','\325','\231','\211','\171','\147',
'\50','\154','\167','\366','\315','\73','\133','\132',
'\310','\271','\341','\243','\114','\326','\152','\272',
'\45','\210','\256','\61','\123','\255','\255','\270',
'\135','\156','\322','\27','\310','\127','\132','\246',
'\30','\372','\250','\55','\237','\166','\360','\167',
'\54','\23','\277','\325','\77','\356','\127','\217',
'\323','\370','\3','\201','\300','\41','\14','\201',
'\124','\174','\116','\240','\233','\344','\150','\121',
'\214','\315','\342','\166','\226','\172','\142','\225',
'\264','\265','\55','\320','\353','\213','\253','\333',
'\32','\126','\310','\50','\273','\165','\75','\305',
'\156','\331','\112','\50','\336','\166','\24','\226',
'\135','\12','\247','\155','\2','\313','\343','\366',
'\336','\254','\270','\227','\145','\225','\103','\215',
'\15','\163','\126','\162','\5','\361','\75','\364',
'\160','\327','\316','\51','\170','\226','\260','\274',
'\325','\200','\27','\124','\45','\25','\230','\213',
'\110','\124','\320','\110','\157','\113','\210','\170',
'\151','\161','\131','\265','\374','\174','\344','\237',
'\64','\46','\125','\255','\245','\315','\136','\123',
'\143','\144','\306','\35','\46','\367','\4','\223',
'\12','\344','\31','\117','\363','\277','\17','\317',
'\243','\247','\322','\311','\360','\342','\136','\202',
'\35','\262','\214','\10','\356','\237','\207','\263',
'\111','\353','\56','\40','\126','\310','\151','\231',
'\314','\260','\341','\63','\213','\2','\101','\261',
'\11','\242','\40','\234','\335','\275','\313','\257',
'\32','\331','\355','\5','\240','\72','\52','\160',
'\27','\150','\154','\300','\63','\15','\215','\361',
'\236','\60','\251','\123','\5','\237','\265','\155',
'\230','\262','\32','\351','\376','\114','\210','\330',
'\324','\305','\262','\225','\225','\134','\13','\332',
'\21','\261','\211','\70','\374','\240','\372','\233',
'\11','\155','\202','\363','\276','\176','\46','\100',
'\5','\50','\122','\12','\244','\315','\144','\73',
'\127','\272','\67','\372','\20','\354','\365','\3',
'\46','\74','\320','\20','\340','\67','\106','\130',
'\255','\360','\226','\363','\246','\272','\164','\230',
'\155','\367','\136','\231','\114','\36','\344','\16',
'\213','\114','\146','\353','\276','\303','\124','\260',
'\276','\220','\134','\311','\326','\153','\312','\153',
'\271','\131','\253','\12','\201','\105','\126','\110',
'\56','\351','\223','\43','\55','\62','\255','\165',
'\174','\77','\262','\27','\137','\332','\214','\101',
'\66','\75','\110','\234','\70','\56','\101','\45',
'\310','\164','\6','\50','\360','\261','\137','\25',
'\317','\47','\71','\247','\143','\171','\43','\174',
'\131','\37','\200','\72','\62','\233','\51','\224',
'\220','\66','\4','\130','\124','\243','\350','\31',
'\31','\154','\351','\204','\315','\64','\121','\376',
'\344','\223','\225','\44','\250','\357','\27','\241',
'\254','\21','\227','\336','\331','\235','\74','\257',
'\132','\231','\76','\154','\343','\52','\303','\317',
'\102','\245','\22','\77','\317','\62','\145','\267',
'\162','\157','\241','\122','\300','\213','\130','\337',
'\333','\152','\31','\346','\276','\275','\124','\111',
'\51','\104','\161','\216','\143','\36','\231','\54',
'\40','\46','\202','\125','\56','\212','\111','\313',
'\65','\55','\304','\362','\335','\175','\132','\21',
'\157','\371','\371','\125','\306','\244','\66','\25',
'\204','\353','\273','\172','\4','\331','\156','\246',
'\71','\164','\214','\315','\255','\331','\205','\311',
'\372','\37','\354','\151','\365','\256','\23','\24',
'\276','\205','\321','\350','\32','\277','\275','\171',
'\353','\15','\307','\140','\334','\156','\106','\152',
'\61','\245','\6','\163','\206','\253','\375','\132',
'\73','\164','\203','\334','\212','\165','\126','\13',
'\376','\256','\261','\242','\374','\376','\140','\316',
'\102','\346','\307','\72','\122','\331','\313','\237',
'\343','\167','\344','\31','\44','\234','\274','\263',
'\205','\42','\66','\166','\302','\276','\31','\35',
'\311','\340','\227','\241','\33','\154','\235','\221',
'\301','\215','\215','\155','\206','\140','\234','\206',
'\355','\140','\246','\23','\262','\227','\361','\341',
'\343','\320','\13','\132','\342','\362','\60','\260',
'\44','\72','\253','\343','\122','\152','\2','\134',
'\211','\204','\161','\20','\350','\45','\301','\243',
'\76','\145','\153','\147','\332','\45','\354','\30',
'\273','\135','\365','\67','\32','\63','\216','\145',
'\3','\11','\373','\166','\167','\167','\304','\112',
'\42','\100','\247','\112','\273','\313','\364','\312',
'\24','\255','\143','\253','\330','\373','\42','\276',
'\66','\241','\347','\332','\362','\212','\226','\76',
'\142','\60','\147','\235','\150','\364','\256','\162',
'\121','\275','\4','\57','\113','\357','\365','\374',
'\46','\124','\116','\76','\77','\3','\315','\304',
'\246','\312','\17','\330','\16','\317','\233','\256',
'\356','\351','\106','\214','\200','\4','\324','\26',
'\206','\17','\135','\232','\274','\36','\360','\161',
'\372','\35','\313','\32','\173','\306','\242','\263',
'\101','\175','\127','\333','\163','\251','\235','\72',
'\66','\364','\320','\260','\63','\326','\13','\263',
'\130','\60','\162','\45','\206','\6','\42','\361',
'\210','\220','\201','\62','\133','\360','\20','\263',
'\262','\121','\272','\317','\320','\374','\13','\322',
'\265','\54','\47','\226','\227','\352','\117','\10',
'\224','\72','\227','\111','\153','\260','\242','\215',
'\311','\6','\40','\243','\330','\237','\57','\143',
'\76','\244','\211','\233','\77','\40','\144','\117',
'\233','\70','\214','\75','\231','\161','\117','\372',
'\305','\173','\21','\266','\30','\157','\203','\117',
'\23','\26','\106','\64','\275','\351','\156','\146',
'\217','\355','\11','\131','\120','\335','\270','\230',
'\163','\34','\67','\264','\227','\62','\11','\172',
'\361','\12','\135','\244','\341','\105','\363','\327',
'\24','\232','\222','\344','\317','\205','\226','\3',
'\303','\157','\5','\344','\330','\35','\361','\113',
'\210','\7','\130','\163','\215','\76','\132','\73',
'\1','\252','\240','\327','\32','\347','\316','\214',
'\51','\230','\123','\152','\5','\232','\246','\47',
'\223','\60','\133','\44','\244','\311','\37','\266',
'\11','\261','\3','\11','\53','\14','\10','\206',
'\160','\370','\220','\144','\200','\372','\133','\234',
'\333','\324','\211','\263','\170','\50','\53','\320',
'\375','\246','\157','\203','\23','\116','\62','\171',
'\72','\247','\107','\264','\146','\323','\60','\214',
'\274','\176','\340','\120','\42','\176','\372','\302',
'\61','\264','\263','\40','\265','\320','\203','\216',
'\167','\132','\301','\165','\244','\63','\125','\342',
'\73','\22','\203','\60','\327','\334','\240','\30',
'\243','\324','\0','\127','\161','\132','\114','\156',
'\56','\24','\301','\235','\250','\376','\43','\126',
'\153','\64','\61','\66','\372','\123','\353','\243',
'\332','\13','\117','\311','\14','\240','\165','\301',
'\205','\236','\136','\30','\50','\204','\153','\41',
'\356','\214','\103','\156','\70','\2','\366','\226',
'\67','\167','\337','\15','\317','\376','\242','\340',
'\117','\140','\370','\131','\62','\306','\3','\15',
'\45','\64','\131','\71','\305','\41','\137','\323',
'\374','\271','\125','\354','\202','\353','\1','\370',
'\273','\356','\320','\262','\156','\256','\276','\167',
'\171','\51','\275','\355','\276','\112','\301','\32',
'\217','\210','\31','\112','\176','\12','\224','\75',
'\272','\360','\74','\34','\220','\362','\41','\326',
'\345','\316','\232','\103','\10','\270','\113','\13',
'\244','\317','\40','\173','\241','\124','\76','\277',
'\376','\143','\260','\145','\132','\16','\326','\344',
'\10','\1','\30','\52','\245','\5','\54','\156',
'\204','\354','\315','\102','\250','\154','\332','\316',
'\211','\264','\230','\31','\256','\206','\145','\324',
'\347','\72','\4','\146','\115','\34','\252','\167',
'\112','\254','\165','\304','\20','\335','\57','\324',
'\106','\152','\341','\207','\244','\266','\131','\51',
'\173','\166','\141','\363','\277','\272','\72','\217',
'\266','\152','\264','\337','\364','\115','\76','\256',
'\246','\143','\137','\40','\307','\43','\165','\137',
'\71','\42','\267','\272','\306','\251','\232','\127',
'\202','\230','\261','\154','\46','\135','\71','\232',
'\6','\22','\145','\357','\26','\14','\137','\345',
'\60','\277','\317','\350','\357','\256','\216','\43',
'\362','\307','\101','\123','\223','\333','\126','\33',
'\144','\242','\172','\5','\66','\62','\117','\277',
'\62','\237','\120','\335','\62','\305','\320','\26',
'\56','\357','\170','\162','\25','\275','\1','\146',
'\63','\17','\360','\41','\46','\7','\112','\105',
'\336','\304','\237','\67','\330','\230','\25','\32',
'\221','\337','\157','\175','\231','\317','\41','\213',
'\223','\62','\141','\162','\265','\117','\320','\37',
'\302','\264','\144','\375','\130','\266','\252','\30',
'\15','\157','\155','\264','\54','\1','\363','\241',
'\133','\106','\20','\242','\10','\261','\225','\24',
'\113','\73','\274','\47','\256','\223','\231','\306',
'\267','\57','\310','\44','\154','\23','\45','\356',
'\203','\51','\154','\365','\306','\157','\365','\177',
'\345','\46','\264','\215','\240','\74','\205','\373',
'\374','\65','\64','\63','\111','\227','\362','\36',
'\353','\104','\107','\66','\72','\204','\253','\333',
'\137','\351','\354','\172','\21','\131','\215','\117',
'\76','\206','\361','\172','\112','\2','\322','\230',
'\103','\204','\334','\66','\330','\237','\102','\246',
'\311','\202','\10','\113','\136','\233','\360','\226',
'\346','\56','\232','\45','\361','\203','\50','\143',
'\361','\257','\125','\23','\275','\122','\332','\104',
'\171','\303','\55','\355','\46','\20','\352','\310',
'\2','\276','\336','\352','\147','\12','\46','\331',
'\224','\102','\44','\66','\103','\200','\275','\375',
'\12','\30','\10','\352','\171','\62','\273','\1',
'\177','\111','\153','\162','\64','\345','\157','\15',
'\231','\127','\0','\272','\77','\246','\245','\362',
'\244','\75','\12','\322','\330','\247','\207','\170',
'\104','\343','\201','\263','\333','\4','\130','\364',
'\45','\207','\352','\253','\312','\176','\63','\240',
'\235','\346','\53','\160','\247','\25','\166','\220',
'\323','\44','\150','\305','\126','\266','\247','\212',
'\43','\343','\206','\207','\33','\357','\330','\367',
'\271','\30','\217','\316','\376','\220','\174','\23',
'\150','\345','\151','\22','\121','\317','\320','\270',
'\277','\234','\54','\123','\350','\150','\33','\54',
'\314','\321','\34','\373','\177','\242','\43','\212',
'\175','\104','\6','\224','\303','\245','\277','\32',
'\143','\312','\230','\2','\315','\77','\157','\366',
'\26','\263','\153','\36','\157','\65','\323','\57',
'\231','\112','\307','\151','\313','\326','\122','\115',
'\201','\235','\171','\116','\2','\220','\111','\307',
'\206','\167','\347','\311','\367','\344','\127','\235',
'\67','\320','\63','\25','\325','\114','\67','\60',
'\302','\147','\33','\164','\243','\56','\320','\233',
'\104','\41','\32','\3','\107','\211','\236','\170',
'\173','\231','\352','\131','\136','\207','\367','\151',
'\207','\137','\115','\127','\62','\43','\174','\345',
'\130','\127','\63','\375','\111','\276','\51','\15',
'\372','\115','\336','\235','\205','\317','\245','\315',
'\111','\323','\306','\176','\31','\23','\262','\333',
'\17','\302','\101','\244','\344','\21','\161','\67',
'\45','\313','\44','\251','\3','\315','\321','\255',
'\120','\27','\221','\337','\155','\343','\265','\252',
'\120','\212','\25','\215','\76','\152','\163','\41',
'\62','\147','\330','\237','\273','\132','\72','\114',
'\374','\307','\13','\245','\142','\300','\27','\235',
'\202','\74','\150','\213','\323','\110','\147','\354',
'\220','\330','\313','\104','\125','\27','\373','\327',
'\315','\134','\235','\135','\153','\21','\226','\261',
'\223','\125','\166','\340','\17','\155','\103','\142',
'\370','\74','\301','\370','\24','\134','\44','\131',
'\230','\174','\1','\135','\356','\51','\265','\113',
'\63','\265','\57','\201','\261','\250','\131','\37',
'\103','\54','\311','\332','\63','\12','\307','\66',
'\100','\116','\103','\366','\47','\321','\42','\202',
'\50','\347','\56','\351','\136','\355','\127','\163',
'\22','\111','\26','\64','\71','\70','\232','\302',
'\372','\162','\265','\35','\64','\122','\67','\25',
'\124','\30','\57','\4','\311','\306','\362','\277',
'\243','\344','\152','\174','\156','\41','\204','\177',
'\61','\126','\44','\26','\325','\165','\376','\52',
'\370','\52','\306','\42','\10','\132','\217','\147',
'\104','\110','\177','\341','\176','\66','\356','\221',
'\30','\276','\20','\43','\142','\44','\303','\317',
'\171','\212','\125','\21','\111','\333','\205','\353',
'\322','\345','\204','\347','\110','\273','\320','\21',
'\313','\252','\154','\113','\275','\137','\273','\140',
'\334','\230','\277','\303','\53','\42','\244','\15',
'\126','\337','\235','\40','\22','\322','\226','\62',
'\174','\323','\165','\62','\167','\216','\324','\267',
'\341','\166','\142','\236','\370','\240','\53','\153',
'\64','\164','\163','\51','\311','\261','\125','\240',
'\151','\260','\161','\25','\103','\101','\74','\24',
'\310','\364','\350','\37','\146','\33','\64','\160',
'\144','\113','\117','\1','\130','\361','\267','\265',
'\166','\353','\255','\373','\217','\124','\365','\147',
'\13','\177','\111','\130','\47','\271','\202','\160',
'\21','\27','\116','\366','\107','\70','\263','\175',
'\375','\206','\173','\314','\42','\0','\251','\2',
'\244','\361','\312','\352','\31','\202','\220','\127',
'\343','\320','\317','\157','\322','\373','\273','\345',
'\114','\307','\205','\212','\317','\266','\13','\164',
'\107','\272','\60','\243','\72','\163','\325','\263',
'\233','\167','\210','\267','\345','\206','\263','\52',
'\311','\206','\161','\274','\150','\373','\161','\211',
'\77','\300','\233','\175','\304','\66','\53','\360',
'\126','\122','\142','\141','\233','\12','\243','\270',
'\7','\111','\374','\26','\222','\277','\164','\302',
'\210','\304','\356','\121','\361','\324','\110','\126',
'\144','\334','\101','\330','\326','\313','\17','\342',
'\160','\77','\271','\317','\362','\363','\265','\6',
'\375','\254','\315','\26','\12','\234','\100','\350',
'\133','\275','\74','\352','\137','\136','\340','\163',
'\341','\43','\246','\313','\26','\257','\2','\0',
'\113','\120','\110','\174','\264','\326','\201','\237',
'\150','\60','\25','\7','\165','\351','\164','\360',
'\372','\36','\62','\12','\264','\137','\337','\160',
'\363','\150','\207','\165','\76','\335','\204','\135',
'\103','\234','\0','\2','\47','\126','\163','\152',
'\302','\35','\270','\124','\250','\201','\222','\161',
'\57','\75','\104','\237','\160','\43','\173','\354',
'\15','\252','\174','\262','\257','\53','\53','\375',
'\154','\367','\116','\40','\100','\361','\135','\211',
'\356','\363','\305','\25','\355','\273','\363','\6',
'\133','\333','\370','\213','\55','\130','\62','\105',
'\60','\7','\263','\6','\26','\200','\14','\71',
'\56','\303','\215','\223','\222','\326','\174','\50',
'\340','\147','\245','\344','\364','\42','\271','\306',
'\171','\164','\375','\251','\25','\251','\360','\62',
'\311','\211','\222','\253','\137','\206','\14','\306',
'\22','\136','\156','\10','\305','\321','\114','\224',
'\141','\204','\214','\203','\103','\207','\124','\374',
'\75','\52','\163','\151','\41','\205','\55','\146',
'\261','\57','\45','\212','\20','\252','\332','\114',
'\5','\41','\127','\161','\335','\61','\162','\174',
'\66','\322','\66','\252','\207','\152','\5','\142',
'\327','\142','\206','\323','\166','\141','\352','\304',
'\130','\114','\260','\124','\362','\32','\41','\344',
'\115','\271','\115','\230','\0','\350','\366','\331',
'\122','\20','\341','\114','\17','\312','\173','\353',
'\326','\327','\330','\272','\266','\300','\52','\40',
'\323','\235','\43','\105','\61','\210','\204','\134',
'\200','\45','\307','\72','\132','\72','\346','\352',
'\212','\255','\371','\272','\251','\52','\231','\67',
'\167','\22','\61','\255','\347','\230','\137','\360',
'\70','\50','\317','\304','\216','\274','\362','\146',
'\205','\64','\121','\124','\41','\120','\121','\236',
'\137','\306','\211','\11','\327','\220','\300','\342',
'\144','\371','\164','\154','\371','\307','\151','\31',
'\237','\223','\347','\6','\31','\225','\230','\353',
'\116','\63','\60','\240','\121','\10','\166','\224',
'\255','\66','\70','\173','\236','\227','\207','\37',
'\210','\74','\333','\207','\243','\134','\315','\230',
'\345','\36','\43','\200','\375','\110','\50','\312',
'\31','\123','\320','\133','\361','\50','\133','\104',
'\161','\70','\217','\161','\43','\15','\113','\6',
'\255','\311','\341','\230','\226','\120','\302','\202',
'\32','\333','\316','\365','\26','\115','\71','\150',
'\303','\376','\353','\55','\163','\172','\323','\202',
'\215','\21','\244','\13','\304','\174','\324','\142',
'\3','\106','\260','\10','\61','\331','\220','\257',
'\217','\320','\157','\303','\264','\340','\160','\356',
'\251','\175','\313','\30','\107','\315','\60','\113',
'\305','\233','\346','\216','\36','\350','\361','\151',
'\114','\375','\205','\122','\230','\130','\367','\324',
'\124','\155','\140','\133','\62','\6','\122','\102',
'\113','\11','\43','\65','\136','\353','\157','\231',
'\370','\262','\202','\336','\101','\70','\267','\271',
'\173','\2','\142','\131','\256','\152','\233','\147',
'\62','\301','\363','\22','\157','\103','\332','\54',
'\241','\64','\14','\76','\247','\270','\260','\222',
'\201','\203','\356','\76','\251','\214','\70','\354',
'\73','\236','\146','\40','\225','\16','\147','\37',
'\147','\356','\51','\37','\104','\266','\347','\146',
'\60','\375','\357','\2','\216','\53','\200','\215',
'\305','\303','\362','\216','\323','\215','\0','\22',
'\132','\5','\347','\117','\376','\43','\103','\72',
'\3','\15','\115','\245','\51','\205','\21','\362',
'\274','\52','\113','\33','\270','\252','\354','\50',
'\103','\140','\67','\106','\16','\103','\333','\55',
'\352','\34','\321','\235','\214','\142','\125','\75',
'\101','\305','\265','\163','\161','\325','\240','\235',
'\177','\221','\126','\214','\176','\273','\66','\55',
'\32','\312','\175','\140','\226','\337','\344','\305',
'\313','\207','\356','\265','\23','\72','\122','\36',
'\251','\271','\303','\53','\53','\60','\361','\26',
'\106','\154','\310','\255','\253','\372','\107','\44',
'\120','\111','\256','\325','\116','\32','\351','\42',
'\33','\372','\256','\106','\252','\157','\253','\150',
'\112','\351','\303','\320','\20','\72','\344','\364',
'\172','\63','\261','\277','\220','\311','\333','\250',
'\153','\362','\7','\314','\217','\342','\162','\100',
'\156','\331','\147','\132','\40','\266','\61','\120',
'\360','\225','\372','\352','\152','\174','\110','\246',
'\335','\272','\316','\201','\211','\7','\256','\332',
'\213','\323','\336','\117','\116','\363','\67','\370',
'\367','\56','\324','\44','\316','\147','\254','\322',
'\125','\366','\360','\17','\77','\323','\370','\220',
'\147','\62','\355','\100','\223','\175','\11','\165',
'\341','\70','\205','\373','\346','\310','\243','\35',
'\77','\4','\371','\133','\74','\164','\254','\114',
'\123','\233','\223','\111','\163','\61','\321','\26',
'\361','\245','\310','\122','\15','\156','\125','\331',
'\355','\317','\51','\163','\53','\70','\46','\10',
'\242','\124','\172','\365','\140','\307','\276','\120',
'\100','\107','\166','\100','\160','\231','\37','\7',
'\111','\265','\321','\163','\234','\153','\366','\67',
'\140','\245','\337','\106','\23','\344','\215','\2',
'\133','\113','\322','\62','\300','\67','\366','\351',
'\133','\23','\200','\230','\320','\351','\341','\266',
'\156','\203','\222','\344','\44','\100','\171','\240',
'\367','\56','\43','\55','\7','\207','\131','\125',
'\352','\142','\276','\326','\310','\65','\44','\222',
'\106','\116','\337','\267','\106','\265','\274','\277',
'\254','\260','\43','\345','\335','\6','\250','\5',
'\140','\157','\163','\350','\155','\301','\304','\310',
'\313','\62','\270','\30','\242','\53','\370','\255',
'\325','\341','\300','\75','\142','\163','\366','\361',
'\315','\262','\271','\35','\31','\226','\217','\42',
'\225','\214','\35','\260','\221','\367','\230','\230',
'\167','\375','\366','\11','\123','\110','\310','\271',
'\342','\157','\54','\277','\354','\21','\165','\75',
'\330','\345','\313','\274','\57','\314','\57','\63',
'\271','\272','\41','\231','\327','\257','\343','\33',
'\376','\121','\20','\241','\206','\344','\174','\220',
'\330','\322','\121','\133','\147','\302','\362','\142',
'\322','\142','\216','\367','\40','\350','\263','\133',
'\375','\244','\367','\260','\301','\343','\74','\121',
'\225','\131','\156','\220','\251','\261','\112','\166',
'\23','\151','\56','\67','\253','\106','\177','\135',
'\273','\173','\240','\46','\355','\0','\336','\254',
'\272','\266','\63','\364','\72','\312','\105','\142',
'\277','\347','\55','\57','\27','\254','\342','\164',
'\245','\23','\127','\370','\371','\263','\174','\121',
'\351','\101','\175','\313','\251','\161','\13','\63',
'\230','\206','\143','\172','\232','\234','\223','\235',
'\302','\32','\337','\262','\113','\312','\340','\152',
'\221','\136','\250','\41','\126','\154','\240','\207',
'\35','\253','\221','\277','\60','\250','\316','\41',
'\44','\22','\117','\143','\140','\27','\223','\73',
'\201','\367','\117','\330','\134','\17','\260','\142',
'\43','\320','\72','\350','\100','\142','\143','\167',
'\142','\120','\52','\66','\73','\173','\12','\207',
'\230','\346','\142','\162','\156','\320','\270','\56',
'\73','\326','\346','\154','\53','\75','\43','\245',
'\111','\100','\224','\371','\154','\122','\26','\233',
'\170','\147','\1','\355','\250','\301','\201','\122',
'\274','\170','\360','\357','\343','\163','\76','\174',
'\206','\370','\123','\106','\365','\21','\140','\305',
'\326','\302','\64','\115','\322','\110','\6','\106',
'\52','\351','\200','\176','\354','\105','\357','\275',
'\120','\354','\30','\344','\106','\22','\105','\203',
'\150','\321','\304','\37','\176','\315','\225','\62',
'\74','\27','\155','\51','\324','\73','\142','\375',
'\50','\4','\326','\375','\54','\54','\67','\3',
'\220','\201','\201','\326','\221','\221','\60','\244',
'\205','\35','\337','\250','\162','\311','\343','\53',
'\301','\174','\246','\355','\31','\302','\205','\114',
'\134','\226','\143','\157','\177','\262','\123','\302',
'\114','\247','\77','\324','\133','\312','\116','\20',
'\125','\357','\202','\346','\42','\236','\170','\332',
'\246','\77','\232','\273','\253','\204','\257','\314',
'\64','\330','\227','\273','\66','\144','\126','\164',
'\306','\356','\46','\212','\247','\105','\262','\345',
'\324','\147','\122','\135','\27','\1','\234','\217',
'\141','\211','\327','\303','\174','\355','\154','\327',
'\337','\120','\255','\320','\42','\10','\135','\156',
'\33','\33','\224','\303','\206','\166','\75','\27',
'\334','\21','\342','\347','\7','\301','\137','\134',
'\367','\106','\15','\145','\101','\115','\26','\127',
'\322','\13','\364','\342','\275','\74','\247','\274',
'\211','\100','\171','\57','\370','\332','\57','\5',
'\220','\220','\265','\26','\253','\36','\300','\210',
'\174','\242','\332','\76','\112','\41','\305','\135',
'\220','\123','\111','\126','\174','\142','\346','\327',
'\200','\121','\61','\360','\25','\101','\233','\157',
'\311','\152','\212','\3','\3','\364','\220','\164',
'\172','\241','\231','\147','\276','\105','\324','\261',
'\277','\47','\375','\10','\304','\355','\151','\40',
'\306','\112','\213','\250','\32','\312','\366','\10',
'\50','\71','\244','\341','\141','\200','\207','\336',
'\223','\41','\130','\164','\234','\370','\155','\137',
'\212','\345','\144','\154','\46','\317','\170','\301',
'\102','\121','\135','\40','\222','\20','\353','\232',
'\15','\167','\16','\111','\247','\333','\103','\137',
'\234','\27','\242','\130','\307','\230','\246','\5',
'\10','\72','\113','\371','\67','\345','\33','\214',
'\53','\55','\322','\30','\327','\253','\232','\64',
'\204','\101','\321','\104','\60','\324','\270','\205',
'\335','\52','\41','\53','\230','\222','\245','\265',
'\45','\57','\114','\172','\70','\100','\376','\327',
'\361','\314','\167','\143','\252','\156','\245','\27',
'\71','\167','\67','\233','\0','\23','\136','\122',
'\24','\33','\254','\273','\254','\332','\214','\122',
'\73','\6','\237','\163','\0','\54','\215','\154',
'\7','\131','\15','\224','\324','\244','\46','\212',
'\26','\5','\176','\333','\344','\312','\50','\321',
'\362','\106','\332','\303','\77','\322','\112','\204',
'\304','\36','\126','\26','\230','\1','\357','\255',
'\141','\61','\35','\43','\56','\222','\51','\101',
'\50','\145','\340','\347','\232','\271','\10','\265',
'\100','\201','\34','\351','\222','\214','\174','\131',
'\367','\332','\167','\260','\166','\222','\141','\323',
'\334','\30','\372','\373','\57','\156','\311','\75',
'\6','\131','\246','\124','\77','\275','\10','\153',
'\46','\173','\236','\346','\212','\135','\114','\214',
'\124','\4','\50','\61','\143','\272','\167','\351',
'\137','\206','\100','\105','\171','\112','\254','\324',
'\132','\240','\113','\253','\2','\37','\15','\152',
'\205','\163','\210','\105','\57','\245','\133','\361',
'\270','\163','\272','\350','\363','\177','\217','\61',
'\375','\6','\11','\223','\273','\350','\135','\272',
'\365','\53','\135','\105','\55','\245','\61','\142',
'\112','\263','\165','\171','\306','\367','\132','\56',
'\210','\72','\74','\345','\215','\303','\166','\237',
'\252','\37','\115','\274','\360','\147','\71','\43',
'\50','\324','\316','\264','\336','\143','\227','\212',
'\352','\200','\254','\15','\360','\313','\271','\136',
'\133','\351','\105','\376','\367','\364','\201','\321',
'\2','\210','\165','\167','\305','\314','\311','\276',
'\310','\364','\314','\270','\7','\216','\163','\52',
'\336','\121','\240','\216','\313','\201','\35','\327',
'\35','\303','\106','\170','\240','\153','\130','\151',
'\14','\331','\347','\46','\133','\11','\26','\277',
'\241','\132','\154','\141','\376','\67','\100','\202',
'\313','\200','\267','\176','\124','\25','\171','\230',
'\174','\232','\217','\237','\161','\300','\250','\10',
'\122','\161','\114','\245','\216','\74','\350','\133',
'\205','\37','\326','\114','\54','\7','\136','\204',
'\107','\252','\350','\0','\370','\227','\300','\3',
'\206','\166','\330','\265','\171','\331','\326','\105',
'\323','\115','\212','\261','\13','\70','\44','\106',
'\71','\14','\224','\123','\33','\340','\326','\170',
'\24','\30','\306','\273','\106','\315','\61','\15',
'\234','\376','\32','\333','\60','\260','\67','\77',
'\344','\311','\267','\330','\352','\174','\254','\315',
'\226','\317','\71','\75','\121','\361','\321','\171',
'\330','\331','\232','\235','\250','\45','\16','\224',
'\147','\300','\371','\61','\114','\246','\230','\152',
'\165','\303','\31','\334','\217','\344','\246','\231',
'\76','\353','\40','\250','\44','\13','\375','\7',
'\226','\131','\341','\376','\143','\220','\245','\171',
'\123','\362','\276','\15','\354','\307','\47','\374',
'\221','\115','\35','\250','\314','\305','\173','\321',
'\102','\101','\273','\251','\301','\175','\356','\354',
'\154','\7','\161','\76','\270','\317','\103','\170',
'\62','\313','\277','\346','\27','\337','\36','\154',
'\251','\335','\353','\172','\111','\200','\27','\251',
'\353','\332','\214','\45','\326','\324','\65','\332',
'\277','\362','\115','\6','\260','\106','\335','\44',
'\243','\77','\273','\122','\174','\371','\271','\230',
'\125','\340','\34','\233','\245','\32','\266','\221',
'\106','\176','\357','\262','\37','\374','\231','\15',
'\50','\200','\53','\303','\142','\43','\261','\20',
'\210','\13','\136','\254','\331','\374','\361','\132',
'\224','\16','\64','\43','\264','\70','\100','\262',
'\311','\314','\331','\353','\342','\272','\162','\222',
'\265','\257','\362','\317','\37','\273','\245','\372',
'\114','\15','\147','\201','\144','\175','\120','\222',
'\263','\43','\327','\62','\31','\336','\232','\225',
'\311','\172','\320','\222','\3','\364','\152','\215',
'\276','\121','\104','\127','\230','\163','\11','\1',
'\132','\21','\113','\202','\37','\141','\27','\64',
'\2','\137','\33','\120','\266','\174','\50','\47',
'\360','\241','\212','\307','\43','\61','\360','\112',
'\261','\126','\66','\341','\136','\356','\172','\176',
'\40','\170','\27','\307','\47','\36','\155','\241',
'\211','\56','\7','\73','\321','\70','\233','\163',
'\236','\355','\41','\266','\25','\32','\205','\333',
'\357','\321','\152','\126','\71','\370','\306','\140',
'\325','\353','\111','\172','\252','\356','\212','\174',
'\3','\234','\246','\154','\34','\326','\304','\325',
'\145','\144','\166','\7','\264','\200','\266','\277',
'\241','\52','\176','\253','\163','\175','\26','\14',
'\53','\152','\224','\144','\33','\304','\256','\307',
'\210','\273','\77','\15','\144','\4','\44','\60',
'\57','\100','\327','\214','\205','\111','\111','\310',
'\300','\335','\345','\105','\323','\305','\315','\72',
'\23','\351','\54','\17','\162','\277','\260','\64',
'\64','\66','\57','\50','\372','\154','\165','\330',
'\150','\237','\305','\55','\250','\120','\76','\357',
'\126','\252','\67','\14','\122','\356','\26','\123',
'\207','\11','\335','\74','\11','\266','\271','\126',
'\326','\67','\306','\127','\237','\367','\212','\320',
'\207','\73','\15','\306','\163','\335','\43','\277',
'\141','\43','\60','\65','\266','\371','\171','\334',
'\136','\366','\140','\242','\314','\174','\116','\227',
'\30','\212','\40','\341','\334','\137','\313','\226',
'\110','\24','\332','\34','\67','\34','\331','\252',
'\251','\346','\332','\145','\351','\75','\10','\77',
'\37','\73','\67','\275','\33','\52','\4','\357',
'\361','\370','\352','\126','\245','\7','\344','\266',
'\171','\110','\132','\120','\17','\74','\324','\37',
'\326','\44','\153','\16','\141','\43','\222','\47',
'\227','\46','\365','\205','\270','\303','\260','\221',
'\231','\221','\301','\346','\14','\235','\325','\211',
'\20','\265','\371','\332','\332','\343','\253','\367',
'\243','\20','\307','\225','\160','\77','\76','\137',
'\136','\130','\224','\276','\372','\107','\340','\356',
'\70','\243','\126','\64','\140','\357','\162','\335',
'\200','\132','\105','\313','\262','\122','\347','\70',
'\307','\234','\360','\371','\310','\151','\260','\252',
'\131','\35','\220','\327','\74','\344','\26','\61',
'\64','\236','\224','\17','\61','\131','\17','\152',
'\110','\53','\243','\166','\166','\324','\256','\263',
'\376','\166','\56','\42','\13','\372','\327','\7',
'\157','\266','\236','\120','\275','\31','\316','\25',
'\22','\32','\47','\73','\333','\22','\210','\23',
'\234','\172','\312','\245','\136','\265','\321','\121',
'\255','\330','\71','\353','\47','\374','\142','\36',
'\213','\172','\347','\365','\174','\374','\341','\64',
'\215','\51','\252','\317','\166','\370','\341','\255',
'\310','\125','\350','\36','\322','\332','\244','\120',
'\53','\43','\150','\307','\234','\320','\63','\334',
'\213','\375','\217','\14','\233','\217','\61','\356',
'\261','\165','\274','\77','\341','\365','\254','\134',
'\254','\25','\202','\137','\137','\220','\351','\256',
'\355','\350','\145','\160','\140','\121','\3','\236',
'\167','\65','\22','\67','\26','\352','\0','\153',
'\247','\323','\233','\163','\172','\23','\145','\174',
'\65','\176','\141','\53','\157','\341','\330','\354',
'\206','\132','\370','\144','\170','\105','\355','\105',
'\314','\350','\243','\155','\173','\230','\365','\60',
'\337','\4','\374','\63','\40','\10','\324','\1',
'\142','\12','\166','\373','\327','\275','\257','\61',
'\147','\51','\222','\250','\157','\147','\165','\166',
'\362','\226','\166','\106','\362','\135','\247','\63',
'\34','\126','\264','\163','\11','\33','\74','\124',
'\362','\316','\320','\317','\325','\206','\317','\125',
'\364','\14','\366','\64','\245','\230','\210','\253',
'\346','\245','\176','\254','\375','\270','\303','\15',
'\27','\247','\53','\7','\347','\50','\141','\21',
'\64','\206','\260','\73','\300','\333','\355','\107',
'\257','\317','\307','\105','\136','\355','\157','\17',
'\77','\15','\107','\2','\63','\23','\362','\340',
'\231','\111','\36','\231','\305','\362','\147','\254',
'\134','\46','\153','\137','\71','\231','\74','\327',
'\146','\161','\144','\373','\205','\257','\32','\65',
'\15','\340','\350','\115','\142','\156','\250','\207',
'\52','\212','\316','\140','\242','\123','\13','\326',
'\151','\125','\337','\142','\150','\276','\364','\11',
'\0','\145','\31','\345','\260','\203','\67','\22',
'\37','\223','\312','\144','\117','\23','\240','\204',
'\273','\25','\176','\326','\61','\174','\342','\135',
'\161','\376','\342','\237','\376','\225','\375','\371',
'\301','\124','\202','\375','\36','\225','\345','\244',
'\315','\270','\341','\313','\42','\216','\362','\14',
'\343','\340','\176','\310','\277','\136','\13','\5',
'\243','\303','\142','\261','\145','\241','\54','\172',
'\304','\21','\201','\203','\360','\22','\221','\257',
'\61','\235','\127','\147','\274','\203','\121','\153',
'\132','\37','\74','\115','\57','\133','\315','\106',
'\217','\161','\173','\232','\244','\322','\23','\35',
'\207','\103','\163','\271','\320','\157','\271','\167',
'\205','\137','\213','\161','\56','\256','\66','\134',
'\64','\360','\65','\312','\374','\130','\147','\246',
'\210','\76','\20','\66','\101','\41','\365','\270',
'\27','\367','\333','\112','\124','\242','\64','\375',
'\351','\113','\27','\144','\136','\164','\336','\347',
'\347','\75','\211','\155','\221','\360','\47','\316',
'\54','\173','\335','\206','\32','\27','\102','\207',
'\164','\100','\131','\311','\120','\301','\174','\114',
'\10','\366','\100','\216','\306','\166','\176','\271',
'\21','\223','\130','\272','\7','\335','\165','\217',
'\202','\232','\357','\212','\121','\162','\200','\235',
'\102','\40','\44','\316','\255','\150','\150','\56',
'\53','\365','\25','\326','\350','\111','\130','\234',
'\367','\103','\225','\217','\304','\317','\6','\325',
'\375','\71','\244','\200','\23','\24','\342','\331',
'\313','\112','\170','\357','\214','\65','\140','\164',
'\233','\24','\44','\167','\30','\245','\210','\253',
'\233','\266','\345','\320','\137','\154','\272','\150',
'\123','\331','\125','\275','\332','\62','\315','\265',
'\132','\0','\261','\161','\60','\262','\23','\113',
'\246','\336','\250','\363','\160','\231','\33','\225',
'\344','\312','\52','\125','\70','\260','\14','\25',
'\357','\170','\75','\33','\4','\254','\141','\117',
'\155','\105','\315','\44','\65','\356','\220','\6',
'\320','\32','\245','\11','\64','\12','\133','\76',
'\351','\334','\177','\277','\55','\40','\124','\12',
'\222','\153','\243','\14','\175','\43','\272','\240',
'\15','\216','\115','\146','\144','\21','\71','\246',
'\337','\114','\324','\167','\202','\352','\277','\344',
'\344','\321','\166','\367','\56','\75','\223','\313',
'\67','\77','\325','\315','\211','\176','\313','\51',
'\71','\131','\64','\273','\206','\160','\175','\4',
'\367','\254','\110','\65','\360','\27','\166','\305',
'\63','\143','\66','\256','\231','\74','\323','\343',
'\124','\363','\61','\250','\211','\310','\50','\332',
'\124','\47','\56','\114','\370','\250','\155','\36',
'\236','\23','\300','\177','\243','\127','\367','\123',
'\352','\340','\233','\340','\363','\3','\1','\136',
'\4','\14','\215','\72','\166','\324','\226','\330',
'\16','\65','\164','\161','\162','\37','\315','\336',
'\170','\125','\22','\267','\107','\320','\341','\365',
'\357','\45','\32','\117','\121','\123','\104','\172',
'\30','\367','\327','\207','\65','\116','\117','\276',
'\231','\153','\154','\352','\17','\46','\161','\363',
'\360','\166','\305','\244','\260','\255','\31','\210',
'\362','\224','\141','\356','\236','\107','\210','\117',
'\233','\10','\5','\207','\20','\247','\155','\203',
'\144','\233','\360','\242','\37','\64','\333','\146',
'\61','\204','\354','\142','\263','\373','\361','\310',
'\134','\316','\201','\255','\76','\247','\121','\25',
'\7','\317','\371','\153','\7','\75','\30','\211',
'\104','\101','\266','\116','\370','\34','\226','\52',
'\70','\222','\154','\357','\11','\340','\374','\205',
'\1','\64','\215','\243','\240','\57','\153','\331',
'\126','\257','\352','\317','\72','\277','\316','\7',
'\114','\132','\364','\303','\361','\20','\312','\325',
'\314','\115','\40','\231','\7','\41','\250','\54',
'\171','\111','\200','\66','\2','\127','\272','\74',
'\363','\134','\131','\65','\51','\205','\136','\141',
'\304','\37','\316','\156','\330','\332','\333','\324',
'\272','\103','\43','\151','\54','\23','\37','\32',
'\273','\216','\166','\5','\225','\110','\250','\351',
'\225','\324','\363','\202','\67','\356','\370','\367',
'\221','\334','\166','\234','\364','\225','\15','\137',
'\215','\360','\276','\374','\300','\161','\223','\11',
'\305','\203','\100','\245','\116','\156','\147','\313',
'\214','\361','\367','\73','\345','\106','\307','\133',
'\265','\61','\252','\56','\121','\254','\205','\261',
'\307','\362','\303','\357','\371','\72','\201','\310',
'\110','\2','\212','\373','\11','\262','\60','\66',
'\46','\105','\361','\305','\103','\137','\330','\332',
'\50','\312','\71','\320','\167','\131','\206','\374',
'\72','\76','\32','\122','\145','\264','\233','\126',
'\312','\262','\111','\40','\3','\147','\202','\331',
'\11','\336','\364','\200','\65','\54','\52','\312',
'\267','\205','\324','\246','\323','\314','\55','\360',
'\66','\327','\261','\347','\257','\307','\263','\113',
'\73','\22','\317','\302','\52','\253','\257','\135',
'\22','\262','\373','\317','\204','\271','\316','\111',
'\7','\102','\237','\174','\5','\322','\242','\177',
'\166','\66','\117','\77','\75','\325','\51','\253',
'\314','\101','\203','\144','\247','\157','\372','\156',
'\175','\222','\7','\126','\306','\132','\116','\266',
'\141','\172','\227','\157','\212','\335','\41','\217',
'\105','\215','\76','\304','\4','\77','\141','\243',
'\65','\150','\310','\55','\362','\164','\371','\120',
'\337','\44','\344','\225','\110','\303','\135','\345',
'\360','\300','\1','\170','\320','\351','\111','\50',
'\14','\32','\233','\171','\174','\206','\164','\54',
'\261','\341','\161','\60','\22','\211','\130','\173',
'\210','\325','\155','\21','\14','\67','\105','\34',
'\201','\43','\50','\212','\302','\331','\57','\10',
'\265','\345','\273','\301','\142','\157','\366','\144',
'\232','\16','\50','\66','\267','\45','\237','\324',
'\77','\335','\120','\105','\46','\157','\370','\3',
'\144','\265','\274','\271','\44','\4','\64','\115',
'\106','\142','\10','\332','\321','\63','\333','\263',
'\226','\217','\300','\305','\20','\325','\42','\176',
'\276','\161','\314','\10','\265','\52','\322','\150',
'\236','\130','\271','\206','\235','\303','\210','\346',
'\167','\227','\164','\360','\24','\310','\261','\277',
'\36','\335','\170','\36','\270','\315','\57','\373',
'\26','\144','\0','\234','\146','\160','\304','\137',
'\352','\157','\145','\340','\36','\22','\1','\326',
'\160','\342','\115','\361','\35','\123','\117','\302',
'\133','\150','\7','\30','\303','\142','\242','\77',
'\155','\111','\40','\274','\245','\165','\344','\240',
'\47','\221','\242','\42','\253','\135','\157','\133',
'\202','\17','\26','\237','\12','\313','\130','\326',
'\365','\266','\174','\251','\311','\117','\171','\112',
'\110','\322','\75','\372','\145','\242','\146','\25',
'\140','\75','\332','\174','\111','\202','\151','\204',
'\161','\120','\233','\32','\227','\123','\304','\43',
'\0','\114','\176','\251','\17','\101','\71','\22',
'\61','\145','\107','\57','\24','\354','\360','\375',
'\234','\305','\4','\357','\63','\5','\21','\322',
'\265','\57','\52','\40','\77','\67','\20','\66',
'\241','\327','\263','\66','\216','\171','\43','\142',
'\237','\24','\170','\131','\117','\144','\162','\236',
'\231','\154','\356','\167','\24','\214','\22','\311',
'\252','\246','\365','\330','\262','\277','\355','\316',
'\304','\62','\266','\272','\234','\214','\212','\375',
'\54','\226','\0','\273','\360','\203','\300','\353',
'\203','\202','\52','\145','\204','\117','\107','\321',
'\23','\200','\176','\323','\333','\225','\326','\240',
'\347','\340','\130','\222','\162','\355','\334','\255',
'\347','\153','\115','\255','\10','\301','\201','\330',
'\223','\164','\131','\336','\343','\130','\367','\43',
'\112','\324','\141','\12','\32','\234','\40','\176',
'\255','\223','\155','\102','\173','\375','\151','\252',
'\356','\264','\13','\245','\153','\356','\331','\356',
'\66','\251','\161','\321','\173','\72','\200','\150',
'\63','\304','\44','\57','\161','\336','\103','\75',
'\174','\227','\242','\141','\126','\222','\376','\243',
'\76','\257','\154','\22','\334','\275','\242','\340',
'\241','\234','\344','\351','\52','\101','\73','\157',
'\27','\325','\237','\31','\246','\326','\253','\112',
'\166','\133','\70','\130','\22','\33','\114','\124',
'\7','\324','\353','\220','\17','\76','\55','\302',
'\254','\116','\216','\127','\70','\153','\221','\61',
'\40','\107','\51','\375','\177','\342','\127','\350',
'\247','\226','\45','\113','\255','\143','\263','\336',
'\356','\225','\132','\321','\10','\316','\335','\345',
'\264','\130','\357','\250','\126','\307','\362','\214',
'\262','\240','\62','\305','\155','\350','\240','\50',
'\344','\126','\65','\74','\65','\136','\343','\306',
'\230','\66','\147','\217','\24','\310','\75','\223',
'\223','\207','\312','\270','\164','\155','\110','\176',
'\261','\316','\25','\160','\55','\125','\371','\50',
'\373','\44','\152','\170','\271','\24','\247','\41',
'\303','\150','\334','\234','\123','\205','\103','\63',
'\356','\143','\260','\373','\273','\113','\56','\273',
'\131','\21','\320','\274','\243','\317','\36','\167',
'\161','\375','\261','\166','\165','\267','\117','\140',
'\227','\146','\10','\167','\261','\374','\120','\364',
'\164','\4','\236','\160','\261','\51','\12','\203',
'\262','\154','\306','\103','\235','\357','\320','\143',
'\317','\166','\224','\332','\322','\106','\341','\371',
'\356','\254','\135','\112','\230','\4','\261','\324',
'\136','\144','\364','\137','\61','\56','\245','\15',
'\56','\73','\341','\232','\343','\11','\175','\235',
'\167','\133','\32','\146','\143','\114','\102','\236',
'\245','\72','\172','\121','\367','\333','\335','\244',
'\76','\213','\140','\3','\13','\307','\345','\62',
'\375','\70','\277','\153','\341','\225','\266','\117',
'\6','\43','\317','\107','\14','\325','\366','\270',
'\46','\151','\376','\56','\42','\335','\47','\372',
'\74','\251','\331','\170','\46','\346','\345','\50',
'\124','\366','\226','\34','\77','\154','\246','\353',
'\23','\21','\122','\311','\270','\376','\175','\102',
'\116','\266','\50','\222','\241','\375','\113','\325',
'\342','\310','\324','\305','\125','\112','\247','\32',
'\17','\275','\364','\101','\34','\366','\73','\343',
'\32','\107','\5','\217','\74','\51','\256','\146',
'\305','\21','\242','\140','\220','\232','\177','\72',
'\347','\327','\46','\374','\117','\257','\260','\57',
'\366','\12','\303','\50','\302','\123','\222','\347',
'\360','\27','\356','\100','\154','\257','\126','\247',
'\340','\373','\236','\127','\372','\367','\150','\337',
'\306','\351','\175','\363','\63','\346','\120','\61',
'\126','\354','\240','\372','\273','\375','\20','\35',
'\60','\246','\301','\70','\237','\207','\321','\7',
'\21','\74','\177','\133','\173','\310','\335','\304',
'\263','\326','\207','\307','\226','\223','\327','\322',
'\346','\107','\176','\63','\20','\315','\325','\330',
'\352','\142','\327','\237','\174','\236','\1','\1',
'\66','\245','\154','\153','\117','\274','\131','\203',
'\6','\130','\174','\210','\355','\101','\107','\126',
'\272','\137','\140','\0','\12','\62','\6','\17',
'\5','\330','\356','\17','\170','\41','\153','\60',
'\317','\56','\112','\16','\244','\226','\212','\217',
'\214','\37','\321','\17','\303','\243','\322','\333',
'\134','\375','\260','\154','\46','\336','\225','\6',
'\324','\75','\40','\101','\27','\301','\367','\165',
'\313','\135','\265','\335','\175','\243','\203','\336',
'\2','\142','\116','\235','\14','\46','\170','\376',
'\3','\260','\355','\67','\12','\250','\70','\132',
'\142','\354','\127','\306','\31','\346','\14','\255',
'\34','\157','\277','\45','\226','\253','\142','\134',
'\371','\376','\175','\355','\324','\353','\314','\172',
'\331','\63','\350','\120','\250','\20','\150','\36',
'\100','\106','\35','\127','\14','\326','\173','\245',
'\150','\117','\40','\35','\120','\214','\61','\210',
'\325','\272','\115','\271','\0','\47','\360','\104',
'\253','\21','\250','\300','\332','\64','\61','\166',
'\7','\160','\5','\171','\132','\324','\52','\160',
'\50','\301','\240','\33','\353','\245','\210','\225',
'\206','\327','\246','\346','\64','\64','\353','\311',
'\77','\254','\122','\221','\174','\245','\172','\366',
'\46','\135','\173','\65','\240','\213','\0','\100',
'\63','\165','\22','\30','\361','\274','\127','\340',
'\321','\373','\155','\111','\267','\212','\372','\17',
'\112','\246','\374','\45','\64','\235','\157','\347',
'\215','\53','\11','\322','\262','\312','\76','\245',
'\6','\252','\171','\352','\173','\141','\316','\60',
'\141','\253','\223','\127','\161','\333','\206','\75',
'\316','\120','\54','\314','\330','\64','\1','\215',
'\156','\146','\236','\7','\100','\364','\22','\241',
'\231','\301','\264','\140','\30','\36','\207','\14',
'\326','\254','\367','\146','\304','\73','\346','\376',
'\273','\241','\64','\253','\211','\110','\222','\144',
'\21','\266','\66','\33','\110','\257','\247','\124',
'\323','\346','\136','\12','\27','\113','\126','\316',
'\223','\211','\230','\55','\251','\270','\372','\12',
'\125','\346','\320','\45','\253','\141','\370','\104',
'\355','\315','\1','\15','\7','\73','\265','\7',
'\30','\155','\315','\227','\220','\56','\4','\261',
'\114','\114','\331','\144','\233','\60','\153','\313',
'\307','\147','\31','\73','\250','\154','\236','\313',
'\315','\357','\74','\176','\272','\50','\312','\341',
'\75','\113','\122','\105','\34','\311','\236','\210',
'\37','\150','\13','\160','\270','\214','\245','\42',
'\376','\20','\320','\207','\200','\215','\74','\66',
'\57','\201','\105','\113','\21','\72','\22','\303',
'\55','\161','\131','\53','\167','\335','\46','\245',
'\150','\11','\151','\40','\352','\20','\216','\227',
'\277','\272','\71','\156','\26','\217','\316','\171',
'\36','\337','\352','\360','\121','\277','\141','\170',
'\0','\2','\6','\230','\240','\224','\222','\147',
'\117','\25','\116','\21','\167','\253','\146','\265',
'\142','\145','\137','\40','\347','\122','\141','\264',
'\161','\127','\44','\336','\75','\220','\23','\0',
'\270','\123','\320','\331','\230','\341','\60','\12',
'\127','\322','\17','\364','\247','\174','\30','\151',
'\155','\223','\313','\162','\202','\320','\145','\376',
'\110','\262','\306','\306','\373','\257','\66','\105',
'\275','\173','\2','\201','\4','\351','\234','\370',
'\366','\313','\123','\165','\316','\217','\116','\162',
'\16','\160','\27','\165','\224','\142','\310','\163',
'\225','\125','\327','\26','\334','\106','\154','\136',
'\333','\102','\240','\61','\254','\207','\32','\372',
'\204','\132','\132','\14','\73','\3','\65','\324',
'\34','\161','\356','\344','\322','\357','\327','\137',
'\236','\276','\340','\322','\74','\25','\335','\373',
'\106','\367','\33','\301','\124','\113','\74','\70',
'\124','\245','\206','\256','\217','\330','\232','\65',
'\265','\30','\65','\12','\12','\127','\177','\337',
'\373','\247','\25','\352','\63','\315','\275','\114',
'\6','\263','\205','\214','\162','\133','\72','\345',
'\272','\325','\257','\307','\344','\200','\337','\276',
'\302','\70','\274','\120','\340','\20','\75','\371',
'\361','\147','\74','\306','\327','\340','\375','\155',
'\351','\257','\376','\306','\363','\32','\125','\366',
'\124','\373','\114','\270','\165','\156','\241','\1',
'\265','\255','\255','\153','\364','\333','\116','\230',
'\310','\325','\215','\253','\147','\37','\220','\303',
'\200','\263','\173','\53','\12','\160','\35','\300',
'\202','\17','\242','\164','\32','\141','\253','\130',
'\235','\146','\71','\54','\171','\335','\334','\46',
'\56','\113','\112','\307','\17','\344','\331','\223',
'\170','\75','\26','\0','\124','\100','\125','\134',
'\17','\107','\62','\310','\142','\157','\273','\241',
'\246','\102','\140','\236','\17','\252','\351','\354',
'\72','\76','\25','\11','\34','\333','\153','\45',
'\376','\132','\102','\121','\375','\73','\233','\372',
'\276','\366','\312','\331','\315','\61','\74','\100',
'\20','\14','\262','\6','\307','\374','\47','\123',
'\75','\272','\274','\2','\101','\31','\300','\24',
'\207','\72','\375','\207','\241','\212','\145','\145',
'\61','\124','\353','\11','\341','\353','\304','\345',
'\0','\131','\113','\31','\232','\121','\144','\352',
'\321','\10','\122','\165','\231','\21','\225','\130',
'\305','\165','\355','\303','\116','\117','\172','\52',
'\14','\130','\271','\256','\141','\325','\355','\122',
'\347','\264','\257','\312','\177','\256','\175','\56',
'\163','\207','\230','\334','\73','\25','\351','\6',
'\107','\222','\117','\63','\211','\356','\366','\254',
'\160','\226','\151','\316','\320','\265','\154','\242',
'\223','\312','\334','\273','\324','\200','\155','\276',
'\56','\133','\163','\32','\304','\213','\62','\137',
'\347','\43','\104','\221','\4','\70','\344','\105',
'\41','\343','\355','\303','\330','\373','\300','\264',
'\131','\75','\150','\41','\117','\136','\323','\347',
'\252','\54','\355','\367','\351','\276','\306','\53',
'\165','\204','\340','\15','\333','\160','\341','\336',
'\160','\236','\122','\263','\61','\17','\352','\105',
'\257','\107','\53','\314','\240','\163','\227','\273',
'\255','\276','\260','\365','\134','\164','\142','\101',
'\162','\117','\160','\45','\312','\260','\310','\23',
'\214','\41','\364','\201','\274','\140','\117','\355',
'\176','\56','\106','\362','\141','\121','\310','\63',
'\21','\221','\234','\266','\244','\61','\25','\135',
'\21','\61','\276','\257','\122','\211','\364','\301',
'\310','\374','\206','\255','\21','\105','\362','\171',
'\255','\37','\131','\245','\316','\60','\54','\25',
'\75','\76','\145','\32','\135','\77','\35','\312',
'\127','\367','\371','\17','\117','\11','\372','\4',
'\200','\46','\231','\273','\202','\227','\7','\373',
'\123','\135','\120','\212','\322','\325','\77','\105',
'\330','\306','\264','\103','\72','\353','\305','\303',
'\173','\137','\23','\124','\302','\132','\162','\260',
'\313','\372','\124','\104','\135','\125','\325','\366',
'\144','\236','\145','\32','\4','\61','\4','\222',
'\173','\232','\372','\371','\105','\237','\221','\316',
'\132','\236','\147','\75','\72','\260','\46','\133',
'\370','\261','\44','\267','\320','\16','\101','\105',
'\112','\32','\107','\100','\245','\45','\62','\253',
'\4','\164','\354','\205','\40','\300','\212','\157',
'\2','\242','\40','\244','\204','\307','\265','\37',
'\123','\113','\321','\303','\212','\336','\140','\317',
'\225','\206','\14','\277','\177','\317','\104','\352',
'\234','\303','\35','\225','\54','\136','\245','\331',
'\135','\330','\345','\14','\202','\141','\344','\171',
'\203','\300','\63','\345','\244','\142','\147','\31',
'\327','\50','\150','\211','\223','\54','\311','\50',
'\105','\106','\276','\167','\104','\123','\71','\330',
'\306','\21','\220','\244','\71','\375','\150','\174',
'\256','\343','\126','\365','\73','\27','\41','\164',
'\176','\325','\326','\264','\31','\121','\237','\203',
'\22','\323','\173','\334','\50','\346','\21','\7',
'\22','\116','\216','\256','\46','\120','\366','\142',
'\250','\221','\262','\336','\23','\170','\207','\300',
'\64','\351','\372','\147','\53','\122','\307','\277',
'\376','\101','\25','\212','\326','\30','\204','\113',
'\72','\303','\102','\134','\3','\250','\353','\275',
'\241','\353','\270','\260','\45','\115','\273','\122',
'\324','\71','\306','\235','\17','\370','\201','\305',
'\24','\312','\154','\311','\114','\116','\166','\352',
'\357','\305','\45','\141','\122','\72','\5','\341',
'\24','\301','\237','\215','\103','\315','\75','\70',
'\55','\112','\36','\15','\312','\66','\323','\116',
'\322','\173','\66','\240','\125','\261','\267','\276',
'\326','\366','\21','\226','\12','\354','\24','\234',
'\14','\173','\177','\1','\56','\113','\361','\305',
'\366','\357','\222','\253','\370','\304','\63','\61',
'\356','\56','\120','\273','\60','\27','\315','\167',
'\265','\43','\16','\101','\237','\361','\106','\146',
'\123','\55','\171','\37','\74','\17','\33','\241',
'\214','\7','\117','\230','\200','\240','\360','\137',
'\101','\354','\207','\307','\115','\42','\126','\244',
'\50','\121','\206','\17','\16','\341','\252','\124',
'\143','\213','\133','\137','\103','\221','\77','\216',
'\354','\143','\63','\147','\310','\252','\40','\20',
'\372','\274','\271','\217','\342','\143','\273','\376',
'\311','\355','\302','\334','\313','\210','\142','\374',
'\31','\353','\144','\31','\326','\105','\76','\60',
'\361','\312','\15','\274','\76','\361','\347','\50',
'\207','\230','\113','\61','\25','\41','\360','\145',
'\162','\157','\371','\52','\272','\370','\117','\274',
'\161','\150','\131','\14','\154','\56','\50','\321',
'\65','\41','\104','\261','\115','\232','\16','\144',
'\172','\122','\341','\241','\314','\151','\255','\134',
'\171','\373','\276','\312','\330','\20','\7','\37',
'\225','\353','\170','\214','\243','\115','\32','\42',
'\35','\367','\117','\163','\123','\207','\267','\26',
'\364','\375','\152','\357','\40','\77','\264','\121',
'\322','\155','\222','\336','\211','\332','\267','\22',
'\3','\113','\141','\331','\106','\205','\254','\144',
'\135','\342','\66','\334','\111','\155','\246','\47',
'\50','\121','\67','\370','\205','\54','\147','\202',
'\347','\54','\52','\311','\6','\223','\165','\267',
'\232','\154','\75','\166','\76','\111','\156','\314',
'\325','\31','\267','\46','\270','\207','\351','\117',
'\204','\145','\264','\276','\323','\43','\233','\112',
'\260','\364','\233','\316','\107','\142','\365','\217',
'\14','\34','\45','\42','\263','\36','\75','\327',
'\313','\321','\104','\305','\130','\103','\346','\235',
'\173','\65','\56','\66','\114','\23','\301','\13',
'\137','\40','\120','\33','\224','\117','\217','\374',
'\115','\344','\340','\362','\57','\245','\150','\346',
'\350','\176','\220','\234','\326','\210','\210','\244',
'\225','\172','\175','\53','\314','\245','\376','\17',
'\316','\77','\100','\300','\356','\226','\156','\344',
'\347','\304','\64','\272','\154','\135','\24','\326',
'\237','\262','\267','\40','\105','\33','\13','\120',
'\121','\21','\332','\323','\140','\370','\215','\147',
'\374','\34','\275','\163','\71','\12','\321','\22',
'\336','\335','\343','\266','\102','\154','\231','\104',
'\51','\154','\120','\367','\105','\6','\200','\146',
'\102','\44','\2','\173','\135','\144','\22','\325',
'\345','\125','\326','\265','\33','\247','\74','\132',
'\210','\113','\316','\114','\325','\5','\317','\351',
'\316','\265','\320','\344','\233','\51','\125','\46',
'\65','\147','\233','\202','\356','\52','\324','\204',
'\125','\230','\346','\364','\225','\223','\164','\77',
'\235','\232','\47','\140','\152','\126','\21','\33',
'\355','\232','\336','\331','\164','\346','\0','\141',
'\0','\76','\315','\46','\21','\250','\344','\111',
'\254','\64','\45','\165','\232','\170','\330','\262',
'\306','\66','\5','\300','\264','\155','\46','\323',
'\147','\135','\263','\310','\31','\54','\236','\14',
'\352','\262','\316','\40','\370','\372','\145','\352',
'\322','\360','\132','\250','\241','\242','\364','\31',
'\62','\200','\45','\253','\315','\323','\301','\362',
'\122','\33','\126','\360','\256','\351','\342','\301',
'\140','\64','\33','\360','\326','\66','\274','\255',
'\3','\204','\233','\10','\237','\122','\21','\163',
'\13','\230','\162','\47','\175','\263','\50','\50',
'\121','\232','\4','\7','\43','\57','\55','\45',
'\354','\325','\14','\230','\3','\242','\7','\276',
'\165','\213','\71','\245','\125','\16','\312','\52',
'\310','\272','\353','\246','\305','\263','\224','\73',
'\107','\105','\344','\76','\63','\32','\245','\104',
'\16','\17','\316','\370','\373','\3','\250','\337',
'\316','\266','\73','\325','\354','\164','\223','\25',
'\123','\276','\42','\343','\372','\23','\176','\250',
'\306','\374','\310','\11','\246','\4','\25','\224',
'\274','\201','\152','\177','\40','\126','\225','\263',
'\147','\54','\203','\341','\267','\200','\273','\235',
'\175','\346','\165','\51','\162','\240','\56','\171',
'\122','\115','\216','\174','\342','\122','\112','\33',
'\116','\16','\121','\13','\273','\22','\130','\264',
'\101','\72','\156','\321','\237','\334','\172','\115',
'\251','\173','\171','\52','\41','\302','\112','\240',
'\62','\133','\55','\24','\231','\202','\132','\261',
'\352','\111','\162','\345','\340','\336','\43','\60',
'\75','\371','\257','\110','\177','\152','\56','\260',
'\32','\63','\116','\315','\267','\306','\277','\234',
'\43','\10','\230','\325','\61','\171','\220','\21',
'\374','\164','\41','\127','\135','\17','\36','\370',
'\140','\106','\332','\123','\12','\203','\340','\225',
'\67','\45','\173','\303','\121','\337','\40','\330',
'\113','\375','\303','\210','\200','\24','\21','\103',
'\230','\150','\154','\54','\137','\252','\317','\203',
'\373','\51','\121','\363','\164','\336','\173','\111',
'\274','\231','\21','\247','\367','\174','\376','\270',
'\332','\267','\351','\60','\256','\162','\51','\333',
'\112','\261','\1','\26','\16','\1','\32','\354',
'\147','\341','\261','\250','\347','\243','\331','\221',
'\42','\13','\235','\120','\144','\235','\32','\273',
'\36','\271','\53','\111','\234','\315','\151','\16',
'\300','\157','\101','\313','\366','\327','\157','\52',
'\166','\5','\11','\247','\27','\140','\236','\36',
'\231','\57','\5','\145','\324','\132','\161','\73',
'\360','\220','\324','\101','\11','\102','\175','\75',
'\152','\254','\236','\77','\213','\170','\362','\4',
'\161','\156','\113','\376','\307','\277','\112','\342',
'\310','\231','\163','\133','\303','\47','\220','\202',
'\5','\224','\55','\266','\113','\365','\67','\162',
'\223','\232','\134','\30','\227','\335','\323','\265',
'\146','\152','\360','\171','\225','\112','\251','\155',
'\213','\100','\24','\122','\365','\206','\6','\367',
'\30','\3','\342','\163','\340','\24','\57','\272',
'\10','\337','\15','\116','\51','\376','\70','\133',
'\15','\63','\106','\161','\61','\12','\174','\14',
'\173','\51','\322','\327','\25','\42','\102','\327',
'\222','\173','\12','\111','\264','\323','\175','\12',
'\135','\275','\63','\44','\31','\237','\357','\176',
'\303','\352','\176','\35','\160','\343','\264','\100',
'\157','\147','\150','\226','\137','\132','\75','\374',
'\202','\117','\33','\5','\345','\227','\153','\215',
'\156','\133','\362','\23','\230','\230','\127','\72',
'\270','\153','\157','\175','\301','\346','\231','\225',
'\370','\272','\56','\256','\355','\344','\236','\316',
'\116','\326','\60','\5','\7','\127','\47','\210',
'\61','\22','\255','\361','\41','\314','\245','\331',
'\155','\254','\150','\260','\154','\127','\136','\262',
'\0','\103','\323','\337','\217','\23','\146','\344',
'\52','\337','\144','\245','\166','\211','\224','\336',
'\374','\301','\25','\21','\172','\101','\67','\106',
'\271','\351','\331','\15','\124','\212','\150','\151',
'\254','\120','\237','\141','\337','\234','\363','\122',
'\24','\353','\334','\20','\122','\143','\242','\47',
'\362','\176','\325','\137','\2','\322','\111','\226',
'\226','\203','\100','\56','\63','\310','\357','\326',
'\312','\170','\307','\146','\66','\60','\73','\325',
'\315','\225','\72','\32','\76','\143','\373','\43',
'\122','\126','\314','\46','\126','\236','\170','\324',
'\205','\46','\230','\170','\145','\303','\26','\76',
'\35','\106','\324','\70','\44','\77','\53','\317',
'\53','\177','\222','\333','\174','\252','\36','\300',
'\375','\31','\224','\271','\142','\54','\277','\214',
'\133','\200','\220','\375','\13','\373','\252','\375',
'\152','\272','\207','\142','\345','\121','\4','\221',
'\130','\262','\221','\233','\73','\113','\24','\367',
'\312','\30','\346','\146','\267','\32','\223','\36',
'\333','\32','\257','\301','\130','\330','\312','\231',
'\46','\336','\112','\337','\156','\322','\253','\5',
'\115','\174','\322','\373','\164','\245','\370','\75',
'\243','\166','\116','\111','\243','\320','\62','\317',
'\223','\250','\30','\6','\341','\61','\344','\123',
'\145','\24','\310','\77','\326','\156','\376','\215',
'\163','\160','\30','\156','\166','\301','\352','\14',
'\205','\7','\250','\121','\366','\231','\263','\322',
'\171','\65','\26','\56','\373','\75','\5','\277',
'\10','\141','\342','\6','\57','\66','\121','\227',
'\360','\106','\240','\121','\331','\227','\34','\277',
'\57','\342','\356','\263','\64','\247','\115','\244',
'\14','\146','\134','\317','\77','\310','\73','\333',
'\176','\321','\315','\45','\347','\122','\357','\304',
'\327','\305','\142','\102','\144','\14','\205','\323',
'\50','\256','\320','\53','\374','\73','\203','\54',
'\341','\20','\70','\12','\116','\375','\247','\53',
'\155','\331','\355','\226','\62','\20','\72','\132',
'\271','\77','\341','\11','\251','\21','\40','\337',
'\204','\113','\227','\65','\172','\75','\63','\123',
'\143','\362','\66','\120','\37','\337','\211','\175',
'\13','\42','\140','\133','\141','\163','\262','\110',
'\52','\205','\222','\307','\300','\106','\367','\216',
'\254','\105','\322','\374','\335','\307','\241','\357',
'\316','\312','\66','\326','\255','\306','\246','\161',
'\24','\134','\165','\341','\120','\331','\346','\364',
'\236','\122','\141','\174','\137','\153','\261','\162',
'\43','\152','\32','\110','\61','\302','\325','\231',
'\16','\4','\344','\177','\43','\330','\141','\65',
'\345','\77','\146','\104','\346','\214','\334','\71',
'\41','\150','\342','\225','\226','\335','\32','\65',
'\274','\62','\361','\56','\243','\315','\17','\336',
'\166','\65','\172','\13','\362','\131','\27','\145',
'\61','\370','\230','\310','\323','\322','\130','\355',
'\254','\346','\126','\2','\36','\7','\211','\57',
'\214','\256','\4','\27','\344','\7','\35','\177',
'\16','\155','\253','\35','\270','\12','\337','\63',
'\272','\372','\76','\223','\276','\22','\253','\215',
'\244','\102','\52','\213','\347','\7','\300','\365',
'\47','\15','\136','\325','\240','\15','\205','\374',
'\75','\234','\211','\100','\23','\106','\102','\37',
'\102','\23','\327','\222','\36','\201','\155','\226',
'\207','\365','\267','\17','\146','\17','\371','\21',
'\57','\17','\222','\167','\365','\231','\360','\233',
'\270','\53','\273','\54','\362','\51','\37','\15',
'\272','\63','\322','\330','\240','\260','\373','\26',
'\334','\117','\102','\50','\26','\352','\14','\241',
'\342','\136','\203','\345','\122','\30','\376','\64',
'\130','\243','\276','\105','\337','\274','\232','\161',
'\30','\150','\313','\76','\367','\16','\4','\215',
'\53','\63','\363','\17','\107','\330','\233','\203',
'\322','\30','\155','\255','\110','\142','\162','\110',
'\33','\152','\14','\142','\267','\354','\21','\117',
'\44','\17','\64','\7','\243','\63','\5','\137',
'\213','\153','\170','\244','\316','\234','\301','\17',
'\207','\123','\70','\175','\4','\254','\140','\310',
'\103','\274','\12','\360','\151','\320','\207','\312',
'\247','\207','\226','\242','\11','\376','\301','\302',
'\217','\31','\164','\312','\135','\304','\347','\122',
'\46','\117','\130','\305','\203','\77','\306','\130',
'\147','\253','\157','\1','\51','\6','\234','\307',
'\333','\137','\57','\4','\76','\373','\46','\65',
'\54','\347','\45','\76','\51','\40','\1','\352',
'\125','\135','\200','\147','\247','\201','\55','\336',
'\253','\67','\162','\5','\300','\304','\234','\131',
'\353','\213','\244','\37','\233','\307','\121','\73',
'\371','\371','\104','\33','\227','\105','\52','\207',
'\0','\70','\216','\43','\105','\327','\53','\201',
'\162','\242','\136','\36','\140','\136','\343','\360',
'\355','\53','\370','\16','\207','\272','\277','\314',
'\135','\110','\330','\203','\74','\125','\263','\51',
'\303','\317','\246','\241','\144','\76','\165','\372',
'\235','\171','\46','\215','\77','\225','\172','\55',
'\234','\373','\101','\23','\10','\323','\323','\325',
'\164','\275','\174','\242','\162','\354','\14','\167',
'\4','\344','\367','\322','\36','\116','\165','\277',
'\215','\373','\313','\164','\261','\353','\367','\206',
'\322','\32','\7','\230','\342','\306','\34','\114',
'\312','\301','\251','\221','\37','\46','\204','\334',
'\352','\51','\300','\312','\342','\216','\64','\26',
'\50','\46','\13','\157','\360','\15','\271','\50',
'\362','\271','\370','\340','\314','\345','\361','\47',
'\370','\315','\21','\123','\153','\346','\213','\167',
'\322','\244','\150','\102','\371','\352','\314','\252',
'\47','\114','\237','\2','\370','\66','\131','\243',
'\307','\332','\200','\313','\272','\235','\121','\162',
'\125','\33','\141','\237','\223','\111','\111','\175',
'\375','\276','\316','\356','\23','\202','\315','\117',
'\361','\35','\137','\0','\52','\124','\325','\205',
'\170','\364','\360','\313','\122','\102','\56','\125',
'\245','\251','\163','\13','\326','\111','\333','\327',
'\132','\201','\241','\370','\13','\5','\314','\101',
'\142','\212','\332','\355','\113','\16','\104','\376',
'\164','\267','\53','\210','\161','\376','\240','\46',
'\143','\114','\115','\32','\166','\310','\201','\257',
'\272','\324','\45','\74','\11','\306','\101','\170',
'\177','\71','\237','\311','\317','\57','\263','\177',
'\105','\46','\275','\274','\321','\22','\126','\251',
'\306','\314','\323','\316','\372','\6','\0','\350',
'\13','\241','\116','\201','\75','\25','\113','\364',
'\161','\4','\311','\263','\54','\216','\317','\270',
'\205','\177','\106','\247','\315','\172','\256','\320',
'\310','\203','\354','\207','\210','\53','\55','\212',
'\306','\232','\376','\34','\2','\45','\257','\213',
'\240','\75','\166','\52','\104','\332','\371','\364',
'\3','\120','\340','\163','\174','\323','\273','\200',
'\331','\262','\201','\41','\23','\45','\104','\344',
'\143','\214','\75','\2','\111','\363','\275','\217',
'\174','\364','\305','\131','\375','\331','\25','\110',
'\250','\71','\5','\147','\30','\10','\105','\340',
'\67','\161','\335','\376','\333','\272','\314','\21',
'\300','\200','\144','\351','\334','\40','\337','\271',
'\246','\247','\156','\25','\273','\263','\73','\370',
'\173','\20','\102','\23','\111','\211','\27','\121',
'\313','\273','\235','\136','\27','\235','\256','\301',
'\116','\36','\352','\347','\36','\140','\15','\257',
'\275','\364','\44','\5','\276','\273','\310','\344',
'\245','\42','\141','\174','\243','\57','\141','\375',
'\62','\105','\376','\363','\1','\177','\253','\366',
'\350','\171','\360','\266','\364','\176','\57','\225',
'\151','\40','\260','\313','\344','\347','\356','\315',
'\211','\127','\136','\43','\234','\57','\26','\126',
'\64','\161','\366','\147','\153','\215','\200','\50',
'\212','\4','\123','\361','\325','\340','\365','\327',
'\230','\11','\35','\200','\221','\62','\347','\332',
'\21','\215','\373','\230','\23','\35','\166','\300',
'\355','\353','\142','\2','\250','\115','\237','\220',
'\207','\246','\255','\144','\125','\274','\345','\255',
'\106','\320','\46','\3','\33','\367','\337','\163',
'\203','\12','\4','\150','\344','\62','\210','\371',
'\174','\253','\103','\160','\212','\70','\361','\351',
'\302','\114','\131','\216','\26','\356','\372','\234',
'\12','\360','\223','\251','\70','\103','\70','\362',
'\157','\266','\275','\67','\302','\146','\235','\234',
'\343','\141','\13','\306','\246','\365','\262','\250',
'\306','\64','\156','\161','\276','\233','\72','\217',
'\231','\115','\251','\254','\157','\41','\133','\336',
'\307','\341','\145','\302','\221','\373','\67','\32',
'\204','\44','\37','\244','\60','\136','\141','\171',
'\126','\106','\0','\212','\323','\26','\227','\220',
'\356','\333','\321','\376','\323','\241','\144','\310',
'\333','\345','\247','\1','\304','\124','\106','\114',
'\335','\50','\160','\63','\307','\44','\367','\11',
'\136','\314','\332','\157','\115','\271','\171','\375',
'\5','\176','\32','\220','\66','\342','\322','\70',
'\106','\365','\63','\373','\225','\75','\201','\115',
'\347','\62','\257','\144','\310','\335','\26','\360',
'\326','\122','\343','\103','\152','\201','\365','\243',
'\237','\261','\315','\310','\275','\11','\312','\217',
'\100','\76','\202','\143','\133','\32','\362','\155',
'\365','\207','\212','\12','\116','\366','\10','\140',
'\335','\150','\301','\155','\44','\311','\154','\265',
'\265','\363','\170','\62','\341','\147','\132','\171',
'\74','\47','\330','\276','\216','\36','\347','\166',
'\163','\172','\200','\373','\135','\53','\67','\351',
'\372','\330','\117','\213','\46','\143','\65','\305',
'\2','\133','\212','\234','\212','\323','\204','\176',
'\24','\17','\210','\121','\231','\333','\51','\363',
'\154','\350','\36','\230','\243','\11','\177','\356',
'\361','\35','\247','\105','\333','\205','\364','\274',
'\23','\310','\65','\75','\203','\14','\115','\176',
'\16','\226','\335','\303','\207','\55','\151','\14',
'\346','\61','\342','\340','\206','\257','\317','\104',
'\107','\300','\146','\115','\117','\167','\61','\224',
'\1','\361','\163','\325','\124','\372','\212','\140',
'\2','\353','\264','\301','\201','\317','\36','\254',
'\324','\345','\335','\215','\65','\174','\213','\176',
'\306','\103','\310','\255','\266','\121','\362','\254',
'\224','\360','\16','\304','\252','\230','\354','\71',
'\133','\374','\3','\370','\172','\123','\270','\310',
'\370','\43','\203','\241','\300','\336','\240','\124',
'\134','\201','\306','\216','\117','\361','\162','\310',
'\217','\267','\275','\210','\125','\312','\363','\14',
'\354','\271','\12','\223','\273','\254','\334','\162',
'\345','\107','\346','\47','\1','\1','\71','\313',
'\140','\126','\6','\42','\103','\42','\105','\117',
'\17','\110','\355','\346','\226','\13','\352','\205',
'\75','\74','\271','\230','\334','\236','\32','\220',
'\40','\157','\207','\55','\255','\120','\202','\100',
'\247','\165','\231','\156','\63','\347','\40','\146',
'\173','\314','\330','\42','\316','\76','\130','\124',
'\46','\166','\237','\323','\122','\350','\110','\40',
'\255','\220','\207','\323','\355','\134','\56','\250',
'\313','\40','\15','\163','\73','\254','\55','\350',
'\25','\103','\370','\11','\113','\60','\110','\0',
'\301','\365','\245','\330','\142','\46','\21','\376',
'\250','\170','\140','\277','\253','\163','\70','\222',
'\267','\225','\51','\305','\361','\47','\13','\45',
'\264','\250','\55','\171','\206','\41','\16','\256',
'\347','\16','\163','\56','\257','\272','\157','\71',
'\150','\167','\112','\165','\271','\5','\334','\356',
'\215','\120','\332','\321','\251','\47','\360','\22',
'\16','\226','\145','\64','\166','\126','\216','\157',
'\20','\347','\164','\66','\270','\62','\265','\307',
'\7','\330','\351','\40','\174','\343','\64','\242',
'\232','\226','\23','\121','\252','\267','\40','\170',
'\375','\212','\270','\171','\1','\75','\203','\237',
'\157','\161','\212','\231','\207','\370','\272','\351',
'\264','\104','\275','\127','\173','\71','\360','\224',
'\227','\331','\172','\373','\124','\121','\352','\157',
'\167','\2','\252','\37','\252','\273','\15','\35',
'\43','\245','\141','\45','\147','\371','\260','\5',
'\375','\71','\326','\255','\142','\316','\214','\216',
'\22','\265','\315','\341','\222','\306','\222','\122',
'\115','\344','\70','\142','\220','\300','\71','\257',
'\105','\204','\200','\314','\325','\42','\7','\11',
'\306','\303','\56','\162','\275','\44','\334','\45',
'\264','\305','\346','\107','\53','\114','\342','\163',
'\140','\333','\243','\232','\43','\216','\374','\364',
'\161','\22','\177','\357','\20','\147','\261','\354',
'\1','\10','\372','\112','\232','\106','\156','\331',
'\203','\133','\276','\365','\234','\251','\261','\117',
'\23','\155','\115','\345','\21','\237','\127','\331',
'\302','\45','\167','\163','\211','\225','\265','\331',
'\101','\212','\210','\65','\257','\71','\301','\20',
'\310','\65','\115','\75','\174','\135','\235','\227',
'\117','\132','\170','\234','\335','\33','\67','\47',
'\306','\157','\301','\144','\226','\65','\30','\163',
'\222','\210','\250','\136','\54','\123','\260','\172',
'\36','\361','\154','\164','\300','\73','\241','\233',
'\136','\7','\260','\44','\376','\177','\301','\315',
'\367','\376','\214','\120','\14','\155','\136','\115',
'\27','\132','\320','\10','\234','\141','\364','\263',
'\372','\160','\352','\271','\33','\266','\144','\367',
'\166','\301','\327','\105','\51','\251','\24','\104',
'\160','\115','\215','\43','\265','\117','\205','\151',
'\365','\373','\210','\6','\40','\352','\66','\145',
'\217','\355','\113','\271','\126','\30','\270','\263',
'\215','\76','\302','\347','\267','\17','\307','\254',
'\143','\324','\177','\226','\101','\11','\175','\225',
'\337','\373','\156','\140','\137','\114','\77','\302',
'\63','\230','\202','\100','\360','\4','\173','\303',
'\351','\270','\243','\37','\15','\173','\353','\140',
'\33','\213','\227','\276','\213','\351','\46','\313',
'\271','\262','\67','\26','\46','\264','\67','\311',
'\123','\11','\104','\227','\1','\114','\277','\2',
'\270','\363','\375','\201','\175','\222','\157','\24',
'\102','\273','\51','\111','\45','\211','\213','\151',
'\104','\170','\373','\131','\2','\57','\3','\347',
'\376','\257','\40','\213','\374','\130','\357','\7',
'\111','\333','\224','\101','\303','\233','\100','\120',
'\155','\275','\32','\314','\261','\10','\305','\153',
'\12','\223','\167','\165','\116','\147','\21','\261',
'\251','\42','\354','\74','\316','\323','\322','\360',
'\115','\153','\57','\130','\327','\203','\242','\327',
'\155','\147','\340','\214','\337','\233','\325','\114',
'\5','\374','\235','\300','\211','\34','\63','\127',
'\215','\51','\22','\152','\303','\274','\254','\224',
'\117','\261','\352','\45','\47','\52','\225','\275',
'\216','\171','\326','\215','\334','\364','\176','\122',
'\224','\302','\51','\144','\17','\25','\103','\270',
'\33','\174','\155','\262','\134','\170','\102','\257',
'\330','\146','\264','\122','\275','\337','\315','\262',
'\150','\332','\226','\44','\14','\36','\23','\323',
'\15','\225','\374','\21','\140','\64','\122','\236',
'\330','\203','\157','\323','\153','\264','\356','\332',
'\21','\373','\365','\267','\347','\137','\76','\277',
'\264','\223','\146','\375','\263','\17','\146','\173',
'\242','\350','\102','\207','\14','\214','\300','\374',
'\125','\267','\134','\5','\367','\155','\163','\247',
'\45','\3','\275','\15','\151','\4','\155','\220',
'\361','\23','\315','\330','\50','\304','\335','\332',
'\55','\267','\167','\146','\274','\201','\146','\150',
'\326','\361','\371','\331','\365','\72','\121','\43',
'\240','\123','\60','\332','\346','\240','\235','\306',
'\272','\211','\20','\245','\317','\76','\225','\353',
'\340','\152','\12','\36','\167','\71','\126','\1',
'\142','\151','\334','\215','\34','\313','\362','\52',
'\127','\103','\307','\202','\364','\236','\105','\100',
'\325','\15','\133','\235','\12','\55','\22','\324',
'\217','\220','\373','\147','\35','\204','\60','\165',
'\372','\353','\23','\144','\253','\375','\201','\171',
'\263','\3','\111','\61','\64','\150','\331','\14',
'\331','\171','\41','\17','\33','\320','\320','\353',
'\155','\360','\172','\267','\40','\364','\250','\334',
'\217','\304','\70','\302','\152','\333','\146','\311',
'\62','\340','\174','\372','\21','\101','\4','\242',
'\42','\13','\354','\144','\144','\72','\314','\36',
'\133','\115','\53','\347','\222','\244','\23','\351',
'\320','\147','\264','\71','\126','\227','\7','\67',
'\126','\26','\41','\266','\153','\225','\36','\117',
'\224','\54','\213','\70','\141','\3','\311','\245',
'\363','\317','\16','\61','\204','\315','\301','\371',
'\212','\145','\330','\132','\324','\155','\203','\322',
'\165','\277','\327','\53','\330','\311','\124','\265',
'\131','\133','\330','\141','\40','\104','\47','\325',
'\334','\116','\137','\373','\302','\230','\262','\202',
'\240','\140','\335','\77','\273','\124','\371','\336',
'\322','\60','\136','\376','\342','\204','\176','\35',
'\362','\277','\320','\210','\13','\51','\107','\7',
'\304','\167','\173','\224','\260','\57','\327','\226',
'\354','\247','\314','\330','\346','\316','\67','\73',
'\150','\31','\217','\227','\71','\375','\216','\324',
'\210','\346','\177','\55','\346','\202','\252','\12',
'\163','\66','\162','\342','\154','\53','\246','\226',
'\306','\105','\156','\112','\112','\177','\216','\35',
'\66','\350','\10','\334','\101','\303','\134','\272',
'\4','\105','\42','\315','\302','\254','\356','\126',
'\337','\207','\151','\233','\204','\266','\253','\113',
'\341','\246','\216','\134','\273','\311','\170','\314',
'\342','\72','\13','\30','\277','\374','\32','\34',
'\125','\351','\170','\317','\245','\357','\351','\173',
'\142','\307','\273','\354','\302','\304','\244','\151',
'\6','\315','\320','\50','\3','\137','\240','\346',
'\357','\67','\125','\361','\6','\376','\153','\143',
'\45','\241','\337','\226','\362','\346','\246','\170',
'\237','\300','\213','\223','\305','\30','\54','\372',
'\112','\323','\302','\315','\365','\33','\257','\360',
'\304','\325','\357','\17','\373','\2','\324','\350',
'\47','\371','\141','\57','\375','\213','\74','\12',
'\202','\373','\145','\65','\151','\204','\66','\375',
'\161','\176','\1','\366','\2','\151','\216','\73',
'\171','\344','\64','\303','\350','\32','\74','\225',
'\75','\217','\56','\217','\256','\247','\303','\113',
'\142','\103','\313','\375','\30','\106','\142','\211',
'\134','\12','\331','\70','\31','\271','\254','\221',
'\273','\321','\0','\155','\333','\224','\105','\67',
'\230','\41','\151','\152','\153','\357','\233','\346',
'\42','\366','\35','\311','\120','\323','\215','\346',
'\51','\142','\302','\173','\76','\246','\117','\6',
'\365','\15','\46','\104','\73','\323','\257','\173',
'\41','\324','\340','\116','\365','\215','\373','\164',
'\340','\166','\164','\235','\173','\6','\102','\264',
'\332','\362','\106','\227','\130','\161','\206','\35',
'\376','\224','\210','\35','\370','\15','\351','\254',
'\25','\245','\124','\262','\130','\206','\244','\163',
'\207','\316','\64','\245','\115','\334','\177','\330',
'\15','\221','\272','\36','\332','\241','\220','\164',
'\276','\73','\370','\103','\64','\164','\317','\155',
'\356','\245','\150','\324','\267','\144','\65','\1',
'\345','\264','\333','\72','\225','\343','\213','\212',
'\324','\115','\43','\264','\330','\165','\333','\247',
'\266','\256','\111','\203','\316','\77','\75','\205',
'\350','\176','\311','\23','\107','\232','\115','\202',
'\273','\154','\371','\45','\310','\33','\230','\27',
'\204','\112','\371','\260','\244','\220','\57','\150',
'\274','\351','\13','\243','\35','\232','\263','\142',
'\370','\157','\100','\157','\33','\157','\361','\275',
'\310','\330','\272','\50','\317','\112','\167','\115',
'\360','\137','\170','\52','\41','\242','\320','\205',
'\205','\131','\322','\326','\176','\334','\220','\374',
'\243','\41','\152','\237','\113','\115','\236','\66',
'\130','\232','\67','\171','\156','\214','\226','\66',
'\144','\260','\320','\106','\33','\326','\321','\373',
'\313','\36','\5','\172','\44','\243','\37','\221',
'\177','\40','\357','\125','\333','\352','\265','\272',
'\372','\307','\131','\332','\207','\371','\262','\210',
'\215','\263','\320','\4','\267','\353','\334','\21',
'\240','\112','\257','\364','\160','\263','\200','\235',
'\236','\157','\40','\222','\4','\234','\202','\332',
'\42','\35','\164','\26','\36','\345','\36','\272',
'\175','\272','\312','\56','\63','\153','\21','\341',
'\200','\177','\64','\110','\345','\30','\67','\37',
'\244','\73','\146','\160','\332','\314','\151','\322',
'\65','\26','\70','\210','\270','\161','\376','\276',
'\310','\57','\300','\50','\315','\147','\145','\177',
'\311','\344','\54','\223','\301','\265','\160','\116',
'\346','\136','\335','\225','\55','\254','\272','\275',
'\360','\233','\155','\13','\235','\251','\173','\1',
'\265','\117','\34','\121','\126','\201','\256','\126',
'\123','\10','\221','\105','\237','\274','\30','\166',
'\44','\233','\33','\255','\323','\125','\102','\307',
'\171','\161','\13','\173','\20','\343','\317','\163',
'\220','\43','\1','\173','\26','\262','\66','\160',
'\71','\101','\171','\177','\327','\145','\173','\51',
'\42','\373','\376','\41','\230','\226','\60','\101',
'\275','\370','\120','\332','\153','\372','\37','\342',
'\315','\120','\312','\40','\343','\261','\372','\371',
'\171','\36','\255','\261','\273','\253','\227','\67',
'\262','\22','\53','\65','\170','\324','\124','\255',
'\124','\260','\257','\364','\315','\203','\164','\76',
'\307','\132','\320','\46','\250','\317','\221','\351',
'\110','\354','\365','\143','\36','\133','\354','\17',
'\243','\236','\52','\165','\54','\233','\155','\216',
'\207','\157','\116','\275','\11','\53','\150','\216',
'\325','\300','\272','\130','\215','\74','\112','\243',
'\155','\213','\13','\3','\247','\267','\77','\304',
'\50','\100','\246','\20','\172','\111','\325','\70',
'\204','\20','\152','\120','\102','\125','\236','\244',
'\176','\151','\130','\32','\262','\64','\316','\335',
'\166','\351','\315','\34','\40','\44','\270','\117',
'\352','\76','\122','\270','\350','\23','\356','\60',
'\102','\121','\0','\162','\210','\26','\323','\217',
'\25','\40','\64','\213','\227','\32','\14','\330',
'\124','\147','\274','\301','\137','\301','\111','\123',
'\315','\31','\117','\316','\313','\210','\320','\355',
'\124','\67','\344','\47','\233','\264','\321','\234',
'\224','\146','\136','\136','\304','\365','\5','\226',
'\160','\71','\150','\32','\106','\146','\137','\123',
'\373','\310','\214','\141','\171','\37','\275','\167',
'\213','\315','\54','\71','\125','\25','\322','\275',
'\366','\203','\130','\105','\173','\341','\363','\171',
'\34','\206','\101','\152','\16','\261','\60','\334',
'\252','\27','\371','\142','\136','\274','\311','\325',
'\156','\332','\5','\176','\172','\43','\241','\365',
'\34','\341','\327','\333','\130','\255','\42','\37',
'\35','\165','\146','\27','\277','\0','\255','\264',
'\137','\120','\63','\325','\231','\257','\33','\0',
'\146','\166','\154','\236','\211','\106','\127','\374',
'\150','\312','\4','\366','\360','\57','\56','\364',
'\266','\251','\56','\207','\42','\32','\271','\155',
'\164','\215','\21','\333','\231','\133','\236','\36',
'\170','\14','\327','\141','\273','\361','\357','\50',
'\275','\273','\347','\67','\335','\222','\371','\360',
'\333','\244','\303','\275','\25','\277','\141','\107',
'\374','\140','\306','\50','\211','\144','\14','\100',
'\15','\107','\16','\270','\121','\232','\371','\305',
'\314','\211','\302','\144','\4','\371','\14','\54',
'\220','\260','\24','\160','\320','\72','\21','\70',
'\376','\217','\122','\2','\240','\124','\55','\367',
'\356','\136','\164','\171','\216','\0','\343','\257',
'\24','\23','\251','\243','\355','\46','\65','\303',
'\140','\66','\325','\207','\302','\267','\22','\25',
'\130','\257','\104','\326','\14','\371','\340','\35',
'\164','\257','\271','\137','\14','\344','\365','\313',
'\117','\334','\232','\111','\166','\131','\45','\351',
'\67','\241','\274','\310','\110','\62','\26','\320',
'\350','\303','\276','\57','\362','\375','\270','\107',
'\302','\313','\11','\243','\257','\333','\236','\166',
'\376','\63','\311','\371','\157','\337','\375','\206',
'\306','\200','\30','\334','\7','\342','\144','\54',
'\353','\354','\126','\147','\65','\32','\55','\327',
'\335','\147','\177','\174','\351','\105','\352','\260',
'\332','\74','\260','\152','\53','\62','\154','\273',
'\326','\320','\327','\212','\307','\235','\32','\351',
'\355','\240','\44','\62','\360','\130','\213','\115',
'\130','\214','\110','\4','\263','\313','\31','\12',
'\250','\106','\4','\173','\32','\305','\22','\1',
'\376','\264','\103','\3','\51','\323','\114','\217',
'\17','\361','\63','\345','\220','\37','\226','\320',
'\223','\151','\146','\256','\30','\246','\251','\325',
'\321','\43','\107','\304','\164','\106','\235','\135',
'\221','\61','\375','\12','\307','\153','\265','\207',
'\367','\335','\234','\4','\262','\162','\137','\321',
'\164','\340','\64','\124','\121','\312','\143','\240',
'\43','\21','\4','\36','\14','\337','\330','\207',
'\371','\125','\110','\213','\153','\44','\160','\300',
'\364','\205','\172','\302','\125','\335','\161','\60',
'\0','\66','\277','\126','\72','\274','\204','\12',
'\67','\250','\300','\73','\0','\221','\26','\66',
'\347','\27','\210','\357','\35','\200','\140','\103',
'\340','\374','\224','\371','\220','\246','\34','\30',
'\330','\133','\134','\317','\337','\106','\160','\302',
'\154','\61','\2','\173','\6','\356','\140','\261',
'\244','\175','\56','\132','\115','\312','\24','\346',
'\134','\143','\370','\151','\334','\75','\231','\163',
'\73','\217','\70','\204','\213','\47','\325','\66',
'\161','\13','\301','\302','\63','\224','\56','\370',
'\240','\60','\1','\46','\311','\335','\176','\144',
'\37','\314','\227','\132','\272','\111','\354','\65',
'\276','\366','\270','\103','\273','\312','\202','\252',
'\333','\374','\130','\46','\1','\66','\150','\366',
'\104','\54','\321','\201','\201','\117','\206','\6',
'\317','\101','\312','\107','\152','\12','\263','\306',
'\56','\303','\151','\14','\223','\367','\220','\250',
'\353','\323','\124','\107','\133','\56','\335','\31',
'\373','\125','\217','\163','\335','\341','\200','\367',
'\307','\364','\330','\257','\67','\211','\32','\63',
'\21','\203','\321','\270','\331','\71','\31','\52',
'\57','\174','\361','\16','\161','\243','\204','\327',
'\171','\325','\155','\272','\322','\52','\253','\372',
'\262','\344','\71','\331','\36','\126','\216','\300',
'\243','\327','\125','\115','\274','\71','\223','\44',
'\333','\271','\5','\42','\17','\126','\113','\314',
'\102','\363','\165','\134','\152','\302','\243','\117',
'\336','\371','\143','\214','\174','\337','\225','\13',
'\237','\70','\11','\116','\224','\56','\213','\254',
'\223','\171','\105','\165','\276','\350','\126','\142',
'\5','\77','\166','\221','\331','\256','\52','\230',
'\67','\35','\103','\155','\244','\145','\227','\57',
'\45','\213','\334','\230','\227','\142','\165','\336',
'\355','\66','\344','\330','\347','\363','\44','\106',
'\302','\105','\110','\16','\73','\57','\162','\160',
'\22','\174','\161','\372','\225','\51','\66','\263',
'\212','\17','\217','\345','\0','\141','\75','\276',
'\174','\61','\111','\312','\163','\371','\45','\305',
'\13','\306','\373','\250','\316','\156','\367','\161',
'\0','\315','\11','\261','\141','\23','\240','\56',
'\271','\263','\217','\35','\200','\51','\40','\373',
'\72','\201','\22','\221','\140','\3','\170','\237',
'\174','\242','\361','\343','\306','\43','\22','\134',
'\125','\333','\125','\342','\340','\23','\236','\177',
'\366','\307','\267','\341','\231','\41','\115','\152',
'\111','\130','\266','\351','\246','\324','\136','\243',
'\167','\54','\133','\176','\62','\215','\217','\71',
'\274','\307','\144','\235','\217','\375','\255','\316',
'\67','\271','\331','\165','\133','\242','\100','\330',
'\52','\304','\210','\110','\357','\204','\313','\33',
'\36','\215','\175','\102','\203','\237','\321','\47',
'\217','\375','\7','\44','\304','\251','\51','\157',
'\55','\361','\141','\136','\343','\331','\326','\134',
'\16','\314','\0','\357','\11','\357','\146','\225',
'\70','\145','\226','\152','\70','\3','\244','\307',
'\213','\347','\23','\374','\276','\146','\2','\315',
'\265','\273','\107','\0','\172','\333','\263','\210',
'\125','\30','\252','\31','\146','\315','\202','\306',
'\161','\36','\20','\307','\352','\302','\155','\136',
'\73','\244','\26','\343','\137','\372','\247','\205',
'\166','\347','\104','\357','\210','\230','\22','\160',
'\362','\167','\131','\160','\27','\340','\132','\55',
'\254','\155','\345','\331','\6','\6','\46','\60',
'\104','\137','\273','\212','\122','\33','\165','\344',
'\44','\63','\276','\365','\266','\240','\154','\104',
'\161','\242','\341','\21','\273','\146','\67','\13',
'\140','\266','\47','\324','\250','\213','\20','\221',
'\302','\21','\60','\222','\341','\161','\3','\256',
'\6','\266','\215','\6','\44','\362','\365','\240',
'\15','\57','\261','\61','\135','\232','\37','\261',
'\234','\252','\131','\277','\71','\4','\340','\164',
'\54','\10','\147','\257','\173','\171','\116','\317',
'\224','\262','\136','\212','\11','\234','\352','\255',
'\31','\174','\172','\350','\121','\14','\257','\115',
'\343','\47','\52','\61','\364','\211','\140','\65',
'\41','\332','\247','\231','\350','\130','\265','\136',
'\25','\64','\266','\164','\110','\23','\2','\116',
'\305','\61','\224','\10','\16','\127','\24','\365',
'\345','\213','\10','\341','\217','\364','\32','\66',
'\234','\51','\322','\224','\123','\327','\253','\312',
'\266','\110','\123','\111','\374','\130','\71','\354',
'\101','\235','\50','\22','\370','\153','\33','\352',
'\75','\175','\306','\124','\5','\335','\261','\111',
'\175','\156','\41','\376','\26','\251','\24','\327',
'\357','\366','\336','\372','\136','\247','\332','\372',
'\115','\212','\33','\73','\356','\74','\203','\123',
'\253','\144','\163','\123','\301','\2','\266','\335',
'\127','\133','\151','\1','\370','\220','\22','\105',
'\117','\235','\366','\275','\173','\336','\146','\351',
'\321','\141','\340','\260','\33','\257','\222','\236',
'\26','\304','\224','\24','\257','\202','\305','\252',
'\247','\101','\4','\240','\54','\372','\224','\143',
'\267','\270','\131','\75','\40','\323','\277','\14',
'\115','\255','\276','\254','\122','\63','\30','\254',
'\263','\50','\374','\130','\144','\26','\260','\353',
'\344','\76','\143','\76','\275','\221','\51','\304',
'\24','\62','\361','\156','\134','\137','\206','\151',
'\20','\154','\27','\126','\36','\256','\371','\272',
'\133','\304','\74','\355','\75','\2','\71','\361',
'\243','\46','\113','\261','\166','\140','\205','\343',
'\234','\121','\225','\372','\25','\151','\261','\367',
'\313','\67','\34','\166','\330','\256','\261','\361',
'\231','\74','\361','\14','\33','\307','\77','\60',
'\225','\240','\327','\141','\111','\34','\266','\30',
'\240','\336','\72','\246','\325','\113','\11','\151',
'\52','\376','\372','\60','\345','\160','\262','\163',
'\220','\251','\330','\25','\27','\150','\54','\204',
'\257','\266','\300','\155','\222','\146','\356','\141',
'\14','\34','\173','\247','\307','\100','\77','\46',
'\253','\110','\63','\5','\260','\325','\4','\20',
'\357','\41','\237','\333','\210','\120','\51','\76',
'\321','\35','\53','\55','\203','\310','\52','\4',
'\123','\262','\373','\340','\277','\333','\157','\177',
'\271','\266','\164','\72','\336','\332','\133','\160',
'\343','\242','\101','\145','\342','\45','\67','\10',
'\310','\310','\253','\345','\200','\142','\266','\53',
'\215','\174','\217','\166','\366','\50','\264','\176',
'\75','\374','\231','\43','\155','\3','\7','\311',
'\124','\142','\201','\234','\107','\306','\115','\47',
'\44','\177','\120','\110','\344','\106','\311','\73',
'\170','\167','\237','\273','\155','\127','\151','\32',
'\116','\336','\343','\105','\335','\157','\272','\240',
'\243','\53','\224','\67','\175','\157','\140','\122',
'\3','\27','\325','\72','\271','\134','\25','\42',
'\255','\272','\76','\47','\45','\226','\16','\262',
'\275','\125','\7','\20','\111','\334','\305','\321',
'\174','\71','\162','\251','\214','\155','\46','\126',
'\311','\344','\175','\106','\62','\321','\13','\232',
'\40','\216','\76','\314','\46','\335','\331','\321',
'\255','\133','\303','\323','\342','\362','\54','\254',
'\115','\360','\274','\163','\52','\102','\27','\23',
'\231','\257','\350','\127','\160','\60','\327','\353',
'\246','\252','\113','\164','\370','\322','\244','\317',
'\352','\200','\353','\30','\103','\342','\77','\265',
'\112','\315','\135','\331','\140','\332','\301','\64',
'\346','\222','\200','\260','\343','\7','\54','\335',
'\160','\103','\346','\311','\257','\110','\155','\226',
'\245','\356','\26','\117','\340','\73','\120','\145',
'\307','\125','\326','\277','\114','\320','\276','\310',
'\72','\10','\215','\57','\43','\133','\142','\124',
'\332','\373','\2','\362','\1','\3','\145','\344',
'\5','\7','\3','\240','\215','\360','\164','\36',
'\127','\170','\43','\264','\343','\320','\173','\33',
'\147','\351','\307','\273','\107','\151','\60','\300',
'\351','\17','\113','\77','\254','\136','\357','\251',
'\375','\136','\233','\227','\353','\355','\114','\36',
'\237','\242','\146','\146','\24','\367','\43','\31',
'\22','\362','\247','\300','\125','\7','\363','\132',
'\10','\244','\5','\365','\5','\206','\320','\70',
'\327','\200','\115','\174','\273','\203','\367','\116',
'\327','\323','\24','\205','\24','\51','\314','\216',
'\6','\246','\72','\265','\131','\151','\327','\305',
'\205','\33','\111','\156','\61','\310','\46','\121',
'\311','\222','\122','\215','\342','\371','\113','\202',
'\354','\1','\205','\236','\255','\211','\201','\102',
'\36','\351','\253','\225','\124','\10','\21','\205',
'\265','\271','\52','\55','\237','\375','\2','\124',
'\346','\22','\376','\164','\172','\103','\14','\303',
'\207','\120','\11','\302','\25','\317','\263','\0',
'\144','\216','\104','\4','\50','\50','\116','\165',
'\252','\77','\216','\156','\176','\10','\166','\50',
'\170','\65','\224','\351','\163','\261','\216','\36',
'\204','\35','\120','\120','\336','\350','\67','\100',
'\17','\351','\236','\61','\176','\371','\335','\275',
'\230','\311','\337','\217','\316','\131','\65','\150',
'\313','\353','\161','\304','\357','\311','\307','\44',
'\177','\132','\144','\366','\250','\124','\227','\353',
'\245','\304','\216','\356','\70','\34','\362','\20',
'\151','\247','\167','\256','\301','\303','\125','\140',
'\234','\365','\352','\347','\77','\341','\250','\42',
'\103','\330','\242','\217','\300','\300','\312','\210',
'\140','\67','\27','\40','\227','\160','\22','\152',
'\117','\173','\331','\7','\211','\344','\172','\135',
'\26','\213','\124','\31','\222','\325','\41','\35',
'\144','\126','\326','\30','\204','\201','\374','\375',
'\130','\373','\324','\271','\71','\160','\344','\77',
'\337','\137','\223','\46','\117','\321','\0','\46',
'\164','\113','\267','\247','\74','\110','\213','\141',
'\335','\33','\113','\334','\67','\241','\205','\20',
'\232','\172','\163','\143','\225','\250','\134','\356',
'\26','\10','\140','\351','\231','\327','\221','\2',
'\170','\273','\314','\247','\112','\133','\170','\110',
'\140','\132','\106','\325','\331','\74','\342','\357',
'\110','\212','\304','\144','\70','\50','\224','\162',
'\322','\373','\332','\224','\375','\57','\350','\225',
'\271','\360','\225','\140','\306','\111','\55','\252',
'\61','\347','\235','\61','\160','\63','\132','\242',
'\65','\143','\17','\5','\331','\141','\30','\341',
'\1','\333','\21','\350','\223','\56','\33','\265',
'\50','\136','\67','\250','\312','\256','\225','\245',
'\125','\140','\125','\62','\115','\63','\132','\357',
'\372','\376','\202','\24','\233','\271','\172','\4',
'\307','\67','\51','\36','\175','\175','\24','\213',
'\222','\75','\333','\75','\344','\46','\110','\104',
'\150','\20','\201','\304','\304','\135','\147','\374',
'\24','\330','\276','\164','\260','\234','\240','\234',
'\313','\371','\314','\215','\310','\156','\135','\20',
'\367','\104','\260','\210','\156','\374','\210','\374',
'\352','\251','\125','\47','\11','\334','\142','\52',
'\153','\57','\121','\357','\264','\244','\55','\201',
'\104','\226','\214','\134','\267','\113','\273','\152',
'\45','\327','\157','\235','\124','\211','\232','\350',
'\155','\217','\43','\300','\347','\372','\156','\111',
'\264','\65','\0','\153','\266','\21','\124','\214',
'\37','\327','\30','\317','\247','\375','\142','\1',
'\73','\304','\3','\110','\364','\146','\272','\41',
'\172','\33','\356','\110','\355','\34','\55','\267',
'\166','\263','\103','\346','\242','\274','\10','\43',
'\233','\251','\231','\14','\224','\63','\55','\113',
'\25','\10','\320','\236','\14','\122','\34','\374',
'\275','\113','\50','\316','\156','\226','\204','\372',
'\73','\156','\66','\233','\151','\57','\322','\356',
'\257','\372','\166','\125','\131','\150','\5','\174',
'\134','\243','\40','\37','\254','\225','\62','\173',
'\167','\324','\260','\6','\123','\327','\117','\253',
'\370','\236','\224','\76','\330','\304','\106','\227',
'\252','\55','\176','\354','\245','\205','\311','\102',
'\322','\122','\36','\150','\103','\253','\253','\216',
'\164','\226','\7','\52','\126','\16','\250','\30',
'\117','\111','\326','\153','\302','\176','\261','\31',
'\317','\314','\1','\104','\122','\205','\316','\347',
'\140','\31','\376','\50','\201','\101','\3','\340',
'\117','\113','\147','\154','\2','\306','\344','\221',
'\135','\260','\353','\140','\63','\103','\162','\351',
'\361','\4','\150','\32','\52','\151','\213','\222',
'\107','\272','\301','\351','\13','\341','\62','\275',
'\352','\45','\177','\141','\113','\156','\245','\344',
'\332','\373','\332','\70','\133','\145','\134','\314',
'\321','\103','\145','\227','\65','\40','\255','\220',
'\137','\102','\354','\240','\172','\166','\352','\170',
'\325','\356','\205','\371','\347','\15','\215','\45',
'\61','\45','\263','\222','\64','\152','\344','\265',
'\133','\177','\373','\152','\146','\273','\322','\162',
'\116','\217','\333','\120','\205','\230','\126','\265',
'\312','\264','\203','\307','\60','\360','\346','\33',
'\347','\345','\232','\311','\270','\360','\275','\353',
'\362','\135','\303','\51','\360','\121','\41','\40',
'\44','\317','\166','\37','\333','\234','\75','\26',
'\75','\315','\366','\151','\261','\362','\367','\156',
'\213','\236','\246','\75','\312','\200','\121','\56',
'\152','\14','\331','\145','\273','\270','\170','\264',
'\167','\264','\317','\305','\77','\363','\43','\170',
'\131','\372','\43','\173','\46','\75','\232','\304',
'\361','\267','\117','\212','\225','\147','\247','\157',
'\127','\300','\116','\25','\135','\110','\361','\303',
'\164','\373','\106','\65','\313','\271','\323','\306',
'\254','\355','\354','\77','\351','\350','\206','\371',
'\263','\30','\26','\245','\133','\57','\301','\262',
'\100','\323','\177','\11','\47','\150','\244','\221',
'\116','\64','\215','\345','\122','\260','\75','\366',
'\4','\66','\163','\356','\14','\167','\254','\244',
'\200','\172','\225','\123','\41','\211','\37','\341',
'\255','\21','\246','\345','\147','\42','\210','\264',
'\351','\266','\313','\372','\334','\16','\115','\16',
'\227','\233','\20','\7','\15','\70','\222','\136',
'\46','\155','\142','\101','\1','\201','\11','\230',
'\112','\326','\32','\127','\323','\57','\251','\364',
'\236','\327','\336','\4','\151','\164','\21','\41',
'\124','\47','\357','\337','\207','\234','\327','\24',
'\306','\316','\313','\20','\116','\131','\13','\363',
'\27','\362','\34','\330','\243','\327','\374','\165',
'\107','\347','\26','\55','\257','\364','\205','\21',
'\246','\223','\165','\360','\63','\257','\215','\125',
'\277','\46','\72','\372','\213','\321','\367','\51',
'\220','\365','\354','\74','\303','\33','\233','\102',
'\200','\134','\152','\226','\151','\50','\206','\160',
'\16','\244','\324','\123','\315','\71','\373','\355',
'\142','\46','\352','\147','\23','\340','\330','\4',
'\147','\271','\302','\30','\213','\134','\212','\65',
'\204','\317','\33','\337','\341','\124','\130','\321',
'\302','\24','\17','\43','\275','\26','\117','\213',
'\216','\63','\170','\253','\216','\77','\267','\112',
'\327','\231','\173','\210','\106','\61','\27','\74',
'\341','\277','\104','\251','\172','\345','\235','\325',
'\226','\267','\355','\147','\121','\231','\113','\34',
'\340','\50','\234','\3','\26','\130','\35','\107',
'\355','\133','\233','\70','\330','\162','\63','\353',
'\174','\126','\65','\126','\303','\264','\100','\161',
'\121','\120','\16','\213','\50','\137','\54','\220',
'\253','\40','\354','\16','\127','\244','\11','\364',
'\102','\354','\163','\44','\42','\347','\225','\45',
'\145','\235','\130','\305','\160','\353','\356','\117',
'\174','\174','\25','\7','\0','\170','\76','\367',
'\30','\255','\303','\264','\120','\113','\126','\127',
'\360','\227','\374','\375','\104','\123','\201','\2',
'\323','\221','\166','\41','\136','\263','\122','\262',
'\7','\76','\112','\242','\124','\275','\236','\352',
'\373','\333','\46','\43','\243','\101','\242','\143',
'\310','\305','\116','\117','\214','\73','\374','\310',
'\164','\326','\274','\135','\44','\200','\131','\55',
'\261','\347','\215','\54','\246','\24','\263','\214',
'\315','\66','\256','\363','\305','\106','\4','\12',
'\34','\237','\143','\1','\216','\320','\130','\331',
'\213','\141','\25','\47','\26','\211','\66','\156',
'\15','\357','\164','\256','\172','\134','\137','\276',
'\251','\256','\352','\213','\253','\53','\157','\232',
'\112','\132','\107','\373','\151','\317','\365','\102',
'\331','\235','\266','\367','\230','\127','\55','\17',
'\214','\376','\11','\72','\231','\327','\232','\201',
'\124','\11','\270','\222','\147','\210','\343','\110',
'\207','\167','\160','\36','\357','\335','\101','\374',
'\142','\165','\244','\100','\301','\206','\43','\47',
'\132','\341','\322','\31','\303','\346','\265','\1',
'\322','\317','\213','\113','\336','\27','\1','\333',
'\72','\101','\122','\130','\225','\245','\352','\145',
'\244','\47','\147','\200','\23','\263','\312','\123',
'\65','\347','\7','\33','\250','\313','\107','\252',
'\74','\56','\312','\227','\110','\232','\22','\374',
'\317','\175','\330','\100','\162','\157','\133','\6',
'\253','\351','\22','\250','\75','\351','\270','\50',
'\223','\134','\177','\107','\315','\146','\311','\75',
'\160','\323','\176','\124','\63','\173','\11','\235',
'\172','\174','\156','\152','\166','\340','\231','\15',
'\124','\251','\232','\141','\374','\340','\337','\376',
'\274','\366','\30','\363','\147','\225','\362','\61',
'\3','\133','\321','\10','\370','\172','\153','\236',
'\225','\147','\15','\332','\353','\327','\107','\224',
'\236','\210','\56','\331','\375','\63','\111','\77',
'\320','\43','\202','\53','\201','\216','\61','\332',
'\346','\7','\135','\232','\363','\301','\304','\277',
'\13','\21','\63','\42','\341','\253','\264','\175',
'\45','\364','\160','\74','\23','\144','\376','\50',
'\163','\311','\251','\135','\205','\251','\331','\324',
'\134','\372','\246','\310','\364','\232','\336','\40',
'\142','\70','\151','\301','\40','\343','\77','\247',
'\256','\267','\200','\44','\10','\206','\1','\21',
'\252','\141','\362','\53','\313','\51','\337','\351',
'\255','\371','\270','\40','\213','\317','\76','\175',
'\105','\270','\273','\356','\134','\167','\46','\142',
'\37','\110','\151','\132','\335','\257','\364','\161',
'\115','\254','\17','\324','\17','\17','\341','\344',
'\362','\257','\5','\60','\207','\306','\201','\114',
'\227','\67','\271','\255','\250','\120','\34','\70',
'\23','\106','\33','\366','\32','\317','\10','\307',
'\7','\10','\32','\133','\107','\167','\372','\370',
'\40','\204','\65','\255','\316','\325','\105','\346',
'\64','\43','\76','\346','\224','\44','\151','\204',
'\244','\6','\271','\262','\306','\151','\350','\370',
'\217','\200','\371','\331','\366','\253','\6','\260',
'\2','\226','\156','\300','\273','\366','\316','\216',
'\240','\24','\74','\147','\11','\235','\336','\15',
'\170','\231','\140','\155','\200','\117','\362','\157',
'\201','\236','\76','\15','\31','\116','\27','\245',
'\201','\52','\251','\55','\33','\312','\245','\226',
'\215','\37','\121','\131','\350','\230','\6','\37',
'\375','\133','\232','\350','\221','\266','\230','\133',
'\25','\104','\27','\215','\332','\4','\350','\12',
'\206','\35','\25','\256','\264','\105','\7','\130',
'\232','\5','\322','\226','\275','\73','\325','\46',
'\211','\371','\365','\213','\163','\23','\274','\166',
'\243','\214','\30','\153','\25','\140','\217','\64',
'\204','\114','\206','\121','\332','\312','\111','\327',
'\321','\240','\362','\5','\103','\205','\236','\235',
'\240','\212','\164','\156','\355','\207','\171','\121',
'\227','\307','\50','\142','\151','\74','\17','\302',
'\63','\61','\243','\270','\366','\120','\206','\353',
'\130','\153','\27','\4','\44','\64','\374','\121',
'\363','\117','\102','\261','\365','\123','\45','\300',
'\31','\124','\255','\41','\322','\367','\120','\345',
'\107','\206','\355','\77','\140','\152','\240','\116',
'\225','\45','\53','\73','\221','\46','\53','\326',
'\256','\323','\263','\110','\150','\26','\161','\115',
'\351','\34','\312','\266','\357','\34','\112','\11',
'\324','\361','\327','\350','\241','\325','\46','\35',
'\237','\156','\374','\255','\351','\16','\117','\312',
'\152','\221','\126','\33','\10','\64','\232','\76',
'\261','\373','\147','\167','\307','\311','\50','\170',
'\204','\274','\75','\343','\74','\34','\206','\343',
'\151','\42','\360','\114','\16','\74','\261','\252',
'\152','\316','\117','\170','\123','\200','\337','\134',
'\154','\32','\252','\275','\230','\11','\225','\273',
'\5','\332','\273','\234','\0','\22','\30','\276',
'\47','\207','\313','\216','\340','\32','\135','\124',
'\216','\114','\314','\12','\60','\37','\341','\307',
'\267','\142','\47','\236','\34','\164','\312','\251',
'\372','\151','\177','\176','\133','\227','\115','\25',
'\77','\2','\253','\110','\117','\277','\172','\243',
'\142','\107','\225','\371','\353','\41','\360','\132',
'\337','\122','\35','\200','\266','\200','\161','\21',
'\107','\311','\342','\227','\312','\234','\330','\351',
'\351','\204','\163','\66','\63','\277','\326','\135',
'\202','\324','\201','\233','\274','\112','\70','\350',
'\41','\265','\233','\302','\314','\257','\73','\124',
'\27','\46','\323','\30','\214','\272','\314','\4',
'\303','\300','\170','\117','\144','\246','\41','\365',
'\61','\146','\344','\363','\262','\301','\325','\362',
'\261','\274','\302','\261','\20','\36','\34','\344',
'\277','\262','\274','\342','\133','\57','\73','\213',
'\243','\32','\266','\310','\226','\166','\23','\56',
'\4','\335','\376','\33','\236','\35','\70','\62',
'\145','\41','\132','\251','\136','\234','\307','\70',
'\35','\135','\230','\232','\17','\251','\342','\2',
'\142','\6','\363','\227','\42','\267','\272','\177',
'\126','\241','\126','\146','\276','\177','\34','\207',
'\225','\137','\111','\277','\40','\221','\250','\240',
'\267','\257','\323','\376','\124','\132','\163','\204',
'\206','\134','\303','\70','\135','\267','\25','\334',
'\212','\372','\137','\236','\174','\66','\216','\173',
'\244','\350','\305','\20','\124','\210','\216','\334',
'\176','\226','\322','\350','\132','\2','\151','\171',
'\26','\132','\6','\40','\2','\345','\367','\31',
'\271','\140','\371','\17','\74','\40','\21','\331',
'\60','\75','\53','\153','\0','\161','\32','\15',
'\106','\150','\61','\75','\265','\57','\343','\161',
'\100','\376','\243','\154','\371','\156','\151','\312',
'\267','\210','\343','\350','\210','\35','\302','\161',
'\120','\151','\266','\35','\113','\71','\224','\244',
'\32','\7','\173','\224','\352','\211','\257','\356',
'\250','\160','\53','\122','\227','\363','\27','\276',
'\353','\301','\263','\341','\45','\337','\170','\240',
'\116','\52','\300','\40','\103','\132','\301','\121',
'\125','\63','\101','\246','\326','\165','\262','\367',
'\234','\42','\64','\245','\366','\341','\367','\257',
'\32','\214','\133','\52','\365','\71','\342','\307',
'\260','\110','\122','\104','\261','\71','\73','\257',
'\323','\10','\35','\360','\373','\65','\305','\73',
'\246','\47','\205','\371','\351','\76','\274','\206',
'\170','\170','\315','\376','\250','\207','\25','\245',
'\136','\70','\210','\344','\170','\175','\61','\140',
'\172','\15','\37','\374','\366','\271','\361','\223',
'\23','\342','\134','\365','\220','\133','\261','\13',
'\350','\31','\272','\314','\300','\274','\175','\162',
'\215','\20','\320','\376','\75','\14','\110','\23',
'\23','\241','\235','\363','\255','\141','\323','\154',
'\216','\263','\104','\333','\306','\130','\47','\353',
'\252','\170','\23','\171','\100','\44','\27','\146',
'\370','\175','\136','\3','\13','\220','\276','\245',
'\124','\247','\261','\264','\224','\131','\342','\374',
'\156','\102','\270','\106','\174','\130','\132','\47',
'\26','\123','\237','\72','\1','\267','\124','\154',
'\115','\364','\102','\370','\314','\172','\211','\220',
'\105','\1','\42','\177','\332','\315','\220','\202',
'\141','\276','\330','\5','\43','\136','\22','\213',
'\60','\151','\277','\22','\302','\103','\34','\302',
'\142','\257','\221','\24','\175','\24','\153','\352',
'\340','\370','\155','\166','\64','\372','\25','\211',
'\43','\66','\171','\355','\364','\20','\117','\211',
'\221','\221','\146','\267','\373','\164','\45','\131',
'\40','\51','\21','\143','\230','\224','\155','\326',
'\233','\242','\213','\65','\151','\1','\74','\245',
'\326','\56','\252','\12','\107','\247','\235','\11',
'\216','\150','\2','\32','\172','\326','\175','\11',
'\135','\312','\352','\376','\221','\23','\115','\174',
'\136','\113','\102','\351','\310','\247','\352','\207',
'\3','\375','\264','\341','\370','\115','\332','\27',
'\313','\106','\170','\231','\3','\131','\210','\36',
'\103','\77','\361','\177','\70','\223','\271','\116',
'\52','\254','\125','\144','\255','\117','\56','\167',
'\233','\273','\241','\340','\322','\44','\200','\76',
'\206','\252','\14','\22','\332','\271','\302','\17',
'\40','\225','\224','\143','\370','\233','\131','\213',
'\141','\370','\100','\27','\172','\265','\225','\214',
'\277','\314','\214','\360','\175','\316','\133','\63',
'\37','\134','\352','\324','\232','\36','\364','\236',
'\112','\133','\44','\226','\324','\167','\340','\16',
'\372','\323','\124','\341','\245','\260','\342','\123',
'\213','\257','\47','\6','\36','\244','\32','\201',
'\112','\70','\76','\142','\306','\340','\251','\122',
'\374','\153','\302','\230','\20','\344','\251','\153',
'\101','\334','\311','\167','\257','\265','\30','\235',
'\21','\261','\225','\67','\127','\23','\364','\366',
'\136','\270','\322','\74','\147','\373','\144','\217',
'\222','\167','\30','\56','\1','\54','\72','\355',
'\235','\164','\157','\345','\115','\6','\213','\122',
'\331','\205','\217','\211','\171','\255','\355','\155',
'\330','\113','\177','\341','\342','\174','\61','\123',
'\76','\124','\7','\310','\150','\22','\21','\276',
'\327','\20','\247','\314','\231','\106','\11','\357',
'\337','\206','\273','\205','\30','\347','\135','\77',
'\70','\302','\46','\67','\302','\50','\333','\336',
'\234','\137','\76','\177','\154','\105','\111','\350',
'\245','\10','\315','\363','\224','\5','\17','\230',
'\23','\274','\272','\252','\270','\11','\166','\137',
'\267','\246','\212','\350','\244','\313','\164','\355',
'\151','\236','\375','\245','\261','\3','\337','\106',
'\162','\375','\311','\155','\56','\124','\310','\327',
'\12','\176','\254','\331','\202','\20','\151','\255',
'\364','\236','\201','\305','\222','\105','\10','\136',
'\225','\316','\52','\47','\326','\272','\317','\72',
'\261','\74','\375','\336','\205','\145','\364','\11',
'\260','\352','\146','\336','\70','\56','\344','\13',
'\11','\56','\173','\212','\53','\45','\165','\70',
'\114','\215','\156','\116','\315','\142','\342','\0',
'\314','\300','\162','\243','\327','\72','\307','\351',
'\224','\122','\112','\10','\32','\342','\347','\134',
'\202','\234','\267','\207','\354','\155','\127','\260',
'\156','\74','\153','\341','\215','\271','\101','\306',
'\126','\23','\223','\57','\24','\2','\42','\253',
'\252','\336','\321','\243','\70','\10','\337','\232',
'\36','\62','\342','\352','\324','\50','\276','\6',
'\152','\137','\17','\247','\46','\305','\307','\350',
'\247','\153','\66','\165','\252','\337','\174','\347',
'\175','\205','\244','\1','\230','\57','\361','\150',
'\320','\14','\322','\344','\223','\172','\376','\154',
'\153','\50','\370','\235','\126','\102','\57','\302',
'\140','\114','\375','\66','\157','\260','\250','\265',
'\140','\326','\27','\356','\223','\104','\7','\374',
'\362','\345','\256','\207','\156','\205','\270','\343',
'\12','\350','\167','\163','\157','\312','\77','\307',
'\44','\314','\350','\250','\233','\254','\141','\323',
'\31','\353','\65','\14','\13','\171','\263','\207',
'\36','\17','\46','\341','\321','\143','\202','\230',
'\47','\60','\114','\306','\25','\372','\124','\214',
'\275','\53','\270','\206','\70','\214','\111','\153',
'\307','\236','\140','\214','\262','\124','\141','\160',
'\300','\106','\46','\250','\265','\104','\162','\227',
'\17','\275','\356','\24','\164','\323','\211','\255',
'\44','\226','\252','\176','\305','\136','\45','\254',
'\115','\337','\273','\51','\255','\76','\45','\217',
'\40','\216','\326','\1','\205','\177','\2','\20',
'\6','\42','\332','\271','\306','\352','\105','\32',
'\121','\177','\317','\211','\246','\42','\112','\61',
'\45','\227','\320','\264','\275','\362','\224','\376',
'\266','\313','\70','\272','\15','\236','\147','\11',
'\105','\53','\255','\74','\266','\3','\141','\34',
'\272','\43','\127','\370','\177','\62','\372','\41',
'\326','\12','\307','\312','\162','\143','\247','\373',
'\334','\257','\13','\211','\63','\101','\267','\64',
'\304','\254','\61','\255','\242','\30','\65','\4',
'\277','\305','\353','\362','\177','\60','\127','\272',
'\356','\210','\256','\202','\152','\45','\121','\20',
'\145','\1','\12','\11','\140','\111','\346','\266',
'\142','\220','\14','\214','\262','\312','\310','\34',
'\33','\77','\33','\357','\367','\201','\252','\313',
'\112','\11','\177','\52','\340','\155','\115','\16',
'\111','\237','\241','\25','\145','\152','\72','\243',
'\252','\254','\326','\217','\67','\325','\17','\43',
'\75','\244','\3','\100','\103','\342','\302','\127',
'\267','\151','\146','\112','\343','\173','\173','\276',
'\345','\136','\264','\43','\222','\361','\14','\366',
'\1','\165','\354','\267','\75','\307','\10','\361',
'\121','\25','\302','\226','\211','\320','\221','\211',
'\271','\102','\137','\101','\327','\110','\226','\262',
'\253','\140','\241','\177','\230','\107','\32','\113',
'\122','\355','\174','\254','\131','\361','\304','\55',
'\257','\43','\351','\241','\243','\63','\47','\62',
'\262','\124','\200','\22','\131','\304','\334','\265',
'\42','\376','\72','\217','\155','\26','\55','\44',
'\144','\237','\253','\346','\54','\136','\142','\206',
'\217','\13','\300','\304','\355','\321','\325','\364',
'\75','\242','\104','\10','\330','\4','\36','\56',
'\210','\334','\0','\162','\304','\146','\311','\50',
'\223','\336','\340','\335','\104','\124','\134','\274',
'\26','\334','\314','\120','\352','\327','\227','\363',
'\165','\135','\217','\372','\120','\305','\131','\73',
'\13','\314','\5','\116','\31','\15','\52','\321',
'\364','\342','\367','\234','\346','\374','\135','\111',
'\54','\346','\170','\335','\263','\127','\323','\114',
'\320','\62','\154','\166','\114','\274','\375','\301',
'\350','\374','\345','\246','\155','\136','\137','\313',
'\257','\345','\327','\15','\44','\77','\37','\51',
'\213','\134','\216','\330','\171','\234','\75','\111',
'\343','\10','\147','\162','\236','\312','\162','\50',
'\62','\21','\304','\331','\330','\35','\51','\3',
'\132','\33','\116','\331','\335','\67','\16','\207',
'\116','\144','\16','\331','\35','\62','\116','\300',
'\311','\323','\341','\223','\205','\101','\242','\254',
'\212','\273','\275','\111','\141','\347','\32','\4',
'\15','\52','\311','\236','\176','\12','\371','\42',
'\75','\223','\217','\13','\104','\60','\205','\205',
'\206','\10','\33','\216','\273','\101','\350','\126',
'\16','\44','\17','\76','\270','\4','\214','\53',
'\47','\172','\111','\22','\355','\36','\72','\104',
'\115','\330','\167','\366','\205','\140','\303','\63',
'\255','\230','\71','\42','\57','\25','\166','\340',
'\3','\117','\125','\153','\34','\332','\347','\367',
'\106','\266','\222','\115','\251','\44','\301','\122',
'\131','\202','\124','\230','\22','\322','\337','\54',
'\41','\277','\350','\114','\347','\261','\103','\56',
'\64','\222','\176','\237','\163','\25','\201','\371',
'\177','\364','\241','\354','\116','\357','\122','\201',
'\166','\171','\71','\341','\55','\304','\235','\323',
'\33','\302','\353','\173','\147','\321','\165','\114',
'\67','\235','\253','\274','\325','\320','\0','\327',
'\13','\332','\136','\57','\244','\33','\215','\114',
'\244','\45','\137','\153','\111','\161','\267','\347',
'\133','\340','\363','\265','\77','\260','\370','\246',
'\350','\306','\215','\36','\113','\66','\362','\52',
'\324','\24','\201','\301','\172','\53','\34','\63',
'\147','\13','\206','\260','\221','\304','\122','\217',
'\264','\343','\57','\266','\204','\310','\343','\340',
'\6','\112','\254','\341','\240','\142','\57','\160',
'\275','\124','\343','\227','\337','\361','\224','\312',
'\54','\17','\206','\207','\53','\60','\302','\10',
'\277','\103','\160','\150','\225','\220','\371','\143',
'\125','\55','\334','\341','\113','\51','\235','\233',
'\336','\166','\52','\11','\230','\41','\314','\253',
'\44','\353','\105','\33','\62','\123','\345','\241',
'\363','\176','\5','\275','\26','\6','\322','\42',
'\164','\154','\155','\325','\131','\162','\375','\177',
'\352','\231','\101','\17','\374','\156','\41','\36',
'\116','\360','\143','\326','\171','\363','\220','\63',
'\270','\157','\164','\165','\317','\52','\234','\136',
'\213','\220','\306','\210','\44','\135','\211','\5',
'\235','\123','\267','\133','\330','\361','\21','\277',
'\305','\224','\111','\240','\65','\247','\204','\26',
'\340','\334','\16','\317','\116','\240','\365','\345',
'\373','\355','\127','\241','\224','\57','\14','\215',
'\53','\315','\214','\327','\276','\347','\256','\241',
'\215','\226','\146','\0','\333','\373','\150','\245',
'\223','\330','\60','\202','\307','\132','\370','\17',
'\50','\350','\315','\1','\63','\51','\21','\44',
'\6','\261','\71','\64','\313','\207','\254','\73',
'\110','\64','\236','\313','\327','\232','\353','\353',
'\304','\156','\323','\223','\373','\342','\174','\6',
'\306','\332','\236','\211','\33','\46','\364','\122',
'\116','\62','\140','\17','\44','\314','\270','\246',
'\24','\301','\255','\64','\135','\25','\222','\41',
'\41','\362','\142','\251','\15','\351','\56','\263',
'\143','\7','\323','\273','\143','\336','\202','\241',
'\40','\333','\35','\243','\74','\326','\131','\242',
'\114','\340','\61','\122','\325','\221','\327','\22',
'\166','\226','\265','\137','\302','\25','\316','\253',
'\21','\203','\361','\177','\217','\340','\115','\161',
'\165','\314','\2','\13','\77','\0','\111','\16',
'\174','\312','\362','\254','\47','\31','\344','\267',
'\364','\117','\311','\342','\117','\240','\124','\255',
'\146','\117','\240','\61','\226','\130','\130','\75',
'\143','\145','\136','\50','\152','\276','\55','\276',
'\240','\145','\225','\336','\305','\55','\127','\143',
'\35','\241','\335','\136','\112','\363','\30','\67',
'\150','\233','\254','\27','\225','\212','\340','\274',
'\52','\215','\27','\107','\122','\266','\137','\221',
'\61','\47','\4','\357','\126','\137','\225','\56',
'\345','\5','\203','\177','\302','\315','\134','\245',
'\73','\320','\20','\42','\153','\165','\136','\16',
'\64','\155','\331','\325','\326','\367','\314','\355',
'\272','\330','\332','\311','\61','\112','\332','\301',
'\140','\333','\6','\163','\37','\153','\136','\34',
'\131','\250','\155','\174','\337','\146','\304','\102',
'\154','\53','\25','\213','\56','\113','\123','\44',
'\244','\247','\170','\116','\256','\357','\107','\37',
'\167','\146','\264','\355','\224','\367','\111','\360',
'\1','\256','\277','\112','\310','\201','\232','\226',
'\63','\62','\121','\341','\233','\351','\171','\100',
'\144','\236','\276','\60','\325','\145','\321','\147',
'\247','\275','\273','\225','\235','\102','\247','\324',
'\366','\341','\72','\114','\257','\150','\251','\143',
'\102','\331','\217','\175','\32','\340','\114','\17',
'\276','\365','\61','\375','\131','\6','\111','\54',
'\266','\335','\17','\174','\321','\62','\354','\316',
'\357','\167','\121','\35','\317','\227','\363','\125',
'\336','\307','\375','\326','\14','\313','\36','\172',
'\7','\335','\205','\40','\160','\164','\347','\47',
'\361','\302','\371','\221','\227','\21','\157','\354',
'\371','\307','\177','\102','\324','\177','\357','\103',
'\355','\123','\167','\71','\225','\323','\104','\216',
'\352','\114','\201','\77','\231','\333','\137','\2',
'\174','\211','\234','\215','\302','\163','\144','\232',
'\330','\252','\136','\321','\66','\112','\66','\167',
'\361','\71','\235','\60','\331','\14','\307','\356',
'\105','\160','\364','\12','\116','\373','\210','\136',
'\355','\324','\342','\2','\111','\245','\15','\166',
'\140','\76','\232','\144','\370','\67','\306','\373',
'\6','\21','\112','\217','\214','\143','\171','\36',
'\232','\307','\251','\277','\51','\275','\215','\142',
'\111','\40','\67','\321','\120','\375','\260','\365',
'\63','\214','\345','\112','\176','\144','\165','\46',
'\325','\55','\137','\216','\321','\311','\317','\212',
'\213','\301','\312','\167','\125','\16','\304','\151',
'\12','\15','\175','\255','\270','\142','\205','\336',
'\73','\35','\57','\375','\220','\257','\116','\17',
'\74','\347','\363','\261','\150','\177','\234','\336',
'\25','\52','\123','\241','\105','\116','\157','\123',
'\121','\310','\33','\177','\322','\305','\327','\305',
'\357','\167','\231','\37','\140','\117','\355','\253',
'\253','\240','\43','\262','\310','\166','\371','\266',
'\272','\264','\222','\71','\216','\277','\7','\164',
'\54','\267','\144','\161','\116','\375','\41','\217',
'\73','\123','\215','\264','\266','\46','\344','\136',
'\156','\235','\17','\13','\276','\257','\114','\334',
'\233','\143','\32','\141','\167','\330','\243','\140',
'\73','\105','\220','\252','\227','\307','\306','\115',
'\61','\205','\243','\274','\267','\111','\151','\335',
'\31','\57','\37','\55','\110','\55','\317','\313',
'\173','\37','\337','\242','\146','\131','\310','\355',
'\42','\10','\21','\371','\156','\232','\363','\220',
'\204','\337','\352','\220','\146','\36','\14','\317',
'\346','\371','\200','\324','\162','\112','\227','\115',
'\42','\207','\231','\235','\131','\54','\130','\237',
'\0','\302','\320','\46','\342','\113','\320','\110',
'\110','\44','\254','\163','\272','\205','\63','\56',
'\373','\127','\11','\307','\0','\234','\224','\201',
'\10','\116','\50','\107','\267','\250','\276','\74',
'\227','\306','\343','\362','\143','\206','\233','\267',
'\375','\50','\313','\252','\112','\371','\267','\240',
'\177','\112','\165','\21','\163','\257','\340','\303',
'\53','\12','\101','\356','\346','\112','\253','\217',
'\45','\203','\247','\20','\72','\71','\166','\133',
'\326','\275','\121','\355','\50','\343','\240','\103',
'\42','\35','\5','\310','\301','\164','\217','\7',
'\240','\137','\113','\123','\113','\142','\100','\242',
'\120','\15','\130','\133','\321','\322','\376','\74',
'\50','\127','\13','\252','\355','\153','\63','\44',
'\150','\341','\242','\222','\375','\111','\116','\202',
'\62','\231','\266','\335','\146','\242','\133','\232',
'\224','\370','\342','\174','\365','\64','\17','\375',
'\263','\43','\45','\56','\165','\106','\267','\35',
'\55','\261','\233','\155','\372','\235','\106','\265',
'\170','\136','\170','\124','\334','\312','\54','\51',
'\242','\346','\204','\277','\365','\354','\153','\112',
'\264','\357','\365','\220','\115','\260','\73','\174',
'\40','\130','\145','\22','\305','\220','\156','\153',
'\126','\372','\367','\66','\76','\42','\221','\150',
'\156','\65','\67','\344','\370','\6','\341','\257',
'\220','\103','\7','\347','\326','\265','\107','\14',
'\10','\174','\134','\245','\213','\11','\226','\72',
'\167','\117','\207','\21','\362','\144','\150','\4',
'\136','\262','\211','\144','\330','\205','\64','\24',
'\120','\350','\265','\37','\4','\144','\323','\33',
'\202','\247','\256','\54','\56','\55','\0','\152',
'\325','\204','\132','\223','\55','\134','\173','\174',
'\362','\36','\62','\267','\261','\341','\70','\353',
'\370','\263','\366','\137','\331','\151','\71','\15',
'\171','\2','\265','\137','\67','\14','\210','\341',
'\276','\152','\62','\41','\323','\122','\332','\143',
'\356','\1','\50','\244','\212','\151','\341','\165',
'\56','\253','\166','\316','\364','\362','\341','\363',
'\305','\357','\141','\13','\207','\50','\241','\216',
'\56','\61','\216','\14','\127','\315','\12','\202',
'\205','\112','\347','\12','\163','\105','\52','\205',
'\74','\305','\374','\54','\203','\137','\77','\60',
'\74','\0','\222','\306','\371','\14','\131','\153',
'\172','\334','\273','\177','\365','\233','\112','\247',
'\157','\260','\321','\310','\157','\202','\263','\225',
'\262','\140','\170','\300','\56','\225','\65','\302',
'\265','\261','\120','\327','\327','\155','\67','\334',
'\331','\67','\70','\251','\325','\247','\320','\246',
'\22','\233','\262','\302','\61','\251','\371','\171',
'\112','\337','\166','\42','\63','\352','\63','\234',
'\2','\204','\107','\113','\304','\242','\326','\310',
'\235','\65','\25','\254','\121','\12','\232','\71',
'\26','\227','\112','\13','\141','\111','\216','\31',
'\256','\312','\217','\214','\250','\224','\6','\21',
'\230','\2','\223','\51','\126','\306','\13','\327',
'\205','\33','\167','\45','\107','\260','\351','\61',
'\61','\252','\221','\245','\102','\236','\226','\360',
'\352','\104','\266','\233','\114','\20','\274','\305',
'\256','\371','\64','\53','\301','\350','\117','\344',
'\146','\236','\170','\272','\7','\357','\201','\275',
'\337','\113','\325','\15','\110','\364','\123','\322',
'\123','\231','\364','\53','\4','\42','\210','\240',
'\263','\330','\254','\214','\251','\371','\256','\162',
'\312','\105','\317','\64','\125','\240','\234','\27',
'\360','\35','\245','\40','\13','\232','\74','\16',
'\105','\132','\175','\65','\77','\125','\65','\362',
'\322','\62','\351','\321','\231','\340','\40','\307',
'\132','\267','\146','\47','\351','\270','\225','\104',
'\123','\231','\254','\316','\342','\63','\152','\300',
'\5','\171','\241','\327','\137','\314','\122','\3',
'\352','\30','\315','\56','\266','\343','\200','\253',
'\311','\14','\163','\74','\267','\11','\242','\204',
'\125','\4','\222','\146','\364','\150','\240','\317',
'\107','\321','\3','\152','\52','\73','\106','\256',
'\355','\131','\262','\254','\1','\214','\240','\315',
'\103','\247','\73','\173','\76','\2','\40','\64',
'\61','\316','\323','\364','\242','\26','\273','\73',
'\51','\163','\172','\11','\342','\363','\12','\372',
'\307','\312','\0','\73','\312','\240','\144','\40',
'\321','\302','\222','\376','\304','\211','\24','\14',
'\250','\217','\356','\41','\130','\121','\243','\362',
'\221','\346','\254','\304','\301','\0','\151','\42',
'\325','\170','\350','\76','\74','\225','\40','\45',
'\46','\103','\5','\200','\275','\151','\227','\66',
'\341','\222','\133','\35','\62','\12','\221','\245',
'\364','\207','\122','\213','\143','\315','\0','\253',
'\233','\116','\371','\104','\313','\251','\164','\161',
'\255','\134','\304','\72','\37','\230','\141','\353',
'\361','\45','\162','\65','\233','\233','\42','\51',
'\333','\231','\103','\367','\311','\165','\260','\263',
'\225','\250','\273','\202','\112','\375','\323','\365',
'\240','\65','\252','\374','\156','\365','\202','\313',
'\253','\270','\110','\153','\3','\212','\372','\347',
'\270','\121','\207','\110','\137','\166','\235','\266',
'\272','\216','\216','\41','\52','\153','\150','\45',
'\171','\270','\354','\57','\227','\227','\55','\362',
'\337','\361','\157','\201','\26','\72','\26','\335',
'\354','\54','\363','\50','\322','\150','\215','\172',
'\273','\361','\20','\76','\135','\352','\235','\267',
'\327','\125','\372','\200','\163','\277','\320','\331',
'\143','\133','\123','\66','\242','\64','\51','\77',
'\322','\234','\21','\353','\15','\174','\207','\301',
'\340','\226','\63','\26','\43','\155','\42','\6',
'\216','\23','\56','\337','\226','\341','\263','\255',
'\131','\126','\27','\346','\132','\20','\362','\371',
'\20','\170','\12','\106','\307','\60','\224','\321',
'\305','\347','\122','\222','\50','\355','\45','\264',
'\2','\373','\36','\273','\330','\30','\114','\327',
'\213','\227','\340','\306','\263','\127','\54','\17',
'\317','\113','\303','\217','\104','\63','\337','\330',
'\21','\61','\300','\332','\10','\232','\141','\200',
'\312','\177','\5','\223','\133','\233','\314','\1',
'\247','\275','\202','\67','\352','\140','\360','\354',
'\326','\41','\321','\115','\345','\212','\123','\201',
'\61','\236','\265','\105','\273','\257','\12','\212',
'\356','\42','\104','\174','\154','\313','\35','\342',
'\164','\242','\275','\366','\115','\125','\204','\216',
'\376','\42','\44','\302','\335','\260','\166','\77',
'\217','\341','\117','\336','\57','\1','\342','\304',
'\111','\336','\171','\362','\22','\233','\147','\107',
'\41','\256','\50','\42','\253','\242','\13','\361',
'\16','\323','\364','\334','\200','\10','\247','\52',
'\330','\144','\2','\170','\270','\201','\173','\55',
'\200','\16','\316','\52','\250','\341','\203','\14',
'\24','\120','\314','\1','\60','\155','\227','\145',
'\10','\351','\333','\265','\110','\141','\111','\33',
'\136','\152','\44','\347','\361','\204','\302','\25',
'\232','\327','\125','\251','\161','\234','\132','\247',
'\312','\12','\0','\311','\276','\106','\337','\275',
'\5','\227','\124','\317','\216','\365','\151','\153',
'\163','\365','\324','\376','\160','\346','\261','\153',
'\204','\301','\1','\145','\255','\121','\204','\112',
'\344','\104','\141','\170','\321','\122','\260','\362',
'\365','\55','\22','\362','\24','\333','\142','\57',
'\123','\222','\247','\14','\355','\11','\42','\200',
'\45','\272','\150','\106','\163','\215','\14','\21',
'\240','\47','\275','\233','\341','\215','\10','\104',
'\242','\101','\126','\144','\367','\250','\110','\3',
'\233','\106','\373','\147','\334','\366','\75','\322',
'\242','\71','\272','\65','\160','\43','\352','\261',
'\211','\320','\233','\257','\343','\271','\345','\117',
'\366','\141','\17','\206','\54','\52','\102','\122',
'\30','\56','\213','\76','\34','\350','\46','\356',
'\234','\353','\0','\61','\251','\115','\136','\123',
'\112','\363','\125','\160','\174','\35','\300','\111',
'\55','\315','\332','\235','\332','\124','\254','\262',
'\234','\315','\65','\76','\170','\234','\123','\66',
'\205','\246','\132','\163','\75','\307','\215','\254',
'\63','\101','\253','\215','\117','\355','\245','\144',
'\221','\275','\11','\334','\157','\210','\233','\146',
'\145','\131','\326','\66','\20','\225','\304','\235',
'\277','\23','\304','\355','\314','\321','\243','\310',
'\73','\35','\224','\374','\343','\43','\27','\65',
'\147','\333','\147','\102','\214','\275','\160','\347',
'\345','\106','\255','\116','\371','\273','\10','\152',
'\5','\215','\367','\122','\12','\201','\115','\134',
'\125','\55','\165','\55','\373','\335','\364','\174',
'\162','\220','\207','\300','\232','\11','\224','\125',
'\16','\66','\365','\172','\307','\41','\173','\336',
'\306','\147','\206','\266','\46','\70','\100','\351',
'\150','\326','\272','\312','\42','\147','\123','\220',
'\70','\167','\127','\177','\362','\345','\356','\46',
'\305','\217','\312','\123','\133','\24','\33','\152',
'\303','\236','\161','\60','\272','\215','\51','\341',
'\127','\17','\150','\32','\25','\256','\145','\346',
'\157','\146','\256','\164','\305','\146','\302','\134',
'\243','\346','\104','\364','\174','\261','\300','\205',
'\176','\252','\142','\7','\121','\243','\52','\120',
'\7','\360','\317','\176','\37','\210','\143','\151',
'\353','\342','\16','\145','\226','\150','\251','\75',
'\40','\306','\315','\262','\73','\204','\267','\7',
'\10','\166','\111','\225','\216','\334','\203','\5',
'\171','\312','\145','\140','\317','\131','\261','\145',
'\306','\174','\47','\311','\166','\73','\315','\166',
'\242','\352','\274','\272','\25','\115','\314','\1',
'\166','\5','\31','\276','\270','\11','\267','\206',
'\34','\220','\76','\314','\361','\156','\156','\250',
'\44','\342','\365','\177','\157','\151','\72','\6',
'\376','\104','\174','\316','\44','\171','\115','\362',
'\256','\323','\41','\2','\147','\103','\365','\35',
'\130','\103','\322','\365','\313','\143','\74','\233',
'\352','\170','\76','\106','\50','\122','\262','\203',
'\126','\104','\235','\265','\241','\163','\217','\375',
'\70','\371','\370','\372','\241','\12','\153','\207',
'\367','\174','\101','\50','\101','\346','\122','\173',
'\355','\70','\61','\155','\327','\331','\143','\257',
'\233','\46','\11','\153','\240','\352','\261','\320',
'\113','\126','\264','\242','\42','\125','\111','\156',
'\212','\6','\24','\271','\172','\330','\40','\245',
'\250','\166','\254','\65','\321','\243','\114','\146',
'\256','\242','\106','\376','\347','\26','\76','\211',
'\361','\73','\23','\206','\154','\170','\133','\107',
'\216','\272','\300','\241','\7','\276','\143','\64',
'\167','\145','\45','\241','\3','\232','\222','\301',
'\143','\170','\216','\233','\347','\317','\22','\102',
'\166','\131','\370','\327','\321','\140','\102','\163',
'\375','\106','\17','\60','\240','\260','\323','\354',
'\76','\314','\157','\113','\55','\156','\166','\257',
'\304','\271','\214','\224','\246','\371','\5','\32',
'\273','\171','\221','\325','\35','\261','\137','\211',
'\46','\130','\61','\267','\154','\44','\112','\263',
'\263','\324','\17','\123','\50','\30','\74','\133',
'\235','\177','\105','\372','\103','\217','\24','\164',
'\40','\266','\375','\350','\104','\257','\241','\212',
'\30','\232','\74','\331','\105','\135','\162','\150',
'\33','\15','\116','\253','\266','\165','\6','\107',
'\131','\332','\327','\64','\155','\257','\347','\233',
'\77','\226','\172','\222','\335','\353','\100','\113',
'\270','\1','\5','\174','\140','\306','\124','\126',
'\45','\151','\43','\16','\42','\144','\214','\203',
'\166','\147','\155','\261','\321','\264','\111','\314',
'\340','\72','\324','\310','\373','\253','\23','\304',
'\346','\343','\207','\247','\123','\241','\272','\324',
'\144','\157','\305','\273','\334','\62','\121','\162',
'\264','\100','\275','\25','\357','\224','\67','\363',
'\32','\40','\363','\270','\162','\262','\233','\120',
'\5','\262','\110','\145','\206','\131','\62','\3',
'\261','\55','\363','\273','\304','\264','\225','\354',
'\40','\263','\323','\103','\70','\140','\51','\226',
'\336','\324','\136','\354','\33','\140','\242','\140',
'\275','\227','\307','\41','\154','\233','\334','\307',
'\37','\11','\100','\253','\133','\325','\311','\202',
'\100','\37','\131','\205','\226','\72','\254','\244',
'\364','\257','\235','\144','\353','\225','\46','\351',
'\16','\351','\172','\30','\130','\276','\42','\234',
'\162','\264','\350','\262','\310','\355','\350','\72',
'\337','\256','\116','\265','\163','\14','\71','\322',
'\211','\334','\245','\123','\102','\36','\244','\201',
'\225','\213','\302','\250','\14','\333','\230','\6',
'\76','\252','\66','\340','\306','\237','\152','\41',
'\45','\300','\276','\227','\122','\10','\105','\213',
'\124','\356','\125','\257','\261','\106','\4','\154',
'\165','\76','\173','\347','\47','\110','\316','\276',
'\210','\215','\366','\332','\271','\61','\215','\375',
'\207','\266','\375','\222','\153','\135','\175','\51',
'\25','\203','\55','\44','\263','\117','\66','\177',
'\41','\250','\351','\352','\345','\71','\222','\47',
'\70','\10','\231','\277','\233','\346','\15','\53',
'\330','\320','\202','\215','\221','\263','\65','\365',
'\34','\321','\375','\73','\142','\225','\54','\141',
'\124','\355','\100','\370','\224','\246','\357','\134',
'\334','\54','\20','\335','\313','\310','\362','\37',
'\147','\132','\45','\170','\354','\240','\123','\276',
'\123','\372','\251','\353','\243','\352','\66','\333',
'\53','\155','\256','\274','\32','\103','\67','\271',
'\151','\47','\147','\31','\306','\224','\242','\134',
'\175','\3','\221','\165','\360','\100','\52','\170',
'\302','\322','\112','\326','\214','\253','\44','\42',
'\302','\347','\166','\262','\2','\136','\173','\12',
'\333','\42','\345','\166','\256','\22','\224','\332',
'\70','\325','\241','\313','\136','\212','\267','\10',
'\251','\305','\106','\147','\101','\330','\75','\137',
'\112','\105','\62','\117','\344','\217','\307','\242',
'\311','\272','\323','\44','\341','\237','\300','\271',
'\50','\203','\220','\44','\152','\112','\343','\143',
'\235','\60','\177','\262','\171','\353','\101','\71',
'\127','\53','\242','\211','\57','\363','\32','\176',
'\254','\36','\73','\71','\37','\154','\44','\106',
'\50','\25','\347','\215','\137','\121','\35','\21',
'\354','\7','\254','\132','\140','\323','\210','\113',
'\15','\264','\350','\250','\200','\124','\225','\341',
'\132','\172','\74','\333','\352','\222','\105','\213',
'\103','\167','\6','\36','\213','\155','\260','\13',
'\35','\354','\240','\206','\163','\305','\350','\312',
'\302','\277','\373','\102','\200','\221','\1','\233',
'\321','\157','\207','\100','\305','\256','\140','\340',
'\350','\276','\270','\330','\140','\301','\142','\101',
'\361','\50','\253','\75','\322','\4','\337','\275',
'\250','\314','\124','\251','\313','\225','\53','\141',
'\216','\106','\144','\261','\15','\355','\250','\332',
'\372','\170','\331','\65','\361','\344','\351','\5',
'\27','\57','\354','\325','\304','\152','\160','\160',
'\340','\100','\333','\24','\241','\161','\70','\15',
'\303','\100','\376','\52','\317','\175','\126','\354',
'\20','\174','\117','\374','\76','\165','\115','\67',
'\353','\24','\315','\267','\362','\270','\50','\200',
'\202','\43','\262','\342','\35','\53','\204','\16',
'\147','\135','\345','\364','\245','\164','\267','\336',
'\102','\330','\7','\15','\166','\266','\107','\137',
'\274','\267','\141','\276','\124','\15','\210','\265',
'\327','\241','\103','\332','\326','\341','\64','\61',
'\71','\166','\364','\61','\376','\225','\357','\213',
'\127','\26','\125','\64','\144','\33','\366','\314',
'\113','\21','\236','\102','\123','\241','\330','\66',
'\43','\157','\376','\164','\165','\41','\356','\265',
'\200','\360','\266','\102','\122','\53','\147','\357',
'\154','\120','\40','\21','\270','\345','\334','\241',
'\175','\323','\332','\142','\154','\22','\102','\202',
'\364','\4','\171','\274','\102','\303','\44','\146',
'\251','\5','\23','\122','\142','\21','\356','\250',
'\222','\311','\147','\360','\65','\305','\363','\237',
'\171','\372','\372','\267','\336','\52','\221','\32',
'\107','\260','\56','\376','\310','\126','\0','\302',
'\274','\125','\251','\307','\257','\26','\254','\335',
'\132','\336','\147','\263','\63','\30','\360','\331',
'\273','\307','\216','\103','\372','\314','\145','\345',
'\110','\275','\250','\232','\140','\330','\373','\264',
'\157','\207','\211','\201','\221','\333','\256','\236',
'\76','\62','\51','\233','\56','\304','\263','\140',
'\150','\157','\126','\56','\364','\47','\306','\10',
'\344','\42','\75','\56','\204','\320','\56','\272',
'\237','\64','\212','\227','\350','\54','\45','\133',
'\136','\36','\354','\314','\326','\76','\37','\77',
'\327','\206','\134','\61','\123','\52','\205','\264',
'\346','\124','\365','\371','\361','\202','\170','\342',
'\322','\12','\256','\140','\354','\165','\100','\51',
'\226','\317','\320','\302','\322','\233','\202','\60',
'\72','\350','\67','\212','\335','\253','\101','\274',
'\140','\177','\350','\207','\76','\256','\33','\65',
'\206','\34','\372','\23','\152','\202','\274','\247',
'\113','\163','\371','\310','\114','\330','\162','\51',
'\362','\166','\20','\152','\103','\310','\306','\235',
'\202','\221','\145','\330','\13','\274','\154','\356',
'\204','\135','\310','\11','\117','\250','\275','\21',
'\364','\300','\171','\163','\150','\170','\365','\302',
'\373','\127','\42','\145','\243','\135','\310','\227',
'\227','\235','\112','\140','\63','\23','\350','\71',
'\176','\70','\117','\327','\160','\371','\63','\60',
'\172','\267','\107','\334','\45','\151','\323','\352',
'\370','\306','\216','\252','\211','\130','\5','\112',
'\336','\220','\271','\166','\70','\302','\335','\303',
'\106','\300','\54','\212','\122','\214','\355','\20',
'\136','\133','\326','\370','\323','\157','\221','\115',
'\276','\270','\245','\273','\331','\46','\133','\271',
'\361','\215','\253','\344','\246','\246','\101','\151',
'\131','\353','\6','\54','\224','\143','\131','\252',
'\10','\324','\4','\314','\175','\336','\355','\44',
'\235','\54','\312','\122','\34','\66','\202','\333',
'\146','\77','\204','\46','\366','\211','\265','\262',
'\123','\2','\363','\6','\360','\261','\136','\232',
'\33','\170','\360','\207','\12','\45','\67','\26',
'\46','\153','\140','\200','\256','\266','\262','\246',
'\230','\4','\67','\215','\162','\247','\113','\155',
'\270','\212','\333','\365','\324','\310','\354','\207',
'\6','\136','\360','\232','\206','\313','\51','\277',
'\246','\177','\355','\151','\27','\16','\146','\345',
'\110','\150','\52','\372','\333','\225','\254','\222',
'\162','\221','\225','\156','\235','\153','\353','\145',
'\304','\337','\222','\105','\236','\244','\77','\262',
'\272','\316','\143','\166','\312','\265','\277','\46',
'\30','\112','\137','\146','\53','\201','\232','\237',
'\231','\20','\10','\101','\54','\203','\177','\363',
'\304','\36','\6','\240','\102','\260','\45','\42',
'\120','\337','\233','\165','\151','\53','\367','\321',
'\366','\205','\146','\133','\76','\147','\345','\317',
'\33','\163','\77','\374','\141','\216','\214','\230',
'\176','\202','\331','\125','\113','\223','\55','\105',
'\122','\330','\0','\317','\102','\62','\271','\311',
'\363','\224','\2','\57','\42','\357','\147','\53',
'\237','\202','\305','\353','\144','\220','\363','\13',
'\215','\301','\164','\231','\347','\304','\30','\55',
'\220','\320','\131','\17','\105','\364','\147','\357',
'\322','\273','\66','\164','\52','\372','\244','\162',
'\364','\30','\337','\260','\253','\150','\55','\44',
'\110','\56','\165','\63','\353','\304','\141','\4',
'\301','\105','\367','\254','\305','\154','\205','\54',
'\214','\255','\13','\262','\344','\213','\334','\150',
'\170','\106','\43','\264','\327','\122','\313','\214',
'\17','\23','\27','\230','\56','\372','\151','\245',
'\56','\210','\150','\166','\25','\31','\225','\214',
'\236','\231','\214','\366','\217','\331','\130','\273',
'\37','\142','\34','\123','\226','\145','\231','\34',
'\26','\156','\103','\117','\126','\154','\226','\0',
'\16','\161','\46','\355','\372','\3','\17','\274',
'\47','\300','\227','\204','\273','\145','\361','\362',
'\373','\247','\124','\335','\244','\374','\16','\56',
'\171','\366','\120','\47','\35','\221','\174','\340',
'\162','\312','\136','\13','\35','\362','\310','\53',
'\145','\163','\376','\267','\300','\234','\15','\347',
'\364','\114','\376','\56','\117','\5','\203','\55',
'\376','\132','\217','\357','\344','\34','\363','\372',
'\133','\43','\205','\4','\244','\125','\344','\24',
'\262','\10','\336','\165','\255','\10','\251','\114',
'\367','\364','\227','\54','\253','\137','\353','\166',
'\157','\72','\205','\31','\115','\162','\106','\121',
'\300','\365','\327','\112','\356','\5','\76','\247',
'\123','\11','\226','\42','\130','\44','\100','\207',
'\351','\124','\152','\313','\315','\201','\264','\12',
'\141','\216','\53','\131','\146','\337','\324','\12',
'\263','\324','\336','\212','\256','\204','\13','\157',
'\13','\102','\301','\336','\220','\152','\301','\145',
'\327','\365','\354','\363','\13','\146','\173','\266',
'\101','\2','\153','\143','\125','\270','\234','\343',
'\375','\124','\140','\171','\223','\325','\304','\204',
'\174','\163','\336','\331','\166','\116','\215','\252',
'\114','\50','\355','\3','\211','\34','\154','\317',
'\326','\211','\51','\65','\360','\65','\251','\205',
'\305','\115','\0','\265','\363','\163','\326','\55',
'\250','\303','\5','\10','\333','\137','\111','\161',
'\53','\340','\71','\37','\307','\146','\71','\167',
'\215','\23','\170','\334','\341','\121','\255','\116',
'\312','\32','\324','\155','\161','\56','\140','\332',
'\374','\15','\310','\152','\124','\6','\226','\136',
'\373','\26','\274','\24','\123','\130','\327','\166',
'\201','\344','\340','\205','\1','\247','\313','\314',
'\306','\367','\76','\161','\13','\103','\331','\313',
'\344','\43','\142','\262','\23','\337','\237','\26',
'\220','\222','\372','\133','\117','\5','\202','\374',
'\123','\333','\56','\0','\43','\35','\1','\60',
'\50','\245','\7','\240','\361','\266','\276','\344',
'\204','\51','\30','\366','\11','\322','\363','\203',
'\221','\4','\126','\1','\203','\256','\313','\225',
'\317','\230','\151','\152','\120','\43','\15','\376',
'\235','\71','\232','\33','\325','\314','\42','\14',
'\1','\365','\5','\116','\200','\322','\61','\275',
'\14','\224','\225','\270','\313','\334','\43','\50',
'\7','\314','\5','\326','\337','\361','\301','\167',
'\47','\41','\361','\272','\121','\40','\122','\322',
'\367','\300','\152','\312','\341','\356','\162','\246',
'\221','\234','\144','\171','\103','\33','\74','\45',
'\21','\171','\50','\32','\172','\2','\271','\55',
'\375','\103','\342','\265','\357','\102','\113','\144',
'\3','\370','\123','\122','\231','\10','\352','\340',
'\273','\212','\255','\304','\166','\277','\275','\371',
'\43','\256','\327','\233','\57','\20','\207','\223',
'\275','\161','\166','\217','\124','\152','\273','\257',
'\337','\331','\312','\175','\243','\173','\147','\27',
'\126','\171','\102','\336','\321','\250','\245','\151',
'\204','\161','\76','\140','\354','\340','\250','\122',
'\30','\171','\206','\34','\143','\156','\171','\243',
'\64','\64','\201','\226','\355','\24','\123','\270',
'\51','\145','\134','\316','\42','\246','\352','\152',
'\130','\374','\10','\70','\6','\114','\66','\74',
'\244','\225','\372','\14','\215','\122','\302','\234',
'\117','\333','\201','\3','\206','\375','\176','\32',
'\30','\101','\127','\264','\331','\134','\34','\333',
'\323','\360','\40','\126','\340','\46','\337','\226',
'\344','\245','\25','\351','\215','\125','\72','\17',
'\114','\33','\166','\155','\314','\216','\116','\174',
'\50','\27','\27','\257','\152','\17','\133','\16',
'\260','\125','\366','\243','\304','\155','\142','\77',
'\217','\146','\213','\30','\61','\356','\312','\233',
'\30','\100','\234','\51','\360','\14','\177','\304',
'\100','\312','\136','\75','\140','\162','\63','\244',
'\25','\255','\135','\164','\44','\112','\104','\260',
'\2','\66','\320','\242','\234','\102','\373','\142',
'\326','\114','\121','\55','\255','\350','\332','\13',
'\77','\111','\105','\120','\111','\27','\372','\336',
'\143','\125','\107','\215','\121','\210','\236','\265',
'\150','\117','\327','\302','\135','\226','\351','\64',
'\212','\25','\366','\20','\51','\33','\251','\66',
'\112','\10','\54','\225','\367','\272','\21','\114',
'\314','\200','\312','\331','\343','\145','\167','\241',
'\300','\74','\65','\235','\276','\276','\141','\10',
'\376','\74','\43','\262','\36','\77','\314','\26',
'\210','\306','\301','\167','\243','\71','\75','\306',
'\215','\232','\214','\277','\326','\53','\321','\201',
'\26','\272','\353','\166','\373','\151','\151','\63',
'\177','\145','\323','\200','\27','\327','\17','\6',
'\361','\60','\114','\44','\215','\113','\317','\77',
'\46','\172','\260','\110','\301','\367','\236','\312',
'\274','\275','\336','\235','\50','\364','\7','\201',
'\323','\221','\52','\177','\252','\360','\366','\116',
'\164','\21','\16','\216','\151','\331','\342','\112',
'\374','\55','\226','\136','\100','\163','\53','\374',
'\220','\102','\125','\226','\54','\35','\17','\242',
'\273','\72','\25','\376','\341','\364','\133','\333',
'\174','\202','\260','\142','\266','\53','\237','\276',
'\55','\234','\110','\221','\310','\351','\162','\61',
'\324','\356','\151','\253','\17','\1','\361','\227',
'\223','\212','\334','\6','\42','\267','\252','\76',
'\0','\371','\26','\256','\252','\101','\212','\137',
'\154','\145','\55','\177','\16','\15','\257','\43',
'\245','\212','\41','\332','\322','\243','\60','\76',
'\202','\265','\114','\10','\177','\135','\64','\315',
'\57','\341','\143','\111','\332','\144','\200','\261',
'\111','\336','\212','\63','\56','\115','\147','\125',
'\10','\47','\270','\32','\67','\114','\11','\61',
'\311','\63','\112','\140','\145','\45','\335','\271',
'\34','\115','\363','\332','\336','\142','\273','\120',
'\201','\370','\41','\267','\154','\343','\65','\120',
'\352','\30','\274','\36','\366','\17','\356','\165',
'\304','\26','\124','\0','\205','\36','\365','\367',
'\134','\220','\21','\165','\340','\202','\40','\266',
'\73','\240','\134','\30','\125','\371','\102','\124',
'\155','\64','\14','\145','\236','\277','\302','\43',
'\357','\142','\311','\150','\200','\70','\60','\311',
'\275','\200','\133','\124','\216','\72','\16','\226',
'\251','\210','\22','\220','\157','\57','\261','\241',
'\45','\361','\65','\34','\234','\275','\307','\221',
'\322','\217','\31','\213','\363','\344','\250','\244',
'\365','\322','\5','\303','\211','\23','\241','\313',
'\16','\31','\57','\201','\43','\5','\133','\135',
'\216','\71','\117','\363','\103','\113','\41','\31',
'\212','\210','\116','\311','\257','\372','\363','\334',
'\172','\367','\140','\354','\13','\253','\105','\307',
'\142','\373','\314','\247','\171','\173','\160','\271',
'\5','\220','\75','\51','\64','\255','\257','\30',
'\130','\14','\142','\234','\7','\364','\206','\106',
'\266','\63','\264','\172','\213','\336','\354','\324',
'\63','\275','\266','\32','\26','\163','\150','\161',
'\60','\102','\147','\23','\362','\375','\126','\63',
'\101','\55','\70','\220','\40','\145','\331','\214',
'\307','\16','\357','\137','\144','\374','\120','\223',
'\140','\371','\301','\40','\271','\260','\301','\44',
'\312','\65','\143','\165','\324','\41','\153','\51',
'\152','\245','\211','\54','\171','\357','\204','\237',
'\135','\273','\127','\365','\271','\267','\256','\14',
'\166','\61','\257','\174','\51','\65','\307','\27',
'\301','\221','\6','\362','\352','\351','\10','\50',
'\257','\64','\7','\156','\103','\230','\343','\123',
'\203','\217','\160','\74','\207','\214','\133','\263',
'\163','\321','\157','\340','\65','\200','\41','\102',
'\365','\125','\44','\13','\254','\353','\273','\11',
'\70','\306','\204','\174','\60','\42','\257','\302',
'\275','\334','\15','\107','\376','\106','\364','\176',
'\233','\260','\316','\337','\75','\34','\153','\245',
'\372','\100','\350','\305','\147','\300','\243','\277',
'\321','\222','\56','\77','\214','\25','\275','\363',
'\66','\276','\22','\125','\103','\5','\325','\117',
'\132','\343','\41','\345','\203','\371','\145','\170',
'\364','\361','\115','\357','\225','\231','\5','\311',
'\246','\116','\65','\251','\3','\356','\247','\113',
'\126','\352','\275','\126','\227','\160','\100','\202',
'\112','\366','\167','\167','\52','\11','\272','\10',
'\353','\337','\11','\341','\22','\315','\113','\351',
'\147','\46','\226','\353','\154','\145','\17','\327',
'\230','\52','\156','\13','\317','\133','\274','\354',
'\6','\155','\345','\263','\207','\66','\175','\274',
'\345','\16','\23','\27','\202','\102','\74','\176',
'\121','\223','\347','\213','\240','\344','\304','\135',
'\360','\11','\220','\166','\201','\203','\212','\56',
'\370','\372','\316','\371','\113','\11','\254','\350',
'\324','\3','\3','\337','\303','\51','\22','\72',
'\71','\144','\172','\131','\331','\71','\141','\326',
'\164','\352','\344','\214','\352','\163','\17','\373',
'\71','\17','\23','\241','\176','\22','\153','\27',
'\131','\202','\21','\107','\12','\212','\373','\31',
'\224','\244','\221','\363','\227','\26','\37','\373',
'\300','\276','\346','\251','\212','\232','\175','\275',
'\250','\177','\316','\174','\37','\207','\76','\37',
'\317','\324','\306','\345','\240','\113','\267','\167',
'\52','\45','\245','\101','\350','\163','\175','\173',
'\145','\222','\321','\2','\362','\343','\327','\71',
'\262','\74','\177','\44','\351','\257','\141','\34',
'\172','\153','\115','\126','\55','\252','\175','\105',
'\131','\124','\326','\210','\246','\231','\344','\156',
'\314','\222','\373','\371','\234','\215','\275','\262',
'\131','\132','\227','\111','\35','\14','\370','\343',
'\312','\121','\152','\140','\314','\343','\225','\365',
'\304','\77','\5','\370','\106','\112','\173','\376',
'\341','\44','\214','\237','\222','\124','\43','\225',
'\240','\373','\353','\314','\335','\231','\256','\142',
'\216','\226','\202','\255','\123','\232','\362','\315',
'\257','\327','\154','\374','\372','\150','\217','\302',
'\271','\233','\135','\120','\56','\136','\65','\150',
'\301','\322','\334','\36','\220','\372','\3','\365',
'\217','\55','\213','\110','\116','\23','\156','\36',
'\151','\360','\101','\101','\46','\303','\140','\67',
'\71','\30','\357','\276','\11','\262','\136','\230',
'\7','\125','\41','\12','\113','\336','\137','\220',
'\155','\173','\272','\270','\101','\313','\142','\202',
'\141','\104','\140','\156','\125','\201','\11','\51',
'\313','\207','\374','\205','\256','\44','\225','\375',
'\277','\20','\215','\224','\320','\241','\273','\346',
'\216','\374','\340','\266','\155','\335','\55','\17',
'\114','\147','\301','\202','\4','\72','\2','\10',
'\225','\20','\174','\266','\15','\274','\324','\206',
'\353','\275','\327','\211','\135','\203','\124','\121',
'\367','\367','\254','\321','\174','\316','\47','\302',
'\360','\353','\11','\367','\26','\330','\50','\70',
'\44','\316','\274','\107','\104','\67','\221','\74',
'\4','\347','\12','\233','\140','\136','\347','\210',
'\166','\325','\171','\14','\43','\50','\337','\134',
'\353','\254','\371','\305','\170','\234','\364','\256',
'\270','\357','\176','\301','\42','\365','\151','\156',
'\201','\326','\127','\160','\252','\375','\342','\224',
'\0','\133','\375','\120','\137','\213','\103','\165',
'\26','\313','\307','\201','\25','\232','\322','\224',
'\146','\63','\103','\357','\262','\350','\244','\340',
'\324','\36','\16','\321','\310','\123','\30','\166',
'\226','\55','\374','\365','\207','\131','\333','\43',
'\210','\113','\175','\62','\102','\200','\311','\237',
'\1','\71','\142','\327','\65','\216','\173','\211',
'\304','\317','\123','\150','\4','\351','\230','\361',
'\102','\174','\4','\342','\127','\121','\312','\305',
'\62','\12','\215','\331','\262','\25','\145','\113',
'\130','\75','\176','\137','\117','\56','\334','\243',
'\351','\177','\361','\243','\1','\211','\127','\302',
'\30','\163','\243','\150','\314','\263','\72','\264',
'\337','\44','\131','\324','\77','\114','\254','\41',
'\345','\322','\165','\137','\347','\17','\231','\177',
'\144','\217','\76','\356','\203','\27','\212','\175',
'\51','\150','\200','\335','\325','\30','\236','\67',
'\127','\253','\351','\357','\106','\314','\11','\152',
'\33','\142','\145','\21','\6','\100','\70','\232',
'\112','\73','\346','\356','\334','\42','\25','\250',
'\314','\254','\352','\116','\6','\336','\27','\173',
'\266','\23','\371','\326','\11','\131','\106','\311',
'\105','\3','\10','\224','\72','\147','\147','\200',
'\233','\322','\362','\206','\21','\277','\40','\333',
'\363','\325','\364','\157','\303','\256','\43','\121',
'\35','\142','\11','\160','\5','\317','\62','\311',
'\237','\23','\372','\125','\20','\124','\207','\360',
'\270','\25','\131','\226','\32','\147','\127','\145',
'\365','\227','\314','\227','\123','\165','\227','\172',
'\142','\363','\230','\343','\127','\23','\367','\251',
'\31','\335','\55','\172','\245','\267','\203','\215',
'\336','\135','\241','\53','\132','\254','\271','\241',
'\376','\25','\326','\41','\170','\36','\321','\364',
'\370','\35','\47','\225','\256','\317','\154','\52',
'\65','\372','\364','\355','\15','\251','\253','\124',
'\230','\267','\300','\62','\374','\170','\250','\154',
'\371','\173','\273','\300','\143','\47','\74','\356',
'\247','\62','\204','\266','\251','\330','\341','\212',
'\304','\351','\152','\50','\257','\101','\367','\221',
'\32','\133','\144','\267','\272','\251','\270','\12',
'\164','\364','\20','\366','\45','\223','\110','\252',
'\66','\13','\203','\343','\371','\151','\133','\6',
'\234','\145','\315','\105','\274','\375','\263','\356',
'\6','\306','\257','\154','\1','\250','\245','\374',
'\372','\171','\63','\304','\335','\170','\56','\275',
'\114','\5','\131','\152','\203','\370','\334','\52',
'\125','\235','\7','\116','\250','\17','\143','\167',
'\370','\226','\302','\137','\264','\316','\173','\167',
'\276','\175','\212','\217','\344','\142','\51','\161',
'\17','\60','\75','\156','\125','\243','\212','\215',
'\314','\31','\250','\240','\362','\326','\201','\203',
'\272','\265','\72','\116','\222','\312','\224','\331',
'\364','\51','\54','\25','\242','\70','\237','\306',
'\317','\355','\245','\7','\34','\214','\364','\165',
'\332','\15','\44','\145','\74','\257','\341','\343',
'\207','\140','\142','\317','\276','\357','\321','\110',
'\137','\116','\310','\115','\207','\374','\227','\126',
'\75','\232','\173','\214','\231','\320','\240','\325',
'\55','\331','\352','\301','\14','\244','\7','\254',
'\311','\334','\113','\125','\257','\241','\56','\170',
'\260','\12','\213','\216','\362','\44','\324','\5',
'\337','\112','\172','\172','\243','\75','\334','\332',
'\226','\204','\204','\151','\77','\60','\32','\345',
'\235','\275','\301','\370','\12','\163','\201','\70',
'\271','\61','\22','\37','\55','\142','\370','\121',
'\103','\145','\75','\242','\144','\326','\222','\342',
'\77','\215','\222','\266','\226','\271','\366','\170',
'\120','\131','\173','\323','\365','\161','\14','\215',
'\6','\24','\75','\0','\330','\6','\110','\262',
'\105','\366','\325','\210','\230','\137','\356','\45',
'\155','\147','\175','\66','\174','\20','\53','\53',
'\52','\375','\341','\54','\211','\361','\11','\220',
'\40','\106','\313','\333','\75','\62','\363','\31',
'\37','\52','\71','\140','\75','\345','\114','\44',
'\57','\174','\361','\311','\347','\31','\204','\165',
'\362','\365','\63','\374','\307','\25','\275','\54',
'\326','\252','\117','\67','\41','\10','\330','\111',
'\70','\0','\104','\262','\50','\22','\246','\263',
'\12','\352','\370','\214','\52','\362','\321','\363',
'\156','\40','\213','\6','\21','\352','\100','\324',
'\361','\70','\34','\211','\275','\72','\174','\172',
'\161','\6','\320','\132','\356','\220','\243','\376',
'\207','\22','\160','\301','\4','\322','\352','\1',
'\270','\27','\32','\227','\112','\236','\313','\274',
'\216','\75','\341','\314','\42','\25','\20','\13',
'\142','\345','\33','\147','\356','\376','\10','\154',
'\147','\164','\352','\170','\240','\241','\10','\64',
'\250','\336','\247','\357','\27','\146','\152','\146',
'\243','\352','\76','\206','\240','\244','\365','\32',
'\252','\372','\325','\212','\61','\110','\367','\203',
'\155','\73','\356','\131','\206','\145','\122','\50',
'\230','\366','\215','\232','\164','\153','\24','\110',
'\72','\40','\221','\144','\314','\277','\164','\342',
'\7','\257','\211','\106','\32','\242','\341','\200',
'\105','\260','\142','\173','\361','\375','\121','\133',
'\257','\36','\364','\74','\326','\66','\245','\306',
'\363','\46','\1','\352','\55','\125','\241','\354',
'\364','\76','\366','\13','\162','\124','\325','\75',
'\106','\157','\245','\67','\223','\334','\241','\162',
'\54','\152','\142','\340','\221','\37','\101','\32',
'\272','\112','\117','\17','\57','\214','\252','\10',
'\151','\362','\325','\272','\152','\164','\270','\74',
'\361','\211','\330','\365','\244','\206','\205','\331',
'\152','\265','\233','\36','\45','\24','\101','\353',
'\324','\4','\64','\267','\351','\140','\204','\12',
'\174','\33','\135','\235','\67','\33','\354','\121',
'\0','\254','\105','\162','\370','\54','\327','\315',
'\165','\112','\331','\167','\242','\111','\202','\136',
'\165','\272','\134','\313','\74','\235','\117','\237',
'\211','\54','\257','\61','\214','\131','\1','\234',
'\100','\120','\103','\254','\160','\367','\151','\165',
'\215','\171','\133','\35','\43','\346','\272','\333',
'\220','\61','\354','\5','\215','\222','\117','\143',
'\154','\363','\57','\113','\272','\337','\144','\376',
'\273','\137','\33','\205','\45','\233','\243','\230',
'\51','\53','\336','\67','\17','\273','\117','\46',
'\277','\273','\2','\100','\236','\41','\67','\100',
'\224','\342','\221','\352','\321','\314','\320','\364',
'\120','\57','\76','\353','\327','\375','\214','\152',
'\160','\61','\3','\115','\363','\56','\53','\53',
'\137','\113','\271','\316','\15','\226','\56','\306',
'\17','\353','\107','\71','\303','\245','\14','\275',
'\271','\254','\64','\300','\307','\41','\371','\223',
'\133','\265','\205','\316','\24','\237','\21','\314',
'\323','\244','\243','\230','\11','\53','\347','\221',
'\164','\220','\267','\26','\66','\152','\157','\333',
'\324','\114','\306','\44','\31','\212','\31','\221',
'\167','\24','\53','\163','\215','\2','\335','\174',
'\166','\265','\200','\255','\156','\240','\116','\335',
'\327','\346','\371','\242','\7','\157','\171','\310',
'\76','\160','\243','\237','\22','\16','\355','\126',
'\56','\354','\27','\133','\273','\107','\103','\226',
'\270','\317','\34','\223','\142','\0','\141','\13',
'\15','\125','\200','\170','\167','\12','\242','\40',
'\345','\75','\133','\62','\106','\226','\71','\336',
'\374','\260','\220','\332','\75','\116','\173','\313',
'\331','\243','\41','\216','\366','\157','\46','\35',
'\212','\146','\103','\337','\307','\127','\34','\117',
'\357','\146','\311','\337','\11','\363','\33','\353',
'\365','\167','\161','\323','\10','\361','\344','\357',
'\312','\31','\64','\227','\102','\113','\3','\226',
'\114','\364','\147','\145','\215','\105','\303','\20',
'\271','\30','\151','\151','\105','\210','\346','\350',
'\327','\7','\131','\136','\54','\331','\31','\363',
'\32','\116','\167','\141','\311','\326','\213','\336',
'\253','\260','\166','\237','\255','\221','\5','\14',
'\237','\151','\172','\250','\305','\12','\325','\326',
'\376','\345','\37','\337','\201','\54','\13','\154',
'\110','\266','\355','\355','\226','\61','\13','\255',
'\170','\376','\355','\234','\302','\333','\23','\1',
'\7','\162','\77','\114','\130','\354','\350','\254',
'\5','\300','\163','\332','\100','\131','\323','\375',
'\70','\167','\260','\21','\316','\51','\367','\236',
'\141','\7','\151','\171','\261','\347','\355','\122',
'\337','\135','\154','\245','\352','\260','\227','\112',
'\175','\166','\74','\204','\142','\341','\264','\331',
'\235','\345','\62','\111','\51','\52','\22','\264',
'\222','\340','\336','\261','\2','\67','\24','\272',
'\102','\373','\235','\272','\366','\361','\110','\135',
'\331','\311','\335','\204','\376','\314','\213','\253',
'\164','\210','\231','\0','\345','\265','\75','\45',
'\33','\160','\7','\60','\271','\245','\2','\40',
'\234','\162','\245','\335','\341','\346','\115','\360',
'\161','\251','\256','\50','\104','\133','\166','\154',
'\305','\156','\251','\321','\175','\155','\270','\42',
'\136','\53','\24','\31','\17','\140','\20','\350',
'\130','\64','\324','\325','\210','\57','\64','\7',
'\301','\53','\126','\24','\70','\316','\154','\367',
'\21','\333','\33','\240','\226','\240','\17','\263',
'\32','\161','\10','\13','\136','\36','\273','\151',
'\346','\203','\321','\60','\43','\247','\0','\314',
'\135','\314','\265','\333','\326','\27','\22','\252',
'\21','\242','\20','\3','\2','\361','\54','\355',
'\114','\120','\4','\130','\14','\214','\6','\364',
'\315','\210','\135','\133','\367','\124','\344','\145',
'\312','\170','\220','\222','\235','\35','\206','\53',
'\44','\266','\7','\5','\21','\56','\142','\361',
'\265','\111','\321','\306','\256','\253','\347','\366',
'\2','\75','\113','\103','\170','\364','\43','\344',
'\175','\130','\373','\373','\337','\375','\212','\132',
'\232','\272','\206','\104','\335','\132','\14','\326',
'\206','\40','\20','\263','\332','\357','\101','\345',
'\371','\67','\226','\241','\74','\207','\240','\310',
'\34','\114','\176','\314','\120','\10','\344','\336',
'\301','\260','\266','\371','\317','\367','\264','\34',
'\102','\24','\253','\326','\267','\232','\104','\263',
'\73','\257','\174','\23','\255','\202','\165','\272',
'\367','\324','\215','\342','\212','\24','\41','\152',
'\335','\373','\7','\4','\67','\245','\240','\106',
'\372','\346','\44','\51','\331','\35','\123','\140',
'\123','\251','\76','\155','\175','\317','\274','\127',
'\322','\272','\361','\313','\277','\16','\172','\373',
'\30','\304','\262','\216','\300','\46','\137','\332',
'\133','\322','\324','\100','\5','\244','\375','\237',
'\1','\322','\357','\275','\124','\23','\357','\340',
'\342','\102','\120','\250','\344','\172','\345','\40',
'\213','\120','\124','\231','\27','\255','\142','\262',
'\247','\153','\175','\276','\173','\251','\147','\360',
'\326','\257','\172','\27','\51','\0','\25','\366',
'\71','\141','\71','\46','\16','\260','\167','\247',
'\52','\150','\302','\106','\51','\123','\70','\20',
'\372','\77','\103','\74','\351','\356','\312','\352',
'\11','\356','\156','\275','\153','\64','\250','\342',
'\46','\352','\326','\320','\204','\17','\172','\227',
'\157','\75','\172','\76','\277','\167','\45','\352',
'\246','\212','\247','\152','\224','\340','\116','\126',
'\73','\115','\117','\75','\54','\150','\55','\222',
'\370','\204','\312','\351','\222','\277','\351','\17',
'\230','\267','\224','\176','\132','\351','\233','\262',
'\263','\40','\156','\55','\10','\77','\264','\222',
'\361','\121','\6','\232','\173','\271','\255','\140',
'\243','\302','\251','\101','\34','\73','\117','\362',
'\326','\206','\373','\146','\330','\17','\241','\247',
'\133','\374','\66','\220','\353','\163','\343','\227',
'\340','\7','\77','\121','\243','\255','\277','\244',
'\337','\15','\202','\363','\21','\214','\20','\13',
'\304','\160','\365','\166','\123','\100','\256','\131',
'\160','\375','\132','\52','\25','\376','\237','\375',
'\341','\234','\275','\50','\52','\301','\342','\312',
'\301','\212','\157','\141','\272','\140','\370','\1',
'\214','\70','\66','\301','\136','\344','\367','\341',
'\45','\242','\223','\173','\203','\374','\301','\25',
'\112','\366','\204','\256','\320','\256','\22','\302',
'\342','\275','\77','\220','\341','\310','\141','\215',
'\330','\266','\226','\231','\172','\241','\74','\314',
'\121','\305','\243','\173','\357','\63','\267','\375',
'\247','\260','\41','\51','\176','\57','\236','\375',
'\136','\176','\53','\303','\151','\125','\245','\130',
'\121','\314','\327','\10','\15','\216','\134','\114',
'\117','\322','\214','\246','\60','\341','\43','\343',
'\136','\300','\145','\35','\136','\14','\347','\200',
'\145','\310','\301','\161','\224','\254','\157','\5',
'\165','\317','\355','\110','\3','\324','\331','\302',
'\122','\140','\303','\140','\320','\222','\26','\335',
'\31','\141','\1','\367','\300','\367','\347','\113',
'\126','\7','\31','\303','\213','\206','\161','\214',
'\200','\53','\342','\316','\21','\220','\315','\177',
'\247','\4','\34','\146','\113','\357','\227','\31',
'\162','\272','\127','\317','\41','\347','\222','\350',
'\117','\301','\144','\63','\303','\111','\364','\321',
'\270','\276','\104','\151','\107','\46','\106','\376',
'\356','\106','\237','\357','\325','\0','\270','\155',
'\266','\126','\136','\127','\302','\47','\226','\255',
'\162','\44','\211','\246','\110','\247','\71','\247',
'\25','\325','\306','\132','\31','\226','\140','\5',
'\4','\0','\160','\310','\45','\153','\121','\241',
'\117','\100','\251','\34','\1','\115','\230','\21',
'\250','\156','\242','\354','\14','\256','\236','\350',
'\101','\236','\176','\14','\331','\115','\33','\24',
'\7','\36','\255','\245','\46','\111','\25','\140',
'\172','\32','\4','\133','\232','\211','\10','\53',
'\223','\50','\332','\14','\274','\23','\350','\356',
'\370','\122','\33','\176','\104','\7','\213','\356',
'\304','\216','\27','\245','\261','\16','\10','\175',
'\332','\172','\21','\0','\361','\213','\0','\122',
'\244','\54','\360','\133','\10','\23','\210','\27',
'\130','\37','\216','\200','\227','\6','\130','\374',
'\77','\211','\111','\232','\226','\174','\53','\100',
'\202','\256','\33','\140','\13','\325','\146','\340',
'\370','\372','\357','\303','\157','\205','\226','\306',
'\116','\215','\133','\27','\166','\212','\121','\231',
'\366','\142','\344','\220','\200','\52','\352','\302',
'\63','\364','\153','\103','\243','\2','\200','\221',
'\317','\0','\44','\200','\216','\314','\253','\245',
'\221','\346','\265','\6','\373','\1','\125','\276',
'\172','\45','\364','\306','\62','\143','\137','\146',
'\52','\356','\236','\356','\365','\310','\33','\250',
'\372','\163','\140','\160','\216','\120','\104','\67',
'\346','\276','\232','\270','\367','\206','\142','\313',
'\311','\173','\244','\263','\13','\267','\165','\47',
'\4','\257','\37','\35','\76','\133','\140','\131',
'\376','\365','\152','\102','\11','\315','\134','\256',
'\132','\170','\42','\175','\130','\247','\360','\3',
'\351','\355','\317','\327','\212','\277','\107','\241',
'\164','\152','\6','\113','\106','\166','\332','\70',
'\144','\41','\226','\171','\56','\271','\332','\306',
'\221','\265','\343','\62','\54','\265','\336','\142',
'\73','\264','\27','\361','\24','\351','\175','\306',
'\221','\110','\253','\216','\236','\241','\322','\222',
'\346','\153','\231','\45','\173','\72','\43','\207',
'\262','\32','\300','\230','\253','\135','\43','\255',
'\330','\274','\236','\267','\326','\71','\144','\320',
'\72','\53','\337','\11','\76','\237','\242','\12',
'\3','\310','\161','\223','\12','\201','\76','\215',
'\333','\131','\122','\133','\22','\104','\141','\277',
'\16','\200','\255','\224','\143','\63','\356','\334',
'\150','\317','\262','\347','\60','\241','\110','\44',
'\76','\146','\105','\117','\112','\336','\113','\31',
'\256','\36','\11','\102','\325','\61','\341','\134',
'\301','\17','\126','\125','\157','\1','\300','\335',
'\211','\13','\116','\272','\67','\136','\101','\302',
'\251','\63','\122','\317','\155','\57','\334','\260',
'\275','\236','\76','\15','\303','\355','\34','\211',
'\106','\247','\231','\165','\152','\350','\342','\173',
'\273','\21','\373','\171','\160','\34','\73','\375',
'\140','\152','\3','\310','\155','\362','\100','\70',
'\310','\255','\273','\370','\226','\326','\372','\15',
'\116','\74','\275','\274','\355','\45','\206','\253',
'\322','\347','\314','\244','\301','\351','\157','\204',
'\324','\220','\120','\311','\127','\104','\216','\70',
'\130','\134','\266','\255','\372','\275','\363','\17',
'\75','\23','\254','\312','\131','\120','\215','\332',
'\224','\162','\167','\243','\233','\372','\251','\220',
'\124','\137','\102','\350','\332','\75','\30','\76',
'\376','\251','\17','\42','\312','\66','\35','\367',
'\122','\131','\55','\36','\116','\164','\332','\103',
'\73','\360','\206','\156','\151','\120','\37','\32',
'\135','\14','\235','\330','\361','\275','\361','\105',
'\300','\122','\235','\73','\2','\31','\376','\121',
'\66','\205','\34','\54','\221','\130','\160','\72',
'\1','\66','\112','\2','\112','\132','\114','\270',
'\350','\274','\204','\217','\63','\4','\367','\1',
'\42','\246','\1','\41','\341','\202','\322','\1',
'\356','\120','\43','\146','\6','\304','\314','\244',
'\42','\337','\124','\142','\212','\354','\115','\276',
'\57','\365','\71','\340','\66','\174','\176','\65',
'\225','\313','\11','\53','\121','\253','\154','\134',
'\23','\152','\255','\347','\316','\137','\153','\51',
'\214','\215','\125','\122','\315','\173','\202','\353',
'\20','\65','\30','\266','\60','\46','\223','\222',
'\74','\270','\270','\36','\364','\155','\70','\370',
'\123','\305','\14','\70','\37','\321','\74','\102',
'\345','\3','\307','\176','\113','\127','\241','\3',
'\216','\117','\161','\63','\265','\376','\43','\32',
'\26','\346','\332','\226','\100','\231','\354','\111',
'\247','\121','\211','\116','\57','\53','\112','\15',
'\245','\307','\104','\373','\160','\224','\156','\41',
'\241','\320','\102','\136','\55','\236','\13','\11',
'\36','\271','\256','\317','\242','\21','\367','\327',
'\135','\0','\71','\117','\346','\15','\223','\155',
'\255','\240','\360','\127','\320','\203','\16','\15',
'\310','\40','\347','\157','\366','\332','\121','\316',
'\363','\210','\24','\337','\103','\75','\170','\70',
'\32','\243','\13','\30','\302','\352','\23','\224',
'\10','\41','\165','\275','\247','\60','\267','\116',
'\305','\45','\273','\301','\314','\272','\137','\44',
'\345','\225','\103','\315','\111','\250','\131','\31',
'\230','\316','\340','\23','\332','\335','\125','\171',
'\132','\302','\263','\64','\15','\225','\157','\316',
'\242','\175','\233','\165','\362','\316','\335','\0',
'\12','\73','\67','\257','\17','\64','\76','\215',
'\306','\262','\12','\163','\42','\41','\336','\365',
'\201','\75','\46','\127','\306','\217','\106','\344',
'\212','\7','\225','\231','\231','\45','\331','\12',
'\10','\55','\11','\147','\272','\63','\231','\144',
'\100','\165','\15','\265','\316','\336','\210','\275',
'\362','\223','\336','\131','\65','\270','\46','\376',
'\63','\11','\246','\246','\170','\13','\362','\142',
'\161','\223','\146','\153','\122','\376','\154','\302',
'\112','\221','\167','\155','\131','\214','\275','\212',
'\323','\214','\362','\361','\253','\316','\27','\245',
'\272','\164','\1','\117','\340','\232','\366','\175',
'\53','\172','\76','\140','\205','\372','\0','\147',
'\277','\122','\146','\102','\14','\127','\347','\214',
'\115','\61','\245','\51','\52','\130','\152','\363',
'\147','\352','\55','\360','\171','\161','\302','\341',
'\321','\45','\7','\26','\200','\146','\242','\106',
'\76','\300','\113','\202','\244','\275','\72','\306',
'\15','\25','\41','\356','\40','\205','\271','\110',
'\247','\30','\131','\156','\172','\230','\304','\170',
'\23','\244','\365','\211','\262','\51','\122','\206',
'\43','\220','\7','\72','\70','\344','\71','\36',
'\102','\24','\226','\13','\365','\51','\35','\131',
'\21','\141','\40','\365','\262','\170','\363','\267',
'\235','\272','\172','\151','\205','\124','\214','\355',
'\322','\14','\130','\24','\216','\217','\52','\116',
'\7','\367','\116','\203','\345','\113','\204','\133',
'\7','\73','\17','\110','\242','\374','\370','\65',
'\275','\372','\340','\43','\126','\250','\362','\362',
'\15','\236','\274','\161','\103','\321','\323','\264',
'\132','\343','\346','\152','\324','\365','\275','\147',
'\374','\177','\210','\142','\375','\343','\371','\336',
'\376','\270','\126','\35','\114','\130','\171','\121',
'\234','\151','\234','\172','\165','\215','\234','\236',
'\75','\216','\277','\100','\376','\25','\72','\335',
'\372','\47','\313','\71','\134','\232','\171','\273',
'\1','\123','\324','\307','\134','\106','\30','\272',
'\246','\127','\252','\224','\365','\250','\54','\36',
'\203','\140','\226','\304','\206','\137','\145','\103',
'\267','\3','\107','\203','\373','\63','\71','\341',
'\301','\104','\317','\337','\227','\363','\161','\36',
'\13','\136','\251','\72','\215','\332','\170','\267',
'\157','\307','\360','\327','\235','\105','\201','\37',
'\220','\217','\367','\173','\255','\44','\110','\210',
'\114','\113','\352','\273','\32','\121','\245','\312',
'\357','\315','\322','\255','\300','\37','\300','\227',
'\161','\267','\213','\177','\204','\25','\230','\71',
'\367','\204','\364','\332','\330','\40','\170','\77',
'\155','\333','\146','\337','\32','\131','\366','\327',
'\126','\70','\114','\36','\277','\154','\206','\127',
'\226','\162','\206','\322','\112','\50','\17','\14',
'\334','\17','\124','\72','\67','\114','\323','\30',
'\166','\231','\141','\242','\137','\123','\20','\32',
'\40','\30','\370','\261','\311','\160','\51','\353',
'\246','\344','\206','\235','\350','\54','\152','\125',
'\144','\314','\175','\166','\370','\242','\306','\73',
'\30','\63','\44','\15','\242','\26','\55','\252',
'\313','\240','\113','\254','\76','\342','\240','\207',
'\252','\207','\274','\277','\122','\61','\121','\72',
'\201','\55','\175','\121','\260','\371','\353','\165',
'\114','\127','\145','\34','\61','\106','\112','\37',
'\72','\376','\340','\51','\331','\71','\3','\51',
'\112','\120','\370','\247','\360','\313','\117','\241',
'\346','\313','\247','\205','\343','\346','\140','\254',
'\145','\114','\313','\25','\11','\147','\203','\200',
'\74','\153','\305','\251','\100','\204','\352','\336',
'\105','\260','\251','\53','\25','\5','\24','\237',
'\361','\30','\120','\33','\75','\341','\232','\270',
'\301','\164','\142','\57','\367','\175','\75','\152',
'\12','\107','\375','\305','\265','\111','\215','\107',
'\331','\216','\64','\307','\250','\176','\21','\253',
'\212','\126','\262','\331','\157','\5','\103','\321',
'\140','\176','\253','\172','\315','\241','\121','\375',
'\326','\32','\365','\114','\133','\225','\215','\105',
'\215','\371','\150','\143','\161','\224','\165','\356',
'\24','\342','\202','\175','\256','\232','\30','\14',
'\160','\342','\316','\151','\354','\366','\270','\224',
'\1','\72','\137','\105','\14','\173','\66','\233',
'\102','\316','\126','\360','\125','\247','\270','\124',
'\16','\310','\252','\144','\264','\131','\104','\101',
'\31','\272','\354','\25','\210','\22','\16','\361',
'\266','\231','\14','\200','\77','\315','\65','\276',
'\324','\271','\321','\112','\345','\275','\306','\56',
'\310','\23','\173','\326','\156','\246','\205','\16',
'\203','\304','\71','\213','\74','\150','\145','\202',
'\370','\233','\37','\13','\300','\372','\214','\336',
'\376','\226','\347','\334','\352','\235','\272','\13',
'\67','\50','\333','\374','\324','\245','\214','\64',
'\240','\157','\140','\63','\14','\72','\62','\365',
'\271','\72','\141','\11','\322','\51','\162','\275',
'\333','\123','\372','\160','\320','\336','\201','\116',
'\43','\310','\52','\10','\133','\52','\120','\2',
'\341','\343','\250','\267','\47','\143','\153','\101',
'\272','\224','\341','\171','\314','\332','\210','\347',
'\4','\22','\321','\313','\112','\141','\342','\235',
'\215','\174','\211','\123','\22','\315','\211','\170',
'\177','\313','\46','\130','\134','\226','\65','\362',
'\116','\347','\236','\4','\224','\40','\153','\204',
'\73','\255','\176','\7','\27','\202','\170','\151',
'\143','\334','\221','\316','\7','\324','\14','\307',
'\106','\371','\332','\307','\106','\150','\174','\176',
'\326','\44','\115','\102','\112','\140','\45','\43',
'\30','\254','\301','\102','\306','\264','\153','\267',
'\205','\54','\174','\223','\120','\314','\166','\316',
'\20','\171','\265','\122','\324','\12','\137','\27',
'\224','\320','\105','\323','\273','\303','\374','\253',
'\126','\145','\122','\162','\203','\77','\120','\166',
'\350','\141','\21','\362','\222','\161','\37','\276',
'\216','\174','\373','\166','\244','\141','\46','\153',
'\126','\134','\357','\123','\60','\54','\146','\16',
'\102','\33','\202','\135','\71','\4','\257','\120',
'\351','\71','\126','\136','\145','\375','\357','\50',
'\171','\350','\342','\147','\222','\317','\105','\107',
'\42','\67','\357','\46','\5','\114','\306','\227',
'\166','\327','\234','\64','\13','\40','\316','\146',
'\2','\23','\222','\74','\315','\125','\46','\16',
'\72','\267','\142','\12','\47','\32','\107','\252',
'\243','\252','\277','\5','\370','\217','\201','\203',
'\66','\102','\313','\50','\57','\231','\2','\153',
'\255','\53','\4','\207','\250','\53','\376','\166',
'\222','\105','\145','\0','\123','\67','\25','\361',
'\240','\131','\150','\341','\357','\64','\143','\226',
'\144','\146','\67','\106','\254','\372','\17','\302',
'\45','\35','\243','\46','\306','\20','\224','\261',
'\11','\145','\233','\26','\236','\115','\37','\237',
'\335','\144','\360','\350','\354','\6','\134','\352',
'\6','\346','\343','\202','\144','\356','\341','\340',
'\73','\0','\143','\74','\227','\307','\321','\1',
'\173','\4','\360','\7','\106','\135','\72','\170',
'\140','\120','\142','\51','\226','\37','\337','\103',
'\341','\243','\301','\211','\167','\314','\336','\10',
'\350','\227','\44','\227','\6','\301','\144','\144',
'\3','\237','\205','\302','\157','\74','\12','\301',
'\5','\77','\277','\331','\166','\25','\2','\62',
'\26','\313','\30','\0','\300','\20','\233','\27',
'\0','\121','\200','\202','\354','\375','\156','\236',
'\113','\42','\150','\273','\250','\312','\101','\256',
'\214','\364','\161','\166','\13','\316','\154','\321',
'\226','\322','\26','\202','\324','\146','\365','\144',
'\123','\143','\131','\262','\223','\76','\200','\125',
'\262','\312','\150','\376','\11','\62','\305','\104',
'\64','\105','\351','\75','\72','\64','\41','\233',
'\374','\34','\51','\71','\151','\57','\35','\364',
'\157','\235','\122','\250','\300','\16','\347','\110',
'\300','\53','\214','\314','\223','\67','\257','\372',
'\324','\370','\131','\232','\373','\305','\106','\136',
'\363','\272','\140','\352','\174','\161','\155','\152',
'\12','\74','\162','\73','\236','\215','\223','\71',
'\157','\357','\343','\237','\67','\132','\124','\236',
'\50','\155','\330','\176','\261','\252','\207','\4',
'\223','\26','\65','\205','\157','\132','\263','\206',
'\273','\234','\350','\363','\0','\125','\245','\66',
'\320','\353','\301','\175','\24','\232','\325','\356',
'\47','\111','\264','\366','\2','\23','\265','\202',
'\200','\324','\122','\252','\215','\75','\223','\71',
'\36','\366','\335','\271','\106','\271','\307','\70',
'\106','\216','\311','\271','\250','\164','\116','\21',
'\224','\140','\62','\303','\313','\165','\361','\202',
'\175','\360','\25','\105','\17','\352','\156','\227',
'\244','\62','\262','\131','\361','\141','\237','\25',
'\264','\124','\303','\75','\106','\77','\217','\375',
'\126','\140','\224','\110','\56','\247','\311','\206',
'\121','\252','\6','\353','\306','\146','\227','\60',
'\331','\165','\364','\152','\55','\17','\200','\226',
'\11','\321','\57','\203','\107','\112','\267','\366',
'\236','\152','\76','\43','\44','\7','\376','\164',
'\130','\75','\144','\206','\63','\220','\224','\36',
'\112','\361','\15','\302','\367','\105','\105','\137',
'\300','\51','\11','\13','\203','\302','\206','\334',
'\122','\310','\206','\100','\62','\7','\232','\272',
'\52','\3','\134','\152','\26','\72','\120','\250',
'\21','\5','\171','\73','\305','\364','\157','\301',
'\27','\164','\233','\343','\324','\367','\330','\234',
'\22','\141','\52','\310','\171','\132','\315','\240',
'\340','\10','\111','\273','\177','\120','\357','\375',
'\22','\223','\157','\237','\354','\216','\3','\264',
'\265','\45','\25','\151','\214','\100','\354','\210',
'\42','\124','\41','\132','\272','\376','\270','\22',
'\362','\344','\313','\46','\7','\340','\275','\141',
'\276','\165','\315','\152','\153','\230','\256','\211',
'\77','\324','\247','\37','\236','\315','\347','\315',
'\24','\143','\357','\130','\111','\17','\100','\257',
'\324','\370','\321','\201','\43','\152','\6','\54',
'\345','\100','\301','\154','\240','\37','\240','\123',
'\77','\274','\151','\141','\265','\251','\321','\103',
'\20','\321','\123','\204','\275','\123','\133','\221',
'\217','\214','\250','\376','\366','\217','\347','\352',
'\224','\351','\264','\75','\27','\75','\373','\320',
'\123','\35','\5','\303','\120','\354','\123','\103',
'\150','\104','\327','\44','\163','\166','\104','\371',
'\252','\326','\266','\143','\7','\142','\126','\156',
'\151','\114','\134','\361','\130','\344','\356','\314',
'\151','\17','\254','\231','\337','\344','\115','\225',
'\341','\320','\107','\222','\207','\205','\224','\46',
'\200','\150','\337','\45','\344','\27','\340','\206',
'\263','\364','\26','\266','\15','\156','\364','\322',
'\142','\7','\333','\337','\355','\275','\145','\204',
'\126','\115','\202','\340','\175','\77','\215','\342',
'\227','\146','\332','\240','\327','\200','\326','\225',
'\254','\245','\274','\376','\366','\216','\57','\134',
'\171','\371','\20','\246','\7','\124','\227','\234',
'\320','\171','\112','\103','\77','\137','\256','\35',
'\110','\160','\371','\15','\166','\146','\242','\5',
'\214','\341','\316','\107','\266','\20','\353','\270',
'\174','\11','\226','\72','\376','\55','\63','\172',
'\36','\151','\246','\311','\11','\160','\362','\246',
'\30','\332','\60','\257','\327','\37','\157','\300',
'\174','\7','\17','\314','\125','\152','\151','\125',
'\133','\337','\233','\164','\126','\76','\52','\240',
'\127','\36','\334','\306','\216','\203','\224','\312',
'\144','\307','\67','\351','\272','\201','\117','\46',
'\255','\266','\235','\212','\326','\363','\117','\174',
'\21','\207','\354','\101','\341','\157','\22','\114',
'\166','\147','\167','\354','\143','\355','\160','\126',
'\131','\127','\56','\311','\222','\165','\6','\335',
'\315','\24','\207','\24','\134','\272','\216','\13',
'\247','\164','\117','\240','\37','\245','\131','\16',
'\156','\337','\272','\357','\355','\330','\40','\316',
'\347','\211','\2','\301','\226','\224','\355','\6',
'\337','\170','\360','\101','\117','\153','\361','\150',
'\370','\226','\62','\163','\376','\227','\66','\211',
'\157','\244','\306','\320','\335','\161','\357','\307',
'\63','\17','\42','\33','\152','\42','\30','\77',
'\173','\267','\155','\253','\246','\301','\327','\343',
'\157','\25','\50','\35','\372','\256','\336','\40',
'\125','\362','\237','\173','\51','\312','\355','\240',
'\373','\71','\322','\243','\304','\60','\217','\24',
'\47','\35','\112','\341','\234','\206','\252','\117',
'\117','\323','\61','\142','\107','\44','\330','\222',
'\22','\350','\107','\306','\135','\346','\256','\147',
'\220','\120','\140','\303','\45','\120','\71','\40',
'\305','\124','\353','\227','\146','\272','\106','\301',
'\340','\265','\63','\142','\155','\6','\360','\224',
'\247','\202','\102','\0','\13','\77','\331','\6',
'\126','\364','\365','\4','\40','\357','\10','\214',
'\361','\172','\56','\344','\266','\322','\240','\174',
'\100','\304','\203','\134','\333','\156','\214','\212',
'\370','\153','\315','\40','\57','\111','\30','\71',
'\142','\161','\150','\201','\3','\163','\41','\143',
'\165','\155','\335','\20','\52','\202','\7','\35',
'\210','\326','\61','\351','\17','\217','\355','\35',
'\222','\203','\232','\235','\353','\224','\152','\216',
'\4','\300','\6','\245','\247','\36','\165','\43',
'\74','\254','\52','\361','\136','\373','\233','\333',
'\327','\111','\204','\221','\135','\202','\315','\322',
'\154','\247','\173','\135','\312','\3','\355','\122',
'\61','\60','\344','\67','\43','\12','\200','\254',
'\375','\346','\144','\175','\61','\152','\117','\276',
'\27','\362','\42','\215','\310','\147','\237','\261',
'\217','\324','\277','\157','\371','\152','\241','\217',
'\256','\103','\214','\337','\311','\61','\213','\31',
'\62','\221','\156','\317','\270','\145','\63','\271',
'\30','\245','\234','\132','\10','\202','\257','\240',
'\201','\241','\252','\103','\77','\117','\40','\344',
'\264','\322','\17','\141','\260','\351','\213','\134',
'\332','\173','\357','\156','\136','\34','\2','\355',
'\122','\35','\304','\265','\52','\122','\156','\310',
'\215','\271','\320','\150','\177','\241','\214','\325',
'\266','\362','\336','\112','\6','\26','\267','\51',
'\123','\176','\225','\121','\130','\134','\61','\267',
'\23','\357','\161','\326','\316','\175','\127','\335',
'\33','\34','\202','\146','\342','\165','\243','\260',
'\50','\160','\330','\14','\303','\244','\273','\231',
'\12','\323','\104','\122','\202','\360','\161','\257',
'\235','\264','\345','\30','\276','\27','\61','\5',
'\311','\156','\161','\370','\121','\23','\154','\270',
'\164','\300','\15','\5','\331','\100','\106','\143',
'\122','\274','\360','\266','\276','\100','\253','\26',
'\121','\24','\364','\351','\43','\217','\103','\157',
'\372','\27','\376','\275','\311','\173','\122','\245',
'\200','\206','\141','\313','\61','\60','\3','\207',
'\142','\320','\43','\324','\362','\241','\300','\131',
'\162','\131','\371','\223','\110','\360','\145','\316',
'\77','\205','\243','\16','\17','\26','\140','\312',
'\131','\37','\223','\351','\115','\12','\124','\366',
'\240','\232','\356','\217','\107','\122','\12','\54',
'\326','\170','\31','\273','\153','\74','\105','\265',
'\55','\101','\70','\107','\341','\367','\62','\152',
'\60','\207','\344','\142','\131','\156','\117','\106',
'\202','\200','\171','\214','\204','\51','\211','\211',
'\130','\142','\325','\306','\64','\147','\230','\355',
'\356','\13','\4','\352','\257','\345','\247','\132',
'\30','\155','\26','\271','\314','\337','\241','\115',
'\40','\266','\345','\244','\10','\334','\35','\77',
'\121','\54','\102','\366','\107','\114','\302','\324',
'\157','\166','\114','\236','\241','\156','\243','\64',
'\140','\301','\245','\337','\41','\254','\153','\31',
'\20','\250','\204','\74','\40','\113','\52','\32',
'\233','\215','\235','\20','\103','\70','\266','\342',
'\14','\123','\23','\353','\276','\364','\340','\7',
'\245','\74','\37','\312','\274','\320','\24','\67',
'\220','\264','\130','\66','\10','\212','\35','\213',
'\151','\367','\212','\264','\256','\255','\123','\131',
'\34','\151','\277','\125','\357','\232','\4','\330',
'\30','\316','\342','\350','\103','\237','\373','\17',
'\43','\245','\204','\216','\210','\10','\217','\262',
'\67','\251','\53','\151','\5','\106','\2','\314',
'\44','\120','\346','\322','\215','\163','\73','\200',
'\221','\262','\260','\171','\1','\24','\373','\266',
'\300','\340','\211','\105','\247','\14','\127','\116',
'\104','\157','\156','\34','\14','\220','\234','\226',
'\75','\60','\67','\311','\236','\346','\23','\153',
'\344','\243','\5','\102','\137','\203','\133','\350',
'\144','\272','\371','\350','\71','\161','\60','\215',
'\22','\265','\232','\303','\324','\124','\155','\45',
'\314','\14','\164','\275','\263','\167','\133','\326',
'\173','\315','\113','\276','\321','\344','\36','\266',
'\135','\267','\200','\110','\272','\110','\51','\27',
'\142','\52','\124','\337','\62','\232','\20','\15',
'\50','\67','\366','\240','\260','\165','\311','\255',
'\135','\102','\147','\360','\265','\346','\143','\371',
'\122','\160','\156','\363','\42','\321','\25','\277',
'\343','\75','\305','\23','\306','\344','\14','\315',
'\157','\10','\127','\340','\56','\66','\122','\142',
'\144','\20','\331','\305','\46','\126','\2','\242',
'\106','\27','\31','\36','\323','\52','\0','\132',
'\306','\355','\110','\316','\106','\223','\132','\371',
'\334','\126','\26','\356','\300','\15','\367','\257',
'\51','\126','\331','\3','\375','\352','\277','\210',
'\165','\326','\312','\366','\236','\170','\162','\164',
'\53','\0','\312','\76','\164','\105','\152','\263',
'\345','\203','\353','\42','\127','\27','\342','\274',
'\241','\102','\122','\333','\213','\207','\346','\122',
'\302','\107','\4','\7','\36','\25','\217','\50',
'\157','\221','\3','\236','\264','\337','\314','\154',
'\264','\227','\260','\41','\136','\67','\241','\43',
'\24','\202','\361','\226','\131','\135','\344','\2',
'\217','\70','\72','\325','\217','\211','\346','\171',
'\316','\123','\161','\65','\267','\252','\33','\176',
'\55','\316','\350','\364','\55','\211','\155','\300',
'\266','\313','\313','\110','\204','\5','\101','\365',
'\302','\257','\54','\176','\50','\42','\121','\66',
'\151','\340','\70','\151','\12','\27','\265','\123',
'\313','\46','\311','\21','\330','\336','\300','\231',
'\164','\212','\250','\21','\141','\304','\50','\230',
'\350','\217','\374','\115','\240','\3','\310','\331',
'\3','\302','\52','\34','\257','\215','\376','\34',
'\126','\101','\236','\162','\243','\173','\315','\52',
'\306','\233','\201','\332','\201','\115','\155','\217',
'\334','\165','\77','\354','\146','\47','\123','\256',
'\316','\207','\161','\145','\73','\104','\313','\112',
'\303','\14','\217','\206','\103','\331','\53','\131',
'\262','\207','\16','\262','\233','\240','\206','\124',
'\362','\242','\252','\112','\121','\272','\120','\14',
'\175','\222','\342','\34','\70','\362','\376','\155',
'\350','\370','\313','\315','\261','\313','\11','\210',
'\201','\233','\102','\34','\324','\230','\135','\302',
'\12','\336','\201','\37','\13','\327','\112','\125',
'\42','\225','\271','\164','\47','\136','\120','\51',
'\5','\161','\160','\177','\312','\354','\365','\60',
'\227','\312','\310','\62','\256','\167','\300','\47',
'\157','\115','\141','\302','\312','\237','\35','\277',
'\363','\132','\135','\62','\0','\33','\222','\121',
'\4','\166','\123','\242','\350','\344','\1','\341',
'\157','\64','\274','\107','\14','\252','\70','\142',
'\241','\137','\27','\364','\135','\64','\226','\262',
'\155','\370','\53','\50','\4','\20','\57','\264',
'\133','\223','\165','\146','\153','\26','\355','\14',
'\253','\2','\253','\325','\101','\347','\51','\100',
'\353','\143','\6','\327','\26','\201','\42','\116',
'\32','\276','\54','\201','\171','\351','\363','\142',
'\210','\271','\233','\66','\150','\120','\71','\140',
'\162','\263','\355','\205','\35','\24','\220','\37',
'\336','\154','\373','\201','\225','\104','\300','\370',
'\13','\272','\312','\64','\254','\235','\322','\112',
'\216','\366','\346','\273','\311','\24','\257','\314',
'\203','\231','\333','\224','\353','\335','\214','\41',
'\263','\160','\266','\105','\74','\125','\216','\76',
'\53','\243','\116','\253','\206','\44','\260','\245',
'\57','\124','\251','\210','\50','\261','\222','\275',
'\270','\212','\311','\120','\7','\153','\151','\314',
'\217','\46','\344','\0','\27','\256','\230','\347',
'\306','\221','\22','\100','\353','\374','\72','\305',
'\340','\120','\311','\232','\326','\41','\266','\345',
'\340','\217','\137','\73','\76','\337','\244','\102',
'\334','\274','\231','\105','\67','\263','\153','\151',
'\163','\130','\275','\214','\332','\160','\6','\321',
'\251','\71','\360','\312','\54','\211','\47','\114',
'\111','\50','\146','\320','\166','\344','\174','\134',
'\36','\270','\202','\137','\41','\310','\276','\17',
'\245','\220','\251','\13','\363','\55','\275','\223',
'\106','\66','\324','\154','\151','\277','\305','\177',
'\235','\340','\301','\327','\221','\241','\252','\144',
'\261','\55','\341','\316','\174','\330','\23','\25',
'\361','\271','\364','\277','\37','\147','\205','\310',
'\212','\204','\17','\236','\122','\331','\261','\271',
'\6','\125','\174','\121','\114','\344','\372','\220',
'\235','\357','\26','\64','\15','\126','\15','\175',
'\253','\117','\135','\311','\155','\352','\314','\342',
'\332','\154','\132','\77','\125','\304','\150','\41',
'\216','\361','\270','\107','\313','\273','\7','\164',
'\16','\244','\141','\125','\371','\232','\373','\246',
'\251','\242','\103','\22','\204','\354','\121','\343',
'\305','\211','\230','\4','\30','\76','\210','\50',
'\341','\160','\321','\267','\134','\162','\201','\4',
'\244','\271','\206','\132','\146','\336','\221','\365',
'\245','\320','\246','\121','\253','\51','\227','\146',
'\20','\107','\165','\24','\137','\223','\362','\11',
'\376','\246','\315','\236','\63','\34','\25','\263',
'\373','\162','\153','\62','\101','\47','\113','\137',
'\171','\120','\215','\4','\336','\363','\65','\327',
'\160','\52','\350','\274','\76','\145','\356','\360',
'\275','\23','\356','\63','\235','\152','\251','\106',
'\127','\125','\302','\247','\112','\242','\60','\70',
'\325','\112','\53','\242','\125','\214','\32','\16',
'\350','\271','\257','\0','\175','\251','\12','\264',
'\33','\25','\157','\231','\333','\250','\57','\100',
'\171','\102','\330','\73','\236','\247','\275','\142',
'\101','\56','\14','\76','\346','\104','\345','\10',
'\367','\330','\303','\4','\215','\1','\27','\26',
'\332','\177','\316','\16','\3','\207','\172','\163',
'\11','\374','\225','\26','\245','\325','\47','\337',
'\301','\325','\73','\170','\34','\11','\235','\5',
'\71','\356','\312','\314','\244','\246','\305','\6',
'\71','\333','\245','\210','\2','\347','\72','\140',
'\153','\66','\3','\156','\47','\334','\152','\362',
'\236','\324','\71','\277','\134','\157','\117','\324',
'\106','\362','\150','\172','\204','\222','\71','\250',
'\220','\136','\23','\67','\255','\111','\165','\316',
'\66','\146','\77','\133','\253','\55','\264','\144',
'\247','\327','\144','\51','\42','\63','\313','\217',
'\112','\106','\165','\312','\71','\360','\352','\234',
'\376','\234','\56','\313','\61','\74','\75','\72',
'\55','\141','\133','\62','\115','\34','\10','\265',
'\247','\42','\5','\270','\130','\110','\275','\362',
'\137','\166','\10','\340','\30','\70','\220','\373',
'\70','\334','\376','\171','\337','\227','\133','\360',
'\57','\102','\44','\150','\332','\203','\321','\127',
'\346','\376','\302','\130','\327','\261','\13','\3',
'\143','\120','\313','\266','\134','\244','\335','\73',
'\61','\31','\15','\344','\101','\203','\343','\120',
'\72','\5','\365','\376','\157','\13','\346','\133',
'\22','\266','\25','\140','\333','\352','\35','\374',
'\253','\345','\153','\371','\126','\327','\15','\336',
'\7','\36','\162','\332','\24','\306','\157','\212',
'\353','\14','\103','\167','\325','\40','\345','\146',
'\244','\344','\113','\141','\116','\6','\56','\121',
'\22','\355','\161','\146','\157','\34','\337','\257',
'\31','\350','\273','\367','\65','\306','\165','\253',
'\21','\57','\25','\171','\234','\27','\315','\131',
'\323','\174','\210','\357','\147','\357','\376','\126',
'\5','\210','\36','\33','\220','\247','\24','\77',
'\371','\23','\141','\134','\203','\316','\227','\112',
'\21','\315','\330','\136','\235','\75','\122','\272',
'\350','\350','\107','\123','\213','\67','\232','\331',
'\356','\247','\225','\330','\103','\246','\24','\306',
'\36','\106','\237','\76','\170','\220','\7','\30',
'\223','\236','\247','\24','\144','\306','\273','\53',
'\230','\165','\205','\267','\144','\155','\53','\62',
'\364','\32','\273','\104','\125','\7','\103','\367',
'\72','\71','\265','\210','\330','\213','\376','\275',
'\117','\321','\315','\322','\322','\54','\256','\273',
'\262','\173','\35','\375','\120','\15','\202','\45',
'\14','\146','\207','\356','\266','\172','\31','\324',
'\117','\325','\106','\220','\362','\216','\174','\333',
'\302','\12','\232','\145','\34','\35','\257','\317',
'\16','\124','\274','\145','\117','\307','\335','\144',
'\351','\337','\102','\107','\227','\160','\101','\162',
'\17','\303','\70','\154','\20','\220','\14','\342',
'\154','\307','\101','\352','\274','\272','\106','\246',
'\173','\247','\254','\157','\342','\250','\223','\333',
'\306','\202','\130','\206','\175','\352','\317','\337',
'\234','\321','\44','\61','\137','\364','\210','\146',
'\327','\25','\267','\336','\344','\65','\320','\16',
'\31','\276','\355','\21','\361','\125','\242','\56',
'\46','\145','\107','\336','\60','\27','\131','\5',
'\356','\54','\62','\335','\324','\21','\324','\155',
'\215','\354','\200','\320','\266','\75','\70','\227',
'\133','\362','\76','\236','\272','\267','\205','\226',
'\342','\367','\375','\201','\37','\166','\221','\152',
'\206','\312','\135','\23','\175','\223','\324','\200',
'\143','\372','\322','\122','\232','\351','\250','\216',
'\202','\117','\337','\255','\366','\342','\163','\204',
'\244','\147','\177','\300','\223','\101','\227','\15',
'\211','\205','\375','\143','\241','\115','\355','\313',
'\157','\356','\275','\14','\300','\241','\43','\136',
'\122','\226','\1','\135','\341','\120','\321','\171',
'\100','\327','\25','\32','\360','\207','\222','\226',
'\212','\343','\224','\216','\51','\6','\153','\44',
'\373','\111','\254','\230','\0','\43','\261','\53',
'\46','\247','\253','\171','\117','\20','\171','\361',
'\371','\32','\316','\23','\44','\346','\140','\113',
'\361','\164','\137','\47','\241','\0','\253','\47',
'\157','\131','\261','\150','\0','\300','\376','\115',
'\334','\371','\112','\335','\142','\252','\66','\76',
'\360','\37','\57','\2','\144','\152','\167','\323',
'\330','\122','\346','\226','\42','\357','\54','\327',
'\312','\260','\56','\160','\254','\316','\273','\273',
'\51','\256','\343','\42','\202','\364','\323','\35',
'\232','\14','\77','\12','\335','\273','\372','\226',
'\216','\0','\72','\217','\24','\60','\220','\41',
'\320','\233','\371','\100','\50','\100','\151','\322',
'\204','\42','\301','\132','\214','\273','\330','\101',
'\303','\345','\134','\221','\333','\372','\106','\250',
'\337','\221','\166','\154','\245','\172','\121','\126',
'\52','\300','\137','\47','\150','\30','\104','\131',
'\110','\131','\255','\113','\336','\376','\120','\344',
'\40','\247','\30','\246','\211','\26','\316','\172',
'\160','\70','\337','\0','\327','\121','\216','\52',
'\331','\147','\254','\100','\333','\331','\360','\335',
'\304','\263','\171','\64','\161','\60','\331','\145',
'\40','\75','\213','\245','\242','\376','\307','\376',
'\346','\366','\172','\47','\74','\151','\147','\151',
'\223','\370','\321','\251','\214','\37','\13','\272',
'\206','\331','\32','\343','\162','\316','\261','\161',
'\121','\347','\204','\260','\25','\44','\302','\157',
'\323','\262','\156','\61','\115','\321','\322','\12',
'\55','\236','\31','\215','\275','\12','\262','\256',
'\353','\243','\40','\360','\351','\242','\73','\167',
'\43','\306','\267','\12','\34','\224','\226','\211',
'\313','\62','\217','\376','\141','\1','\10','\327',
'\355','\166','\302','\201','\344','\17','\65','\331',
'\304','\274','\370','\12','\10','\55','\327','\112',
'\346','\74','\240','\115','\160','\320','\1','\125',
'\260','\242','\312','\73','\44','\242','\170','\21',
'\335','\143','\367','\337','\231','\16','\236','\327',
'\337','\173','\151','\315','\44','\305','\207','\166',
'\201','\354','\64','\45','\333','\144','\136','\333',
'\230','\137','\210','\4','\164','\47','\72','\353',
'\334','\175','\237','\72','\11','\22','\302','\160',
'\271','\262','\31','\113','\364','\234','\365','\323',
'\11','\152','\262','\40','\56','\121','\326','\360',
'\205','\172','\100','\137','\5','\213','\25','\156',
'\300','\266','\264','\327','\317','\352','\177','\254',
'\342','\310','\161','\273','\66','\14','\5','\246',
'\330','\16','\303','\24','\42','\255','\326','\155',
'\273','\350','\205','\60','\101','\326','\360','\14',
'\316','\120','\330','\71','\145','\350','\316','\142',
'\221','\243','\133','\32','\57','\50','\32','\30',
'\275','\126','\331','\324','\112','\75','\175','\263',
'\346','\143','\376','\226','\204','\225','\335','\200',
'\67','\315','\137','\34','\334','\334','\213','\300',
'\120','\154','\71','\114','\143','\230','\357','\10',
'\1','\41','\60','\252','\122','\132','\56','\116',
'\30','\12','\330','\111','\14','\360','\326','\317',
'\123','\356','\337','\341','\303','\200','\47','\265',
'\125','\270','\243','\236','\352','\42','\204','\227',
'\113','\363','\274','\72','\210','\353','\250','\146',
'\172','\157','\73','\371','\31','\12','\230','\231',
'\27','\177','\250','\303','\255','\314','\202','\150',
'\325','\73','\354','\27','\320','\102','\244','\261',
'\217','\151','\16','\21','\327','\313','\67','\270',
'\326','\137','\71','\376','\322','\132','\63','\353',
'\342','\147','\261','\322','\121','\102','\362','\230',
'\144','\123','\54','\167','\23','\334','\57','\301',
'\227','\332','\300','\31','\67','\75','\146','\53',
'\233','\100','\22','\332','\173','\77','\165','\74',
'\326','\230','\352','\131','\56','\307','\202','\315',
'\320','\62','\120','\276','\224','\370','\121','\253',
'\303','\367','\354','\366','\336','\362','\165','\246',
'\220','\30','\346','\366','\24','\52','\25','\225',
'\202','\370','\46','\124','\107','\65','\157','\353',
'\27','\310','\54','\77','\136','\163','\264','\364',
'\257','\370','\362','\203','\52','\112','\47','\100',
'\26','\361','\62','\331','\177','\114','\135','\343',
'\33','\207','\222','\125','\231','\335','\366','\315',
'\107','\362','\3','\132','\307','\52','\177','\117',
'\2','\364','\303','\72','\251','\307','\67','\235',
'\334','\236','\301','\361','\233','\220','\364','\110',
'\200','\70','\73','\4','\252','\211','\165','\140',
'\365','\375','\277','\357','\237','\136','\10','\167',
'\24','\223','\30','\215','\260','\314','\356','\321',
'\7','\131','\31','\225','\63','\154','\130','\346',
'\374','\5','\232','\327','\211','\40','\106','\212',
'\337','\57','\41','\34','\26','\152','\256','\12',
'\227','\147','\327','\5','\135','\207','\357','\231',
'\73','\66','\151','\322','\376','\124','\32','\153',
'\36','\217','\315','\203','\107','\235','\307','\11',
'\373','\257','\264','\253','\156','\146','\341','\155',
'\36','\342','\54','\140','\154','\3','\314','\220',
'\345','\12','\11','\6','\330','\250','\301','\325',
'\62','\161','\147','\152','\373','\107','\50','\52',
'\111','\364','\330','\374','\32','\354','\313','\251',
'\117','\136','\310','\367','\173','\160','\250','\107',
'\134','\134','\34','\333','\240','\200','\340','\321',
'\146','\277','\321','\161','\114','\364','\240','\163',
'\223','\314','\226','\163','\206','\133','\347','\236',
'\254','\253','\351','\373','\7','\101','\364','\140',
'\13','\246','\227','\211','\372','\63','\210','\324',
'\141','\13','\145','\245','\145','\160','\300','\257',
'\261','\34','\16','\176','\151','\146','\121','\373',
'\121','\164','\103','\26','\204','\254','\110','\220',
'\136','\306','\13','\245','\216','\57','\261','\342',
'\33','\171','\145','\10','\360','\200','\115','\260',
'\132','\165','\267','\357','\273','\110','\27','\171',
'\110','\320','\176','\125','\243','\131','\43','\175',
'\357','\353','\100','\135','\176','\77','\142','\0',
'\301','\227','\305','\171','\252','\11','\343','\326',
'\53','\162','\227','\57','\11','\356','\154','\337',
'\326','\37','\144','\61','\257','\161','\177','\32',
'\267','\314','\101','\167','\325','\103','\313','\203',
'\112','\105','\127','\213','\76','\50','\15','\150',
'\371','\136','\357','\161','\250','\260','\374','\324',
'\64','\162','\33','\7','\151','\37','\64','\76',
'\115','\200','\41','\36','\311','\273','\311','\302',
'\170','\70','\77','\226','\105','\272','\341','\1',
'\352','\361','\26','\250','\10','\140','\345','\170',
'\177','\300','\126','\267','\171','\20','\270','\206',
'\314','\37','\114','\247','\342','\121','\205','\367',
'\163','\374','\133','\0','\250','\234','\356','\272',
'\362','\132','\257','\260','\245','\43','\204','\305',
'\263','\2','\321','\41','\327','\170','\74','\177',
'\303','\101','\72','\104','\22','\105','\22','\57',
'\375','\272','\370','\126','\272','\301','\362','\270',
'\15','\124','\344','\127','\70','\100','\111','\114',
'\132','\12','\155','\235','\76','\77','\344','\104',
'\376','\240','\7','\121','\340','\54','\306','\72',
'\115','\133','\312','\115','\352','\104','\173','\146',
'\247','\307','\262','\71','\71','\374','\231','\177',
'\102','\230','\75','\277','\144','\344','\147','\303',
'\155','\234','\212','\155','\325','\256','\74','\245',
'\254','\35','\243','\375','\6','\114','\261','\331',
'\306','\323','\140','\175','\332','\0','\333','\72',
'\63','\151','\305','\163','\126','\27','\374','\246',
'\256','\254','\251','\354','\75','\57','\263','\314',
'\0','\335','\330','\272','\41','\231','\243','\10',
'\145','\322','\371','\357','\216','\120','\305','\305',
'\355','\334','\302','\350','\161','\15','\360','\336',
'\35','\304','\44','\113','\52','\27','\66','\276',
'\156','\214','\203','\351','\160','\213','\163','\154',
'\173','\347','\257','\154','\245','\357','\265','\316',
'\217','\343','\307','\361','\4','\234','\225','\100',
};

static long int callnum = 0;
static int rndim = 0;

/*
   Restart the random number package.
 */
void
vdt_restart_rand_0_to_1 (void)
{
  callnum = 0;
  rndim   = sizeof (rn) / sizeof (rn[0]);
}

/*
  Generates a pseudo-random number in the interval 0...1
  */
double 
vdt_rand_0_to_1 (int dim)
{
  int j = (callnum++) % rndim;
  j += dim;
  if (j >= rndim) j = 0;
  return (double) rn[j] / 255.0;
}

#else

/* Pseudo random number generator data structures */
/* Knuth's lagged Fibonacci-based generator: See "Seminumerical Algorithms:
   The Art of Computer Programming" sections 3.2-3.3 */

#define MZ	0L
#define RAND_0_TO_1_SEED  3127

static long mrand_list[56];
static int  started = 0;
static int  inext = 0, inextp = 31; /* this is why we need 32-bit int's */


double 
vdt_rand_0_to_1 (int dim) ;


static int
maxint (void)
{
  int i = 1, old_i;
  while (i > 0) {
    old_i = i;
    i = (i << 1) | 1;
  }
  return old_i;
}

static double  factor;
static int modulus = 0;
#define INT_MODULUS 2147483647  /* for 32-bit int's */

static void 
vdt_smrand (int seed) 
{
  int i;

  modulus = maxint ();
  /* we need 32-bit int's */
  if (modulus < INT_MODULUS) {
    char buf[128];
    sprintf (buf, "Not enough precision for rand_0_to_1 (): max int = %d",
             modulus);
    errexit (buf);
  }
  /* this is what we want to use across all platforms */
  modulus = INT_MODULUS;

  factor =  1.0 / ((double) modulus);
  
  mrand_list[0] = (123413*seed) % modulus;
  for ( i = 1; i < 55; i++ ) 
    mrand_list[i] = (123413*mrand_list[i-1]) % modulus;
  
  started = 1;
  
  /* run mrand () through the list sufficient times to
     thoroughly randomise the array */
  for ( i = 0; i < 55*55; i++ ) 
    vdt_rand_0_to_1 (0);
} 

/*
   Restart the random number package.
 */
void
vdt_restart_rand_0_to_1 (void)
{
  vdt_smrand (RAND_0_TO_1_SEED);
}

/*
  Generates a pseudo-random number in the interval 0...1
  */
double 
vdt_rand_0_to_1 (int dim) 
{
  long	lval;
  
  if ( ! started ) 
    vdt_restart_rand_0_to_1 ();
  
  inext = (inext >= 54) ? 0 : inext+1;
  inextp = (inextp >= 54) ? 0 : inextp+1;
  
  lval = mrand_list[inext] - mrand_list[inextp];
  if ( lval < 0L ) 
    lval += modulus;
  mrand_list[inext] = lval;
  
  return (double) lval*factor;
} 


#undef MODULUS
#undef MZ
#undef FAC
#endif /* VDT_REPRODUCIBLE_RANDOM */
